#include <assert.h>
#include <math.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "libmesh.h"
#include "struct.h"

static int idir[5]  = {0,1,2,0,1};

extern ubyte quiet;


/* very sioux! (09/2002) */
int hashTria(pMesh mesh) {
  pTria     pt,pt1;
  pPoint    p1,p2;
  int       k,kk,l,ll,mins,maxs,mins1,maxs1,hsize;
  int      *hcode,*link,*adj,inival,iadr,pp;
  char     *hvoy;
  ubyte     i,i1,i2,ii;
  unsigned int key;

  if ( !quiet )  fprintf(stdout," Setting topology\n");

  /* memory alloc */
  hcode = (int*)calloc(max(1,3*mesh->nt/4)+1,sizeof(int));
  link  = (int*)calloc(3*mesh->nt+1,sizeof(int));
  hsize = max(2,3*mesh->nt/4-1);
  if ( !hcode || !link ) {
    fprintf(stdout,"  Unable to hash edges. Exit.\n");
    exit(1);
  }
  hvoy = (char*)hcode;

  /* init */
  inival = 2 << 30;
  for (k=0; k<=3*mesh->nt/4; k++)
    hcode[k] = -inival;

  /* build hash table */
  for (k=1; k<=mesh->nt; k++) {
    pt = &mesh->tria[k];
    if ( !pt->v[0] )  continue;
    for (i=0; i<3; i++) {
      i1 = idir[i+1];
      i2 = idir[i+2];
      mins = min(pt->v[i1],pt->v[i2]);
      maxs = max(pt->v[i1],pt->v[i2]);
      p1   = &mesh->point[pt->v[i]];
      p1->b = 0;
      if ( !p1->s )  p1->s = k;

      /* compute key */
      key = KA*mins + KB*maxs;
      key = key % hsize + 1;

      /* insert */
      iadr = 3*(k-1) + i+1;
      link[iadr] = hcode[key];
      hcode[key] = -iadr;
    }
  }

  /* set adjacency */
  for (l=3*mesh->nt; l>0; l--) {
    if ( link[l] >= 0 )  continue;
    k = (l-1) / 3 + 1;
    i = (l-1) % 3;
    i1 = idir[i+1];
    i2 = idir[i+2];
    pt = &mesh->tria[k];

    mins = min(pt->v[i1],pt->v[i2]);
    maxs = max(pt->v[i1],pt->v[i2]);

    /* accross link */
    ll = -link[l];
    pp = 0;
    link[l] = 0;
    hvoy[l] = 0;
    while ( ll != inival ) {
      kk = (ll-1) / 3 +1;
      ii = (ll-1) % 3;
      i1 = idir[ii+1];
      i2 = idir[ii+2];
      pt1   = &mesh->tria[kk];
      mins1 = min(pt1->v[i1],pt1->v[i2]);
      maxs1 = max(pt1->v[i1],pt1->v[i2]);
      
      /* adjacent found */
      if ( mins1 == mins && maxs1 == maxs ) {
        if ( pp != 0 )  link[pp] = link[ll];
        link[l] = kk;
        hvoy[l] = ii;
        link[ll]= k;
        hvoy[ll]= i;
        break;
      }
      pp = ll;
      ll = -link[ll];
    }
  }
  mesh->adja = (int*)link;
  mesh->voy  = (ubyte*)hcode;

  for (k=1; k<=mesh->nt; k++) {
    pt   = &mesh->tria[k];
    iadr = 3*(k-1)+1;
    adj  = &mesh->adja[iadr];
    for (i=0; i<3; i++) {
      i1 = idir[i+1];
      i2 = idir[i+2];
      if ( !adj[i] ) {
	p1 = &mesh->point[pt->v[i1]];
        p2 = &mesh->point[pt->v[i2]];
	p1->b = 1;
	p2->b = 1;
      }
    }
  }

  /* set simplex */
  for (k=1; k<=mesh->nt; k++) {
    pt   = &mesh->tria[k];
    iadr = 3*(k-1)+1;
    adj  = &mesh->adja[iadr];
    for (i=0; i<3; i++) {
      i2 = idir[i+2];
      p1 = &mesh->point[pt->v[i]];
      if ( !adj[i2] )  p1->s = k;
    }
  }

  return(1);
}


