#ifdef __cplusplus
extern "C" {
#endif

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "libmesh.h"
#include "struct.h"

#define MAXTYP    20

extern double  delta;
extern ubyte   quiet,iso;
extern int     eigen2(double m[3],double lambda[2],double vp[2][2]);


int loadMesh(char *name,pMesh mesh) {
  FILE   *in;
  pTria   pt;
  pPoint  p0,p1,p2,p3;
  double  dx,dy,ux,uy,h1,h2,h3,pe,ar,rins;
  float   xd,yd,xmin,ymin,xmax,ymax,dmax;
  int     k,vers,key,dim,dummy;

  if ( !quiet )  fprintf(stdout," Checking %s\n",name);
  in = ouvrir_mesh(name,"r",&vers);
  if ( !in ) {
    fprintf(stdout,"  File not found. Exit.\n");
    exit(1);
  }

  mesh->ns = mesh->nt = 0;
  xmin = ymin =  1.0e30;
  xmax = ymax = -1.0e30;
  dx   = dy   = 1.0;

  do {
    key = mot_clef_suivant(in);
    if ( key == End )  break;
    switch(key) {
      case MeshDimension:
        dim = lire_int(in);
	if ( dim != 2 ) {
	  fprintf(stdout,"  Wrong dimension %d. Exit.\n",dim);
          exit(1);
	}
	break;

      case Vertices:
        mesh->ns = lire_int(in);
	if ( !mesh->ns ) {
	  fprintf(stdout,"  No vertex found. Exit.\n");
	  exit(1);
	}
        mesh->point = (pPoint)calloc(mesh->ns+1,sizeof(Point));
        if ( !mesh->point ) {
	  fprintf(stdout,"  Unable to store coordinates. Exit.\n");
	  exit(1);
	}
	dmax = 0.0;
	for (k=1; k<=mesh->ns; k++) {
	  p0    = &mesh->point[k];
	  p0->x = lire_reel(in);
	  p0->y = lire_reel(in);
	  dummy = lire_int(in);
	  p0->r = 0.0;
	  p0->a = 0.0;
	  if ( p0->x < xmin )  xmin = p0->x;
	  if ( p0->x > xmax )  xmax = p0->x;
	  if ( p0->y < ymin )  ymin = p0->y;
	  if ( p0->y > ymax )  ymax = p0->y;
	}

	/* normalize coords */
        delta = 1.0;
        /*	
	xd = xmax - xmin;
	yd = ymax - ymin;
	dx = (double)ALPHA / xd;
	dy = (double)ALPHA / yd;
	delta = max(dx,dy);
	for (k=1; k<=mesh->ns; k++) {
	  p0    = &mesh->point[k];
	  p0->x = (p0->x - xmin) * delta;
	  p0->y = (p0->y - ymin) * delta;
	}
        */
	break;

      case Triangles:
        mesh->nt = lire_int(in);
	if ( !mesh->nt ) {
	  fprintf(stdout,"  No triangle found. Exit.\n");
	  exit(1);
	}
	mesh->tria = (pTria)calloc(mesh->nt+1,sizeof(Tria));
        if ( !mesh->tria ) {
	  fprintf(stdout,"  Unable to store triangles. Exit.\n");
	  exit(1);
	}
	for (k=1; k<=mesh->nt; k++) {
	  pt       = &mesh->tria[k];
	  pt->v[0] = lire_int(in);
	  pt->v[1] = lire_int(in);
	  pt->v[2] = lire_int(in);
	  dummy    = lire_int(in);
	  
	  /* compute in-radius */
	  p1 = &mesh->point[pt->v[0]];
          p2 = &mesh->point[pt->v[1]];
          p3 = &mesh->point[pt->v[2]];

          ux = p2->x - p1->x;
          uy = p2->y - p1->y;
          h1 = sqrt(ux*ux + uy*uy);
          ux = p3->x - p1->x;
          uy = p3->y - p1->y;
          h2 = sqrt(ux*ux + uy*uy);
          ux = p3->x - p2->x;
          uy = p3->y - p2->y;
          h3 = sqrt(ux*ux + uy*uy);

          pe    = 0.5 * (h1 + h2 + h3);
          ar    = pe * (pe-h1) * (pe-h2) * (pe-h3);
          ar    = sqrt(ar);
	  rins  = 2.0 * ar / pe; 

          pt->s  = ar;
	  p1->a += ar;
	  p1->r += rins;
	  
	  p2->a += ar;
          p2->r += rins;

	  p3->a += ar;
          p3->r += rins;
	}
        break;
    }
  }
  while ( key != End && (!mesh->ns || !mesh->nt) );
  fclose(in);
  if ( !mesh->ns || !mesh->nt ) {
    fprintf(stdout,"  Missing entities. Exit.\n");
    exit(1);
  }

  return(1);
}


/* read solution from file */
int loadSolu(char *name,pSol sol) {
  FILE    *in;
  double   m[3],lambda[2],vp[2][2];
  float    solu[3];
  int      typ[MAXTYP+1],j,k,l,big,argc,dim,dummy,nbs,typage;
  char    *ptr,*val,tmp[256],data[256],line[256];

  strcpy(tmp,name);
  ptr = strstr(tmp,".mesh");
  if ( ptr )  *ptr = '\0';
  sprintf(data,"%s.bb",tmp);
  if ( !quiet )  fprintf(stdout," Checking %s\n",data);
  big = 0;
  in = fopen(data,"r");
  if ( !in ) {
    big = 1;
    sprintf(data,"%s.BB",tmp);
    if ( !quiet )  fprintf(stdout," Checking %s\n",data);
    in = fopen(data,"r");
    if ( !in ) {
      fprintf(stdout,"  Unable to open solution. Exit.\n");
      exit(1);
    }
  }
  ptr = fgets(line,255,in);
  if ( !ptr )  exit(1);
  argc = 1;
  val = strtok(ptr," \t");
  while ( (val = strtok(NULL," \t\n")) != NULL )  argc++;
  big = argc > 4;
  rewind(in);
  fscanf(in,"%d",&dim);
  if ( dim != 2 ) {
    fprintf(stdout,"  Wrong dimension. Exit.\n");
    exit(1);
  }

  if ( big ) {
    fscanf(in,"%d",&sol->nbtyp);
    if ( sol->nbtyp > MAXTYP ) {
      fprintf(stdout,"  Max. solutions exceeded. Bye\n");
      exit(1);
    }
    for (k=1; k<=sol->nbtyp; k++) {
      fscanf(in,"%d",&dummy);
      typ[k] = (ubyte)dummy;
    }
    fscanf(in,"%d %d\n",&nbs,&typage);
    if ( sol->nbs != nbs ) {
      fprintf(stdout,"  Wrong solution number. Exit\n");
      exit(1);
    }
    if ( typage != 2 ) {
      fprintf(stdout,"  Wrong typage. Exit.\n");
      exit(1);
    }

    sol->sol = (float*)calloc(nbs+1,sol->nbtyp*sizeof(float));
    assert(sol);
    l = 1;
    for (k=1; k<=nbs; k++)
      for (j=1; j<=sol->nbtyp; j++) {
        switch (typ[j]) {
	case 1:  /* scalar */
          fscanf(in,"%f",&solu[0]);
          sol->sol[l++] = solu[0]*delta;
	  break;

	case 2:  /* L2-norm of vector */
	  fscanf(in,"%f %f",&solu[0],&solu[1]);
          sol->sol[l++] = sqrt(solu[0]*solu[0]+solu[1]*solu[1])*delta;
	  break;
	  
	case 3:  /* eigenvalues of matrix */
	  fscanf(in,"%f %f %f",&solu[0],&solu[1],&solu[2]);
	  m[0] = solu[0]*delta;
	  m[1] = solu[1]*delta;
	  m[2] = solu[2]*delta;
	  eigen2(m,lambda,vp);
	  sol->sol[l++] = min(lambda[0],lambda[1]);
	  break;
	}
      }
  }
  else {
    fscanf(in,"%d %d %d\n",&typ[1],&nbs,&typage);
    if ( sol->nbs != nbs ) {
      fprintf(stdout,"  Wrong solution number. Exit\n");
      printf("nbs %d  nbs %d\n",sol->nbs,nbs);
      exit(1);
    }
    if ( typage != 2 ) {
      fprintf(stdout,"  Wrong typage. Exit.\n");
      exit(1);
    }
    sol->sol = (float*)calloc(nbs+1,sizeof(float));
    assert(sol);

    sol->nbtyp  = 1;
    switch (typ[1]) {
    case 1:  /* scalar */
      for (k=1; k<=nbs; k++) {
        fscanf(in,"%f",&sol->sol[k]);
        sol->sol[k] *= delta;
      }
      break;
  
    case 2:  /* L2-norm of vector */
      for (k=1; k<=nbs; k++) {
        fscanf(in,"%f %f",&solu[0],&solu[1]);
        sol->sol[k] = sqrt(solu[0]*solu[0]+solu[1]*solu[1])*delta;
      }
      break;
      
    case 3:  /* eigenvalues of matrix */
      for (k=1; k<=nbs; k++) {
        fscanf(in,"%f %f %f",&solu[0],&solu[1],&solu[2]);
	m[0] = solu[0]*delta;
	m[1] = solu[1]*delta;
	m[2] = solu[2]*delta;
	eigen2(m,lambda,vp);
	sol->sol[k] = min(lambda[0],lambda[1]);
      }
      break;
    }
  }
  fclose(in);

  /* memory for metric */
  if ( iso )
    sol->met = (float*)calloc(nbs+1,sizeof(float));
  else
    sol->met = (float*)calloc(nbs+1,3*sizeof(float));
  assert(sol->met);

  return(1);
}


int saveMetric(char *name,pSol sol) {
  FILE    *out;
  int      k,l;
  char    *ptr,data[128];

  strcpy(data,name);
  ptr = strstr(data,".mesh");
  if ( ptr )  *ptr = '\0';
  ptr = strstr(data,".bb");
  if ( ptr )  *ptr = '\0';
  strcat(data,".new.bb");
  out = fopen(data,"w");
  if ( !quiet )  fprintf(stdout," Writing %s\n",data);
  if ( !out ) {
    fprintf(stdout,"  Unable to save metric. Bye\n");
    return(0);
  }

  if ( iso ) {
    fprintf(out,"2 1 %d 2\n",sol->nbs);
    for (l=0,k=1; k<=sol->nbs; k++) {
      fprintf(out,"%g ",sol->met[k]);
      if ( ++l == 8 ) {
        l = 0;
	fprintf(out,"\n");
      }
    }
    fprintf(out,"\n");
  }
  else {
    fprintf(out,"2 3 %d 2\n",sol->nbs);
    for (l=1; l<=3*sol->nbs; l+=3)
      fprintf(out,"%g %g %g\n",sol->met[l],sol->met[l+1],sol->met[l+2]);
  }
  
  fclose(out);
  return(1);
}


#ifdef __cplusplus
}
#endif
