#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define EPS   1.e-06
#define EPSD  1.e-15
#define EPSD2 1.e-10
#define EPSDD 1.e-12

#define   max(a,b)   ( ((a) < (b)) ? (b) : (a) )
#define   min(a,b)   ( ((a) < (b)) ? (a) : (b) )

/* check if numbers are equal */ 
#define egal(x,y)   ( \
  (  ((x) == 0.0f) ? (fabs(y) < EPS) : \
   ( ((y) == 0.0f) ? (fabs(x) < EPS) : \
     (fabs((x)-(y)) / (fabs(x) + fabs(y)) < EPSDD) )  ) )


/* eigen value + vector extraction */
int eigen2(double m[3],double lambda[2],double vp[2][2]) {
  double   dd,a1,xn,ddeltb,rr1,rr2,ux,uy;

  /* init */
  ux = 1.0;
  uy = 0.0;

  /* normalize */
  xn = fabs(m[0]);
  if ( fabs(m[1]) > xn )  xn = fabs(m[1]);
  if ( fabs(m[2]) > xn )  xn = fabs(m[2]);
  if ( xn < EPSD2 ) {
    lambda[0] = lambda[1] = 0.0;
    vp[0][0] = 1.0; 
    vp[0][1] = 0.0;
    vp[1][0] = 0.0;
    vp[1][1] = 1.0;
    return(1);
  }
  xn = 1.0 / xn;
  m[0] *= xn;
  m[1] *= xn;
  m[2] *= xn;

  /* check for diagonal matrix */
  if ( egal(m[1],0.0) ) {
    rr1 = m[0];
    rr2 = m[2];
    goto vect;
  }

  /* eigenvalues of jacobian */
  a1	 = -(m[0] + m[2]);
  ddeltb = a1*a1 - 4.0 * (m[0]*m[2] - m[1]*m[1]);

  if ( ddeltb < 0.0 ) {
    fprintf(stderr,"  Delta: %f\n",ddeltb);
    ddeltb = 0.0;
  }
  ddeltb = sqrt(ddeltb);

  if ( fabs(a1) < EPS ) {
    rr1 = 0.5 * sqrt(ddeltb);
    rr2 = -rr1;
  } 
  else if ( a1 < 0.0 ) {
    rr1 = 0.5 * (-a1 + ddeltb);
    rr2 = (-m[1]*m[1] + m[0]*m[2]) / rr1;
  }
  else if ( a1 > 0.0 ) {
    rr1 = 0.5 * (-a1 - ddeltb);
    rr2 = (-m[1]*m[1] + m[0]*m[2]) / rr1;
  }
  else {
    rr1 = 0.5 * ddeltb;
    rr2 = -rr1;
  }

vect:
  lambda[0] = rr1 / xn;
  lambda[1] = rr2 / xn;

  /* eigenvectors */
  a1 = m[0] - rr1;

  if ( fabs(a1)+fabs(m[1]) < EPS ) {
    if (fabs(lambda[1]) > fabs(lambda[0]) ) {
      ux = 1.0;
      uy = 0.0;    
    }
    else {
      ux = 0.0;
      uy = 1.0;
    }
  }
  else if ( fabs(a1) < fabs(m[1]) ) {
    ux = 1.0;
    uy = -a1 / m[1];
  }
  else if ( fabs(a1) > fabs(m[1]) ) {
    ux = -m[1] / a1;
    uy = 1.0;
  }
  else if ( fabs(lambda[1]) > fabs(lambda[0]) ) {
    ux = 0.0;
    uy = 1.0;
  }
  else {
    ux = 1.0;
    uy = 0.0;
  }

  dd = sqrt(ux*ux + uy*uy);
  dd = 1.0 / dd;
  if ( fabs(lambda[0]) > fabs(lambda[1]) ) {
    vp[0][0] =  ux * dd;
    vp[0][1] =  uy * dd;
  }
  else {
    vp[0][0] =  uy * dd;
    vp[0][1] = -ux * dd;
  }

  /* orthogonal vector */
  vp[1][0] = -vp[0][1];
  vp[1][1] =  vp[0][0];

  return(1);
}


#define tuMv(u,m,v) (\
  (u[0]*m[0]+u[1]*m[1])*v[0] + \
  (u[0]*m[1]+u[1]*m[2])*v[1])

/* find non orthogonal basis (v1,v2) where m1,m2 are diag */
int reducSim(float *m1,double *m2,double *mr) {
  double     n[3],v1[3],v2[3],lambda[2],vnorm,dd,det1,inv1,tracn;
  double     lambda1[2],lambda2[2],ip[4],tip[4],disc,det,vp1,vp2;

  /* eigenvect of n = m1(-1) m2 */
  det1 = m1[0]*m1[2]-m1[1]*m1[1];
  if ( det1 == 0.0 )  return(0);
  inv1 = 1.0 / det1;

  /* verif idem vp tq : M1v = vp*M2v <=> (M1-vp*M2)v = 0 */ 
  n[0] = (m1[2]*m2[0] - m1[1]*m2[1]) * inv1;
  n[1] = (m1[2]*m2[1] - m1[1]*m2[2]) * inv1;
  n[2] = (m1[0]*m2[1] - m1[1]*m2[0]) * inv1;
  n[3] = (m1[0]*m2[2] - m1[1]*m2[1]) * inv1;

  /* delta >=0 as N  diag in IR */
  dd    = n[0] - n[3];
  disc  = sqrt(fabs(dd*dd + 4.0*n[1]*n[2]));
  tracn = n[0] + n[3];

  lambda[0] = 0.5 * (tracn + disc);

  /* check for ellipse included in other */
  if ( disc / lambda[0] < EPS ) {
    /* delta<<Tr => delta~0 : eigenvalue double, P not invertible */
     if ( lambda[0] <= 1.0 )
  	memcpy(mr,m1,3*sizeof(double));
     else
  	memcpy(mr,m2,3*sizeof(double));
     return(1);
  }
  else {
    lambda[1] = 0.5 * (tracn - disc);

    v1[0] = -n[1];
    v1[1] =  n[0] - lambda[0];
    vnorm = sqrt(v1[0]*v1[0] + v1[1]*v1[1]);
    if ( vnorm < EPS ) {
       v1[0] = lambda[0] - n[3];
       v1[1] = n[2];   
       vnorm = sqrt(v1[0]*v1[0] + v1[1]*v1[1]);         
    }
    vnorm = 1.0 / vnorm;
    v1[0] *= vnorm;
    v1[1] *= vnorm;

    v2[0] = -n[1];
    v2[1] =  n[0] - lambda[1];
    vnorm = sqrt(v2[0]*v2[0] + v2[1]*v2[1]);
    if ( vnorm < EPS ) {
       v2[0] = lambda[1] - n[3];
       v2[1] = n[2];   
       vnorm = sqrt(v2[0]*v2[0] + v2[1]*v2[1]);         
    }
    vnorm = 1.0 / vnorm;
    v2[0] *= vnorm;
    v2[1] *= vnorm;

    lambda1[0] = tuMv(v1,m1,v1);
    lambda2[0] = tuMv(v1,m2,v1);
    lambda1[1] = tuMv(v2,m1,v2);
    lambda2[1] = tuMv(v2,m2,v2);

    /* intersect (m1,m2)= tP^(-1)&*L&*P^(-1) */
    det = v1[0]*v2[1] - v2[0]*v1[1];
    det = 1.0 / det;

    ip[0] =  v2[1]*det;
    ip[1] = -v2[0]*det;
    ip[2] = -v1[1]*det;
    ip[3] =  v1[0]*det;

    /* tP^(-1)&*L */
    vp1 = max(lambda1[0],lambda2[0]);
    vp2 = max(lambda1[1],lambda2[1]);

    tip[0] = ip[0] * vp1;
    tip[1] = ip[2] * vp2;
    tip[2] = ip[1] * vp1;
    tip[3] = ip[3] * vp2;

    mr[0] = tip[0]*ip[0] + tip[1]*ip[2];
    mr[1] = tip[0]*ip[1] + tip[1]*ip[3];
    mr[2] = tip[2]*ip[1] + tip[3]*ip[3]; 
  }

  return(1);
}

