 /* mhes2d.c
 * build metric tensor from mesh and solutions
 *
 * Pascal J. FREY
 * Copyright (c) LJLL, 2004. All rights reserved.
*/

#include <assert.h>
#include <math.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "libmesh.h"
#include "struct.h"
#include "sproto.h"
#include "compil.date"

#define VERSION  "1.1"
#define RELEASE  "(June 21, 2004)"

#define EPS   1.e-06
#define EPSD  1.e-15
#define EPSDD 1.e-12

static int idir[5]  = {0,1,2,0,1};

double   delta,omega;
float    eps,hmin,hmax;
int      nbmi,nb3;
ubyte    quiet,ddebug,iso,nnu;
char     meshname[256],solname[256];


/* compute gradient */
int femGradient(pMesh mesh,pSol sol,int basesol,pDeriv du) {
  
  return(1);
}


void outVois(pMesh mesh) {
  pTria    pt;
  pPoint   ppt;
  FILE    *outs,*outt;
  int      k,l,nsdep,adj,it,iadr,nv;
  ubyte    i,i1,voy;
  char    *ptr,data[256];

  strcpy(data,meshname);
  ptr = strstr(data,".mesh");
  if ( ptr )  *ptr = '\0';
  ptr = strcat(data,".som");
  outs = fopen(data,"w");

  strcpy(data,meshname);
  ptr = strstr(data,".mesh");
  if ( ptr )  *ptr = '\0';
  ptr = strcat(data,".tri");
  outt = fopen(data,"w");

  for (k=1; k<=mesh->ns; k++) {
    ppt  = &mesh->point[k];
    nsdep = ppt->s;
    pt    = &mesh->tria[nsdep];
    if ( pt->v[0] == k )       i = 0;
    else if ( pt->v[1] == k )  i = 1;
    else                       i = 2;

    fprintf(outs,"%d  ",ppt->nv);
    if ( !ppt->b )
      fprintf(outt,"%d  ",ppt->nv);
    else
      fprintf(outt,"%d  ",ppt->nv-1);

    adj = nsdep;
    voy = i;
    nv  = 0;
    do {
      pt    = &mesh->tria[adj];
      i1    = idir[voy+1];
      it    = pt->v[i1];
      fprintf(outs,"%d ",it);
      fprintf(outt,"%d ",adj);
      nv++;

      iadr = 3*(adj-1)+1;
      adj  = mesh->adja[iadr+i1];
      voy  = mesh->voy[iadr+i1];
      voy  = idir[voy+1];
    }
    while ( adj && adj != nsdep );

    if ( !adj ) {
      if ( pt->v[0] == k )       i = 0;
      else if ( pt->v[1] == k )  i = 1;
      else                       i = 2;
      voy   = idir[i+2];
      it    = pt->v[voy];
      fprintf(outs,"%d ",it);
      nv++;
    }
    for (l=nv; l<10; l++)  fprintf(outs,"0 ");
    for (l=nv; l<10+ppt->b; l++)  fprintf(outt,"0 ");
    fprintf(outs,"\n");
    fprintf(outt,"\n");
  }
  fclose(outs);
  fclose(outt);
}


/* compute gradient using least squares approach */ 
int leastSquaresGradient(pMesh mesh,pSol sol,int basesol,pDeriv du) {
  pPoint       ppt,p1;
  pTria        pt;
  double       dd,x,y,u,u1,a1,a2,a3,b1,b2;
  double       x2,xy,y2,ax,ay,ddu;
  int          i,k,adj,iadr,iasol,it,nsdep,nb;
  ubyte        i1,voy;

  if ( !quiet )  fprintf(stdout,"   Gradient computation\n");
  nb3 = 0;
  for (k=1; k<=mesh->ns; k++) {
    ppt     = &mesh->point[k];
    ppt->h  = 0;
    ppt->nv = 0;
    nsdep   = ppt->s;
    if ( !nsdep ) {
      fprintf(stdout," No simplex stored. Exit\n");
      exit(1);
    }
    iasol = (k-1)*sol->nbtyp+basesol+1;
    u = sol->sol[iasol];
    x = ppt->x;
    y = ppt->y;

    a1 = 0.0;
    a2 = 0.0;
    a3 = 0.0;
    b1 = 0.0;
    b2 = 0.0;

    pt = &mesh->tria[nsdep];
    if ( pt->v[0] == k )       i = 0;
    else if ( pt->v[1] == k )  i = 1;
    else                       i = 2;

    /* Gradient : ui = u + <grad u,PPi> */
    adj = nsdep;
    voy = i;
    nb  = 0;
    do {
      pt    = &mesh->tria[adj];
      i1    = idir[voy+1];
      it    = pt->v[i1];
      iasol = (it-1)*sol->nbtyp+basesol+1;

      u1  = sol->sol[iasol];
      p1  = &mesh->point[it];
      ax  = (double)p1->x - x;
      ay  = (double)p1->y - y;
      x2  = ax * ax;
      xy  = ax * ay;
      y2  = ay * ay;
 
      /* M = At*A symmetric definite positive */
      a1 += x2;
      a2 += xy;
      a3 += y2;
 
      /* b = At*du */
      ddu = u1 - u;
      b1 += ax * ddu;
      b2 += ay * ddu;
      nb++;
      ppt->nv++;

      iadr = 3*(adj-1)+1;
      adj  = mesh->adja[iadr+i1];
      voy  = mesh->voy[iadr+i1];
      voy  = idir[voy+1];
    }
    while ( adj && adj != nsdep );

    /* open ball */
    if ( !adj ) {
      if ( pt->v[0] == k )       i = 0;
      else if ( pt->v[1] == k )  i = 1;
      else                       i = 2;
      voy   = idir[i+2];
      it    = pt->v[voy];
      iasol = (it-1)*sol->nbtyp+basesol+1;

      u1  = sol->sol[iasol];
      p1  = &mesh->point[it];
      ax  = (double)p1->x - x;
      ay  = (double)p1->y - y;
      x2  = ax * ax;
      xy  = ax * ay;
      y2  = ay * ay;

      a1 += x2;
      a2 += xy;
      a3 += y2;

      /* b = At*du */
      ddu = u1 - u;
      b1 += ax * ddu;
      b2 += ay * ddu;
      nb++;
      ppt->nv++;
      ppt->b = 1;   /* boundary */
    }

    /* solution of mat(2,2)*grad(1,2) = b(1,2) */ 
    dd = a1*a3 - a2*a2;
    if ( fabs(dd) == 0.0 ) {
      fprintf(stdout," Invalid matrix (%E). Exit\n",dd);
      exit(1);
    }
    dd = 1.0 / dd;
    du[k].grd[0] = (a3*b1 - a2*b2) * dd;
    du[k].grd[1] = (a1*b2 - a2*b1) * dd;
  }

  return(1);
}


void saveDerivatives(pMesh mesh,pDeriv du) {
  pPoint       ppt;
  int          k;
  FILE        *outg,*outh;

  outg = fopen("gradiant.bb","w");
  assert(outg);
  outh = fopen("hessian.bb","w");
  assert(outh);

  fprintf(outg,"2 2 %d 2\n",mesh->ns);
  fprintf(outh,"2 3 %d 2\n",mesh->ns);

  for (k=1; k<=mesh->ns; k++) {
    ppt   = &mesh->point[k];
    fprintf(outg,"%f %f\n",du[k].grd[0],du[k].grd[1]);
    fprintf(outh,"%f %f %f\n",du[k].hes[0],du[k].hes[1],du[k].hes[2]);
  }
  fclose(outg);
  fclose(outh);
}


/* compute hessian using least squares approach + weight */ 
int leastSquaresHessian(pMesh mesh,pSol sol,int basesol,pDeriv du) {
  pPoint       ppt,p1;
  pTria        pt;
  double       dd,x,y,u,u1,a1,a2,a3,aa,bb,cc,ee,ff,det;
  double       ma[6],mb[3],ax,ay,hesxx,hesxy,hesyy,weight;
  int          i,j,k,adj,iadr,iasol,it,nsdep,nb,nc,nex;
  ubyte        i1,voy;

  if ( !quiet )  fprintf(stdout,"   Hessian computation\n");
  nex = 0;

  for (k=1; k<=mesh->ns; k++) {
    ppt   = &mesh->point[k];
    if ( ppt->nv < 3 ) {
      nex++;
      continue;
    }
    nsdep = ppt->s;
    iasol = (k-1)*sol->nbtyp+basesol+1;
    u = sol->sol[iasol];
    x = ppt->x;
    y = ppt->y;

    for (j=0; j<6; j++)  ma[j] = 0.0;
    mb[0] = mb[1] = mb[2] = 0.0;

    pt = &mesh->tria[nsdep];
    if ( pt->v[0] == k )       i = 0;
    else if ( pt->v[1] == k )  i = 1;
    else                       i = 2;

    /* Hessian  : Ui = U + <gradU,PPi> + 0.5*(tPPi.Hess.PPi) */
    adj = nsdep;
    voy = i;
    nb  = 0;
    weight = 1.0;
    do {
      pt    = &mesh->tria[adj];
      i1    = idir[voy+1];
      it    = pt->v[i1];
      iasol = (it-1)*sol->nbtyp+basesol+1;

      u1  = sol->sol[iasol];
      p1  = &mesh->point[it];
      ax  = (double)p1->x - x;
      ay  = (double)p1->y - y;

      a1  = 0.5 * ax*ax;
      a2  =       ax*ay;
      a3  = 0.5 * ay*ay;
      dd  = (u1-u) - (ax*du[k].grd[0] + ay*du[k].grd[1]);

      /* M = At*A symmetric definite positive */
      ma[0] += a1*a1 * weight;
      ma[1] += a1*a2 * weight;
      ma[2] += a2*a2 * weight;
      ma[3] += a1*a3 * weight;
      ma[4] += a2*a3 * weight;
      ma[5] += a3*a3 * weight;

      /* c = At*b */
      mb[0] += a1*dd * weight;
      mb[1] += a2*dd * weight;
      mb[2] += a3*dd * weight;

      nb++;
      iadr = 3*(adj-1)+1;
      adj  = mesh->adja[iadr+i1];
      voy  = mesh->voy[iadr+i1];
      voy  = idir[voy+1];
    }
    while ( adj && adj != nsdep );

    if ( !adj ) {
      if ( pt->v[0] == k )       i = 0;
      else if ( pt->v[1] == k )  i = 1;
      else                       i = 2;
      voy   = idir[i+2];
      it    = pt->v[voy];
      iasol = (it-1)*sol->nbtyp+basesol+1;
      nb++;

      u1  = sol->sol[iasol];
      p1  = &mesh->point[it];
      ax  = p1->x - x;
      ay  = p1->y - y;

      a1  = 0.5 * ax*ax;
      a2  =       ax*ay;
      a3  = 0.5 * ay*ay;
      dd  = (u1-u) - (ax*du[k].grd[0] + ay*du[k].grd[1]);

      /* M = At*A symmetric definite positive */
      ma[0] += a1*a1;
      ma[1] += a1*a2;
      ma[2] += a2*a2;
      ma[3] += a1*a3;
      ma[4] += a2*a3;
      ma[5] += a3*a3;

      /* c = At*b */
      mb[0] += a1*dd;
      mb[1] += a2*dd;
      mb[2] += a3*dd;
    }
    if ( fabs(ma[0]) == 0.0 ) {
      fprintf(stdout," Ill cond'ed matrix (d, %E). Exit\n",k,ma[0]);
      exit(1);
    }

    /* direct solving */
    aa = ma[2]*ma[5] - ma[4]*ma[4];
    bb = ma[4]*ma[3] - ma[1]*ma[5];
    cc = ma[1]*ma[4] - ma[2]*ma[3];
    dd = ma[0]*aa + ma[1]*bb + ma[3]*cc;

    /* singular matrix */
    if ( fabs(dd) == 0.0 ) {
      /*fprintf(stderr," Singular matrix (%d, %E).\n",k,dd);*/
      nex++;
      continue;
    }
    det = 1.0 / dd;

    dd = ma[0]*ma[5] - ma[3]*ma[3];
    ee = ma[1]*ma[3] - ma[0]*ma[4];
    ff = ma[0]*ma[2] - ma[1]*ma[1];

    du[k].hes[0] = (mb[0]*aa + mb[1]*bb + mb[2]*cc) * det;
    du[k].hes[1] = (mb[0]*bb + mb[1]*dd + mb[2]*ee) * det;
    du[k].hes[2] = (mb[0]*cc + mb[1]*ee + mb[2]*ff) * det;

    if ( !ppt->b && nb == 4 )  nex++;
    else  ppt->h = 1;
  }

  if ( nex == 0 )  return(1);

  /* average values of neighbors */
  if ( !quiet )  fprintf(stdout,"   Correction (%d)\n",nex);
  do {
    nc = 0;
    for (k=1; k<=mesh->ns; k++) {
      ppt = &mesh->point[k];
      if ( ppt->h )  continue;

      nsdep = ppt->s;
      pt = &mesh->tria[nsdep];
      if ( pt->v[0] == k )       voy = 0;
      else if ( pt->v[1] == k )  voy = 1;
      else                       voy = 2;

      adj = nsdep;
      nb  = 0;
      hesxx = 0.0;
      hesxy = 0.0;
      hesyy = 0.0;
      do {
        pt    = &mesh->tria[adj];
        i1    = idir[voy+1];
        it    = pt->v[i1];
	if ( mesh->point[it].h ) {
  	  hesxx += du[it].hes[0];
	  hesxy += du[it].hes[1];
	  hesyy += du[it].hes[2];
          nb++;
	}
	iadr = 3*(adj-1)+1;
        adj  = mesh->adja[iadr+i1];
        voy  = mesh->voy[iadr+i1];
        voy  = idir[voy+1];
      }
      while ( adj && adj != nsdep );
      if ( !adj ) {
        if ( pt->v[0] == k )       i = 0;
        else if ( pt->v[1] == k )  i = 1;
        else                       i = 2;
        voy   = idir[i+2];
        it    = pt->v[voy];
	if ( mesh->point[it].h ) {
	  hesxx += du[it].hes[0];
	  hesxy += du[it].hes[1];
	  hesyy += du[it].hes[2];
          nb++;
        }
      }
      if ( nb > 0 ) {
        dd = 1.0 / nb;
        du[k].hes[0] = hesxx * dd;
        du[k].hes[1] = hesxy * dd;
        du[k].hes[2] = hesyy * dd;
	ppt->h = 1;
        nc++;
	nex--;
      }
    }
  }
  while ( nc > 0 && nex > 0 );

  if ( nex == 0 )  return(1);

  /* value of first neighbor */
  if ( !quiet )  fprintf(stdout,"   Correction (%d)\n",nex);
  for (k=1; k<=mesh->ns; k++) {
    ppt = &mesh->point[k];
    if ( ppt->h )  continue;

    nsdep = ppt->s;
    if ( !nsdep ) {
      fprintf(stdout," No simplex stored. Exit\n");
      exit(1);
    }
    pt = &mesh->tria[nsdep];
    if ( pt->v[0] == k )       voy = 0;
    else if ( pt->v[1] == k )  voy = 1;
    else		       voy = 2;

    i1    = idir[voy+1];
    it    = pt->v[i1];
    du[k].hes[0] = du[it].hes[0];
    du[k].hes[1] = du[it].hes[1];
    du[k].hes[2] = du[it].hes[2];

    ppt->h = 1;
    nex--;
  }

  return(1);
}


/* compute local relative error */
int normHessian(pMesh mesh,pSol sol,int basesol,pDeriv du) {
  pPoint   ppt;
  double   err,err1,errs,u,norm;
  double   maxu,maxg;
  int      k,iasol;

  if ( !quiet )  fprintf(stdout,"   Hessian normalization\n");

  /* no normalisation */
  if ( nnu == 0 ) {
    err = 1.0 / eps;
  printf("err = %f\n",err);
    for (k=1; k<=mesh->ns; k++) {
      du[k].hes[0] *= err;
      du[k].hes[1] *= err;
      du[k].hes[2] *= err;    
    }
  }

  /* relative value: M(u)= |H(u)| / err*||u||_inf */
  else if ( nnu == 1 ) {
    maxu = 0.0;
    for (k=1; k<=mesh->ns; k++) {
      iasol = (k-1)*sol->nbtyp+basesol+1;
      u     = sol->sol[iasol];
      maxu  = max(fabs(u),maxu);
    }
    maxu =  1.0 / (eps * maxu);
    for (k=1; k<=mesh->ns; k++) {
      du[k].hes[0] *= maxu; 
      du[k].hes[1] *= maxu;
      du[k].hes[2] *= maxu;
    }
  }

  /* local norm: M(u)= |H(u)| / err*|u| */
  else if ( nnu == 2 ) {
    maxu = 0.0;
    for (k=1; k<=mesh->ns; k++) {
      iasol = (k-1)*sol->nbtyp+basesol+1;
      u     = sol->sol[iasol];
      maxu  = max(fabs(u),maxu);
    }
    if ( maxu == 0.0 )
      errs = 0.01;
    else
      errs = maxu * 0.01;
    for (k=1; k<=mesh->ns; k++) {
      iasol = (k-1)*sol->nbtyp+basesol+1;
      u     = sol->sol[iasol];
      err1  = max(errs,fabs(u));
      maxu  = 1.0 / err1; 
      du[k].hes[0] *= maxu; 
      du[k].hes[1] *= maxu;
      du[k].hes[2] *= maxu;
    }
  }

  /* local norm: M(u)= |H(u)| / err*(e1*|u|+e2*h*|du|) */
  else if ( nnu == 3 ) {
    maxu = 0.0;
    maxg = 0.0;
    for (k=1; k<=mesh->ns; k++) {
      iasol = (k-1)*sol->nbtyp+basesol+1;
      u     = sol->sol[iasol];
      maxu  = max(fabs(u),maxu);
      norm  = du[k].grd[0]*du[k].grd[0] + du[k].grd[1]*du[k].grd[1];
      maxg  = max(sqrt(norm),maxg);
    }
    if ( maxu == 0.0 )
      errs = 0.01;
    else
      errs = maxu * 0.01;
    err1 = maxg * 0.01;

    for (k=1; k<=mesh->ns; k++) {
      ppt   = &mesh->point[k];
      iasol = (k-1)*sol->nbtyp+basesol+1;
      u     = sol->sol[iasol];
      norm  = du[k].grd[0]*du[k].grd[0] + du[k].grd[1]*du[k].grd[1];
      norm  = max(err1,sqrt(norm));
      norm *= ppt->r / ppt->nv;
      err1  = max(errs,fabs(u));
      
      /* variable weight */
      err1  = 0.01*err1 + 0.01*norm;

      maxu  = 1.0 / (eps*err1);
      du[k].hes[0] *= maxu;
      du[k].hes[1] *= maxu;
      du[k].hes[2] *= maxu;
    } 
  }
  
  return(1);
}


int wallLaplace(pMesh mesh,pSol sol,int basesol,pDeriv du) {
  pTria      pt;
  pPoint     ppt;
  double     dd;
  float     *hess;
  int        k,l,it,maxtou,iv0,iv1,iv2;
  ubyte      i;

  if ( !quiet )  fprintf(stdout,"   Laplacian correction\n");

  maxtou = 3;
  hess = (float*)malloc((3*mesh->nt+1)*sizeof(float));
  assert(hess);
  
  for (it=0; it<maxtou; it++) {

    /* average hessian */
    l = 1;
    for (k=1; k<=mesh->nt; k++) {
      pt  = &mesh->tria[k];

      iv0 = pt->v[0];
      iv1 = pt->v[1];
      iv2 = pt->v[2];
      hess[l++] = (du[iv0].hes[0] + du[iv1].hes[0] + du[iv2].hes[0]) / 3.0;
      hess[l++] = (du[iv0].hes[1] + du[iv1].hes[1] + du[iv2].hes[1]) / 3.0;
      hess[l++] = (du[iv0].hes[2] + du[iv1].hes[2] + du[iv2].hes[2]) / 3.0;

      for (i=0; i<3; i++) {
        iv0 = pt->v[i];        
        ppt = &mesh->point[iv0];
        if ( ppt->b ) {
	  du[iv0].hes[0] = 0.0;
	  du[iv0].hes[1] = 0.0;
	  du[iv0].hes[2] = 0.0;
        }
      }
    }

    for (k=1; k<=mesh->nt; k++) {
      pt = &mesh->tria[k];

      for (i=0; i<3; i++) {
        iv0 = pt->v[i];
        ppt = &mesh->point[iv0];
	if ( !ppt->b )  continue;
	dd  = pt->s / ppt->a;
	dd *= (ppt->nv-1) / 3.0;
	du[iv0].hes[0] += dd * hess[3*(k-1)+1];
	du[iv0].hes[1] += dd * hess[3*(k-1)+2];
	du[iv0].hes[2] += dd * hess[3*(k-1)+3];
      }
    }
  }
  
  free(hess);
  return(1);
}


int defineMetric(pMesh mesh,pSol sol,int basesol,pDeriv du) {
  double    lambda[2],vp[2][2],mat[3],mr[3],dd,hlmin,hlmax;
  float    *tmet;
  int       k,iasol;

  if ( !quiet )  fprintf(stdout,"   Metric definition\n");
  hlmin = 1.0 / (hmax*hmax);
  hlmax = 1.0 / (hmin*hmin);

  for (k=1; k<=mesh->ns; k++) {
    if ( !eigen2(du[k].hes,lambda,vp) ) {
      fprintf(stdout,"  Unable to define metric (%d). Bye\n",k);
      return(0);
    }

    /* truncation */
    dd        = max(fabs(lambda[0]),hlmin);
    lambda[0] = min(dd,hlmax);
    dd        = max(fabs(lambda[1]),hlmin);
    lambda[1] = min(dd,hlmax);

    if ( iso ) {
      tmet = &sol->met[k];
      if ( lambda[0] < lambda[1] ) 
        lambda[0] = lambda[1];
      else
        lambda[1] = lambda[0];
      if ( basesol == 0 )
        tmet[0] = 1.0 / sqrt(lambda[0]);
      else
        tmet[0] = min(1.0/sqrt(lambda[0]),tmet[0]);
    }

    else {
      iasol = (k-1)*sol->nbtyp+basesol+1;
      iasol = (k-1)*3 + 1;
      tmet  = &sol->met[iasol];

      /* compute Mat = B M Bt */
      mat[0] = lambda[0]*vp[0][0]*vp[0][0] + lambda[1]*vp[1][0]*vp[1][0];
      mat[1] = lambda[0]*vp[0][0]*vp[0][1] + lambda[1]*vp[1][0]*vp[1][1];
      mat[2] = lambda[0]*vp[0][1]*vp[0][1] + lambda[1]*vp[1][1]*vp[1][1];

     if ( basesol == 0 ) {
        tmet[0] = mat[0];
	tmet[1] = mat[1];
	tmet[2] = mat[2];
      }
      else {  /* intersection */
        reducSim(tmet,mat,mr);
	tmet[0] = mr[0];
	tmet[1] = mr[1];
	tmet[2] = mr[2];
      }
    }
  }

  return(1);
}
 

int doMetric(pMesh mesh,pSol sol,int method) {
  pDeriv    du;
  int       basesol;

  if ( !quiet )
    fprintf(stdout," Metric computation: %d field(s)\n",sol->nbtyp);
  du = (pDeriv)malloc((mesh->ns+1)*sizeof(Deriv));
  if ( !du ) {
    fprintf(stderr,"  Unable to allocate memory. Bye\n");
    return(0);
  }

  ddebug = 1;
  for (basesol=0; basesol<sol->nbtyp; basesol++) {
    if ( !quiet )  fprintf(stdout,"  Metric %d\n",basesol+1); 
    if ( !leastSquaresGradient(mesh,sol,basesol,du) ) return(0);
    if ( !leastSquaresHessian(mesh,sol,basesol,du) )  return(0);
    
    if ( ddebug )  saveDerivatives(mesh,du);

    if ( !normHessian(mesh,sol,basesol,du) )   return(0);
    if ( !wallLaplace(mesh,sol,basesol,du) )   return(0);
    if ( !defineMetric(mesh,sol,basesol,du) )  return(0);
  }

  return(1);
}


void stats(pMesh mesh) {
  fprintf(stdout,"\n Statistics:\n");
  fprintf(stdout,"  %8d vertices   %8d triangles\n",mesh->ns,mesh->nt);
  if ( nbmi )
    fprintf(stdout,"  %8d metric intersections\n",nbmi);
}


static void excfun(int sigid) {
  fprintf(stdout,"\n Unexpected error:");  fflush(stdout);
  switch(sigid) {
    case SIGABRT:
      fprintf(stdout,"  Abnormal stop\n");  exit(1);
    case SIGFPE:
      fprintf(stdout,"  Floating-point exception\n"); exit(1);
    case SIGILL:
      fprintf(stdout,"  Illegal instruction\n"); exit(1);
    case SIGSEGV:
      fprintf(stdout,"  Segmentation fault\n");  exit(1);
    case SIGTERM:
    case SIGINT:
      fprintf(stdout,"  Program killed\n");  exit(1);
  }
  exit(1);
}


int main(int argc,char **argv) {
  pMesh       mesh;
  pSol        sol;
  int         method;
  clock_t     ct0,ct1,ct2;

  /* trap exceptions */
  signal(SIGABRT,excfun);
  signal(SIGFPE,excfun);
  signal(SIGILL,excfun);
  signal(SIGSEGV,excfun);
  signal(SIGTERM,excfun);
  signal(SIGINT,excfun);

  ct0    = clock();
  quiet  = 1;
  ddebug = 0;
  iso    = 0;
  nbmi   = 0;
  method = 1;
  meshname[0] = '\0';
  solname[0]  = '\0';
  hmin = hmax = -1.0;
  eps  = 0.01;
  hmin = 0.01;
  hmax = 25.0;
  nnu  = 0;

  fprintf(stdout,"  -- MHES2d, Release %s (%s)\n",VERSION,RELEASE);
  fprintf(stdout,"     Copyright (C) LJLL, 2004\n");
  fprintf(stdout,"     compiled: %s\n",COMPIL);
  parsar(argc,argv);

  /* load files */
  fprintf(stdout,"\n Loading data\n");
  ct1  = clock();
  mesh = (pMesh)calloc(1,sizeof(Mesh));
  sol  = (pSol)calloc(1,sizeof(Sol));
  if ( !mesh || !sol ) {
    fprintf(stdout,"  Not enough memory. Exit.\n");
    exit(1);
  }
  loadMesh(meshname,mesh);
  sol->nbs = mesh->ns;
  loadSolu(solname,sol);
  parsop(solname);
  hashTria(mesh);
  ct1  = difftime(clock(),ct1);
  fprintf(stdout,"  Input seconds:         %7.2f\n",ct1/(double)CLOCKS_PER_SEC);

  /* metric */
  fprintf(stdout,"\n Computing metric\n");
  ct2 = clock();
  doMetric(mesh,sol,method);
  /*outVois(mesh);*/
  ct2 = difftime(clock(),ct2);
  fprintf(stdout,"  Metric seconds:        %7.2f\n",ct2/(double)CLOCKS_PER_SEC);

  /* save file */
  fprintf(stdout,"\n Saving result\n");
  ct2 = clock();
  saveMetric(solname,sol);
  ct2 = difftime(clock(),ct2);
  fprintf(stdout,"  Output seconds:        %7.2f\n",ct2/(double)CLOCKS_PER_SEC);

  if ( !quiet )  stats(mesh);
  
  ct0 = difftime(clock(),ct0);
  fprintf(stdout,"\n Total running seconds:%8.2f\n",ct0/(double)CLOCKS_PER_SEC);

  return(0);
}

