#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "hde.h"

using  namespace detri2;

int main(int argc, char* argv[])
{
  HDE_Triangulation *Tr = new HDE_Triangulation();

  Tr->parse_commands(argc, argv);
  if (!Tr->HDE_parse_commands(argc, argv)) {
    // No input or wrong parameters.
    printf("Usage: detri2HDE [options]\n");
    printf("List of basic options:\n");
    printf("  input_mesh=filename\n");
    printf("  s1=# s2=# s3=#, set the s-factors.\n");
    printf("  length=#, set target length.\n");
    printf("  angle=#, set target angle (degree).\n");
    printf("  max_iter=#, set maximum iterations.\n");
    printf("  smooth_iter=#, set maximum smoothing iterations.\n");
    printf("  use_intpoints=#, use # interior points for calculating distance.\n");
    delete Tr;
    return 0;
  }

  if (!Tr->read_mesh()) {
    printf("Failed to read input mesh.\n");
    delete Tr;
    return 0;
  }
  if (!Tr->io_with_sol || !Tr->io_with_grd) {
    printf("Missing solution and gradients (.grd)\n");
    delete Tr;
    return 0;
  }
  if (!Tr->io_with_metric) {
    printf("Missing mesh sizing function (.mtr)\n");
    delete Tr;
    return 0;
  }

  if (!Tr->reconstruct_mesh(0)) {
    printf("Failed to reconstruct the input mesh.\n");
    delete Tr;
    return 0;
  }

   //if (!Tr->HDE_reconstruct_OMT_domain()) {
  if (!Tr->HDE_prepare()) {
    printf("Failed to prepare HDE.\n");
    delete Tr;
    return 0;
  }

  /*
  // Read solution and gradient from file *.txt
  char filename[256];
  strcpy(filename, Tr->io_infilename);
  strcat(filename, ".grd");
  FILE *fdata = fopen(filename, "r");
  if (fdata == NULL) {
    printf("Failed to read vertex data.\n");
    delete Tr;
    return 0;
  }
  printf("Reading solution and gradient from %s.\n", filename);

  REAL *sol, *grdx, *grdy;
  sol  = new double[Tr->ct_in_vrts];
  grdx = new double[Tr->ct_in_vrts];
  grdy = new double[Tr->ct_in_vrts];
  for (int i = 0; i < Tr->ct_in_vrts; i++) {
    fscanf(fdata, "%lf %lf %lf\n", &(sol[i]), &(grdx[i]), &(grdy[i]));
  }
  fclose(fdata);

  if (!Tr->HDE_reconstruct_OMT_domain(sol, grdx, grdy)) {
    printf("Failed to reconstruct the background mesh.\n");
    delete Tr;
    return 0;
  }

  delete [] sol;
  delete [] grdx;
  delete [] grdy;
  */

  // Only used by wave equation to interpolating solution uh0, uh1.
  char filename[256];
  strcpy(filename, "../files/uh0.txt");
  FILE *fdata = fopen(filename, "r");
  if (fdata != NULL) {
    printf("Reading uh0 from %s\n", filename);
    for (int i = 0; i < Tr->ct_in_vrts; i++) {
      VertexData *data = Tr->OMT_domain->in_vrts[i].Data;
      fscanf(fdata, "%lf\n", &(data->uh0));
    }
    fclose(fdata);
  } else {
    printf("Warning: not found file uh0.txt\n");
  }

  strcpy(filename, "../files/uh1.txt");
  fdata = fopen(filename, "r");
  if (fdata != NULL) {
    printf("Reading uh1 from %s\n", filename);
    for (int i = 0; i < Tr->ct_in_vrts; i++) {
      VertexData *data = Tr->OMT_domain->in_vrts[i].Data;
      fscanf(fdata, "%lf\n", &(data->uh1));
    }
    fclose(fdata);
  } else {
    printf("Warning: not found file uh1.txt\n");
  }

  // Prepare the HDE vertex data from the solution and gradients.
  //if (!Tr->HDE_prepare()) {
  //  printf("Failed in HDE_prepare().\n");
  //  delete Tr;
  //  return 0;
  //}

  //std::cout << "--------------------------------\n";
  //std::cout << "Statistics before the improvment\n";
  //std::cout << "--------------------------------\n";
  //Tr->mesh_statistics();

  //Tr->HDE_mesh_adapt();

  //Tr->mesh_statistics();
  //Tr->HDE_quality_statistic();

  //int meshidx = 0;

  if (Tr->op_target_length == 0.) {
    Tr->op_target_length = Tr->geo_average_length; // HDE_get_average_length();
  }

  printf("Mesh adaptation via HDE.\n");
  printf("  - Target HDE edge length = %g.\n", Tr->op_target_length);
  //printf("  - Target HDE min angle = %g (degree).\n", op_target_angle/PI*180.);
  printf("  - s1 = %g, s2 = %g, s3 = %g\n", Tr->op_s1, Tr->op_s2, Tr->op_s3);
  //printf("  - max size iteration = %d.\n", Tr->op_max_size_iter);
  //printf("  - max quality iteration = %d.\n", Tr->op_max_quality_iter);
  //printf("  - use %d integral points.\n", Tr->op_use_intpoints);
  printf("  - number of smooth iterations: %d\n", Tr->op_smooth_iter);

  arraypool *fqueue = new arraypool(sizeof(TriEdge), 10);
  int count, meshidx = 0;

  Tr->op_round_flip = 1; // to avoid degenerate triangles produced by flips.
  Tr->op_check_min_angle = 1; // to avoid bad quality triangles produced by flips.

  // Save the initial mesh with solution.
  Tr->writeEmbeddedMeshLifted(meshidx); meshidx++;

  count = Tr->HDE_lawson_flip(NULL, fqueue);
  printf("Fliped %d edges\n", count);
  Tr->writeEmbeddedMeshLifted(meshidx); meshidx++;
  
  count = Tr->HDE_coarsen_mesh();
  printf("Removed %d vertices.\n", count);
  Tr->writeEmbeddedMeshLifted(meshidx); meshidx++;

  //if (count > 0) {
  //  Tr->HDE_lawson_flip(NULL, fqueue);
  //}

  count = Tr->HDE_split_edges();
  printf("Split %d edges.\n", count);
  Tr->writeEmbeddedMeshLifted(meshidx); meshidx++;

  //if (count > 0) {
  //count = Tr->HDE_lawson_flip(NULL, fqueue);
  //printf("Fliped %d edges\n", count);
  //Tr->writeEmbeddedMeshLifted(meshidx); meshidx++;
  //}

  delete fqueue;

  // Smooth vertices.
  for (int i = 0; i < Tr->op_smooth_iter; i++) {
    count = Tr->HDE_smooth_vertices();
    printf("i=%d, smoothed %d vertices\n", i+1, count);
    Tr->writeEmbeddedMeshLifted(meshidx); meshidx++;
  }

  //Tr->HDE_repair_triangles();
  //Tr->writeEmbeddedMeshLifted(meshidx); meshidx++;

  Tr->mesh_statistics();
  Tr->HDE_quality_statistic();

  //std::cout << "-------------------------------\n";
  //std::cout << "Statistics after the improvment\n";
  //std::cout << "-------------------------------\n";
  //Tr->mesh_statistics();

  //Tr->writeEmbeddedEdgeLength(meshidx);
  //Tr->save_triangulation();
  Tr->save_inria_mesh(); // for FreeFem++

  // Save uh0, uh1
  FILE *fuh0 = fopen("../files/uh0.1.txt", "w");
  if (fuh0 != NULL) {
    FILE *fuh1 = fopen("../files/uh1.1.txt", "w");
    for (int i = 0; i < Tr->ct_in_vrts; i++) {
      Vertex *v = &(Tr->in_vrts[i]);
      if (v->typ == UNUSEDVERTEX) continue;
      fprintf(fuh0, "%.17g\n", v->Data->uh0);
      fprintf(fuh1, "%.17g\n", v->Data->uh1);
    }
    if (Tr->tr_steiners) {
      for (int i = 0; i < Tr->tr_steiners->used_items; i++) {
        Vertex *v = (Vertex *) Tr->tr_steiners->get(i);
        if (v->is_deleted()) continue;
        fprintf(fuh0, "%.17g\n", v->Data->uh0);
        fprintf(fuh1, "%.17g\n", v->Data->uh1);
      }
    }
    fclose(fuh0);
    fclose(fuh1);
  }

  delete Tr;
  return 1;
}
