// ********** DO NOT REMOVE THIS BANNER **********
//
// SUMMARY:  Bamg: Bidimensional Anisotrope Mesh Generator
// RELEASE: 0 
// USAGE  : You may copy freely these files and use it for    
//          teaching or research. These or part of these may   
//          not be sold or used for a commercial purpose with- 
//          out our consent : fax (33) 1 39 63 55 14       
//
// AUTHOR:   F. Hecht,    
// ORG    :  INRIA
// E-MAIL :   Frederic.Hecht@Inria.fr   
//
// ORIG-DATE:     Dec 97

#include <stdio.h>
#include <math.h>

typedef float  Real4;
typedef double Real8;
typedef short  Int1;
typedef short  Int2;
typedef long   Int4;

typedef int  Icoor1;  
typedef long   Icoor2;
const Icoor1 MaxICoor = 1073741823; // 2^30-1
const  Icoor2 MaxICoor22 = Icoor2(2)*Icoor2(MaxICoor) * Icoor2(MaxICoor) ;

#include "R2.h"

typedef P2<Icoor1,Icoor2> I2;    // a 2d vector integers
typedef P2<Real8,Real8> R2;      // a 2d vector doubles
typedef P2xP2<Int2,Int4> I2xI2;
typedef P2<Real4,Real8> R2xR2;

typedef P2<double,double> D2;
typedef P2xP2<double,double> D2xD2;

#include <limits.h>
template<class T> inline T Square (const T &a) { return a*a;} 
template<class T> inline T Min (const T &a,const T &b){return a < b ? a : b;}
template<class T> inline T Max (const T &a,const T & b){return a > b ? a : b;}
template<class T> inline T Abs (const T &a){return a <0 ? -a : a;}
template<class T> inline double Norme (const T &a){return sqrt(a*a);}
template<class T> inline void Exchange (T& a,T& b) {T c=a;a=b;b=c;}
// for pb on microsoth compiler 
template<class T> inline T Max3 (const T &a,const T & b,const T & c)
  {return Max(Max(a,b),c);}
template<class T> inline T Min3 (const T &a,const T & b,const T & c)
  {return Min(Min(a,b),c);}


class MatVVP2x2;

class MetricAnIso
{ 
public:
  friend class MatVVP2x2;

  Real8 a11,a21,a22; // double

  MetricAnIso(Real8 a): a11(1/(a*a)),a21(0),a22(1/(a*a)) {}
  MetricAnIso(Real8 a,Real8 b,Real8 c) :a11(a),a21(b),a22(c){}
  MetricAnIso()  {}; // 
  MetricAnIso(const Real8  a[3],const  MetricAnIso m0,
	      const  MetricAnIso m1,const  MetricAnIso m2 );

  R2 mul(const R2 x)const { 
    // matrix [2x2] mul vector [x.x, x.y]^T
    return R2(a11 * x.x + a21 * x.y, a21 * x.x + a22 * x.y);
  }
  R2 Orthogonal(const R2 x){
    return R2(-(a21 * x.x + a22 * x.y), a11 * x.x + a21 * x.y);
  }
  R2 Orthogonal(const I2 x){
    return R2(-(a21 * x.x + a22 * x.y), a11 * x.x + a21 * x.y);
  }
  
//  D2 Orthogonal(const D2 x){return D2(-(a21*x.x+a22*x.y),a11*x.x+a21*x.y);}

  MetricAnIso(Real8  a, const  MetricAnIso ma,
	          Real8  b, const  MetricAnIso mb);

  int IntersectWith(const MetricAnIso M);

  MetricAnIso operator*(Real8 c) const {
    Real8 c2 = c * c;
    return  MetricAnIso(a11*c2, a21*c2, a22*c2);
  }

  MetricAnIso operator/(Real8 c) const {
    Real8 c2= 1. / (c * c);
    return  MetricAnIso(a11*c2, a21*c2, a22*c2);
  } 

  Real8 norm(R2 x) { // H. Si, norm (x^T M x)
    return x.x*x.x*a11+2*x.x*x.y*a21+x.y*x.y*a22;
  }

  Real8 operator()(R2 x) const { // distance sqrt(x^T M x)
    return sqrt(x.x*x.x*a11+2*x.x*x.y*a21+x.y*x.y*a22);
  };

//  Real8 operator()(D2 x) const { return sqrt(x.x*x.x*a11+2*x.x*x.y*a21+x.y*x.y*a22);};

  Real8 operator()(R2 x, R2 y) const { 
    return x.x * y.x * a11 + (x.x * x.y + x.y * y.x) * a21 + x.y * y.y * a22;
  };
  inline void  Box(Real4 &hx,Real4 &hy) const ;

  // Debug
  void print() {
    printf("  Mtr a11=%g, a21=%g, a22=%g\n", a11, a21, a22);
  }
 //friend ostream& operator <<(ostream& f, const  MetricAnIso & M)
 // {f << " mtr a11=" << M.a11 << " a21=a12=" << M.a21 << " a22=" << M.a22 << ";" ;   return f;}
#ifdef DRAWING
  void Draw(R2 ) const;
#endif
  MetricAnIso(const MatVVP2x2);
};


class MatVVP2x2 
{
  friend  class MetricAnIso;
public:
  double lambda1,lambda2;
  D2 v; // A vector (double)

  MatVVP2x2(double r1,double r2,const D2 vp1): lambda1(r1),lambda2(r2), v(vp1) {}
  
  void  Abs(){
    lambda1=::Abs(lambda1),
    lambda2=::Abs(lambda2);
  }
  void  pow(double p){
    lambda1=::pow(lambda1,p);
    lambda2=::pow(lambda2,p);
  }
  void  Min(double a) { 
    lambda1=::Min(a,lambda1); 
    lambda2=::Min(a,lambda2) ;
  }
  void  Max(double a) { 
    lambda1=::Max(a,lambda1); 
    lambda2=::Max(a,lambda2) ;
  }

  void Minh(double h) {
    Max(1.0/(h*h));
  }
  void Maxh(double h) {
    Min(1.0/(h*h));
  }
  void print() {
    printf("{%g, %g, %g, %g}\n", 1./sqrt(lambda1), 1./sqrt(lambda2), v.x, v.y);
  }
  //friend ostream& operator <<(ostream& f, const MatVVP2x2 & c)
  //{ f << '{' << 1/sqrt(c.lambda1)<< ',' << 1/sqrt(c.lambda2) << ','<< c.v << '}' <<flush ; return f; }
  //friend istream& operator >>(istream& f,  MatVVP2x2 & c)
  //{ f >> c.lambda1 >>c.lambda2 >> c.v.x >> c.v.y ;c.v /= Norme2(c.v); return f; }

  MatVVP2x2(const MetricAnIso );
  Real8 hmin() const {
    return sqrt(1./::Max3(lambda1,lambda2,1e-30));
  }
  Real8 hmax() const {
    return sqrt(1/::Max(::Min(lambda1,lambda2),1e-30));
  }
  Real8 lmax() const {
    return::Max3(lambda1,lambda2,1e-30);
  }
  Real8 lmin() const {
    return ::Max(::Min(lambda1,lambda2),1e-30);
  }
  Real8 Aniso2() const  { 
    return lmax()/lmin();
  }

  Real8 Aniso() const  { 
    return sqrt( Aniso2());
  }
  inline void BoundAniso2(const Real8 coef);
  void BoundAniso(const Real8 c){ 
    BoundAniso2(1/(c*c));
  }
  void operator *=(double coef){ 
    lambda1*=coef;lambda2*=coef;
  }
};

inline MetricAnIso::MetricAnIso(const MatVVP2x2 M)  
{
 //     recompose M in: M = V^t lambda V 
  //     V = ( v,v^\perp)
  //  where v^\perp = (-v_1,v_0)
  double v00=M.v.x*M.v.x;
  double v11=M.v.y*M.v.y;
  double v01=M.v.x*M.v.y;
  a11=v00*M.lambda1+v11*M.lambda2;
  a21=v01*(M.lambda1-M.lambda2);
  a22=v00*M.lambda2+v11*M.lambda1;
}

inline void  MatVVP2x2::BoundAniso2(const Real8 coef) 
{
  if (coef<=1.00000000001) 
    if (lambda1 < lambda2)
      lambda1 = ::Max(lambda1,lambda2*coef);
    else
      lambda2 = ::Max(lambda2,lambda1*coef);
  else  // a verifier 
    if (lambda1 > lambda2)
      lambda1 = ::Min(lambda1,lambda2*coef);
    else
      lambda2 = ::Min(lambda2,lambda1*coef);
}