#include "test_functions.h"

//==============================================================================
// Gaussian
// see an example from BossenHeckbert96-anisoMesh.pdf, page 8.

double fun_gaussian(const detri2::Vertex * vert)
{
  double x = vert->crd[0];
  double y = vert->crd[1];

  return 4.0 * exp(-(x * x + y * y) / 2.0);
}

double funX_gaussian(const detri2::Vertex * vert)
{
  double x = vert->crd[0];
  double y = vert->crd[1];

  return -x * 4.* exp(-(x * x + y * y) / 2.0);
}

double funY_gaussian(const detri2::Vertex * vert)
{
  double x = vert->crd[0];
  double y = vert->crd[1];

  return -y * 4.* exp(-(x * x + y * y) / 2.0);
}

// Hessian
double funXX_gaussian(const detri2::Vertex * vert)
{
  double x = vert->crd[0];
  double y = vert->crd[1];

  double f = exp(-(x*x + y*y) / 2.0);

  return (x*x - 1.) * f;
}

double funXY_gaussian(const detri2::Vertex * vert)
{
  double x = vert->crd[0];
  double y = vert->crd[1];

  double f = exp(-(x*x + y*y) / 2.0);

  return (x * y) * f;
}

double funYY_gaussian(const detri2::Vertex * vert)
{
  double x = vert->crd[0];
  double y = vert->crd[1];

  double f = exp(-(x*x + y*y) / 2.0);

  return (y*y - 1.) * f;
}

//==============================================================================
// lennard
// f2(x,y)=tanh(60x)-tanh(60(x-y)-30)

double fun_lennard(const detri2::Vertex * vert)
{
    double x = vert->crd[0];
    double y = vert->crd[1];

    return(tanh(60.*x)-tanh(60.*x-60.*y-30.));
}

double funX_lennard(const detri2::Vertex * vert)
{
    double x = vert->crd[0];
    double y = vert->crd[1];

    return(60.*tanh(60.*x-60.*y-30.)*tanh(60.*x-60.*y-30.)-60.*tanh(60.*x)*tanh(60.*x));
}

double funY_lennard(const detri2::Vertex * vert)
{
    double x = vert->crd[0];
    double y = vert->crd[1];

    return(-60.+60.*tanh(60.*x-60.*y-30.)*tanh(60.*x-60.*y-30.));
}

double funXX_lennard(const detri2::Vertex * vert)
{
  double x = vert->crd[0];
  double y = vert->crd[1];

  double ins1 = exp(120 * x);
  double ins2 = exp(120 * (x - y) - 60);
  double value1 = 2 * (14400 * ins1*(ins1 + 1)*(ins1 + 1) - 28800 * ins1*ins1*(ins1 + 1)) / pow(ins1 + 1, 4) -
                2 * (14400 * ins2*(ins2 + 1)*(ins2 + 1) - 28800 * ins2*ins2*(ins2 + 1)) / pow(ins2 + 1, 4);
//  double value2 = 2 * (14400 * ins2*(ins2 + 1)*(ins2 + 1) - 28800 * ins2*ins2*(ins2 + 1)) / pow(ins2 + 1, 4);
//  double value3 = -value2;

  return value1;
}

double funXY_lennard(const detri2::Vertex * vert)
{
    double x = vert->crd[0];
    double y = vert->crd[1];

  double ins1 = exp(120 * x);
  double ins2 = exp(120 * (x - y) - 60);
  //double value1 = 2 * (14400 * ins1*(ins1 + 1)*(ins1 + 1) - 28800 * ins1*ins1*(ins1 + 1)) / pow(ins1 + 1, 4) -
//				2 * (14400 * ins2*(ins2 + 1)*(ins2 + 1) - 28800 * ins2*ins2*(ins2 + 1)) / pow(ins2 + 1, 4);
  double value2 = 2 * (14400 * ins2*(ins2 + 1)*(ins2 + 1) - 28800 * ins2*ins2*(ins2 + 1)) / pow(ins2 + 1, 4);
  //double value3 = -value2;

  return value2;
}

double funYY_lennard(const detri2::Vertex * vert)
{
    double x = vert->crd[0];
    double y = vert->crd[1];

  double ins1 = exp(120 * x);
  double ins2 = exp(120 * (x - y) - 60);
  //double value1 = 2 * (14400 * ins1*(ins1 + 1)*(ins1 + 1) - 28800 * ins1*ins1*(ins1 + 1)) / pow(ins1 + 1, 4) -
//				2 * (14400 * ins2*(ins2 + 1)*(ins2 + 1) - 28800 * ins2*ins2*(ins2 + 1)) / pow(ins2 + 1, 4);
  double value2 = 2 * (14400 * ins2*(ins2 + 1)*(ins2 + 1) - 28800 * ins2*ins2*(ins2 + 1)) / pow(ins2 + 1, 4);
  double value3 = -value2;

  return value3;
}

//==============================================================================
// f3(x,y)=sin(5 * pow(x - 0.2, 3)*(y*y - y + 1))

double fun_alauzet(const detri2::Vertex * vert)
{
    double x = vert->crd[0];
    double y = vert->crd[1];

    return(sin(5.*(x-0.2)*(x-0.2)*(x-0.2)*(y*y-y+1.)));
}

double funX_alauzet(const detri2::Vertex * vert)
{
    double x = vert->crd[0];
    double y = vert->crd[1];

    return(cos(5.*(x-0.2)*(x-0.2)*(x-0.2)*(y*y-y+1.))*3.*(x-0.2)*(x-0.2));
}

double funY_alauzet(const detri2::Vertex * vert)
{
    double x = vert->crd[0];
    double y = vert->crd[1];

    return(cos(5.*(x-0.2)*(x-0.2)*(x-0.2)*(y*y-y+1.))*(2.*y-1.));
}

double funXX_alauzet(const detri2::Vertex * vert)
{
    double x = vert->crd[0];
    double y = vert->crd[1];

  double value1 = -sin(5 * pow(x - 0.2, 3)*(y*y - y + 1)) * 225 * pow(x - 0.2, 4)*(y*y - y + 1)*(y*y - y + 1) +
                cos(5 * pow(x - 0.2, 3)*(y*y - y + 1)) * 30 * (x - 0.2)*(y*y - y + 1);

  return value1;
}

double funXY_alauzet(const detri2::Vertex * vert)
{
    double x = vert->crd[0];
    double y = vert->crd[1];

  double value2 = -sin(5 * pow(x - 0.2, 3)*(y*y - y + 1)) * 75 * pow(x - 0.2, 5)*(y*y - y + 1)*(2 * y - 1) +
                cos(5 * pow(x - 0.2, 3)*(y*y - y + 1)) * 15 * (x - 0.2)*(x - 0.2)*(2 * y - 1);

  return value2;
}

double funYY_alauzet(const detri2::Vertex * vert)
{
    double x = vert->crd[0];
    double y = vert->crd[1];

  double value3 = -sin(5 * pow(x - 0.2, 3)*(y*y - y + 1)) * 25 * pow(x - 0.2, 6)*(2 * y - 1)*(2 * y - 1) +
                cos(5 * pow(x - 0.2, 3)*(y*y - y + 1)) * 10 * pow(x - 0.2, 3);

  return value3;
}

//==============================================================================
// This function is an example in Huang's paper (Huang05-AnisoMetric.pdf example 4.2, page 23)
// Use the following website to get the derivatives of f_x and f_y.
// https://www.derivative-calculator.net
// tanh(-100*(y-0.5-0.4*sin(2pix))^2)

// tanh(-relax*(y-0.5-m*sin(omega*pi*x))^2)

double huang_coeff_relax = 10.0; // 100.0;
double huang_coeff_omega = 1.0; // 2.0;
double huang_coeff_m = 0.4;

double fun_Huang(const detri2::Vertex * vert)
{
  double x = vert->crd[0];
  double y = vert->crd[1];

  // //double w = y - 0.5 - 0.4*sin(2.*detri2::PI*x);
  // double w = y - 0.4*sin(2.*detri2::PI*x);
  double w = y - 0.4*sin(huang_coeff_omega *detri2::PI*x);

  return tanh(-huang_coeff_relax * w * w);
}

double funX_Huang(const detri2::Vertex * vert)
{
  double x = vert->crd[0];
  double y = vert->crd[1];

  // //double w = -2.*sin(2.*detri2::PI*x) / 5. + y - 0.5;
  // double w = -2.*sin(2.*detri2::PI*x) / 5. + y;
  double w = -huang_coeff_omega*sin(huang_coeff_omega*detri2::PI*x) / 5. + y;

  return -tanh(-huang_coeff_relax * w * w);
}

double funY_Huang(const detri2::Vertex * vert)
{
  double x = vert->crd[0];
  double y = vert->crd[1];

  // //double w = y - 2.*sin(2.*detri2::PI*x) / 5. - 0.5;
  // double w = y - 2.*sin(2.*detri2::PI*x) / 5.;
  double w = y - huang_coeff_omega*sin(huang_coeff_omega*detri2::PI*x) / 5.;

  return -tanh(-huang_coeff_relax * w * w);
}

//==============================================================================
// Du, Q.; Faber, V. & Gunzburger, M.
// Centroidal Voronoi Tessellations: Applications and Algorithms SIAM Rev.,
// Society for Industrial and Applied Mathematics, 1999, 41, 637-676
// Page 670, Figure 7.5

double alpha = 40;

double fun_du(const detri2::Vertex * vert)
{
  double x = vert->crd[0];
  double y = vert->crd[1];

  double sx = sin(detri2::PI * x);
  double sy = sin(detri2::PI * y);

  return exp(-alpha * x * x -alpha * y * y) + 0.05 * sx * sx * sy * sy;
}

double funX_du(const detri2::Vertex * vert)
{
  double x = vert->crd[0];
  double y = vert->crd[1];

  double sx = 2.0 * detri2::PI * cos(detri2::PI * x) * sin(detri2::PI * x);
  double sy = sin(detri2::PI * y);

  return -2.*alpha * x * exp(-alpha * x * x -alpha * y * y) + 0.05 * sx * (sy * sy);
}

double funY_du(const detri2::Vertex * vert)
{
  double x = vert->crd[0];
  double y = vert->crd[1];

  double sx = sin(detri2::PI * x);
  double sy = 2.0 * detri2::PI * cos(detri2::PI * y) * sin(detri2::PI * y);

  return -2.*alpha * y * exp(-alpha * x * x -alpha * y * y) + 0.05 * (sx * sx) * sy;
}

//==============================================================================
// This function is an example of the BAMG program (at manual.pdf page 12)
// Use the following website to get the derivatives of f_x and f_y.
// https://www.derivative-calculator.net

//f5=(10x^3+y^3)+atan2(0.001,sin(5y)-2x)
//    +(10y^3+x^3+atan2(0.01,sin(5x)-2y))

double fun_bamg(const detri2::Vertex * vert)
{
  double x = vert->crd[0];
  double y = vert->crd[1];

  double f1, f2;

  f1  = 10.0 * x * x * x + y * y * y;
  f1 += atan2(0.001, (sin(5.0 * y) - 2.0 * x)); // atan2(0.001,sin(5*y)-2*x)

  f2  = 10.0 * y * y * y + x * x * x;
  f2 += atan2(0.01, (sin(5.0 * x) - 2.0 * y));  // atan2(0.01,sin(5*x)-2*y)

  return (f1 + f2); // * 0.05; // add * 0.01 for testing conformal mapping;
}

double funX_bamg(const detri2::Vertex * vert)
{
  double x = vert->crd[0];
  double y = vert->crd[1];

  double tmp1 = 500. * ((sin(5.*y)-2.*x) * (sin(5.*y)-2.*x) + 1.0 / 1000000.0);
  double f1x = 1.0 / tmp1;

  double tmp2 = 20.0 * ((sin(5.*x)-2.*y) * (sin(5.*x)-2.*y) + 1.0 / 10000.0);
  double f2x = -cos(5.*x) / tmp2;

  return 30.0 * x * x + f1x + 3.0 * x * x + f2x;
}

double funY_bamg(const detri2::Vertex * vert)
{
  double x = vert->crd[0];
  double y = vert->crd[1];


  double tmp1 = 200.0 * ((sin(5.*y)-2.*x) * (sin(5.*y)-2.*x) + 1.0 / 1000000.0);
  double f1y = -cos(5.*y) / tmp1;

  double tmp2 = 50. * ((sin(5.*x)-2.*y) * (sin(5.*x)-2.*y) + 1.0 / 10000.0);
  double f2y = 1.0 / tmp2;

  return 3.0 * y * y + f1y + 30.0 * y * y + f2y;
}

//==============================================================================

double x0 = 0.25;
double x1 = -0.25;

double fun_Gaussian2(const detri2::Vertex * vert)
{
  double x = vert->crd[0];
  double y = vert->crd[1];

  double f1 =  exp(-20.0 * (x - x0) * (x - x0) -20 * (y) * (y));
  double f2 = -exp(-20.0 * (x - x1) * (x - x1) -20 * (y) * (y));

  return f1 + f2;
}

double funX_Gaussian2(const detri2::Vertex * vert)
{
  double x = vert->crd[0];
  double y = vert->crd[1];

  double f1 =  exp(-20.0 * (x - x0) * (x - x0) -20 * (y) * (y));
  double f2 = -exp(-20.0 * (x - x1) * (x - x1) -20 * (y) * (y));

  return -40.0 * x * f1 + 40.0 * x * f2;
}

double funY_Gaussian2(const detri2::Vertex * vert)
{
  double x = vert->crd[0];
  double y = vert->crd[1];

  double f1 =  exp(-20.0 * (x - x0) * (x - x0) -20 * (y) * (y));
  double f2 = -exp(-20.0 * (x - x1) * (x - x1) -20 * (y) * (y));

  return -40.0 * y * f1 + 40.0 * y * f2;
}

//==============================================================================
// Rosenbrock function
// see https://en.wikipedia.org/wiki/Rosenbrock_function
// f = (a-x)^2+b*(y-x^2)^2
// fx = -4*b*x*(y-x^2)-2*(a-x)
// fy = 2*b*(y-x^2);

double coef_a = 1.;
double coef_b = 100.;

double fun_Rosenbrock(const detri2::Vertex * vert)
{
  double x = vert->crd[0];
  double y = vert->crd[1];

  double v1 = (coef_a - x);
  double v2 = (y - x*x);
  double f = v1*v1 + coef_b*v2*v2;

  return f;
}

double funX_Rosenbrock(const detri2::Vertex * vert)
{
  double x = vert->crd[0];
  double y = vert->crd[1];

  double v1 = (coef_a - x);
  double v2 = (y - x*x);
  double fx = -4.*coef_b*x*v2-2.*v1;

  return fx;
}

double funY_Rosenbrock(const detri2::Vertex * vert)
{
  double x = vert->crd[0];
  double y = vert->crd[1];

  double fy = 2*coef_b*(y-x*x);

  return fy;
}

//==============================================================================

double fun_thanks(const detri2::Vertex * vert)
{
  double x = vert->crd[0];
  double y = vert->crd[1];

  //double f = 100.* tanh(1.*(y - 100.* cos(detri2::PI * x / 400.)+200.));
  double f= 4.*(1.-exp(-100.*x) - (1.- exp(-100.))*x)*y*(1.-y);

  return f;
}

double funX_thanks(const detri2::Vertex * vert)
{
  double x = vert->crd[0];
  double y = vert->crd[1];

  //double f = 100.* tanh(100.*(y - 100.* cos(detri2::PI * x / 400.)+200.));
  double f = 4. * (1. + 100. * exp(-100. * x) - (1. - exp(-100.)))*y*(1. - y);

  return f;
}

double funY_thanks(const detri2::Vertex * vert)
{
  double x = vert->crd[0];
  double y = vert->crd[1];

  //double f = 100.* tanh(1.*(y - 100.* cos(detri2::PI * x / 400.)+200.));
  double f = 4. * (1. - exp(-100. * x) - (1. - exp(-100.))*x)*(1. - 2. * y);

  return f;
}

//==============================================================================
//f1(x,y)=tanh(2(x-y)-1)

//==============================================================================
// The quadratic function, z = x^+y^2

double fun_Delaunay(const detri2::Vertex * vert)
{
  double x = vert->crd[0];
  double y = vert->crd[1];
  return x*x + y*y;
}

double funX_Delaunay(const detri2::Vertex * vert)
{
  double x = vert->crd[0];
  //double y = vert->crd[1];
  return 2.*x;
}

double funY_Delaunay(const detri2::Vertex * vert)
{
  //double x = vert->crd[0];
  double y = vert->crd[1];
  return 2.*y;
}

//==============================================================================
// A saddle surface is a smooth surface containing one or more saddle points.
// The hyperbolic paraboloid (hyperboloid of one sheet): z = x^2 - y^2 or z = xy

double fun_Saddle(const detri2::Vertex * vert)
{
  double x = vert->crd[0];
  double y = vert->crd[1];
  return x*x - y*y;
}

double funX_Saddle(const detri2::Vertex * vert)
{
  double x = vert->crd[0];
  double y = vert->crd[1];
  //return 2.*x;
  return y;

}

double funY_Saddle(const detri2::Vertex * vert)
{
  double x = vert->crd[0];
  double y = vert->crd[1];
  //return -2.*y;
  return x;
}

//==============================================================================
