#include <iostream>
#include <stdio.h>
#include <math.h>
#include <assert.h>

#include "hde.h"
#include "test_functions.cpp"

/*
//==============================================================================
// Gaussian
// see an example from BossenHeckbert96-anisoMesh.pdf, page 8.

REAL fun_gaussian(const detri2::Vertex * vert)
{
  REAL x = vert->crd[0];
  REAL y = vert->crd[1];

  return exp(-(x * x + y * y) / 2.0);
}

REAL funX_gaussian(const detri2::Vertex * vert)
{
  REAL x = vert->crd[0];
  REAL y = vert->crd[1];

  return -x * exp(-(x * x + y * y) / 2.0);
}

REAL funY_gaussian(const detri2::Vertex * vert)
{
  REAL x = vert->crd[0];
  REAL y = vert->crd[1];

  return -y * exp(-(x * x + y * y) / 2.0);
}

// Hessian
REAL funXX_gaussian(const detri2::Vertex * vert)
{
  REAL x = vert->crd[0];
  REAL y = vert->crd[1];

  double f = exp(-(x*x + y*y) / 2.0);

  return (x*x - 1.) * f;
}

REAL funXY_gaussian(const detri2::Vertex * vert)
{
  REAL x = vert->crd[0];
  REAL y = vert->crd[1];

  double f = exp(-(x*x + y*y) / 2.0);

  return (x * y) * f;
}

REAL funYY_gaussian(const detri2::Vertex * vert)
{
  REAL x = vert->crd[0];
  REAL y = vert->crd[1];

  double f = exp(-(x*x + y*y) / 2.0);

  return (y*y - 1.) * f;
}

//==============================================================================
// lennard
// f2(x,y)=tanh(60x)-tanh(60(x-y)-30)

REAL fun_lennard(const detri2::Vertex * vert)
{
    REAL x = vert->crd[0];
    REAL y = vert->crd[1];
    
    return(tanh(60.*x)-tanh(60.*x-60.*y-30.));
}

REAL funX_lennard(const detri2::Vertex * vert)
{
    REAL x = vert->crd[0];
    REAL y = vert->crd[1];
    
    return(60.*tanh(60.*x-60.*y-30.)*tanh(60.*x-60.*y-30.)-60.*tanh(60.*x)*tanh(60.*x));
}

REAL funY_lennard(const detri2::Vertex * vert)
{
    REAL x = vert->crd[0];
    REAL y = vert->crd[1];
    
    return(-60.+60.*tanh(60.*x-60.*y-30.)*tanh(60.*x-60.*y-30.));
}

REAL funXX_lennard(const detri2::Vertex * vert)
{
  REAL x = vert->crd[0];
  REAL y = vert->crd[1];

  double ins1 = exp(120 * x);
  double ins2 = exp(120 * (x - y) - 60);
  double value1 = 2 * (14400 * ins1*(ins1 + 1)*(ins1 + 1) - 28800 * ins1*ins1*(ins1 + 1)) / pow(ins1 + 1, 4) -
				2 * (14400 * ins2*(ins2 + 1)*(ins2 + 1) - 28800 * ins2*ins2*(ins2 + 1)) / pow(ins2 + 1, 4);
//  double value2 = 2 * (14400 * ins2*(ins2 + 1)*(ins2 + 1) - 28800 * ins2*ins2*(ins2 + 1)) / pow(ins2 + 1, 4);
//  double value3 = -value2;
  
  return value1;
}

REAL funXY_lennard(const detri2::Vertex * vert)
{
    REAL x = vert->crd[0];
    REAL y = vert->crd[1];
    
  double ins1 = exp(120 * x);
  double ins2 = exp(120 * (x - y) - 60);
  //double value1 = 2 * (14400 * ins1*(ins1 + 1)*(ins1 + 1) - 28800 * ins1*ins1*(ins1 + 1)) / pow(ins1 + 1, 4) -
//				2 * (14400 * ins2*(ins2 + 1)*(ins2 + 1) - 28800 * ins2*ins2*(ins2 + 1)) / pow(ins2 + 1, 4);
  double value2 = 2 * (14400 * ins2*(ins2 + 1)*(ins2 + 1) - 28800 * ins2*ins2*(ins2 + 1)) / pow(ins2 + 1, 4);
  //double value3 = -value2;
  
  return value2;
}

REAL funYY_lennard(const detri2::Vertex * vert)
{
    REAL x = vert->crd[0];
    REAL y = vert->crd[1];
    
  double ins1 = exp(120 * x);
  double ins2 = exp(120 * (x - y) - 60);
  //double value1 = 2 * (14400 * ins1*(ins1 + 1)*(ins1 + 1) - 28800 * ins1*ins1*(ins1 + 1)) / pow(ins1 + 1, 4) -
//				2 * (14400 * ins2*(ins2 + 1)*(ins2 + 1) - 28800 * ins2*ins2*(ins2 + 1)) / pow(ins2 + 1, 4);
  double value2 = 2 * (14400 * ins2*(ins2 + 1)*(ins2 + 1) - 28800 * ins2*ins2*(ins2 + 1)) / pow(ins2 + 1, 4);
  double value3 = -value2;
  
  return value3;
}

//==============================================================================
// f3(x,y)=sin(5 * pow(x - 0.2, 3)*(y*y - y + 1))

REAL fun_alauzet(const detri2::Vertex * vert)
{
    REAL x = vert->crd[0];
    REAL y = vert->crd[1];
    
    return(sin(5.*(x-0.2)*(x-0.2)*(x-0.2)*(y*y-y+1.)));
}

REAL funX_alauzet(const detri2::Vertex * vert)
{
    REAL x = vert->crd[0];
    REAL y = vert->crd[1];
    
    return(cos(5.*(x-0.2)*(x-0.2)*(x-0.2)*(y*y-y+1.))*3.*(x-0.2)*(x-0.2));
}

REAL funY_alauzet(const detri2::Vertex * vert)
{
    REAL x = vert->crd[0];
    REAL y = vert->crd[1];
    
    return(cos(5.*(x-0.2)*(x-0.2)*(x-0.2)*(y*y-y+1.))*(2.*y-1.));
}

REAL funXX_alauzet(const detri2::Vertex * vert)
{
    REAL x = vert->crd[0];
    REAL y = vert->crd[1];
    
  double value1 = -sin(5 * pow(x - 0.2, 3)*(y*y - y + 1)) * 225 * pow(x - 0.2, 4)*(y*y - y + 1)*(y*y - y + 1) +
				cos(5 * pow(x - 0.2, 3)*(y*y - y + 1)) * 30 * (x - 0.2)*(y*y - y + 1);

  return value1;
}

REAL funXY_alauzet(const detri2::Vertex * vert)
{
    REAL x = vert->crd[0];
    REAL y = vert->crd[1];
    
  double value2 = -sin(5 * pow(x - 0.2, 3)*(y*y - y + 1)) * 75 * pow(x - 0.2, 5)*(y*y - y + 1)*(2 * y - 1) +
				cos(5 * pow(x - 0.2, 3)*(y*y - y + 1)) * 15 * (x - 0.2)*(x - 0.2)*(2 * y - 1);

  return value2;
}

REAL funYY_alauzet(const detri2::Vertex * vert)
{
    REAL x = vert->crd[0];
    REAL y = vert->crd[1];
    
  double value3 = -sin(5 * pow(x - 0.2, 3)*(y*y - y + 1)) * 25 * pow(x - 0.2, 6)*(2 * y - 1)*(2 * y - 1) +
				cos(5 * pow(x - 0.2, 3)*(y*y - y + 1)) * 10 * pow(x - 0.2, 3);

  return value3;
}

//==============================================================================
// This function is an example in Huang's paper (Huang05-AnisoMetric.pdf example 4.2, page 23)
// Use the following website to get the derivatives of f_x and f_y.
// https://www.derivative-calculator.net
// tanh(-100*(y-0.5-0.4*sin(2pix))^2)

REAL fun_Huang(const detri2::Vertex * vert)
{
  REAL x = vert->crd[0];
  REAL y = vert->crd[1];

  //REAL w = y - 0.5 - 0.4*sin(2.*detri2::PI*x);
  REAL w = y - 0.4*sin(2.*detri2::PI*x);

  return tanh(-100.0 * w * w);
}

REAL funX_Huang(const detri2::Vertex * vert)
{
  REAL x = vert->crd[0];
  REAL y = vert->crd[1];

  //REAL w = -2.*sin(2.*detri2::PI*x) / 5. + y - 0.5;
  REAL w = -2.*sin(2.*detri2::PI*x) / 5. + y;

  return -tanh(100.0 * w * w);
}

REAL funY_Huang(const detri2::Vertex * vert)
{
  REAL x = vert->crd[0];
  REAL y = vert->crd[1];

  //REAL w = y - 2.*sin(2.*detri2::PI*x) / 5. - 0.5;
  REAL w = y - 2.*sin(2.*detri2::PI*x) / 5.;

  return -tanh(100. * w * w);
}

//==============================================================================
// Du, Q.; Faber, V. & Gunzburger, M. 
// Centroidal Voronoi Tessellations: Applications and Algorithms SIAM Rev., 
// Society for Industrial and Applied Mathematics, 1999, 41, 637-676
// Page 670, Figure 7.5

REAL alpha = 40;

REAL fun_du(const detri2::Vertex * vert)
{
  REAL x = vert->crd[0];
  REAL y = vert->crd[1];

  REAL sx = sin(detri2::PI * x);
  REAL sy = sin(detri2::PI * y);

  return exp(-alpha * x * x -alpha * y * y) + 0.05 * sx * sx * sy * sy;  
}

REAL funX_du(const detri2::Vertex * vert)
{
  REAL x = vert->crd[0];
  REAL y = vert->crd[1];

  REAL sx = 2.0 * detri2::PI * cos(detri2::PI * x) * sin(detri2::PI * x);
  REAL sy = sin(detri2::PI * y);

  return -2.*alpha * x * exp(-alpha * x * x -alpha * y * y) + 0.05 * sx * (sy * sy);
}

REAL funY_du(const detri2::Vertex * vert)
{
  REAL x = vert->crd[0];
  REAL y = vert->crd[1];

  REAL sx = sin(detri2::PI * x);
  REAL sy = 2.0 * detri2::PI * cos(detri2::PI * y) * sin(detri2::PI * y);

  return -2.*alpha * y * exp(-alpha * x * x -alpha * y * y) + 0.05 * (sx * sx) * sy;
}

//==============================================================================
// This function is an example of the BAMG program (at manual.pdf page 12)
// Use the following website to get the derivatives of f_x and f_y.
// https://www.derivative-calculator.net

//f5=(10x^3+y^3)+atan2(0.001,sin(5y)-2x)
//    +(10y^3+x^3+atan(0.01,sin(5x)-2y))

REAL fun_bamg(const detri2::Vertex * vert)
{
  REAL x = vert->crd[0];
  REAL y = vert->crd[1];

  REAL f1, f2;

  f1  = 10.0 * x * x * x + y * y * y; 
  f1 += atan2(0.001, (sin(5.0 * y) - 2.0 * x)); // atan2(0.001,sin(5*y)-2*x)

  f2  = 10.0 * y * y * y + x * x * x;
  f2 += atan2(0.01, (sin(5.0 * x) - 2.0 * y));  // atan2(0.01,sin(5*x)-2*y)

  return f1 + f2;
}

REAL funX_bamg(const detri2::Vertex * vert)
{
  REAL x = vert->crd[0];
  REAL y = vert->crd[1];

  REAL tmp1 = 500. * ((sin(5.*y)-2.*x) * (sin(5.*y)-2.*x) + 1.0 / 1000000.0);
  REAL f1x = 1.0 / tmp1;

  REAL tmp2 = 20.0 * ((sin(5.*x)-2.*y) * (sin(5.*x)-2.*y) + 1.0 / 10000.0);
  REAL f2x = -cos(5.*x) / tmp2;

  return 30.0 * x * x + f1x + 3.0 * x * x + f2x;
}

REAL funY_bamg(const detri2::Vertex * vert)
{
  REAL x = vert->crd[0];
  REAL y = vert->crd[1];


  REAL tmp1 = 200.0 * ((sin(5.*y)-2.*x) * (sin(5.*y)-2.*x) + 1.0 / 1000000.0);
  REAL f1y = -cos(5.*y) / tmp1;

  REAL tmp2 = 50. * ((sin(5.*x)-2.*y) * (sin(5.*x)-2.*y) + 1.0 / 10000.0);
  REAL f2y = 1.0 / tmp2;

  return 3.0 * y * y + f1y + 30.0 * y * y + f2y;
}

//==============================================================================

REAL x0 = 0.25;
REAL x1 = -0.25;

REAL fun_Gaussian2(const detri2::Vertex * vert)
{
  REAL x = vert->crd[0];
  REAL y = vert->crd[1];

  REAL f1 =  exp(-20.0 * (x - x0) * (x - x0) -20 * (y) * (y));
  REAL f2 = -exp(-20.0 * (x - x1) * (x - x1) -20 * (y) * (y));

  return f1 + f2;
}

REAL funX_Gaussian2(const detri2::Vertex * vert)
{
  REAL x = vert->crd[0];
  REAL y = vert->crd[1];

  REAL f1 =  exp(-20.0 * (x - x0) * (x - x0) -20 * (y) * (y));
  REAL f2 = -exp(-20.0 * (x - x1) * (x - x1) -20 * (y) * (y));

  return -40.0 * x * f1 + 40.0 * x * f2;
}

REAL funY_Gaussian2(const detri2::Vertex * vert)
{
  REAL x = vert->crd[0];
  REAL y = vert->crd[1];

  REAL f1 =  exp(-20.0 * (x - x0) * (x - x0) -20 * (y) * (y));
  REAL f2 = -exp(-20.0 * (x - x1) * (x - x1) -20 * (y) * (y));

  return -40.0 * y * f1 + 40.0 * y * f2;
}

//==============================================================================
// Rosenbrock function
// see https://en.wikipedia.org/wiki/Rosenbrock_function
// f = (a-x)^2+b*(y-x^2)^2
// fx = -4*b*x*(y-x^2)-2*(a-x)
// fy = 2*b*(y-x^2);

REAL coef_a = 1.;
REAL coef_b = 100.;

REAL fun_Rosenbrock(const detri2::Vertex * vert)
{
  REAL x = vert->crd[0];
  REAL y = vert->crd[1];

  REAL v1 = (coef_a - x);
  REAL v2 = (y - x*x);
  REAL f = v1*v1 + coef_b*v2*v2;

  return f;
}

REAL funX_Rosenbrock(const detri2::Vertex * vert)
{
  REAL x = vert->crd[0];
  REAL y = vert->crd[1];

  REAL v1 = (coef_a - x);
  REAL v2 = (y - x*x);
  REAL fx = -4.*coef_b*x*v2-2.*v1;

  return fx;
}

REAL funY_Rosenbrock(const detri2::Vertex * vert)
{
  REAL x = vert->crd[0];
  REAL y = vert->crd[1];

  REAL fy = 2*coef_b*(y-x*x);

  return fy;
}

//==============================================================================

REAL fun_thanks(const detri2::Vertex * vert)
{
  REAL x = vert->crd[0];
  REAL y = vert->crd[1];

  REAL f = 100.* tanh(1.*(y - 100.* cos(detri2::PI * x / 400.)+200.));

  return f;
}

REAL funX_thanks(const detri2::Vertex * vert)
{
  REAL x = vert->crd[0];
  REAL y = vert->crd[1];

  REAL f = 100.* tanh(100.*(y - 100.* cos(detri2::PI * x / 400.)+200.));

  return f;
}

REAL funY_thanks(const detri2::Vertex * vert)
{
  REAL x = vert->crd[0];
  REAL y = vert->crd[1];

  REAL f = 100.* tanh(1.*(y - 100.* cos(detri2::PI * x / 400.)+200.));

  return f;
}

//==============================================================================
//f1(x,y)=tanh(2(x-y)-1)

//==============================================================================
// The quadratic function, z = x^+y^2

REAL fun_Delaunay(const detri2::Vertex * vert)
{
  REAL x = vert->crd[0];
  REAL y = vert->crd[1];
  return x*x + y*y;
}

REAL funX_Delaunay(const detri2::Vertex * vert)
{
  REAL x = vert->crd[0];
  //REAL y = vert->crd[1];
  return 2.*x;
}

REAL funY_Delaunay(const detri2::Vertex * vert)
{
  //REAL x = vert->crd[0];
  REAL y = vert->crd[1];
  return 2.*y;
}

//==============================================================================
// A saddle surface is a smooth surface containing one or more saddle points.
// The hyperbolic paraboloid (hyperboloid of one sheet): z = x^2 - y^2 or z = xy

REAL fun_Saddle(const detri2::Vertex * vert)
{
  REAL x = vert->crd[0];
  REAL y = vert->crd[1];
  return x*x - y*y;
}

REAL funX_Saddle(const detri2::Vertex * vert)
{
  REAL x = vert->crd[0];
  REAL y = vert->crd[1];
  //return 2.*x; 
  return y;

}

REAL funY_Saddle(const detri2::Vertex * vert)
{
  REAL x = vert->crd[0];
  REAL y = vert->crd[1];
  //return -2.*y; 
  return x;
}
*/

//==============================================================================

using  namespace detri2;

int main(int argc, char* argv[])
{
  HDE_Triangulation *Tr = new HDE_Triangulation();

  if (!Tr->HDE_parse_commands(argc, argv)) {
    // No input or wrong parameters.
    printf("Usage: detri2HDE [options]\n");
    printf("List of basic options:\n");
    printf("  input_mesh=filename, filename without .ele.\n");
    printf("  s1=# s2=# s3=#, set the s-factors.\n");
    printf("  length=#, set target length.\n");
    printf("  angle=#, set target angle (degree).\n");
    printf("  max_iter=#, set maximum iterations.\n");
    printf("  use_intpoints=#, use # interior points for calculating distance.\n");
    delete Tr;
    return 0;
  }

  if (Tr->op_test_function == 1) {
    Tr->Func = &fun_gaussian;
    Tr->GradX = &funX_gaussian;
    Tr->GradY = &funY_gaussian; 
    Tr->FunXX = &funXX_gaussian; // Hessian
    Tr->FunXY = &funXY_gaussian;
    Tr->FunYY = &funYY_gaussian;
  } else if (Tr->op_test_function == 2) {
    Tr->Func = &fun_lennard;
    Tr->GradX = &funX_lennard;
    Tr->GradY = &funY_lennard;
    Tr->FunXX = &funXX_lennard; // Hessian
    Tr->FunXY = &funXY_lennard;
    Tr->FunYY = &funYY_lennard;
  } else if (Tr->op_test_function == 3) {
    Tr->Func = &fun_alauzet;
    Tr->GradX = &funX_alauzet;
    Tr->GradY = &funY_alauzet;
    Tr->FunXX = &funXX_alauzet; // Hessian
    Tr->FunXY = &funXY_alauzet;
    Tr->FunYY = &funYY_alauzet;
  } else if (Tr->op_test_function == 4) {
    Tr->Func = &fun_Huang;
    Tr->GradX = &funX_Huang;
    Tr->GradY = &funY_Huang;
  } else if (Tr->op_test_function == 5) {
    Tr->Func = &fun_du;
    Tr->GradX = &funX_du;
    Tr->GradY = &funY_du;
  } else if (Tr->op_test_function == 6) {
    Tr->Func = &fun_Gaussian2;
    Tr->GradX = &funX_Gaussian2;
    Tr->GradY = &funY_Gaussian2;
  } else if (Tr->op_test_function == 7) {
    // use unitcircle.1.ele
    Tr->Func = &fun_bamg;
    Tr->GradX = &funX_bamg;
    Tr->GradY = &funY_bamg;  
  } else if (Tr->op_test_function == 8) {
    Tr->Func = &fun_Rosenbrock;
    Tr->GradX = &funX_Rosenbrock;
    Tr->GradY = &funY_Rosenbrock;  
  } else if (Tr->op_test_function == 100) {
    Tr->Func = &fun_Delaunay;
    Tr->GradX = &funX_Delaunay;
    Tr->GradY = &funY_Delaunay;
  } else if (Tr->op_test_function == 101) {
    Tr->Func = &fun_Saddle;
    Tr->GradX = &funX_Saddle;
    Tr->GradY = &funY_Saddle;
  } else {
    printf("Please choose a test function by set test_fun=#:\n");
    printf("  1, Gaussian.\n");
    printf("  2, Lennard.\n");
    printf("  3, Alauzet.\n");
    printf("  4, Huang.\n");
    printf("  5, Du.\n");
    printf("  6, Gaussian2.\n");
    printf("  7, BAMG.\n");
    printf("  8, Rosenbrock.\n");
    return 0;
  }

  if (!Tr->read_mesh()) {
    printf("Failed to read input mesh.\n");
    delete Tr;
    return 0;
  }

  Tr->reconstruct_mesh(0); // do not check delaunay.

  std::cout << "--------------------------------\n";
  std::cout << "Statistics before the improvment\n";
  std::cout << "--------------------------------\n";
  Tr->mesh_statistics();

  Tr->HDE_mesh_adapt();

  std::cout << "-------------------------------\n";
  std::cout << "Statistics after the improvment\n";
  std::cout << "-------------------------------\n";
  Tr->mesh_statistics();

  Tr->save_triangulation();

  delete Tr;
  return 1;
}
