#define MACOSX

#ifdef MACOSX
  #include "OpenGL/glu.h"
#else
  #include <gl/gl.h>
  #include <gl/glu.h>
#endif

#include "view3d.h"
#include "scribblearea.h"
//#include "glhls.h"  // pick different colors for materials

using namespace detri2;

//=============================================================================
// 'shrinkpoint'  Calculate a scaled point 'newpoint' (an REAL[3] array) for
//   'oldpoint' (point *) with respect to the 'center' (an REAL[3] array)
//   and 'scale' (REAL, [0, 1]).
#define shrinkpoint(newpoint, oldpoint, center, scale) \
  (newpoint)->crd[0] = center[0] + scale * ((oldpoint)->crd[0] - center[0]); \
  (newpoint)->crd[1] = center[1] + scale * ((oldpoint)->crd[1] - center[1]); \
  (newpoint)->crd[2] = center[2] + scale * ((oldpoint)->crd[2] - center[2])

#define drawedge(pt1, pt2) \
  glVertex3d((pt1)->crd[0], (pt1)->crd[1], (pt1)->crd[2]); \
  glVertex3d((pt2)->crd[0], (pt2)->crd[1], (pt2)->crd[2])

// 'drawface'  Primitive to draw a triangles.
/*
#define drawface(pt1, pt2, pt3) \
  ax = (pt1)->crd[0] - (pt2)->crd[0];\
  ay = (pt1)->crd[1] - (pt2)->crd[1];\
  az = (pt1)->crd[2] - (pt2)->crd[2];\
  bx = (pt1)->crd[0] - (pt3)->crd[0];\
  by = (pt1)->crd[1] - (pt3)->crd[1];\
  bz = (pt1)->crd[2] - (pt3)->crd[2];\
  nm[0] = ay * bz - by * az;\
  nm[1] = az * bx - bz * ax;\
  nm[2] = ax * by - bx * ay;\
  le = sqrt(nm[0] * nm[0] + nm[1] * nm[1] + nm[2] * nm[2]);\
  if (le > 0.0) {nm[0] /= le; nm[1] /= le; nm[2]/= le;}\
  glNormal3dv(nm);\
  glVertex3d((pt1)->crd[0], (pt1)->crd[1], (pt1)->crd[2]); \
  glVertex3d((pt2)->crd[0], (pt2)->crd[1], (pt2)->crd[2]); \
  glVertex3d((pt3)->crd[0], (pt3)->crd[1], (pt3)->crd[2])
*/
/*
  if (pv->switch_backface == tetview::ON) {\
  glGetFloatv(GL_CURRENT_COLOR, curr_clr); \
  glColor3dv(pv->clr.backface);\
  glVertex3d((pt1)->x, (pt1)->y, (pt1)->z); \
  glVertex3d((pt3)->x, (pt3)->y, (pt3)->z); \
  glVertex3d((pt2)->x, (pt2)->y, (pt2)->z); \
  glColor4fv(curr_clr);\
  } */

#define drawface(pt1, pt2, pt3) \
  glVertex3d((pt1)->crd[0], (pt1)->crd[1], (pt1)->crd[2]); \
  glVertex3d((pt2)->crd[0], (pt2)->crd[1], (pt2)->crd[2]); \
  glVertex3d((pt3)->crd[0], (pt3)->crd[1], (pt3)->crd[2])

/*
// --------------------------- helpers --------------------------- //

inline void glhlsColor3d_(double h, double  l, double s)
{
  float hls[3],rgb[3];
  hls[0]=h;
  hls[1]=l;
  hls[2]=s;
  glColor3fv(glhlsConvertToRGB(hls,rgb));
}

inline void glhlsColor4d_(double h, double  l, double s, double color_transparency)
{
  float hls[3],rgb[4];
  hls[0]=h;
  hls[1]=l;
  hls[2]=s;
  glhlsConvertToRGB(hls,rgb);
  rgb[3]=color_transparency;
  glColor4fv(rgb);

}

inline void glcolori(int icol, int maxcol, double color_transparency)
{
  const double h=(double)(--icol/3)/(double)(maxcol/3+1);
  //glhlsColor3d_(0.333*(icol%3)+0.333*h,0.5,0.5);
  float hls[3],rgb[4];
  hls[0]=0.333*(icol%3)+0.333*h;
  hls[1]=0.5;
  hls[2]=0.5;
  glhlsConvertToRGB(hls,rgb);
  rgb[3]=color_transparency;
  glColor4fv(rgb);
}
*/

//=============================================================================
//                                                                           //
// cuttri()    Cut a triangle by a cutting plane.                            //
//                                                                           //
// The orient3d plane is given by four coefficients, a, b, c, and d for the  //
// plane equation. We keep the whole (or part of the) triangle on the posi-  //
// tive side of the plane, i.e., any keeping point (x,y,z) satisfy,          //
//   a * x + b * y + c * z + d >= 0                                          //
//                                                                           //
// The return value is one of the following:                                 //
//   0, the whole triangle is cut away;                                      //
//   1, the whole triangle is kept;                                          //
//   3, the triangle is cut by three points returned in the array 'polypt';  //
//   4, the triangle is cut by four points returned in the array 'polypt';   //
//                                                                           //
// See sphere_surface_draw() for an example of using this function.          //
//                                                                           //
// This function will create the intersecting points between the plane and   //
// the triangle. All created points will have a default marker '-99'. Hope-  //
// fully, this will make the distinction of existing and new points in the   //
// array 'polypt'.                                                           //

void tv2Point_copy(Vertex& p_dest, Vertex& p_src)
{
  p_dest.crd[0] = p_src.crd[0];
  p_dest.crd[1] = p_src.crd[1];
  p_dest.crd[2] = p_src.crd[2];
}

// The returning point is in 'intp'.
int planelineinter(double a, double b, double c, double d,
                   Vertex *p0, Vertex *p1, Vertex *intp)
{
    double vx = p1->crd[0] - p0->crd[0];
    double vy = p1->crd[1] - p0->crd[1];
    double vz = p1->crd[2] - p0->crd[2];

    double u = -(a * p0->crd[0] + b * p0->crd[1] + c * p0->crd[2] + d)
               /(a * vx + b * vy + c * vz);
    // Debug code
    //if ((u <= 0) || (u >= 1)) {
    //  printf("  !! Wrong point, u = %g.\n", u);
    //}

    intp->crd[0] = p0->crd[0] + u * vx;
    intp->crd[1] = p0->crd[1] + u * vy;
    intp->crd[2] = p0->crd[2] + u * vz;

    return 1;
}

int cutseg(double a, double b, double c, double d,
           Vertex *p0, Vertex *p1, Vertex polypt[2], double dmax)
{
    REAL s0 = a * p0->crd[0] + b * p0->crd[1] + c * p0->crd[2] + d;
    REAL s1 = a * p1->crd[0] + b * p1->crd[1] + c * p1->crd[2] + d;
    // Round the result.
    if (fabs(s0) / dmax < 1e-4) s0 = 0;
    if (fabs(s1) / dmax < 1e-4) s1 = 0;

    //printf("    [%d]: %g, %g, %g, s0(%g)\n", p0->idx, p0->crd[0], p0->crd[1], p0->crd[2], s0);
    //printf("    [%d]: %g, %g, %g, s1(%g)\n", p1->idx, p1->crd[0], p1->crd[1], p1->crd[2], s1);

    if (s0 >= 0) {
      if (s1 >= 0) {
        // Fully visible;
        tv2Point_copy(polypt[0], *p0);
        tv2Point_copy(polypt[1], *p1);
      } else {
        tv2Point_copy(polypt[0], *p0);
        planelineinter(a,b,c,d,p0,p1,&(polypt[1]));
      }
      //printf("    -[%d]: %g, %g, %g, s0(%g)\n", p0->idx, polypt[0].crd[0], polypt[0].crd[1], polypt[0].crd[2], s0);
      //printf("    -[%d]: %g, %g, %g, s1(%g)\n", p1->idx, polypt[1].crd[0], polypt[1].crd[1], polypt[1].crd[2], s1);
      return 2;
    } else { // s1 < 0
      if (s1 >= 0) {
        planelineinter(a,b,c,d,p0,p1,&(polypt[0]));
        tv2Point_copy(polypt[1], *p1);
        //printf("    -[%d]: %g, %g, %g, s0(%g)\n", p0->idx, polypt[0].crd[0], polypt[0].crd[1], polypt[0].crd[2], s0);
        //printf("    -[%d]: %g, %g, %g, s1(%g)\n", p1->idx, polypt[1].crd[0], polypt[1].crd[1], polypt[1].crd[2], s1);
        return 2;
      } else {
        return 0;
      }
    }
}

int cuttri(double a, double b, double c, double d,
           Vertex *p0, Vertex *p1, Vertex *p2, Vertex polypt[4], double dmax)
{
    // Evaulate the point positions (signs) w.r.t. the plane.
      REAL s0 = a * p0->crd[0] + b * p0->crd[1] + c * p0->crd[2] + d;
      REAL s1 = a * p1->crd[0] + b * p1->crd[1] + c * p1->crd[2] + d;
      REAL s2 = a * p2->crd[0] + b * p2->crd[1] + c * p2->crd[2] + d;
      // Round the result.
      if (fabs(s0) / dmax < 1e-4) s0 = 0;
      if (fabs(s1) / dmax < 1e-4) s1 = 0;
      if (fabs(s2) / dmax < 1e-4) s2 = 0;

      // Debug.
      //printf("    [%d]: %g, %g, %g, s0(%g)\n", p0->idx, p0->crd[0], p0->crd[1], p0->crd[2], s0);
      //printf("    [%d]: %g, %g, %g, s1(%g)\n", p1->idx, p1->crd[0], p1->crd[1], p1->crd[2], s1);
      //printf("    [%d]: %g, %g, %g, s2(%g)\n", p2->idx, p2->crd[0], p2->crd[1], p2->crd[2], s2);

      ///////////////////////////////////////////////////////////////////////
      // First process degenerate case.
      if (s0 == 0) {
        if (s1 > 0) {
          if (s2 > 0) {
            return 1; // Fully visible.
          } else { // s2 <= 0
            if (s2 < 0) {
              tv2Point_copy(polypt[0], *p0);
              tv2Point_copy(polypt[1], *p1);
              planelineinter(a,b,c,d,p1,p2,&(polypt[2]));
              polypt[2].tag = -99;
              return 3;
            } else { // s2 == 0
              return 1; // Fully visible.
            }
          }
        } else { // s1 <= 0
          if (s1 < 0) {
            if (s2 > 0) {
              tv2Point_copy(polypt[0], *p2);
              tv2Point_copy(polypt[1], *p0);
              planelineinter(a,b,c,d,p1,p2,&(polypt[2]));
              polypt[2].tag = -99;
              return 3;
            } else {
              if (s2 < 0) {
                return 0; // Not visisble.
              } else { // s2 == 0
                return 0; // Not visisble.
              }
            }
          } else { // s1 == 0
            if (s2 > 0) {
              return 1; // Fully visible.
            } else {
              if (s2 < 0) {
                return 0; // Not visisble.
              } else { // s2 == 0
                return 0; // !!! Not possible.
              }
            }
          }
        }
      }

      if (s1 == 0) { // s0 != 0
        if (s0 > 0) {
          if (s2 > 0) {
            return 1; // Fully visible.
          } else {
            if (s2 < 0) {
              tv2Point_copy(polypt[0], *p0);
              tv2Point_copy(polypt[1], *p1);
              planelineinter(a,b,c,d,p0,p2,&(polypt[2]));
              polypt[2].tag = -99;
              return 3;
            } else { // s2 == 0
              return 1; // Fully visible.
            }
          }
        } else {  // s0 < 0
          if (s2 > 0) {
            tv2Point_copy(polypt[0], *p1);
            tv2Point_copy(polypt[1], *p2);
            planelineinter(a,b,c,d,p2,p0,&(polypt[2]));
            polypt[2].tag = -99;
            return 3;
          } else { // s2 <= 0
            if (s2 < 0) {
              return 0; // Not visible.
            } else { // s2 == 0
              return 0; // Not visible.
            }
          }
        }
      }

      if (s2 == 0) { // s0 != 0 && s1 != 0
        if (s0 > 0) {
          if (s1 > 0) {
            return 1; // Fully visible.
          } else { // s1 < 0
            tv2Point_copy(polypt[0], *p2);
            tv2Point_copy(polypt[1], *p0);
            planelineinter(a,b,c,d,p0,p1,&(polypt[2]));
            polypt[2].tag = -99;
            return 3;
          }
        } else { // s0 < 0
          if (s1 > 0) {
            tv2Point_copy(polypt[0], *p1);
            tv2Point_copy(polypt[1], *p2);
            planelineinter(a,b,c,d,p0,p1,&(polypt[2]));
            polypt[2].tag = -99;
            return 3;
          } else { // s1 < 0
            return 0;
          }
        }
      }
      ///////////////////////////////////////////////////////////////////////

      ///////////////////////////////////////////////////////////////////////
      // Process non-degenrate case.
      if (s0 > 0) { // p0 is visible.
        if (s1 > 0) { // p1 is visible.
          if (s2 > 0) { // p2 is visible.
            return 1; // fully visible
          } else { // s2 < 0
            tv2Point_copy(polypt[0], *p0);
            tv2Point_copy(polypt[1], *p1);
            planelineinter(a,b,c,d,p1,p2,&(polypt[2]));
            planelineinter(a,b,c,d,p2,p0,&(polypt[3]));
            polypt[2].tag = -99;
            polypt[3].tag = -99;
            return 4;
          }
        } else {  // s1 < 0
          if (s2 > 0) {
            tv2Point_copy(polypt[0], *p2);
            tv2Point_copy(polypt[1], *p0);
            planelineinter(a,b,c,d,p0,p1,&(polypt[2]));
            planelineinter(a,b,c,d,p1,p2,&(polypt[3]));
            polypt[2].tag = -99;
            polypt[3].tag = -99;
            return 4;
          } else { // s2 < 0
            tv2Point_copy(polypt[0], *p0);
            planelineinter(a,b,c,d,p0,p1,&(polypt[1]));
            planelineinter(a,b,c,d,p2,p0,&(polypt[2]));
            polypt[1].tag = -99;
            polypt[2].tag = -99;
            return 3;
          }
        }
      } else { // s0 < 0
        if (s1 > 0) {
          if (s2 > 0) {
            tv2Point_copy(polypt[0], *p1);
            tv2Point_copy(polypt[1], *p2);
            planelineinter(a,b,c,d,p2,p0,&(polypt[2]));
            planelineinter(a,b,c,d,p0,p1,&(polypt[3]));
            polypt[2].tag = -99;
            polypt[3].tag = -99;
            return 4;
          } else { // s2 < 0
            tv2Point_copy(polypt[0], *p1);
            planelineinter(a,b,c,d,p1,p2,&(polypt[1]));
            planelineinter(a,b,c,d,p0,p1,&(polypt[2]));
            polypt[1].tag = -99;
            polypt[2].tag = -99;
            return 3;
          }
        } else { // s1 < 0
          if (s2 > 0) {
            tv2Point_copy(polypt[0], *p2);
            planelineinter(a,b,c,d,p2,p0,&(polypt[1]));
            planelineinter(a,b,c,d,p1,p2,&(polypt[2]));
            polypt[1].tag = -99;
            polypt[2].tag = -99;
            return 3;
          } else { // s2 < 0
            return 0; // Invisible.
          }
        }
      }
      ///////////////////////////////////////////////////////////////////////

      //assert(0); // Should not be here.
      return 0;
}

//=============================================================================
/*
int GLViewer::get_plane(double *a,double *b,double *c, double *d)
{
  *a=0.0;
  *b=0.0;
  *c=0.0;
  int  plane_dir=0;
  double plane_d, dummy;

  if (submenu_entry_state("View","X Section")) plane_dir=0;
  else if (submenu_entry_state("View","Y Section")) plane_dir=1;
  else if (submenu_entry_state("View","Z Section")) plane_dir=2;
  glvMouseControl *mc=mouse_control("plane");
  if (mc==0) return -1;

  mc->get_state(&dummy, &plane_d);
  plane_d=_min[plane_dir]-plane_d*(_max[plane_dir]-_min[plane_dir]);

  switch(plane_dir)
    {
    case 0:
      *a=1.0; break;
    case 1:
      *b=1.0; break;
    case 2:
    default:
      *c=1.0; break;
    }

  // why negativ ?
  *d=-plane_d;

  return plane_dir;
}
*/
//=============================================================================

void OGLWidget::draw_triangulation()
{
    ScribbleArea *object_data = (ScribbleArea *) _scribbleArea;
    Triangulation *tria = object_data->_tria;
    if (tria == NULL) return;

    // Used by shrinkpoint()
    Vertex Vshrink[3]; double tcenter[3];
    double shrink = _shrink;
    qDebug()<<" shrink="<<shrink;

    printf("Draw %d triangles\n", tria->tr_tris->objects - tria->ct_hullsize);
    //double ax,ay,az,bx,by,bz,nm[3],le; // Used by drawface()

    if (showFacets->isChecked()) {
      double r, g, b, a;
      _facetcolor = object_data->_v3d_facetcolor;
      _facetcolor.getRgbF(&r, &g, &b, &a);
      glColor4f(r, g, b, 0.45); // a=1 no transparency
      if (showTransparent->isChecked()) {
        glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
        glAlphaFunc(GL_GREATER,0.1);
        glEnable(GL_BLEND);
        glDepthMask(GL_FALSE);
      }
      glBegin(GL_TRIANGLES);
      if (enableShrink->isChecked()) {
        for (int i = 0; i < tria->tr_tris->used_items; i++) {
          Triang *t = (Triang *) tria->tr_tris->get(i);
          if (t->is_deleted()) continue;
          if (t->is_hulltri()) continue;
          for (int j = 0; j < 3; j++) {
            tcenter[j] = (t->vrt[0]->crd[j] + t->vrt[1]->crd[j] + t->vrt[2]->crd[j]) / 3.;
          }
          for (int j = 0; j < 3; j++) {
            shrinkpoint(&(Vshrink[j]), t->vrt[j], tcenter, shrink);
          }
          drawface(&(Vshrink[0]), &(Vshrink[1]), &(Vshrink[2]));
        }
      } else {
        for (int i = 0; i < tria->tr_tris->used_items; i++) {
          Triang *t = (Triang *) tria->tr_tris->get(i);
          if (t->is_deleted()) continue;
          if (t->is_hulltri()) continue;
          drawface(t->vrt[0], t->vrt[1], t->vrt[2]);
        }
      }
      glEnd();
      if (showTransparent->isChecked()) {
        //glClear(GL_DEPTH_BUFFER_BIT); // Clear Z-Buffer
        //glDepthMask(GL_FALSE);
        glDepthMask(GL_TRUE);
        glDisable(GL_BLEND);
      }
    } // _facetfill

    if (showEdges->isChecked()) {
      double r, g, b, a;
      _edgecolor = object_data->_v3d_edgecolor;
      _edgecolor.getRgbF(&r, &g, &b, &a);
      glColor4f(r, g, b, a); //glColor3dv(clr);
      glLineWidth(_edge_width);
      glBegin(GL_LINES);
      TriEdge E, N;
      for (int i = 0; i < tria->tr_tris->used_items; i++) {
        E.tri = (Triang *) tria->tr_tris->get(i);
        if (E.tri->is_deleted()) continue;
        if (E.tri->is_hulltri()) continue;
        for (E.ver = 0; E.ver < 3; E.ver++) {
          N = E.esym();
          if (N.apex()->idx < E.apex()->idx) {
            drawedge(E.org(), E.dest());
          }
        }
      }
      glEnd();
    } // _show_edge

    /*
    if (showVertices->isChecked()) {
      double r, g, b, a;
      _vertexcolor.getRgbF(&r, &g, &b, &a);
      glColor4f(r, g, b, a); //glColor3dv(clr);
      glPointSize(_vertex_size);
      glEnable(GL_POINT_SMOOTH);
      glBegin(GL_POINTS);
      for (int i = 0; i < object_data->_tria->ct_in_vrts; i++) {
        Vertex *v = &(object_data->_tria->in_vrts[i]);
        //if (v->typ == UNUSEDVERTEX) continue;
        glVertex3d(v->crd[0], v->crd[1], v->crd[2]);
      }
      if (object_data->_tria->tr_steiners != NULL) {
        for (int i = 0; i < object_data->_tria->tr_steiners->objects; i++) {
          Vertex *v = (Vertex *) object_data->_tria->tr_steiners->get(i);
          if (v->is_deleted()) continue;
          glVertex3d(v->crd[0], v->crd[1], v->crd[2]);
        }
      }
      glEnd();
      glDisable(GL_POINT_SMOOTH);
    }
    */
}

void OGLWidget::draw_vertices()
{
    ScribbleArea *object_data = (ScribbleArea *) _scribbleArea;
    Triangulation *tria = object_data->_tria;
    if (object_data->_tria == NULL) {
      tria = object_data->_geom;
    }

    double r, g, b, a;
    _vertexcolor = object_data->_v3d_vertexcolor;
    _vertexcolor.getRgbF(&r, &g, &b, &a);
    glColor4f(r, g, b, a); //glColor3dv(clr);
    glPointSize(_vertex_size);
    glEnable(GL_POINT_SMOOTH);
    glBegin(GL_POINTS);

    //double pz;
    int i;
    for (i = 0; i < tria->ct_in_vrts; i++) {
      Vertex *v = &(tria->in_vrts[i]);
      if (v->typ == UNUSEDVERTEX) continue;
      if (v->idx <= 0) continue; // skip bbox vertices
      //pz = v->crd[0]*v->crd[0] + v->crd[1]*v->crd[1] - v->wei + _zshift;
      glVertex3d(v->crd[0], v->crd[1], 0.);
    }
    if (tria->tr_steiners != NULL) {
      for (i = 0; i < tria->tr_steiners->objects; i++) {
        Vertex *v = (Vertex *) tria->tr_steiners->get(i);
        if (v->is_deleted()) continue;
        //pz = v->crd[0]*v->crd[0] + v->crd[1]*v->crd[1] - v->wei + _zshift;
        glVertex3d(v->crd[0], v->crd[1], 0.);
      }
    }

    glEnd();
    glDisable(GL_POINT_SMOOTH);
}

void  OGLWidget::draw_poles()
{
    ScribbleArea *object_data = (ScribbleArea *) _scribbleArea;
    Triangulation *tria = object_data->_tria;
    if (object_data->_tria == NULL) {
      tria = object_data->_geom;
    }

    double r, g, b, a;
    _polecolor = object_data->_v3d_polecolor;
    _polecolor.getRgbF(&r, &g, &b, &a);
    glColor4f(r, g, b, a); //glColor3dv(clr);
    glPointSize(_vertex_size);
    glEnable(GL_POINT_SMOOTH);
    glBegin(GL_POINTS);

    double x, y, pz;
    int i;
    for (i = 0; i < tria->ct_in_vrts; i++) {
      Vertex *v = &(tria->in_vrts[i]);
      if (v->typ == UNUSEDVERTEX) continue;
      if (v->idx <= 0) continue; // skip bbox vertices
      //pz = v->crd[0]*v->crd[0] + v->crd[1]*v->crd[1] - v->wei + _zshift;
      x = v->crd[0];
      y = v->crd[1];
      pz = object_data->_a11*x*x + 2.*object_data->_a21 *x*y + object_data->_a22*y*y - v->wei + _zshift;
      glVertex3d(v->crd[0], v->crd[1], pz);
    }
    if (tria->tr_steiners != NULL) {
      for (i = 0; i < tria->tr_steiners->objects; i++) {
        Vertex *v = (Vertex *) tria->tr_steiners->get(i);
        if (v->is_deleted()) continue;
        //pz = v->crd[0]*v->crd[0] + v->crd[1]*v->crd[1] - v->wei + _zshift;
        x = v->crd[0];
        y = v->crd[1];
        pz = object_data->_a11*x*x + 2.*object_data->_a21 *x*y + object_data->_a22*y*y - v->wei + _zshift;
        glVertex3d(v->crd[0], v->crd[1], pz);
      }
    }

    glEnd();
    glDisable(GL_POINT_SMOOTH);
}

//==============================================================================
// draw dual (vooronoi) and polar dual.

void  OGLWidget::draw_polars()
{
    ScribbleArea *object_data = (ScribbleArea *) _scribbleArea;
    Triangulation *geom = object_data->_geom;
    Triangulation *tria = object_data->_tria;    

    double a11 = object_data->_a11;
    double a21 = object_data->_a21;
    double a22 = object_data->_a22;
    double x,y; // z

    if (geom->tr_steiners->objects == 0) {
      return; // nothing to show
    } else if (geom->tr_steiners->objects == 1) {
      // draw_polar_1pt()
      Vertex *vt = NULL;
      for (int i = 0; i < geom->tr_steiners->used_items; i++) {
        vt = (Vertex *) geom->tr_steiners->get(i);
        if (vt->is_deleted()) continue;
        break;
      }
      if (vt != NULL) {
        double r, g, b, a;
        _polarcolor = object_data->_v3d_polarcolor;
        _polarcolor.getRgbF(&r, &g, &b, &a);
        glColor4f(r, g, b, a); //glColor3dv(clr);
        glLineWidth(_edge_width);
        glBegin(GL_LINES);
        // The polar plane of vt is:
        //   z = 2ax + 2by - (a^2+b^2) + wei,
        // with metric,
        //   z = (2*a11*a+2*a21*b)*x + (2*a22*b+2*a21*a)*y -
        //       (a11*x*x + 2.*a21*x*y + a22*y*y) + wei
        // where (a,b), wei are the coordinates of vt's x, y, and weight.
        a = vt->crd[0];
        b = vt->crd[1];
        double wei = vt->wei;
        Vertex proj1, proj2;
        proj1.init(); proj2.init();
        TriEdge E, N;
        for (int i = 0; i < geom->tr_tris->used_items; i++) {
          E.tri = (Triang *) geom->tr_tris->get(i);
          if (E.tri->is_deleted()) continue;
          if (E.tri->is_hulltri()) continue;
          for (E.ver = 0; E.ver < 3; E.ver++) {
            N = E.esym();
            if (N.tri->is_hulltri()) {
              if (viewVoronoiAct->isChecked()) {
                proj1.crd[0] = E.org()->crd[0];
                proj1.crd[1] = E.org()->crd[1];
                proj1.crd[2] = 0.;
                proj2.crd[0] = E.dest()->crd[0];
                proj2.crd[1] = E.dest()->crd[1];
                proj2.crd[2] = 0.;
                drawedge(&proj1, &proj2);
              }
              if (viewPolarsAct->isChecked()) {
                // Draw this hull edge by projecting its endpoints to the polar plane of vt.
                x = proj1.crd[0] = E.org()->crd[0];
                y = proj1.crd[1] = E.org()->crd[1];
                //proj1.crd[2] = 2*a*(proj1.crd[0]) + 2*b*(proj1.crd[1]) - (a*a+b*b) + wei;
                proj1.crd[2] = (2*a11*a+2*a21*b)*x + (2*a22*b+2*a21*a)*y
                             - (a11*a*a+ 2.*a21*a*b + a22*b*b) + wei;
                proj1.crd[2] += _zshift;
                x = proj2.crd[0] = E.dest()->crd[0];
                y = proj2.crd[1] = E.dest()->crd[1];
                //proj2.crd[2] = 2*a*(proj2.crd[0]) + 2*b*(proj2.crd[1]) - (a*a+b*b) + wei;
                proj2.crd[2] = (2*a11*a+2*a21*b)*x + (2*a22*b+2*a21*a)*y
                             - (a11*a*a+ 2.*a21*a*b + a22*b*b) + wei;
                proj2.crd[2] += _zshift;
                drawedge(&proj1, &proj2);
              }
            }
          }
        }
        glEnd();
      } // if (vt != NULL)
      return;
    } // else if (geom->tr_steiners->objects == 1)
    else if (geom->tr_steiners->objects == 2) {
      int i;
      Vertex *v1 = NULL;
      for (i = 0; i < geom->tr_steiners->used_items; i++) {
        v1 = (Vertex *) geom->tr_steiners->get(i);
        if (v1->is_deleted()) continue;
        break;
      }
      Vertex *v2 = NULL;
      i++;
      for (; i < geom->tr_steiners->used_items; i++) {
        v2 = (Vertex *) geom->tr_steiners->get(i);
        if (v2->is_deleted()) continue;
        break;
      }
      if ((v1 != NULL) && (v2 != NULL)) {
        // detri2_draw_voronoi_2pts(...)
        Vertex mid;
        mid.init();
        // set the metric.
        geom->_a11 = a11;
        geom->_a21 = a21;
        geom->_a22 = a22;
        //
        get_bissector(v1->crd[0], v1->crd[1], v1->wei,
                            v2->crd[0], v2->crd[1], v2->wei,
                            &(mid.crd[0]), &(mid.crd[1]), &(mid.crd[2]),
                            a11, a21, a22);

        double outerN[2], len;
        /*
        outerN[0] =  (v1->crd[1] - v2->crd[1]);
        outerN[1] = -(v1->crd[0] - v2->crd[0]);
        len = sqrt(outerN[0]*outerN[0] + outerN[1]*outerN[1]);
        outerN[0] /= len;
        outerN[1] /= len;
        */

        // Calculate the bisector line (may not be orthogonal if metric is not Euclidean).
        // The line direction vector.
        double Ux = v1->crd[0], Uy = v1->crd[1];
        double Vx = v2->crd[0], Vy = v2->crd[1];
        outerN[0] =   2.*a22*(Uy-Vy) + 2.*a21*(Ux-Vx);
        outerN[1] = -(2.*a11*(Ux-Vx) + 2.*a21*(Uy-Vy));
        len = sqrt(outerN[0]*outerN[0] + outerN[1]*outerN[1]);
        outerN[0] /= len;
        outerN[1] /= len;
        // Choose the line direction towards exterior.
        Vertex tmp_pt; // Construct a point on the bisector line.
        tmp_pt.init();
        tmp_pt.crd[0] = mid.crd[0] + 2 * len * outerN[0];
        tmp_pt.crd[1] = mid.crd[1] + 2 * len * outerN[1];
        REAL check_sign = Orient2d(v1, v2, &tmp_pt);
        if (check_sign < 0) {
          // not ccw orientation. reverse the direction.
          outerN[0] = -outerN[0];
          outerN[1] = -outerN[1];
        } else if (check_sign == 0) {
          assert(0); // this is a bug.
        }

        Vertex far1, far2;
        far1.init();
        far2.init();
        double far_dist = geom->io_diagonal * 2;
        far1.crd[0] = mid.crd[0] + outerN[0] * far_dist;
        far1.crd[1] = mid.crd[1] + outerN[1] * far_dist;
        far1.crd[2] = 0.;
        far2.crd[0] = mid.crd[0] - outerN[0] * far_dist;
        far2.crd[1] = mid.crd[1] - outerN[1] * far_dist;
        far2.crd[2] = 0.;

        TriEdge S_cut1, S_cut2; // The bdry edges of geom contain cut points
        Vertex cut1, cut2;
        cut1.init();
        cut2.init();

        int loc = geom->locate_point(&mid, S_cut1, 0);
        //assert(loc == LOC_IN_TRI);
        loc = geom->locate_point(&far1, S_cut1, 0);
        geom->get_boundary_cut_dualedge(mid, far1, S_cut1);

        // Calculate the cut point.
        Vertex *e1 = S_cut1.org();
        Vertex *e2 = S_cut1.dest();
        double X0 = mid.crd[0];
        double Y0 = mid.crd[1];
        double X1 = far1.crd[0];
        double Y1 = far1.crd[1];
        double X2 = e1->crd[0];
        double Y2 = e1->crd[1];
        double X3 = e2->crd[0];
        double Y3 = e2->crd[1];
        double t1 = 0., t2 = 0.;
        line_line_intersection(X0, Y0, X1, Y1, X2, Y2, X3, Y3, &t1, &t2);
        cut1.crd[0] = X0 + t1 * (X1 - X0);
        cut1.crd[1] = Y0 + t1 * (Y1 - Y0);
        cut1.crd[2] = 0.;

        loc = geom->locate_point(&mid, S_cut2, 0);
        //assert(loc == LOC_IN_TRI);
        loc = geom->locate_point(&far2, S_cut2, 0);
        geom->get_boundary_cut_dualedge(mid, far2, S_cut2);

        e1 = S_cut2.org();
        e2 = S_cut2.dest();
        X0 = mid.crd[0];
        Y0 = mid.crd[1];
        X1 = far2.crd[0];
        Y1 = far2.crd[1];
        X2 = e1->crd[0];
        Y2 = e1->crd[1];
        X3 = e2->crd[0];
        Y3 = e2->crd[1];
        line_line_intersection(X0, Y0, X1, Y1, X2, Y2, X3, Y3, &t1, &t2);
        cut2.crd[0] = X0 + t1 * (X1 - X0);
        cut2.crd[1] = Y0 + t1 * (Y1 - Y0);
        cut2.crd[2] = 0.;

        double r, g, b, a;
        _polarcolor = object_data->_v3d_polarcolor;
        _polarcolor.getRgbF(&r, &g, &b, &a);
        glColor4f(r, g, b, a); //glColor3dv(clr);
        glLineWidth(_edge_width);
        glBegin(GL_LINES);

        // for debugging.
        //drawedge(&far1, &far2);

        // The polar plane of vt is:
        //   z = 2ax + 2by - (a^2+b^2) + wei,
        // where (a,b), wei are the coordinates of vt's x, y, and weight.
        a = v1->crd[0];
        b = v1->crd[1];
        double wei = v1->wei;
        Vertex proj1, proj2;
        proj1.init(); proj2.init();

        if (viewVoronoiAct->isChecked()) {
          drawedge(&cut1, &cut2);
        }
        if (viewPolarsAct->isChecked()) {
          x = proj1.crd[0] = cut1.crd[0];
          y = proj1.crd[1] = cut1.crd[1];
          //proj1.crd[2] = 2*a*(proj1.crd[0]) + 2*b*(proj1.crd[1]) - (a*a+b*b) + wei;
          proj1.crd[2] = (2*a11*a+2*a21*b)*x + (2*a22*b+2*a21*a)*y -
                         (a11*a*a + 2.*a21*a*b + a22*b*b) + wei;
          proj1.crd[2] += _zshift;
          x = proj2.crd[0] = cut2.crd[0];
          y = proj2.crd[1] = cut2.crd[1];
          //proj2.crd[2] = 2*a*(proj2.crd[0]) + 2*b*(proj2.crd[1]) - (a*a+b*b) + wei;
          proj2.crd[2] = (2*a11*a+2*a21*b)*x + (2*a22*b+2*a21*a)*y -
                         (a11*a*a + 2.*a21*a*b + a22*b*b) + wei;
          proj2.crd[2] += _zshift;
          drawedge(&proj1, &proj2);
        }

        // Get the power cell of vertex v1.
        // Take a walk on the convex hull.
        // Starting from startE and ending at endE.
        TriEdge startEdge, endEdge, searchEdge;
        startEdge = S_cut1.esym();
        endEdge = S_cut2.esym();
        Vertex *ptlist = new Vertex[geom->ct_in_vrts+2];
        ptlist[0].init();
        ptlist[0].crd[0] = cut1.crd[0];
        ptlist[0].crd[1] = cut1.crd[1];
        ptlist[0].crd[2] = 0.;
        int ptcount = 1;
        searchEdge = startEdge;
        while (searchEdge.tri != endEdge.tri) {
          if (geom->op_db_verbose > 3) {
            printf("  Hull searchE: (%d,%d,%d)\n", searchEdge.org()->idx,
                   searchEdge.dest()->idx, searchEdge.apex()->idx);
          }
          // Found a corner vertex.
          Vertex *pt = &(ptlist[ptcount]);
          pt->init();
          pt->crd[0] = searchEdge.dest()->crd[0];
          pt->crd[1] = searchEdge.dest()->crd[1];
          pt->crd[2] = 0.;
          ptcount++;
          // Go to the next boundary edge.
          TriEdge workedge = searchEdge.enext();
          while (true) {
            searchEdge = workedge.esym_enext(); // CW
            if (searchEdge.is_segment()) break;
            if ((searchEdge.esym()).tri->is_hulltri()) break;
            if (geom->op_db_verbose > 3) {
              printf("  Hull workE: (%d,%d,%d)\n", workedge.org()->idx,
                     workedge.dest()->idx, workedge.apex()->idx);
            }
            workedge = searchEdge;
          }
        }
        ptlist[ptcount].init();
        ptlist[ptcount].crd[0] = cut2.crd[0];
        ptlist[ptcount].crd[1] = cut2.crd[1];
        ptlist[ptcount].crd[2] = 0.;
        ptcount++;

        for (i = 0; i < ptcount - 1; i++) {
          if (viewVoronoiAct->isChecked()) {
            drawedge(&(ptlist[i]), &(ptlist[i+1]));
          }
          if (viewPolarsAct->isChecked()) {
            x = proj1.crd[0] = ptlist[i].crd[0];
            y = proj1.crd[1] = ptlist[i].crd[1];
            //proj1.crd[2] = 2*a*(proj1.crd[0]) + 2*b*(proj1.crd[1]) - (a*a+b*b) + wei;
            proj1.crd[2] = (2.*a11*a + 2.*a21*b)*x +
                           (2.*a22*b + 2.*a21*a)*y -
                           (a11*a*a+ 2*a21*a*b +a22*b*b) + wei;
            proj1.crd[2] += _zshift;
            x = proj2.crd[0] = ptlist[i+1].crd[0];
            y = proj2.crd[1] = ptlist[i+1].crd[1];
            //proj2.crd[2] = 2*a*(proj2.crd[0]) + 2*b*(proj2.crd[1]) - (a*a+b*b) + wei;
            proj2.crd[2] = (2.*a11*a + 2.*a21*b)*x +
                           (2.*a22*b + 2.*a21*a)*y -
                           (a11*a*a+ 2*a21*a*b +a22*b*b) + wei;
            proj2.crd[2] += _zshift;
            drawedge(&proj1, &proj2);
          }
        }

        // The polar plane of vt is:
        //   z = 2ax + 2by - (a^2+b^2) + wei,
        // where (a,b), wei are the coordinates of vt's x, y, and weight.
        a = v2->crd[0];
        b = v2->crd[1];
        wei = v2->wei;
        //Vertex proj1, proj2;
        //proj1.init(); proj2.init();

        // Get the power cell of vertex v2.
        // Take a walk on the convex hull.
        // Starting from startE and ending at endE.
        startEdge = S_cut2.esym();
        endEdge = S_cut1.esym();
        //Vertex *ptlist = new Vertex[geom->ct_in_vrts+2];
        ptlist[0].init();
        ptlist[0].crd[0] = cut2.crd[0];
        ptlist[0].crd[1] = cut2.crd[1];
        ptlist[0].crd[2] = 0.;
        ptcount = 1;
        searchEdge = startEdge;
        while (searchEdge.tri != endEdge.tri) {
          if (geom->op_db_verbose > 3) {
            printf("  Hull searchE: (%d,%d,%d)\n", searchEdge.org()->idx,
                   searchEdge.dest()->idx, searchEdge.apex()->idx);
          }
          // Found a corner vertex.
          Vertex *pt = &(ptlist[ptcount]);
          pt->init();
          pt->crd[0] = searchEdge.dest()->crd[0];
          pt->crd[1] = searchEdge.dest()->crd[1];
          pt->crd[2] = 0.;
          ptcount++;
          // Go to the next boundary edge.
          TriEdge workedge = searchEdge.enext();
          while (true) {
            searchEdge = workedge.esym_enext(); // CW
            if (searchEdge.is_segment()) break;
            if ((searchEdge.esym()).tri->is_hulltri()) break;
            if (geom->op_db_verbose > 3) {
              printf("  Hull workE: (%d,%d,%d)\n", workedge.org()->idx,
                     workedge.dest()->idx, workedge.apex()->idx);
            }
            workedge = searchEdge;
          }
        }
        ptlist[ptcount].init();
        ptlist[ptcount].crd[0] = cut1.crd[0];
        ptlist[ptcount].crd[1] = cut1.crd[1];
        ptlist[ptcount].crd[2] = 0.;
        ptcount++;

        for (i = 0; i < ptcount - 1; i++) {
          if (viewVoronoiAct->isChecked()) {
            drawedge(&(ptlist[i]), &(ptlist[i+1]));
          }
          if (viewPolarsAct->isChecked()) {
            x = proj1.crd[0] = ptlist[i].crd[0];
            y = proj1.crd[1] = ptlist[i].crd[1];
            //proj1.crd[2] = 2*a*(proj1.crd[0]) + 2*b*(proj1.crd[1]) - (a*a+b*b) + wei;
            proj1.crd[2] = (2*a11*a+2*a21*b)*x + (2*a22*b+2*a21*a)*y -
                           (a11*a*a + 2.*a21*a*b + a22*b*b) + wei;
            proj1.crd[2] += _zshift;
            x = proj2.crd[0] = ptlist[i+1].crd[0];
            y = proj2.crd[1] = ptlist[i+1].crd[1];
            //proj2.crd[2] = 2*a*(proj2.crd[0]) + 2*b*(proj2.crd[1]) - (a*a+b*b) + wei;
           proj2.crd[2] = (2*a11*a+2*a21*b)*x + (2*a22*b+2*a21*a)*y -
                          (a11*a*a + 2.*a21*a*b + a22*b*b) + wei;
            proj2.crd[2] += _zshift;
            drawedge(&proj1, &proj2);
          }
        }

        delete [] ptlist;
        glEnd();
        // restore the background metric.
        geom->_a11 = 1.0;
        geom->_a21 = 0.0;
        geom->_a22 = 1.0;
      } // if ((v1 != NULL) && (v2 != NULL))
    } // else if (geom->tr_steiners->objects == 2)
    else {
      // There are more than 3 vertices.
      if (tria == NULL) {
        detri2_Delaunay(geom, &tria, 0, 1, 0,
                        a11, a21, a22); // furtherst = 0, weighted = 1, cdt = 0
      }
      tria->OMT_domain = object_data->_geom;

      // Calculate Voronoi vertices for triangles.
      int i, idx;
      idx = tria->io_firstindex;
      // Calculate circumcenters for hull triangles.
      for (i = 0; i < tria->tr_tris->used_items; i++) {
        Triang* tri = (Triang *) tria->tr_tris->get(i);
        // Ignore exterior triangles.
        if (tri->is_deleted() || tri->is_hulltri()) continue;
        tria->get_tri_orthocenter(tri);
        tri->idx = idx;
        idx++;
      }
      //Tr->op_db_verbose=10; // debug
      // Calculate bisectors for hull triangles.
      for (i = 0; i < tria->tr_tris->used_items; i++) {
        Triang* tri = (Triang *) tria->tr_tris->get(i);
        // Ignore exterior triangles.
        if (tri->is_deleted()) continue;
        if (tri->is_hulltri()) { // A hull triangle.
          tria->get_hulltri_orthocenter(tri);
          tri->idx = idx; // hulltri is also indexed.
          idx++;
        }
      }

      double r, g, b, a;
      _polarcolor = object_data->_v3d_polarcolor;
      _polarcolor.getRgbF(&r, &g, &b, &a);
      glColor4f(r, g, b, a); //glColor3dv(clr);
      glLineWidth(_edge_width);
      glBegin(GL_LINES);

      for (i = 0; i < tria->ct_in_vrts; i++) {
        Vertex *mesh_vertex = &(tria->in_vrts[i]);
        if (mesh_vertex->typ == UNUSEDVERTEX) continue;
        Vertex *ptlist = NULL;
        int ptnum = 0;
        if (tria->get_powercell(mesh_vertex, &ptlist, &ptnum)) {
          /*
          if (viewVoronoiCellAct->isChecked()) {
            double r, g, b, a;
            _voronoicellcolor.getRgbF(&r, &g, &b, &a);
            glColor4f(r, g, b, 0.45); // a=1 no transparency
            if (showTransparent->isChecked()) {
              glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
              glAlphaFunc(GL_GREATER,0.1);
              glEnable(GL_BLEND);
              glDepthMask(GL_FALSE);
            }
            glBegin(GL_TRIANGLES);
            // To draw...
            glEnd
            if (showTransparent->isChecked()) {
              //glClear(GL_DEPTH_BUFFER_BIT); // Clear Z-Buffer
              //glDepthMask(GL_FALSE);
              glDepthMask(GL_TRUE);
              glDisable(GL_BLEND);
            }
          } // if (showFacets->isChecked())
          */
          //if (viewPolarsAct->isChecked()) {            
            //double r, g, b, a;
            //_polarcolor.getRgbF(&r, &g, &b, &a);
            //glColor4f(r, g, b, a); //glColor3dv(clr);
            //glLineWidth(_edge_width);
            //glBegin(GL_LINES);
            a = mesh_vertex->crd[0];
            b = mesh_vertex->crd[1];
            double wei = mesh_vertex->wei;

            Vertex *pt1, *pt2;
            Vertex Vproj1, Vproj2;
            Vproj1.init();
            Vproj2.init();
            double x, y, w;
            int j;
            if (viewVoronoiAct->isChecked()) {
              for (j = 0; j < ptnum - 1; j++) {
                pt1 = &(ptlist[j]);
                pt2 = &(ptlist[j+1]);
                Vproj1.crd[0] = pt1->crd[0];
                Vproj1.crd[1] = pt1->crd[1];
                Vproj1.crd[2] = 0.0;
                Vproj2.crd[0] = pt2->crd[0];
                Vproj2.crd[1] = pt2->crd[1];
                Vproj2.crd[2] = 0.0;
                drawedge(&Vproj1, &Vproj2);
              }
              pt1 = &(ptlist[j]);
              pt2 = &(ptlist[0]);
              Vproj1.crd[0] = pt1->crd[0];
              Vproj1.crd[1] = pt1->crd[1];
              Vproj1.crd[2] = 0.0;
              Vproj2.crd[0] = pt2->crd[0];
              Vproj2.crd[1] = pt2->crd[1];
              Vproj2.crd[2] = 0.0;
              drawedge(&Vproj1, &Vproj2);
            } // if (viewVoronoiAct->isChecked())
            if (viewPolarsAct->isChecked()) {
              for (j = 0; j < ptnum - 1; j++) {
                pt1 = &(ptlist[j]);
                pt2 = &(ptlist[j+1]);
                x = Vproj1.crd[0] = pt1->crd[0];
                y = Vproj1.crd[1] = pt1->crd[1];
                w = pt1->crd[2]; // weight
                //Vproj1.crd[2] = x*x + y*y - w +_zshift;
                Vproj1.crd[2] = (2*a11*a+2*a21*b)*x + (2*a22*b+2*a21*a)*y -
                                (a11*a*a + 2.*a21*a*b + a22*b*b) + wei;
                Vproj1.crd[2] += _zshift;
                x = Vproj2.crd[0] = pt2->crd[0];
                y = Vproj2.crd[1] = pt2->crd[1];
                //w = pt2->crd[2]; // weight
                //Vproj2.crd[2] = x*x + y*y - w +_zshift;
                Vproj2.crd[2] = (2*a11*a+2*a21*b)*x + (2*a22*b+2*a21*a)*y -
                                (a11*a*a + 2.*a21*a*b + a22*b*b) + wei;
                Vproj2.crd[2] += _zshift;
                drawedge(&Vproj1, &Vproj2);
              }
              pt1 = &(ptlist[j]);
              pt2 = &(ptlist[0]);
              x = Vproj1.crd[0] = pt1->crd[0];
              y = Vproj1.crd[1] = pt1->crd[1];
              w = pt1->crd[2]; // weight
              //Vproj1.crd[2] = x*x + y*y - w +_zshift;
              Vproj1.crd[2] = (2*a11*a+2*a21*b)*x + (2*a22*b+2*a21*a)*y -
                              (a11*a*a + 2.*a21*a*b + a22*b*b) + wei;
              Vproj1.crd[2] += _zshift;
              x = Vproj2.crd[0] = pt2->crd[0];
              y = Vproj2.crd[1] = pt2->crd[1];
              //w = pt2->crd[2]; // weight
              //Vproj2.crd[2] = x*x + y*y - w +_zshift;
              Vproj2.crd[2] = (2*a11*a+2*a21*b)*x + (2*a22*b+2*a21*a)*y -
                      (a11*a*a + 2.*a21*a*b + a22*b*b) + wei;
              Vproj2.crd[2] += _zshift;
              drawedge(&Vproj1, &Vproj2);
            } // if (viewPolarsAct->isChecked())
            //glEnd();
          //} // if (viewPolarsAct->isChecked()) {
          delete [] ptlist;
        }
      }

      if (tria->tr_steiners != NULL) {
        for (i = 0; i < tria->tr_steiners->used_items; i++) {
          Vertex *mesh_vertex = (Vertex *) tria->tr_steiners->get(i);
          if (mesh_vertex->is_deleted()) continue;
          Vertex *ptlist = NULL;
          int ptnum = 0;
          if (tria->get_powercell(mesh_vertex, &ptlist, &ptnum)) {
            /*
            if (viewVoronoiCellAct->isChecked()) {
              double r, g, b, a;
              _voronoicellcolor.getRgbF(&r, &g, &b, &a);
              glColor4f(r, g, b, 0.45); // a=1 no transparency
              if (showTransparent->isChecked()) {
                glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
                glAlphaFunc(GL_GREATER,0.1);
                glEnable(GL_BLEND);
                glDepthMask(GL_FALSE);
              }
              glBegin(GL_TRIANGLES);
              // To draw...
              glEnd
              if (showTransparent->isChecked()) {
                //glClear(GL_DEPTH_BUFFER_BIT); // Clear Z-Buffer
                //glDepthMask(GL_FALSE);
                glDepthMask(GL_TRUE);
                glDisable(GL_BLEND);
              }
            } // if (showFacets->isChecked())
            */
            //if (viewPolarsAct->isChecked()) {
              //double r, g, b, a;
              //_polarcolor.getRgbF(&r, &g, &b, &a);
              //glColor4f(r, g, b, a); //glColor3dv(clr);
              //glLineWidth(_edge_width);
              //glBegin(GL_LINES);
              a = mesh_vertex->crd[0];
              b = mesh_vertex->crd[1];
              double wei = mesh_vertex->wei;

              Vertex *pt1, *pt2;
              Vertex Vproj1, Vproj2;
              Vproj1.init();
              Vproj2.init();
              double x, y, w;
              int j;
              if (viewVoronoiAct->isChecked()) {
                for (j = 0; j < ptnum - 1; j++) {
                  pt1 = &(ptlist[j]);
                  pt2 = &(ptlist[j+1]);
                  Vproj1.crd[0] = pt1->crd[0];
                  Vproj1.crd[1] = pt1->crd[1];
                  Vproj1.crd[2] = 0.0;
                  Vproj2.crd[0] = pt2->crd[0];
                  Vproj2.crd[1] = pt2->crd[1];
                  Vproj2.crd[2] = 0.0;
                  drawedge(&Vproj1, &Vproj2);
                }
                pt1 = &(ptlist[j]);
                pt2 = &(ptlist[0]);
                Vproj1.crd[0] = pt1->crd[0];
                Vproj1.crd[1] = pt1->crd[1];
                Vproj1.crd[2] = 0.0;
                Vproj2.crd[0] = pt2->crd[0];
                Vproj2.crd[1] = pt2->crd[1];
                Vproj2.crd[2] = 0.0;
                drawedge(&Vproj1, &Vproj2);
              } // if (viewVoronoiAct->isChecked())
              if (viewPolarsAct->isChecked()) {
                for (j = 0; j < ptnum - 1; j++) {
                  pt1 = &(ptlist[j]);
                  pt2 = &(ptlist[j+1]);
                  x = Vproj1.crd[0] = pt1->crd[0];
                  y = Vproj1.crd[1] = pt1->crd[1];
                  w = pt1->crd[2]; // weight
                  //Vproj1.crd[2] = x*x + y*y - w +_zshift;
                  Vproj1.crd[2] = (2*a11*a+2*a21*b)*x + (2*a22*b+2*a21*a)*y -
                          (a11*a*a + 2.*a21*a*b + a22*b*b) + wei;
                  Vproj1.crd[2] += _zshift;
                  x = Vproj2.crd[0] = pt2->crd[0];
                  y = Vproj2.crd[1] = pt2->crd[1];
                  w = pt2->crd[2]; // weight
                  //Vproj2.crd[2] = x*x + y*y - w +_zshift;
                  Vproj2.crd[2] = (2*a11*a+2*a21*b)*x + (2*a22*b+2*a21*a)*y -
                          (a11*a*a + 2.*a21*a*b + a22*b*b) + wei;
                  Vproj2.crd[2] += _zshift;
                  drawedge(&Vproj1, &Vproj2);
                }
                pt1 = &(ptlist[j]);
                pt2 = &(ptlist[0]);
                x = Vproj1.crd[0] = pt1->crd[0];
                y = Vproj1.crd[1] = pt1->crd[1];
                w = pt1->crd[2]; // weight
                //Vproj1.crd[2] = x*x + y*y - w +_zshift;
                Vproj1.crd[2] = (2*a11*a+2*a21*b)*x + (2*a22*b+2*a21*a)*y -
                        (a11*a*a + 2.*a21*a*b + a22*b*b) + wei;
                Vproj1.crd[2] += _zshift;
                x = Vproj2.crd[0] = pt2->crd[0];
                y = Vproj2.crd[1] = pt2->crd[1];
                w = pt2->crd[2]; // weight
                //Vproj2.crd[2] = x*x + y*y - w +_zshift;
                Vproj2.crd[2] = (2*a11*a+2*a21*b)*x + (2*a22*b+2*a21*a)*y -
                        (a11*a*a + 2.*a21*a*b + a22*b*b) + wei;
                Vproj2.crd[2] += _zshift;
                drawedge(&Vproj1, &Vproj2);
              } // if (viewPolarsAct->isChecked())
              //glEnd();
            //} // if (viewPolarsAct->isChecked()) {
            delete [] ptlist;
          }
        }
      } // if (tr_steiners != NULL)
      glEnd();
    } // tria
}

void  OGLWidget::draw_weights()
{
    ScribbleArea *object_data = (ScribbleArea *) _scribbleArea;
    Triangulation *tria = object_data->_tria;
    if (object_data->_tria == NULL) {
      tria = object_data->_geom;
    }

    glBegin(GL_LINES);

    double R, G, B, A;
    _weightcolor = object_data->_v3d_weightcolor;
    _weightcolor.getRgbF(&R, &G, &B, &A);
    glColor3f(R, G, B); // a=1 no transparency
    glLineWidth(2); // _edge_width

    int segnum = 36;
    int np = segnum + 1;
    double ang_start = 0.; // ang1 / 180. * detri2::PI;
    double dfi = 360. / segnum; // fabs(ang2 - ang1) / segnum;
    dfi = dfi / 180.0 * detri2::PI;
    double px, py, zlift;

    int i, j;

    for (i = 0; i < tria->ct_in_vrts; i++) {
      Vertex *mesh_vertex = &(tria->in_vrts[i]);
      if (mesh_vertex->typ == UNUSEDVERTEX) continue;
      if (mesh_vertex->idx <= 0) continue; // Skip bounding box vertices.
      double wei = fabs(mesh_vertex->wei);
      if (wei > 0) {
        double r = sqrt(wei);
        for (j = 0; j < np; j++) {
          px = r * cos(ang_start + dfi * (double) j) + mesh_vertex->crd[0];
          py = r * sin(ang_start + dfi * (double) j) + mesh_vertex->crd[1];
          glVertex3d(px, py, 0.0);
          px = r * cos(ang_start + dfi * (double) (j+1)) + mesh_vertex->crd[0];
          py = r * sin(ang_start + dfi * (double) (j+1)) + mesh_vertex->crd[1];
          glVertex3d(px, py, 0.0);
        }
        if (viewPolarsAct->isChecked()) {
          for (j = 0; j < np; j++) {
            px = r * cos(ang_start + dfi * (double) j) + mesh_vertex->crd[0];
            py = r * sin(ang_start + dfi * (double) j) + mesh_vertex->crd[1];
            zlift = px*px + py*py + _zshift;
            glVertex3d(px, py, zlift);
            px = r * cos(ang_start + dfi * (double) (j+1)) + mesh_vertex->crd[0];
            py = r * sin(ang_start + dfi * (double) (j+1)) + mesh_vertex->crd[1];
            zlift = px*px + py*py + _zshift;
            glVertex3d(px, py, zlift);
          }
        }
      } // if (wei > 0)
    }

    if (tria->tr_steiners != NULL) {
      for (i = 0; i < tria->tr_steiners->used_items; i++) {
        Vertex *mesh_vertex = (Vertex *) tria->tr_steiners->get(i);
        if (mesh_vertex->is_deleted()) continue;
        double wei = fabs(mesh_vertex->wei);
        if (wei > 0) {
          double r = sqrt(wei);
          int np = segnum + 1;
          for (j = 0; j < np; j++) {
            px = r * cos(ang_start + dfi * (double) j) + mesh_vertex->crd[0];
            py = r * sin(ang_start + dfi * (double) j) + mesh_vertex->crd[1];
            glVertex3d(px, py, 0.0);
            px = r * cos(ang_start + dfi * (double) (j+1)) + mesh_vertex->crd[0];
            py = r * sin(ang_start + dfi * (double) (j+1)) + mesh_vertex->crd[1];
            glVertex3d(px, py, 0.0);
          }
          if (viewPolarsAct->isChecked()) {
            for (j = 0; j < np; j++) {
              px = r * cos(ang_start + dfi * (double) j) + mesh_vertex->crd[0];
              py = r * sin(ang_start + dfi * (double) j) + mesh_vertex->crd[1];
              zlift = px*px + py*py + _zshift;
              glVertex3d(px, py, zlift);
              px = r * cos(ang_start + dfi * (double) (j+1)) + mesh_vertex->crd[0];
              py = r * sin(ang_start + dfi * (double) (j+1)) + mesh_vertex->crd[1];
              zlift = px*px + py*py + _zshift;
              glVertex3d(px, py, zlift);
            }
          }
        } // if (wei > 0)
      }
    } // if (tria->tr_steiners != NULL)

    glEnd();
}

void  OGLWidget::draw_circle()
{
    qDebug()<<"draw circle: cx="<<_ccent[0]<<",cy="<<_ccent[1]<<",r="<<_radius;
    if (_radius <= 0.0) {
      return; // Do not draw it.
    }

    glBegin(GL_LINES);

    double R, G, B, A;
    _circlecolor = ((ScribbleArea *) _scribbleArea)->_v3d_circlecolor;
    _circlecolor.getRgbF(&R, &G, &B, &A);
    glColor3f(R, G, B); // a=1 no transparency
    glLineWidth(_edge_width);

    int segnum = 72;
    double ang_start = 0.; // ang1 / 180. * detri2::PI;
    double dfi = 360. / segnum; // fabs(ang2 - ang1) / segnum;
    dfi = dfi / 180.0 * detri2::PI;

    int i;
    int np = segnum + 1;
    double *ptlist = new double[np * 2];
    for (i = 0; i < np; i++) {
      double *pp = &(ptlist[i*2]);
      pp[0] = _radius * cos(ang_start + dfi * (double) i) + _ccent[0];
      pp[1] = _radius * sin(ang_start + dfi * (double) i) + _ccent[1];
    }

    //int *elist = new int[segnum * 2];
    //for (int i = 0; i < segnum; i++) {
    //  elist[i*2] = i;
    //  elist[i*2+1] = i+1;
    //}

    if (viewprojectionAct->isChecked()) {
      for (i = 0; i < segnum; i++) {
        double *pp1 = &(ptlist[i*2]);
        double *pp2 = &(ptlist[(i+1)*2]);
        glVertex3d(pp1[0], pp1[1], 0.0);
        glVertex3d(pp2[0], pp2[1], 0.0);
      }
    }

    if (viewlifttransformAct->isChecked()) {
      for (i = 0; i < segnum; i++) {
        double *pp1 = &(ptlist[i*2]);
        double *pp2 = &(ptlist[(i+1)*2]);
        glVertex3d(pp1[0], pp1[1], pp1[0]*pp1[0]+pp1[1]*pp1[1]+_zshift);
        glVertex3d(pp2[0], pp2[1], pp2[0]*pp2[0]+pp2[1]*pp2[1]+_zshift);
      }
    }

    glEnd();
}

void OGLWidget::draw_paraboloid()
{
    ScribbleArea *SA = (ScribbleArea *)_scribbleArea;
    if (_disc == NULL) return;

    glEnable(GL_BLEND);
    //glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
    glAlphaFunc(GL_GREATER,0.1);
    glEnable(GL_BLEND);
    glDepthMask(GL_FALSE);
    glColor4d(0.5,0.5,0.5,0.5); // a = 1.0 (no transparency)

    Vertex vrt[3];
    double x, y;

    glBegin(GL_TRIANGLES);
    for (int i = 0; i < _disc->tr_tris->used_items; i++) {
      Triang *t = (Triang *) _disc->tr_tris->get(i);
      if (t->is_deleted()) continue;
      if (t->is_hulltri()) continue;
      //qDebug()<<"draw tri: "<<t->vrt[0]->idx<<","<<t->vrt[1]->idx<<","<<t->vrt[2]->idx;
      for (int j = 0; j < 3; j++) {
        x = vrt[j].crd[0] = t->vrt[j]->crd[0];
        y = vrt[j].crd[1] = t->vrt[j]->crd[1];
        //vrt[j].crd[2] = t->vrt[j]->crd[2] + _zshift;
        vrt[j].crd[2] = SA->_a11*x*x + 2*SA->_a21*x*y + SA->_a22*y*y + _zshift;
      }
      //drawface(t->vrt[0], t->vrt[1], t->vrt[2]);
      drawface(&(vrt[0]), &(vrt[1]), &(vrt[2]));
    }
    glEnd();

    glDepthMask(GL_TRUE);
    glDisable(GL_BLEND);
    //glDisable(GL_BLEND);
}

void OGLWidget::draw_scene()
{
    printf("debug:  +++++draw_scene()+++++ is called.\n");

    if (viewParaboildAct->isChecked()) {
      draw_paraboloid();
    }

    if (showVertices->isChecked()) {
      draw_vertices();
    }

    if (viewPolesAct->isChecked()) {
      draw_poles();
    }

    if (viewVoronoiAct->isChecked() ||
        viewPolarsAct->isChecked()) {
      draw_polars();
    }

    if (viewWeightsAct->isChecked()) {
      draw_weights();
    }

    if (viewCircleAct->isChecked()) {
      draw_circle();
    }

    ScribbleArea *object_data = (ScribbleArea *) _scribbleArea;
    if (object_data->_tria == NULL) {
      printf("No triangulation to draw.\n");
      return;
    }

    Triangulation *tria = object_data->_tria;

    if (viewlifttransformAct->isChecked()) { // f
      for (int i = 0; i < tria->ct_in_vrts; i++) {
        Vertex *v = &(tria->in_vrts[i]);
        double x = v->crd[0];
        double y = v->crd[1];
        double z = tria->_a11 *x*x + 2.*tria->_a21*x*y + tria->_a22*y*y;
        v->crd[2] = z + _zshift;
      }
      if (tria->tr_steiners != NULL) {
        for (int i = 0; i < tria->tr_steiners->used_items; i++) {
          Vertex *v = (Vertex *) tria->tr_steiners->get(i);
          //if (v->is_deleted()) continue;
          double x = v->crd[0];
          double y = v->crd[1];
          double z = tria->_a11 *x*x + 2.*tria->_a21*x*y + tria->_a22*y*y;
          v->crd[2] = z + _zshift;
        }
      }
      _facetcolor = object_data->_v3d_facetcolor; // Qt::green;
      draw_triangulation();
      //_facetcolor = object_data->_v3d_facetcolor; // Qt::cyan; // default
    } /*
    else {
      if (viewfunctionAct->isChecked()) {
        for (int i = 0; i < tria->ct_in_vrts; i++) {
          Vertex *v = &(tria->in_vrts[i]);
          v->crd[2] = v->mtr[0] * object_data->_tria->op_hde_s1;
        }
        if (tria->tr_steiners != NULL) {
          for (int i = 0; i < tria->tr_steiners->used_items; i++) {
            Vertex *v = (Vertex *) tria->tr_steiners->get(i);
            //if (v->is_deleted()) continue;
            v->crd[2] = v->mtr[0] * object_data->_tria->op_hde_s1;
          }
        }
        draw_triangulation();
      }
      if (viewfunctionXAct->isChecked()) { // fx
        for (int i = 0; i < tria->ct_in_vrts; i++) {
          Vertex *v = &(tria->in_vrts[i]);
          v->crd[2] = v->mtr[1] * object_data->_tria->op_hde_s2;
        }
        if (tria->tr_steiners != NULL) {
          for (int i = 0; i < tria->tr_steiners->used_items; i++) {
            Vertex *v = (Vertex *) tria->tr_steiners->get(i);
            //if (v->is_deleted()) continue;
            v->crd[2] = v->mtr[1] * object_data->_tria->op_hde_s2;
          }
        }
        _facetcolor = Qt::green;
        draw_triangulation();
      }
      if (viewfunctionYAct->isChecked()) { // fy
        for (int i = 0; i < tria->ct_in_vrts; i++) {
          Vertex *v = &(tria->in_vrts[i]);
          v->crd[2] = v->mtr[2] * object_data->_tria->op_hde_s2;
        }
        if (tria->tr_steiners != NULL) {
          for (int i = 0; i < tria->tr_steiners->used_items; i++) {
            Vertex *v = (Vertex *) tria->tr_steiners->get(i);
            //if (v->is_deleted()) continue;
            v->crd[2] = v->mtr[2] * object_data->_tria->op_hde_s2;
          }
        }
        _facetcolor = Qt::yellow;
        draw_triangulation();
      }
      _facetcolor = Qt::cyan; // default
    }
    */
    // if (viewer->viewprojectionAct)

    if (viewprojectionAct->isChecked()) {
      for (int i = 0; i < tria->ct_in_vrts; i++) {
        Vertex *v = &(tria->in_vrts[i]);
        v->crd[2] = 0.; //_min[2]; // zmin;
      }
      if (tria->tr_steiners != NULL) {
        for (int i = 0; i < tria->tr_steiners->used_items; i++) {
          Vertex *v = (Vertex *) tria->tr_steiners->get(i);
          //if (v->is_deleted()) continue;
          v->crd[2] = 0.; // _min[2]; // zmin;
        }
      }
      draw_triangulation();
    } // if (viewer->viewprojectionAct)

    //if (viewfunctionAct->isChecked() ||
    //    viewprojectionAct->isChecked()) {
      // Restor heights
      for (int i = 0; i < tria->ct_in_vrts; i++) {
        Vertex *v = &(tria->in_vrts[i]);
        v->crd[2] = v->crd[0]*v->crd[0]+v->crd[1]*v->crd[1]-v->wei;
      }
      if (tria->tr_steiners != NULL) {
        for (int i = 0; i < tria->tr_steiners->used_items; i++) {
          Vertex *v = (Vertex *) tria->tr_steiners->get(i);
          //if (v->is_deleted()) continue;
          v->crd[2] = v->crd[0]*v->crd[0]+v->crd[1]*v->crd[1]-v->wei;
        }
      }
    //}    
}

/*
void OGLWidget::draw_mesh()
{
    Triangulation* m = _mesh;

    Vertex *vrt[4], Vshrink[4]; double tcenter[3];
    double shrink = _shrink;
    qDebug()<<" shrink="<<shrink;
    //double ax,ay,az,bx,by,bz,nm[3],le; // Used by drawface()

    // get cut plane
    double Pa, Pb, Pc, Pd;
    Pa = Pb = Pc = 0.;
    if (_plane_dir == 2) Pc = -1.; // z-section
    else if (_plane_dir == 1) Pb = -1.; // y
    else Pa = -1.0; // x
    Pd = _min[_plane_dir]+_plane_d*(_max[_plane_dir]-_min[_plane_dir]);
    //Pd = -Pd; // // why negativ ?
    if (invertCutPlane->isChecked()) {
      Pa = -Pa; Pb = -Pb; Pc = -Pc; Pd = -Pd;
    }
    qDebug()<<"Plane a,b,c,d ="<<Pa<<","<<Pb<<","<<Pc<<","<<Pd;
    int negative = 0;

    Vertex polypt[4]; // for cutting triangles
    //double cent[3];
    int corners;

    if (showTransparent->isChecked()) {
      glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
      glAlphaFunc(GL_GREATER,0.1);
      glEnable(GL_BLEND);
      glDepthMask(GL_FALSE);
      //glEnable(GL_DEPTH_TEST); // Enable Z-Buffer
      //glDepthMask(GL_TRUE); // Read from Z-Buffer
      ///////////////////////////////////////////////////
      //glEnable(GL_BLEND);
      // //glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
    }
    glBegin(GL_TRIANGLES);

    if (m->tr_tets && showElements->isChecked() && enableCutPlane->isChecked()) {
      //glBegin(GL_TRIANGLES);
      TEdge E; //, N;
      for (int i = 0; i < m->tr_tets->used_items; i++) {
        E.t = (Tetra *) m->tr_tets->get(i);
        if (E.t->is_deleted() || E.t->is_hulltet()) continue;
        // draw faces if this tetrahedron is cut by the plane.
        negative = 0;
        for (int k = 0; k < 4; k++) {
          if ((Pa*E.t->vrt[k]->crd[0]+Pb*E.t->vrt[k]->crd[1]+
               Pc*E.t->vrt[k]->crd[2]+Pd)<=0.) {negative++;}
        }
        if ((negative > 0) && (negative < 4)) {
          glcolori(abs(E.t->tag) + 1, 256, _transparency);
          for (int j = 0; j < 4; j++)
            vrt[j] = E.t->vrt[j];
          if (enableShrink->isChecked()) {
            for (int j = 0; j < 3; j++) {
              tcenter[j] = (vrt[0]->crd[j]+vrt[1]->crd[j]+vrt[2]->crd[j]+vrt[3]->crd[j]) / 4.;
            }
            for (int j = 0; j < 4; j++) {
              shrinkpoint(&(Vshrink[j]), vrt[j], tcenter, shrink);
            }
            drawface(&(Vshrink[0]), &(Vshrink[1]), &(Vshrink[2]));
            drawface(&(Vshrink[1]), &(Vshrink[0]), &(Vshrink[3]));
            drawface(&(Vshrink[2]), &(Vshrink[1]), &(Vshrink[3]));
            drawface(&(Vshrink[0]), &(Vshrink[2]), &(Vshrink[3]));
          } else {
            drawface(vrt[0], vrt[1], vrt[2]);
            drawface(vrt[1], vrt[0], vrt[3]);
            drawface(vrt[2], vrt[1], vrt[3]);
            drawface(vrt[0], vrt[2], vrt[3]);
          }
        }
      } // i
      //glEnd();
    }

    if (m->tr_tets && showTestedEles->isChecked()) {
      double R, G, B, A;
      //_hullfacecolor.getRgbF(&R, &G, &B, &A);
      _testedelecolor.getRgbF(&R, &G, &B, &A);
      glColor4f(R, G, B, _transparency); // a=1 no transparency
      TEdge E; //, N;
      for (int i = 0; i < m->tr_tets->used_items; i++) {
        E.t = (Tetra *) m->tr_tets->get(i);
        if (E.t->is_deleted() || E.t->is_hulltet()) continue;
        if (E.t->is_tested()) {
          for (int j = 0; j < 4; j++) vrt[j] = E.t->vrt[j];
          if (enableShrink->isChecked()) {
            for (int j = 0; j < 3; j++) {
              tcenter[j] = (vrt[0]->crd[j]+vrt[1]->crd[j]+vrt[2]->crd[j]+vrt[3]->crd[j]) / 4.;
            }
            for (int j = 0; j < 4; j++) {
              shrinkpoint(&(Vshrink[j]), vrt[j], tcenter, shrink);
            }
            drawface(&(Vshrink[0]), &(Vshrink[1]), &(Vshrink[2]));
            drawface(&(Vshrink[1]), &(Vshrink[0]), &(Vshrink[3]));
            drawface(&(Vshrink[2]), &(Vshrink[1]), &(Vshrink[3]));
            drawface(&(Vshrink[0]), &(Vshrink[2]), &(Vshrink[3]));
          } else {
            drawface(vrt[0], vrt[1], vrt[2]);
            drawface(vrt[1], vrt[0], vrt[3]);
            drawface(vrt[2], vrt[1], vrt[3]);
            drawface(vrt[0], vrt[2], vrt[3]);
          }
        } // if (E.t->is_tested())
      } // i
    }

    if (m->tr_tets && showHullfaces->isChecked() &&
        !showSubfaces->isChecked() &&
        !(m->tr_tris && showFacets->isChecked())
       ) {
      double R, G, B, A;
      //_hullfacecolor.getRgbF(&R, &G, &B, &A);
      _hullfacecolor.getRgbF(&R, &G, &B, &A);
      glColor4f(R, G, B, _transparency); // a=1 no transparency
      // if (showTransparent->isChecked()) {
      //   glColor4f(R, G, B, A); // a=1 no transparency
      //   glEnable(GL_BLEND);
      //   //glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
      // } else {
      //   glColor3f(R, G, B);
      // }
      //glBegin(GL_TRIANGLES);
      TEdge E;
      for (int i = 0; i < m->tr_tets->used_items; i++) {
        E.t = (Tetra *) m->tr_tets->get(i);
        if (E.t->is_deleted() || !E.t->is_hulltet()) continue;
        for (E.v = 0; E.v < 4; E.v++) {
          if (E.oppo() == m->tr_infvrt) break;
        }
        // E is a hull face.
        vrt[0] = E.org();
        vrt[1] = E.dest();
        vrt[2] = E.apex();
        if (enableCutPlane->isChecked()) {
          negative = 0; // default draw it.
          for (int j = 0; j < 3; j++) {
            if ((Pa*vrt[j]->crd[0]+Pb*vrt[j]->crd[1]+
                 Pc*vrt[j]->crd[2]+Pd)<=0.) {negative++;}
          }
          // ``negative == 3" means it lies completely below the plane.
          if (negative < 3) {
            if (enableShrink->isChecked()) {
              // Draw this face (normal points to outside).
              for (int j = 0; j < 3; j++) {
                tcenter[j] = (vrt[0]->crd[j] + vrt[1]->crd[j] + vrt[2]->crd[j]) / 3.;
              }
              for (int j = 0; j < 3; j++) {
                shrinkpoint(&(Vshrink[j]), vrt[j], tcenter, shrink);
              }
              drawface(&(Vshrink[0]), &(Vshrink[1]), &(Vshrink[2]));
            } else {
              drawface(vrt[0], vrt[1], vrt[2]);
            }
          }
        } else { // no cutplane
          if (enableShrink->isChecked()) {
            // Draw this face (normal points to outside).
            for (int j = 0; j < 3; j++) {
              tcenter[j] = (vrt[0]->crd[j] + vrt[1]->crd[j] + vrt[2]->crd[j]) / 3.;
            }
            for (int j = 0; j < 3; j++) {
              shrinkpoint(&(Vshrink[j]), vrt[j], tcenter, shrink);
            }
            drawface(&(Vshrink[0]), &(Vshrink[1]), &(Vshrink[2]));
          } else {
            drawface(vrt[0], vrt[1], vrt[2]);
          }
        }
      } // i
      //glEnd();
      // if (showTransparent->isChecked()) {
      //   glDisable(GL_BLEND);
      // }
    }

    if (m->tr_tets && showSubfaces->isChecked() &&
        !showHullfaces->isChecked()
       ) {
      double R, G, B, A;
      _subfacecolor.getRgbF(&R, &G, &B, &A);
      glColor4f(R, G, B, _transparency); // a=1 no transparency
      //glBegin(GL_TRIANGLES);
      TEdge E, N;
      for (int i = 0; i < m->tr_tets->used_items; i++) {
        E.t = (Tetra *) m->tr_tets->get(i);
        if (E.t->is_deleted() || E.t->is_hulltet()) continue;
        for (E.v = 0; E.v < 4; E.v++) {
          if (E.is_subface()) {
            N = E.fsym();
            if (N.oppo()->idx < E.oppo()->idx) {
              // Draw this face (normal points to outside).
              vrt[0] = E.org();
              vrt[1] = E.dest();
              vrt[2] = E.apex();
              if (enableCutPlane->isChecked()) {
                negative = 0;
                for (int j = 0; j < 3; j++) {
                  if ((Pa*vrt[j]->crd[0]+Pb*vrt[j]->crd[1]+
                       Pc*vrt[j]->crd[2]+Pd)<=0.) {negative++;}
                }
                if (negative < 3) {
                  if (enableShrink->isChecked()) {
                    for (int j = 0; j < 3; j++) {
                      tcenter[j] = (vrt[0]->crd[j] + vrt[1]->crd[j] + vrt[2]->crd[j]) / 3.;
                    }
                    for (int j = 0; j < 3; j++) {
                      shrinkpoint(&(Vshrink[j]), vrt[j], tcenter, shrink);
                    }
                    drawface(&(Vshrink[0]), &(Vshrink[1]), &(Vshrink[2]));
                  } else {
                     drawface(vrt[0], vrt[1], vrt[2]);
                  }
                }
              } else { // no cutplane
                // Draw this face (normal points to outside).
                if (enableShrink->isChecked()) {
                  for (int j = 0; j < 3; j++) {
                    tcenter[j] = (vrt[0]->crd[j] + vrt[1]->crd[j] + vrt[2]->crd[j]) / 3.;
                  }
                  for (int j = 0; j < 3; j++) {
                    shrinkpoint(&(Vshrink[j]), vrt[j], tcenter, shrink);
                  }
                  drawface(&(Vshrink[0]), &(Vshrink[1]), &(Vshrink[2]));
                } else {
                  drawface(vrt[0], vrt[1], vrt[2]);
                }
              }
            }
          } // if (E.is_subface())
        } // E.v
      } // i
      //glEnd();
      // if (showTransparent->isChecked()) {
      //   glDisable(GL_BLEND);
      // }
    }

    if (m->tr_tris && showFacets->isChecked()
        && !showMissingFacets->isChecked()
        && !showIntersectFacets->isChecked()) {
      double A = _transparency;
      //if (!showTransparent->isChecked()) A = 1.0; // no transparency
      //glBegin(GL_TRIANGLES);
      if (enableCutPlane->isChecked()) {
        if (enableShrink->isChecked()) {
          for (int i = 0; i < m->tr_tris->used_items; i++) {
            Tetra *f = (Tetra *) m->tr_tris->get(i);
            if (f->is_deleted()) continue;
            glcolori(abs(f->tag) + 1, 256, A); // maxcol=256, trans_value=1.0 no transparency
            vrt[0] = f->vrt[0];
            vrt[1] = f->vrt[1];
            vrt[2] = f->vrt[2];
            corners = cuttri(Pa,Pb,Pc,Pd,vrt[0],vrt[1],vrt[2],polypt,_dmax);
            if (corners == 1) {
              for (int j = 0; j < 3; j++) {
                tcenter[j] = (vrt[0]->crd[j] + vrt[1]->crd[j] + vrt[2]->crd[j]) / 3.;
              }
              for (int j = 0; j < 3; j++) {
                shrinkpoint(&(Vshrink[j]), vrt[j], tcenter, shrink);

              }
              drawface(&(Vshrink[0]), &(Vshrink[1]), &(Vshrink[2]));
            } else if (corners == 3) {
              for (int j = 0; j < 3; j++) {
                tcenter[j] = (polypt[0].crd[j] + polypt[1].crd[j] + polypt[2].crd[j]) / 3.;
              }
              for (int j = 0; j < 3; j++) {
                shrinkpoint(&(Vshrink[j]), &(polypt[j]), tcenter, shrink);
              }
              drawface(&(Vshrink[0]), &(Vshrink[1]), &(Vshrink[2]));
            } else if (corners == 4) {
              for (int j = 0; j < 3; j++) {
                tcenter[j] = (polypt[0].crd[j] + polypt[1].crd[j] + polypt[2].crd[j]) / 3.;
              }
              for (int j = 0; j < 3; j++) {
                shrinkpoint(&(Vshrink[j]), &(polypt[j]), tcenter, shrink);
              }
              drawface(&(Vshrink[0]), &(Vshrink[1]), &(Vshrink[2]));
              for (int j = 0; j < 3; j++) {
                tcenter[j] = (polypt[0].crd[j] + polypt[2].crd[j] + polypt[3].crd[j]) / 3.;
              }
              shrinkpoint(&(Vshrink[0]), &(polypt[0]), tcenter, shrink);
              shrinkpoint(&(Vshrink[1]), &(polypt[2]), tcenter, shrink);
              shrinkpoint(&(Vshrink[2]), &(polypt[3]), tcenter, shrink);
              drawface(&(Vshrink[0]), &(Vshrink[1]), &(Vshrink[2]));
            }
          }
        } else { // only cutplane
          for (int i = 0; i < m->tr_tris->used_items; i++) {
            Tetra *f = (Tetra *) m->tr_tris->get(i);
            if (f->is_deleted()) continue;
            glcolori(f->tag + 1, 256, A); // maxcol=256, trans_value=0.6
            corners = cuttri(Pa,Pb,Pc,Pd,f->vrt[0],f->vrt[1],f->vrt[2],polypt,_dmax);
            if (corners == 1) {
              drawface(f->vrt[0],f->vrt[1],f->vrt[2]);
            } else if (corners == 3) {
              drawface(&(polypt[0]), &(polypt[1]), &(polypt[2]));
            } else if (corners == 4) {
              drawface(&(polypt[0]), &(polypt[1]), &(polypt[2]));
              drawface(&(polypt[0]), &(polypt[2]), &(polypt[3]));
            }
          }
        }
      } else {
        // without using cutplane
        for (int i = 0; i < m->tr_tris->used_items; i++) {
          Tetra *f = (Tetra *) m->tr_tris->get(i);
          if (f->is_deleted()) continue;
          glcolori(f->tag + 1, 256, A); // maxcol=256, trans_value=0.6
          if (enableShrink->isChecked()) {
            vrt[0] = f->vrt[0];
            vrt[1] = f->vrt[1];
            vrt[2] = f->vrt[2];
            for (int j = 0; j < 3; j++) {
              tcenter[j] = (vrt[0]->crd[j] + vrt[1]->crd[j] + vrt[2]->crd[j]) / 3.;
            }
            for (int j = 0; j < 3; j++) {
              shrinkpoint(&(Vshrink[j]), vrt[j], tcenter, shrink);
            }
            drawface(&(Vshrink[0]), &(Vshrink[1]), &(Vshrink[2]));
          } else {
            drawface(f->vrt[0], f->vrt[1], f->vrt[2]);
          }
        }
      }
    }

    if (m->tr_tris && showMissingFacets->isChecked()) {
      double R, G, B, A;
      _missingfacetcolor.getRgbF(&R, &G, &B, &A);
      glColor4f(R, G, B, _transparency); // a=1 no transparency
      for (int i = 0; i < m->tr_tris->used_items; i++) {
        Tetra *f = (Tetra *) m->tr_tris->get(i);
        if (f->is_deleted()) continue;
        if (f->is_infected()) continue;
        if (f->is_hulltet()) continue;
        vrt[0] = f->vrt[0];
        vrt[1] = f->vrt[1];
        vrt[2] = f->vrt[2];
        if (enableCutPlane->isChecked()) {
          corners = cuttri(Pa,Pb,Pc,Pd,vrt[0],vrt[1],vrt[2],polypt,_dmax);
          if (corners == 1) {
            for (int j = 0; j < 3; j++) {
              tcenter[j] = (vrt[0]->crd[j] + vrt[1]->crd[j] + vrt[2]->crd[j]) / 3.;
            }
            for (int j = 0; j < 3; j++) {
              shrinkpoint(&(Vshrink[j]), vrt[j], tcenter, shrink);
            }
            drawface(&(Vshrink[0]), &(Vshrink[1]), &(Vshrink[2]));
          } else if (corners == 3) {
            for (int j = 0; j < 3; j++) {
              tcenter[j] = (polypt[0].crd[j] + polypt[1].crd[j] + polypt[2].crd[j]) / 3.;
            }
            for (int j = 0; j < 3; j++) {
              shrinkpoint(&(Vshrink[j]), &(polypt[j]), tcenter, shrink);
            }
            drawface(&(Vshrink[0]), &(Vshrink[1]), &(Vshrink[2]));
          } else if (corners == 4) {
            for (int j = 0; j < 3; j++) {
              tcenter[j] = (polypt[0].crd[j] + polypt[1].crd[j] + polypt[2].crd[j]) / 3.;
            }
            for (int j = 0; j < 3; j++) {
              shrinkpoint(&(Vshrink[j]), &(polypt[j]), tcenter, shrink);
            }
            drawface(&(Vshrink[0]), &(Vshrink[1]), &(Vshrink[2]));
            for (int j = 0; j < 3; j++) {
              tcenter[j] = (polypt[0].crd[j] + polypt[2].crd[j] + polypt[3].crd[j]) / 3.;
            }
            shrinkpoint(&(Vshrink[0]), &(polypt[0]), tcenter, shrink);
            shrinkpoint(&(Vshrink[1]), &(polypt[2]), tcenter, shrink);
            shrinkpoint(&(Vshrink[2]), &(polypt[3]), tcenter, shrink);
            drawface(&(Vshrink[0]), &(Vshrink[1]), &(Vshrink[2]));
          }
        } else { // no cutplane
          if (enableShrink->isChecked()) {
            for (int j = 0; j < 3; j++) {
              tcenter[j] = (vrt[0]->crd[j] + vrt[1]->crd[j] + vrt[2]->crd[j]) / 3.;
            }
            for (int j = 0; j < 3; j++) {
              shrinkpoint(&(Vshrink[j]), vrt[j], tcenter, shrink);
            }
            drawface(&(Vshrink[0]), &(Vshrink[1]), &(Vshrink[2]));
          } else {
            drawface(f->vrt[0], f->vrt[1], f->vrt[2]);
          }
        }
      }
    }

    if (m->tr_tris && showIntersectFacets->isChecked()) {
      double R, G, B, A;
      _intersectfacetcolor.getRgbF(&R, &G, &B, &A);
      glColor4f(R, G, B, _transparency); // a=1 no transparency
      for (int i = 0; i < m->tr_tris->used_items; i++) {
        Tetra *f = (Tetra *) m->tr_tris->get(i);
        if (f->is_deleted()) continue;
        if (f->is_infected()) {
          vrt[0] = f->vrt[0];
          vrt[1] = f->vrt[1];
          vrt[2] = f->vrt[2];
          if (enableCutPlane) {
            corners = cuttri(Pa,Pb,Pc,Pd,vrt[0],vrt[1],vrt[2],polypt,_dmax);
            if (corners == 1) {
              for (int j = 0; j < 3; j++) {
                tcenter[j] = (vrt[0]->crd[j] + vrt[1]->crd[j] + vrt[2]->crd[j]) / 3.;
              }
              for (int j = 0; j < 3; j++) {
                shrinkpoint(&(Vshrink[j]), vrt[j], tcenter, shrink);
              }
              drawface(&(Vshrink[0]), &(Vshrink[1]), &(Vshrink[2]));
            } else if (corners == 3) {
              for (int j = 0; j < 3; j++) {
                tcenter[j] = (polypt[0].crd[j] + polypt[1].crd[j] + polypt[2].crd[j]) / 3.;
              }
              for (int j = 0; j < 3; j++) {
                shrinkpoint(&(Vshrink[j]), &(polypt[j]), tcenter, shrink);
              }
              drawface(&(Vshrink[0]), &(Vshrink[1]), &(Vshrink[2]));
            } else if (corners == 4) {
              for (int j = 0; j < 3; j++) {
                tcenter[j] = (polypt[0].crd[j] + polypt[1].crd[j] + polypt[2].crd[j]) / 3.;
              }
              for (int j = 0; j < 3; j++) {
                shrinkpoint(&(Vshrink[j]), &(polypt[j]), tcenter, shrink);
              }
              drawface(&(Vshrink[0]), &(Vshrink[1]), &(Vshrink[2]));
              for (int j = 0; j < 3; j++) {
                tcenter[j] = (polypt[0].crd[j] + polypt[2].crd[j] + polypt[3].crd[j]) / 3.;
              }
              shrinkpoint(&(Vshrink[0]), &(polypt[0]), tcenter, shrink);
              shrinkpoint(&(Vshrink[1]), &(polypt[2]), tcenter, shrink);
              shrinkpoint(&(Vshrink[2]), &(polypt[3]), tcenter, shrink);
              drawface(&(Vshrink[0]), &(Vshrink[1]), &(Vshrink[2]));
            }
          } else { // no cutplane
            if (enableShrink->isChecked()) {
              for (int j = 0; j < 3; j++) {
                tcenter[j] = (vrt[0]->crd[j] + vrt[1]->crd[j] + vrt[2]->crd[j]) / 3.;
              }
              for (int j = 0; j < 3; j++) {
                shrinkpoint(&(Vshrink[j]), vrt[j], tcenter, shrink);
              }
              drawface(&(Vshrink[0]), &(Vshrink[1]), &(Vshrink[2]));
            } else {
              drawface(f->vrt[0], f->vrt[1], f->vrt[2]);
            }
          }
        } // if (f->is_infected())
      }
    }

    if (m->tr_tets && showFixedFaces->isChecked()) {
      double R, G, B, A;
      _fixedcolor.getRgbF(&R, &G, &B, &A);
      glColor4f(R, G, B, _transparency); // a=1 no transparency
      //glBegin(GL_TRIANGLES);
      TEdge E; //, N;
      for (int i = 0; i < m->tr_tets->used_items; i++) {
        E.t = (Tetra *) m->tr_tets->get(i);
        if (E.t->is_deleted() || E.t->is_hulltet()) continue;
        for (E.v = 0; E.v < 4; E.v++) {
          if (E.is_face_fixed()) {
            //N = E.fsym();
            //if (N.oppo()->idx < E.oppo()->idx) {
              // Draw this face (normal points to outside).
              vrt[0] = E.org();
              vrt[1] = E.dest();
              vrt[2] = E.apex();
              if (enableCutPlane->isChecked()) {
                negative = 0;
                for (int j = 0; j < 3; j++) {
                  if ((Pa*vrt[j]->crd[0]+Pb*vrt[j]->crd[1]+
                       Pc*vrt[j]->crd[2]+Pd)<=0.) {negative++;}
                }
                if (negative < 3) {
                  if (enableShrink->isChecked()) {
                    for (int j = 0; j < 3; j++) {
                      tcenter[j] = (vrt[0]->crd[j] + vrt[1]->crd[j] + vrt[2]->crd[j]) / 3.;
                    }
                    for (int j = 0; j < 3; j++) {
                      shrinkpoint(&(Vshrink[j]), vrt[j], tcenter, shrink);
                    }
                    drawface(&(Vshrink[0]), &(Vshrink[1]), &(Vshrink[2]));
                  } else {
                     drawface(vrt[0], vrt[1], vrt[2]);
                  }
                }
              } else { // no cutplane
                // Draw this face (normal points to outside).
                if (enableShrink->isChecked()) {
                  for (int j = 0; j < 3; j++) {
                    tcenter[j] = (vrt[0]->crd[j] + vrt[1]->crd[j] + vrt[2]->crd[j]) / 3.;
                  }
                  for (int j = 0; j < 3; j++) {
                    shrinkpoint(&(Vshrink[j]), vrt[j], tcenter, shrink);
                  }
                  drawface(&(Vshrink[0]), &(Vshrink[1]), &(Vshrink[2]));
                } else {
                  drawface(vrt[0], vrt[1], vrt[2]);
                }
              }
            //}
          } // if (E.is_face_fixed())
        } // E.v
      } // i
      //glEnd();
    }

    glEnd();
    if (showTransparent->isChecked()) {
      //glClear(GL_DEPTH_BUFFER_BIT); // Clear Z-Buffer
      //glDepthMask(GL_FALSE);
      glDepthMask(GL_TRUE);
      glDisable(GL_BLEND);
    }
    //// end of drawing triangles

    glBegin(GL_LINES);

    if (m->tr_tets && showEdges->isChecked()) {
      double R, G, B, A;
      _edgecolor.getRgbF(&R, &G, &B, &A);
      glColor3f(R, G, B); // a=1 no transparency
      glLineWidth(_edge_width);
      TEdge E, N;
      if (showElements->isChecked() && enableCutPlane->isChecked()) {
        // Only draw edges of cut elements.
        for (int i = 0; i < m->tr_tets->used_items; i++) {
          E.t = (Tetra *) m->tr_tets->get(i);
          if (E.t->is_deleted() || E.t->is_hulltet()) continue;
          // draw edges if this tetrahedron is cut by the plane.
          negative = 0;
          for (int k = 0; k < 4; k++) {
            if ((Pa*E.t->vrt[k]->crd[0]+Pb*E.t->vrt[k]->crd[1]+
                 Pc*E.t->vrt[k]->crd[2]+Pd)<=0.) {negative++;}
          }
          if ((negative > 0) && (negative < 4)) {
            for (int j = 0; j < 6; j++) {
              E.v = _e2v[j];
              drawedge(E.org(), E.dest());
            }
          } // j
        } // i
      } else {
        // Draw all edges of the mesh.
        for (int i = 0; i < m->tr_tets->used_items; i++) {
          E.t = (Tetra *) m->tr_tets->get(i);
          if (E.t->is_deleted() || E.t->is_hulltet()) continue;
          for (int j = 0; j < 6; j++) {
            E.v = _e2v[j];
            int idx = E.apex()->idx;
            N = E;
            do {
              if (N.apex()->idx > idx) break;
            } while (N.t != E.t);
            if (N.t == E.t) {
              if (enableCutPlane->isChecked()) {
                vrt[0] = E.org();
                vrt[1] = E.dest();
                negative = 0;
                for (int k = 0; k < 2; k++) {
                  if ((Pa*vrt[k]->crd[0]+Pb*vrt[k]->crd[1]+
                       Pc*vrt[k]->crd[2]+Pd)<=0.) {negative++;}
                }
                if (negative < 2) {
                  drawedge(E.org(), E.dest());
                }
              } else {
                drawedge(E.org(), E.dest());
              }
            }
          } // j
        } // i
      } //
    } // if (m->tr_tets && showEdges->isChecked())

    if (m->tr_tets
        && (showElements->isChecked() && enableCutPlane->isChecked())
        && showEdges->isChecked()) {
      double R, G, B, A;
      _edgecolor.getRgbF(&R, &G, &B, &A);
      glColor3f(R, G, B); // a=1 no transparency
      glLineWidth(_edge_width);
      //glBegin(GL_LINES);
      TEdge E, N;
      for (int i = 0; i < m->tr_tets->used_items; i++) {
        E.t = (Tetra *) m->tr_tets->get(i);
        if (E.t->is_deleted() || E.t->is_hulltet()) continue;
        // draw edges if this tetrahedron is cut by the plane.
        negative = 0;
        for (int k = 0; k < 4; k++) {
          if ((Pa*E.t->vrt[k]->crd[0]+Pb*E.t->vrt[k]->crd[1]+
               Pc*E.t->vrt[k]->crd[2]+Pd)<=0.) {negative++;}
        }
        if ((negative > 0) && (negative < 4)) {
          for (int j = 0; j < 6; j++) {
            E.v = _e2v[j];
            drawedge(E.org(), E.dest());
          }
        } // j
      } // i
      //glEnd();
    }

    if (m->tr_tets && showEdges->isChecked() &&
        !(showElements->isChecked() && enableCutPlane->isChecked())
       ) {
      double R, G, B, A;
      _edgecolor.getRgbF(&R, &G, &B, &A);
      glColor3f(R, G, B); // a=1 no transparency
      glLineWidth(_edge_width);
      //glBegin(GL_LINES);
      TEdge E, N;
      for (int i = 0; i < m->tr_tets->used_items; i++) {
        E.t = (Tetra *) m->tr_tets->get(i);
        if (E.t->is_deleted() || E.t->is_hulltet()) continue;
        for (int j = 0; j < 6; j++) {
          E.v = _e2v[j];
          int idx = E.apex()->idx;
          N = E;
          do {
            if (N.apex()->idx > idx) break;
          } while (N.t != E.t);
          if (N.t == E.t) {
            if (enableCutPlane->isChecked()) {
              vrt[0] = E.org();
              vrt[1] = E.dest();
              negative = 0;
              for (int k = 0; k < 2; k++) {
                if ((Pa*vrt[k]->crd[0]+Pb*vrt[k]->crd[1]+
                     Pc*vrt[k]->crd[2]+Pd)<=0.) {negative++;}
              }
              if (negative < 2) {
                drawedge(E.org(), E.dest());
              }
            } else {
              drawedge(E.org(), E.dest());
            }
          }
        } // j
      } // i
      //glEnd();
    }

    if (m->tr_tris && showFacetEdges->isChecked()) {
      double R, G, B, A;
      _edgecolor.getRgbF(&R, &G, &B, &A);
      glColor3f(R, G, B); // a=1 no transparency
      glLineWidth(_edge_width);
      //glBegin(GL_LINES);
      if (enableCutPlane->isChecked()) {
        for (int i = 0; i < m->tr_tris->used_items; i++) {
          Tetra *f = (Tetra *) m->tr_tris->get(i);
          if (f->is_deleted()) continue;
          for (int j = 0; j < 3; j++) {
            if (cutseg(Pa,Pb,Pc,Pd,f->vrt[j], f->vrt[(j+1)%3],polypt,_dmax) > 0) {
              //printf("    ---[%d]: %g, %g, %g\n", f->vrt[j]->idx, polypt[0].crd[0], polypt[0].crd[1], polypt[0].crd[2]);
              //printf("    ---[%d]: %g, %g, %g\n", f->vrt[(j+1)%3]->idx, polypt[1].crd[0], polypt[1].crd[1], polypt[1].crd[2]);
              drawedge(&(polypt[0]), &(polypt[1]));
            }
          }
        }
      } else {
        for (int i = 0; i < m->tr_tris->used_items; i++) {
          Tetra *f = (Tetra *) m->tr_tris->get(i);
          if (f->is_deleted()) continue;
          for (int j = 0; j < 3; j++) {
            drawedge(f->vrt[j], f->vrt[(j+1)%3]);
          }
        }
      }
      //glEnd();
    }

    if (m->tr_segs && showSegments->isChecked()) {
      double R, G, B, A;
      _segmentcolor.getRgbF(&R, &G, &B, &A);
      glColor3f(R, G, B); // a=1 no transparency
      glLineWidth(_segment_width);
      for (int i = 0; i < m->tr_segs->used_items; i++) {
        Tetra *f = (Tetra *) m->tr_segs->get(i);
        if (f->is_deleted()) continue;
        if (enableCutPlane->isChecked()) {
          if (cutseg(Pa,Pb,Pc,Pd,f->vrt[0], f->vrt[1],polypt,_dmax) > 0) {
            drawedge(&(polypt[0]), &(polypt[1]));
          }
        } else {
          drawedge(f->vrt[0], f->vrt[1]);
        }
      }
    }

    if (m->tr_tets && showSubsegs->isChecked()) {
      double R, G, B, A;
      _fixedcolor.getRgbF(&R, &G, &B, &A);
      glColor3f(R, G, B); // a=1 no transparency
      glLineWidth(_edge_width+1);
      // Draw all edges of the mesh.
      TEdge E, N;
      for (int i = 0; i < m->tr_tets->used_items; i++) {
        E.t = (Tetra *) m->tr_tets->get(i);
        if (E.t->is_deleted() || E.t->is_hulltet()) continue;
        for (int j = 0; j < 6; j++) {
          E.v = _e2v[j];
          if (E.is_segment()) {
            int idx = E.apex()->idx;
            N = E;
            do {
              if (N.apex()->idx > idx) break;
            } while (N.t != E.t);
            if (N.t == E.t) {
              if (enableCutPlane->isChecked()) {
                vrt[0] = E.org();
                vrt[1] = E.dest();
                negative = 0;
                for (int k = 0; k < 2; k++) {
                  if ((Pa*vrt[k]->crd[0]+Pb*vrt[k]->crd[1]+
                       Pc*vrt[k]->crd[2]+Pd)<=0.) {negative++;}
                }
                if (negative < 2) {
                  drawedge(E.org(), E.dest());
                }
              } else {
                drawedge(E.org(), E.dest());
              }
            }
          } // if (E.is_segment())
        } // j
      } // i
    }

    glEnd(); // end of lines
    //////////////////////////////

    if (showVertices->isChecked()) {
      double R, G, B, A;
      _vertexcolor.getRgbF(&R, &G, &B, &A);
      glColor3f(R, G, B); //glColor3dv(clr);
      glPointSize(_vertex_size);
      glEnable(GL_POINT_SMOOTH);
      glBegin(GL_POINTS);
      if (enableCutPlane->isChecked()) {
        for (int i = 0; i < m->ct_in_vrts; i++) {
          Vertex *v = &(m->in_vrts[i]);
          //if (v->typ == UNUSEDVERTEX) continue;
          if (Pa*v->crd[0]+Pb*v->crd[1]+Pc*v->crd[2]+Pd > 0) {
            glVertex3d(v->crd[0], v->crd[1], v->crd[2]);
          }
        }
        if (m->tr_steiners != NULL) {
          for (int i = 0; i < m->tr_steiners->objects; i++) {
            Vertex *v = (Vertex *) m->tr_steiners->get(i);
            if (v->is_deleted()) continue;
            if (Pa*v->crd[0]+Pb*v->crd[1]+Pc*v->crd[2]+Pd > 0) {
              glVertex3d(v->crd[0], v->crd[1], v->crd[2]);
            }
          }
        }
      } else {
        for (int i = 0; i < m->ct_in_vrts; i++) {
          Vertex *v = &(m->in_vrts[i]);
          //if (v->typ == UNUSEDVERTEX) continue;
          glVertex3d(v->crd[0], v->crd[1], v->crd[2]);
        }
        if (m->tr_steiners != NULL) {
          for (int i = 0; i < m->tr_steiners->objects; i++) {
            Vertex *v = (Vertex *) m->tr_steiners->get(i);
            if (v->is_deleted()) continue;
            glVertex3d(v->crd[0], v->crd[1], v->crd[2]);
          }
        }
      }
      glEnd();
      glDisable(GL_POINT_SMOOTH);
    }
}
*/

/*
void OGLWidget::draw_labels()
{
    if (!_mesh) return;
    qDebug()<<"---draw labels---";

    // get cut plane
    double Pa, Pb, Pc, Pd;
    Pa = Pb = Pc = 0.;
    if (_plane_dir == 2) Pc = 1.; // z-section
    else if (_plane_dir == 1) Pb = 1.; // y
    else Pa = 1.0; // x
    if (invertCutPlane->isChecked()) {
      Pa = -Pa; Pb = -Pb; Pc = -Pc;
    }
    Pd = _min[_plane_dir]+_plane_d*(_max[_plane_dir]-_min[_plane_dir]);
    Pd = -Pd; // // why negativ ?

    if (showVertexLabels->isChecked()) {
      QPainter painter(this);
      painter.setPen(Qt::gray);
      painter.setFont(QFont("Arial", 12)); //setFont(QFont("Times", 24));
      if (enableCutPlane->isChecked()) {
        for (int i = 0; i < _mesh->ct_in_vrts; i++) {
          Vertex *v = &(_mesh->in_vrts[i]);
          //if (v->typ == UNUSEDVERTEX) continue;
          if (Pa*v->crd[0]+Pb*v->crd[1]+Pc*v->crd[2]+Pd > 0) {
            renderText(&painter, v->crd[0], v->crd[1], v->crd[2], QString::number(v->idx));
          }
        }
        if (_mesh->tr_steiners != NULL) {
          for (int i = 0; i < _mesh->tr_steiners->objects; i++) {
            Vertex *v = (Vertex *) _mesh->tr_steiners->get(i);
            if (v->is_deleted()) continue;
            if (Pa*v->crd[0]+Pb*v->crd[1]+Pc*v->crd[2]+Pd > 0) {
              renderText(&painter, v->crd[0], v->crd[1], v->crd[2], QString::number(v->idx));
            }
          }
        }
      } else {
        for (int i = 0; i < _mesh->ct_in_vrts; i++) {
          Vertex *v = &(_mesh->in_vrts[i]);
          //if (v->typ == UNUSEDVERTEX) continue;
          renderText(&painter, v->crd[0], v->crd[1], v->crd[2], QString::number(v->idx));
        }
        if (_mesh->tr_steiners != NULL) {
          for (int i = 0; i < _mesh->tr_steiners->objects; i++) {
            Vertex *v = (Vertex *) _mesh->tr_steiners->get(i);
            if (v->is_deleted()) continue;
            renderText(&painter, v->crd[0], v->crd[1], v->crd[2], QString::number(v->idx));
          }
        }
      }
      painter.end();
    }
}
*/

void OGLWidget::draw_selection()
{
    /*
    if (_sel_vrts->objects > 0) {
      double R, G, B, A;
      _selectcolor.getRgbF(&R, &G, &B, &A);
      glColor3f(R, G, B); //glColor3dv(clr);
      glPointSize(_vertex_size + 6);
      glEnable(GL_POINT_SMOOTH);
      glBegin(GL_POINTS);
      for (int i = 0; i < _sel_vrts->objects; i++) {
        Vertex *v = (Vertex *) _sel_vrts->get(i);
        glVertex3d(v->crd[0], v->crd[1], v->crd[2]);
      }
      glEnd();
      glDisable(GL_POINT_SMOOTH);
    }

    if (_sel_edges->objects > 0) {
      double R, G, B, A;
      _selectcolor.getRgbF(&R, &G, &B, &A);
      glColor3f(R, G, B); //glColor3dv(clr);
      glLineWidth(_edge_width+1);
      glBegin(GL_LINES);
      for (int i = 0; i < _sel_edges->objects; i++) {
        Vertex *vv = (Vertex *) _sel_edges->get(i);
        drawedge(&(vv[0]), &(vv[1]));
      }
      glEnd();
    }
    */

    // Show labels.
    QPainter painter(this);
    QPen pen;
    pen.setColor(Qt::red);
    pen.setWidth(2);
    painter.setPen(pen);
    painter.setFont(QFont("Times", 16));
    if (_sel_vrts->objects > 0) {
      for (int i = 0; i < _sel_vrts->objects; i++) {
        Vertex *v = (Vertex *) _sel_vrts->get(i);
        renderText(&painter, v->crd[0], v->crd[1], v->crd[2], QString::number(v->idx));
      }
    }

    if (_pos.tag > 0) {
      int ox, oy;
      double Ox = _pos.crd[0];
      double Oy = _pos.crd[1];
      double Oz = _pos.crd[2];
      projection(Ox, Oy, Oz, &ox, &oy);
      painter.drawEllipse(QPoint(ox,oy), 4, 4);
      int px, py;
      double Px = Ox + _dmax * _ray.crd[0];
      double Py = Oy + _dmax * _ray.crd[1];
      double Pz = Oz + _dmax * _ray.crd[2];
      projection(Px, Py, Pz, &px, &py);
      painter.drawLine(QPoint(ox, oy), QPoint(px,py));
    }

    painter.end();
}

//============================================================

void OGLWidget::plot3d() // tetviewmain
{
    //Triangulation* _mesh = ((ScribbleArea*) _scribbleArea)->_tria;
    //if (_mesh == NULL) return;
    printf("debug: plot3d() is called.\n");

    reset_volume();

    // set the callback function.
    //_draw_object = &draw_scene;
    //if (_mesh != NULL) {
      _draw_object = &OGLWidget::draw_scene;
      _object_valid=0;
    //} else {
    //  // no mesh to draw.
    //  _draw_object = NULL;
    //  _object_valid=1;
    //}
    update();
}

