#ifndef SCRIBBLEAREA_H
#define SCRIBBLEAREA_H

#include <QColor>
#include <QImage>
#include <QWidget>
#include <QMenuBar>
#include <QStatusBar>
#include <QToolBar>
#include <QDebug>

#include "view3d.h"
#include "detri2/detri2.h"
#include "widgets/newfilepanel.h"
#include "widgets/meshoptionpanel.h"
#include "widgets/hdeoptionpanel.h"
#include "widgets/sortoptionpanel.h"
#include "widgets/vertexpanel.h"
#include "widgets/segmentpanel.h"
#include "widgets/subdomainpanel.h"
#include "widgets/imageview.h"
#include "widgets/preferdialog.h"
#include "widgets/muParserdialog.h"
#include "widgets/querydialog.h"
#include "widgets/metricpanel.h"

class ScribbleArea : public QWidget
{
    Q_OBJECT

public:
    ScribbleArea(QToolBar *toolbar_top, QToolBar *toolbar_left,
                 QStatusBar *statusbar, QMenuBar *menubar,
                 QWidget *parent = 0);

    detri2::Triangulation *_geom; // the geometry
    detri2::Triangulation *_tria; // the triangulation

    QImage _canvas[2]; // [0] foreground, [1] background (copy).
    int    _object_valid;

    QImage *_background_img;

    double _viewport_min[2], _viewport_max[2]; // the viewport in (x,y) physical coordinates.
    double _cx, _cy, _scale_xy; // affine transformation from (x,y) to (X,Y)
                                // (x,y) are input coordinates,
                                // (X,Y) are screen coordinates, such that
                                // X =   _scale_xy * x + _cx, and
                                // Y = - _scale_xy * y + _cy,
                                // (_cx, _cy) is the center in screen coordinates.

    double  _mouse_sens;
    int     _mouse_last_x;  /**< last mouse x position in window */
    int     _mouse_last_y;  /**< last mouse y position in window */
    int     _mouse_down;    /**< mouse button pressed flag */
    int     _mouse_down_move;

    //bool    _zoom_selected_region;
    int     _select_x, _select_y, _select_width, _select_height;

    // Quadratic form.
    double _a11, _a21, _a22;

    int     _color_number;
    int     *_rgb_color_bar; // 3 x color_number

    detri2::Vertex*     _active_vertex;
    detri2::Triang*     _active_segment;
    detri2::arraypool*  _active_trilist;
    detri2::Vertex*     _active_sdm_vertex;

    newfilePanel*       _newfilepabel;
    meshoptionPanel*    _meshoptionPanel;
    hdeoptionPanel*     _hdeoptionPanel;
    sortOptionPanel*    _sortoptionpanel;
    vertexPanel*        _vertexPanel;
    segmentPanel*       _segmentPanel;
    subdomainPanel*     _subdomainPanel;
    ImageView*          _imgView;
    PreferDialog*       _preferdialog;

    View3dWidget*       _view3dWindow; // Opengl window

    // Pointers to mainwindow
    QMenuBar*           _menubar;
    QToolBar*           _toolbar_top;
    QToolBar*           _toolbar_left;
    QStatusBar*         _statusbar;
    QLabel*             m_statusLeft;
    QLabel*             m_statusMiddle;
    QLabel*             m_statusRight;

    // View options
    int                 _draw_vertex_size; // 4
    int                 _draw_segment_width; // 2
    int                 _draw_edge_width; // 1
    int                 _draw_dual_edge_width; // 1
    int                 _draw_ccircle_width; // 1
    QColor              _color_canvas_background; // Qt::white
    QColor              _color_vertex;
    QColor              _color_segment; // Qt::blue
    QColor              _color_edge;
    QColor              _color_dual_edge;
    QColor              _color_bgmesh_edge;
    QColor              _color_ccircle;
    QColor              _color_weight_positive;
    QColor              _color_weight_negative;
    QColor              _color_mesh_size;
    QColor              _color_selected;
    QFont               _font_bbox_label;
    QFont               _font_vertex_label;
    QFont               _font_segment_label;
    QFont               _font_subdomain_label;
    QFont               _font_edgelength_label;
    QString             _file_last_used_path;

    // view3d preferences
    int                 _v3d_vertex_size;
    int                 _v3d_edge_width;
    int                 _v3d_segment_width;
    QColor              _v3d_bgcolor;
    QColor              _v3d_vertexcolor;
    QColor              _v3d_edgecolor;
    QColor              _v3d_facetcolor;
    QColor              _v3d_selectcolor;
    QColor              _v3d_circlecolor;
    QColor              _v3d_weightcolor;
    QColor              _v3d_polecolor;
    QColor              _v3d_polarcolor;

    // Debug options
    int                 _enable_debug;
    QString             _file_debug;

public slots:
    void clean_scribble();
    void reset_scribble();
    void reset_preference();

    void ShowContextMenu(const QPoint& pos);

//private slots:
    void newFile();
    void openFile();
    void addFile();
    void saveFile();
    void closeFile();
    void open_preferDialog();
    void on_user_exit();

    // Edit
    void edit_mouse_act();
    void edit_attributes_act();
    void edit_delete_act();
    void edit_scale_act();

    // View
    void view_update_act();
    void view_trans_act(int i);
    void view_zoom_act(int i);
    void view_reset();

    void detri2_options_act();
    void set_metric_act();
    void metric_transform_act();
    void generate_mesh_act();
    void weighted_Delaunay_act();
    void furthest_Delaunay_act();
    void detri2_HDE_option_act();
    void detri2_HDE_act();
    void detri2_Sort_option_act();
    void detri2_Sort_act();
    void detri2_mesh_smooth_act();
    void clean_mesh_act();
    void test_muParser_act();
    void query_act();    
    void viewgl_act();
    void draw_extra_points_act();

    // tests
    void draw_Bezier_curve_act();

private:
    // file menu
    QAction*    newAct;
    QAction*    openAct;
    QAction*    addAct;
    QAction*    saveAct;
    QAction*    closeAct;
    QAction*    UserPreferAct;
    QAction*    UserExitAct;

    // edit menu
    QAction*    transAct;
    //QAction *insertAct;
    QAction*    insertVertexAct;
    QAction*    insertSegmentAct;
    QAction*    insertRectAct;
    QAction*    insertArcAct;
    //QAction *selectAct;
    QAction*    assignVertexAct;
    QAction*    assignVertexWeightAct;
    QAction*    assignMeshSizeAct;
    QAction*    assignSegmentAct;
    QAction*    assignSubdomainAct;
    QAction*    setAttributeAct;
    QAction*    selectMeshVertexAct;
    QAction*    selectMeshEdgeAct;
    QAction*    selectMeshEleAct;
    QAction*    deleteAct;
    QAction*    scaleAct;

    // view menu
    QAction*    viewVerticesAct;
    QAction*    viewWeightsAct;
    QAction*    viewMeshSizeAct;
    QAction*    viewVertexLabelsAct;
    QAction*    viewSegmentsAct;
    QAction*    viewSegmentLabelsAct;
    QAction*    viewSubdomainsAct;
    QAction*    viewConvexHullAct;
    QAction*    viewTrianglesAct;
    QAction*    viewEdgeLengthsAct;
    QAction*    viewTriangleInfoAct;
    QAction*    viewOthocirclesAct;
    QAction*    viewOthocircleAct;
    QAction*    viewOthocenterAct;
    QAction*    viewPowerDiagramAct;
    QAction*    viewMassCentersAct;
    QAction*    viewLaplacianCentersAct;
    QAction*    viewDistmeshPointsAct;
    QAction*    viewPowerCellAct;
    QAction*    viewMassCenterAct;
    QAction*    viewLaplaceCenterAct;
    QAction*    viewBboxAct;
    QAction*    viewBgMeshAct;
    QAction*    viewGLAct;

    QAction*    MoveLeftAct;
    QAction*    MoveRightAct;
    QAction*    MoveUpAct;
    QAction*    MoveDownAct;
    QAction*    zoomInAct;
    QAction*    zoomOutAct;
    QAction*    zoomSelectAct;
    QAction*    viewResetAct;

    // mesh Menu    
    QAction*    detri2OptionsAct;
    QAction*    setMetricAct;
    QAction*    weightedDelaunayAct;
    QAction*    furthestDelaunayAct;
    QAction*    detri2HDEAct;
    QAction*    detri2SortAct;
    QAction*    cleanMeshAct;
    QAction*    muParserAct;
    QAction*    queryAct;
    QAction*    drawextrapointsAct;

    QAction*    testGLAct; // test
    QAction*    drawBezierCurveAct; // test
    QAction*    viewLaplaceCenter2Act; // test
    QAction*    viewMetricTransformAct; // test

    void createActions();
    void createMenus();

protected:
    void mousePressEvent(QMouseEvent *event) override;
    void mouseMoveEvent(QMouseEvent *event) override;
    void mouseReleaseEvent(QMouseEvent *event) override;
    void wheelEvent(QWheelEvent *e) override;
    //void keyPressEvent(QKeyEvent *event) override;

    void paintEvent(QPaintEvent *event) override;
    void resizeEvent(QResizeEvent *event) override;

public:
    int  set_active_vertex(detri2::Vertex *thevertex);
    int  is_vertex_active(detri2::Vertex *thevertex);
    int  remove_active_vertex(detri2::Vertex *thevertex);
    int  add_active_vertex(detri2::Vertex *thevertex);
    int  set_active_segment(detri2::Triang *theseg);
    int  is_segment_active(detri2::Triang *theseg);
    int  remove_active_segment(detri2::Triang *theseg);
    int  add_active_segment(detri2::Triang *theseg);
    int  set_active_trilist(detri2::arraypool *newtrilist);

    void calc_trans(double xmin, double xmax, double ymin, double ymax);
    void get_selection_region(double& xmin, double& xmax, double& ymin, double& ymax);
    void generate_color_bar(int colornumber);
    void draw_scene(QPainter &painter);
};

#endif
