#include <QtWidgets>

#include "scribblearea.h"
//#include "adaptoptiondialog.h"

//==============================================================================
//#include "muParser.h"

mu::Parser User_Func; // function
mu::Parser User_Func_X, User_Func_Y;
double user_x, user_y;

double callback_user_func(const detri2::Vertex * vert)
{
    user_x = vert->crd[0];
    user_y = vert->crd[1];
    return User_Func.Eval();
}

double callback_user_func_X(const detri2::Vertex * vert)
{
    user_x = vert->crd[0];
    user_y = vert->crd[1];
    return User_Func_X.Eval();
}

double callback_user_func_Y(const detri2::Vertex * vert)
{
    user_x = vert->crd[0];
    user_y = vert->crd[1];
    return User_Func_Y.Eval();
}

//==============================================================================

FuncInterpolationDialog::FuncInterpolationDialog(void *ptr, QWidget *parent)
    : QDialog(parent)
{
    _scribblearea = ptr;
    ScribbleArea *sa = (ScribbleArea *) _scribblearea;

    func = new QLineEdit;
    func_x = new QLineEdit;
    func_y = new QLineEdit;

    func_template = new QComboBox();
    func_template->addItem("none"); // 0
    func_template->addItem("Gaussian"); // 1
    func_template->addItem("Lennard");  // 2
    func_template->addItem("Alauzet");  // 3
    func_template->addItem("Huang");  // 4
    //func_template->addItem("Frey"); // 5 Differential Geometry book, p. 279
    func_template->addItem("BAMG"); // 5

    //QRadioButton *length_measure;
    //QRadioButton *L2err_measure;

    QLabel *lbl_s1 = new QLabel(tr("s1"));
    hde_s1 = new QLineEdit;
    hde_s1->setText(tr("1.0"));
    QLabel *lbl_s2 = new QLabel(tr("s2"));
    hde_s2 = new QLineEdit;
    hde_s2->setText(tr("1.0"));

    btnOptions = new QPushButton(tr("More options"));
    btnView3d = new QPushButton(tr("View 3d..."));
    btnTest = new QPushButton(tr("Evaulate..."));
    btnGenerate = new QPushButton(tr("Generate"));

    // Layout
    QGridLayout *layout_func = new QGridLayout;
    QLabel *lbl_func = new QLabel(tr("f"));
    QLabel *lbl_fx = new QLabel(tr("fx"));
    QLabel *lbl_fy = new QLabel(tr("fy"));
    layout_func->addWidget(lbl_func, 0, 0, 1, 1);
    layout_func->addWidget(func, 0, 1, 1, 4);
    layout_func->addWidget(lbl_fx, 1, 0, 1, 1);
    layout_func->addWidget(func_x, 1, 1, 1, 4);
    layout_func->addWidget(lbl_fy, 2, 0, 1, 1);
    layout_func->addWidget(func_y, 2, 1, 1, 4);
    layout_func->addWidget(func_template, 3, 0, 1, 2);
    layout_func->addWidget(lbl_s1, 4,0,1,1);
    layout_func->addWidget(hde_s1, 4,1,1,1);
    layout_func->addWidget(lbl_s2, 4,2,1,1);
    layout_func->addWidget(hde_s2, 4,3,1,1);
    QGroupBox *group_func = new QGroupBox(tr("Function and gradients"));
    group_func->setLayout(layout_func);

    QGridLayout *layout_btn = new QGridLayout;
    layout_btn->addWidget(btnOptions, 0, 0, 1, 1);
    layout_btn->addWidget(btnView3d, 0, 1, 1, 1);
    layout_btn->addWidget(btnTest, 1, 0, 1, 1);
    layout_btn->addWidget(btnGenerate, 1, 1, 1, 1);
    QGroupBox *group_btn = new QGroupBox;
    group_btn->setLayout(layout_btn);

    // Signals
    connect(func_template, SIGNAL(currentIndexChanged(int)), SLOT(setFunction(int)));

    connect(btnView3d, SIGNAL(clicked()), this, SLOT(view3d()));
    connect(btnTest, SIGNAL(clicked()), this, SLOT(evaluate_function()));
    connect(btnGenerate, SIGNAL(clicked()), this, SLOT(generate()));

    // Final layout
    QGridLayout *mainlayout = new QGridLayout(this);
    mainlayout->addWidget(group_func);
    //layout->addWidget(group_method);
    //layout->addWidget(group_operations);
    //layout->addWidget(group_options);
    mainlayout->addWidget(group_btn);

    //group_options->setVisible(false);
    //setLayout(layout);
}

void FuncInterpolationDialog::setFunction( int idx ) {
    qDebug()<<"idx="<<idx;
    // do another stuff
    if (idx == 1) {
      func->setText(tr("4.0 * exp(-(x * x + y * y) / 2.0)"));
      func_x->setText(tr("-x * 4.* exp(-(x * x + y * y) / 2.0)"));
      func_y->setText(tr("-y * 4.* exp(-(x * x + y * y) / 2.0)"));
    } else if (idx == 2) {
      func->setText(tr("tanh(60.*x)-tanh(60.*x-60.*y-30.)"));
      func_x->setText(tr("60.*tanh(60.*x-60.*y-30.)*tanh(60.*x-60.*y-30.)-60.*tanh(60.*x)*tanh(60.*x)"));
      func_y->setText(tr("-60.+60.*tanh(60.*x-60.*y-30.)*tanh(60.*x-60.*y-30.)"));
    } else if (idx == 3) {
      func->setText(tr("sin(5.*(x-0.2)*(x-0.2)*(x-0.2)*(y*y-y+1.))"));
      func_x->setText(tr("cos(5.*(x-0.2)*(x-0.2)*(x-0.2)*(y*y-y+1.))*3.*(x-0.2)*(x-0.2)"));
      func_y->setText(tr("cos(5.*(x-0.2)*(x-0.2)*(x-0.2)*(y*y-y+1.))*(2.*y-1.)"));
    } else if (idx == 4) {
      func->setText(tr("-10*((y-0.4*sin(1.0*pi*x))^2)"));
      func_x->setText(tr("-tanh(-10*((-1.0.*sin(1.0*pi*x) / 5. + y))^2)"));
      func_y->setText(tr("-tanh(-10*((y-1.0*sin(1.0*pi*x) / 5.)^2))"));
    } else if (idx == 5) {
      func->setText(tr(""));
      func_x->setText(tr(""));
      func_y->setText(tr(""));
    } else {
      func->setText(tr(""));
      func_x->setText(tr(""));
      func_y->setText(tr(""));
    }
}

void FuncInterpolationDialog::view3d() {
    qDebug()<<"view3d";
    ScribbleArea *sa = (ScribbleArea *) _scribblearea;
    if (sa->_tria == NULL) {
      QMessageBox msg;
      msg.setText("Please first generate a mesh!");
      msg.exec();
      return;
    }

    if (sa->_tria->OMT_domain != NULL) {
      // The function vlaue will be interpolated from a background mesh (.mtr).
      // ....
    } else {
      if (func->text().isEmpty()) {
        return;
      }
      // Use the function given by the user
      User_Func.SetExpr(std::string(func->text().toStdString()));
      User_Func.DefineVar("x", &user_x);
      User_Func.DefineVar("y", &user_y);
      sa->_tria->Func = &callback_user_func;

      User_Func_X.SetExpr(std::string(func_x->text().toStdString()));
      User_Func_X.DefineVar("x", &user_x);
      User_Func_X.DefineVar("y", &user_y);
      sa->_tria->GradX = &callback_user_func_X;

      User_Func_Y.SetExpr(std::string(func_y->text().toStdString()));
      User_Func_Y.DefineVar("x", &user_x);
      User_Func_Y.DefineVar("y", &user_y);
      sa->_tria->GradY = &callback_user_func_Y;
    }

    sa->_tria->op_hde_s1 = hde_s1->text().toDouble();
    //sa->_tria->op_hde_s2 = hde_s2->text().toDouble();
    //sa->_tria->op_target_length = target_length->text().toDouble();
    //sa->_tria->op_smooth_deltat = 0.1;

    // set the function values on nodes.
    sa->_tria->op_metric = METRIC_HDE;
    sa->_tria->pre_adapt();
    //sa->_tria->op_metric = METRIC_Euclidean;

    sa->viewgl_act();
}

void FuncInterpolationDialog::evaluate_function() {
    qDebug()<<"Evaluate";
    if (!func->text().isEmpty()) {
      MuParserDialog _testdialog(_scribblearea); // = new MuParserDialog();
      _testdialog.textLabel->setText(func->text());
      _testdialog.exec();
    }
}

void FuncInterpolationDialog::generate()
{
    qDebug()<<"generate";
    if (func->text().isEmpty()) {
      return;
    }
    /*
    ScribbleArea *sa = (ScribbleArea *) _scribblearea;
    if (sa->_tria == NULL) {
      return;
    }
    if (sa->_tria->tr_tris->objects == 0) {
      return;
    }

    // Set parameters.
    if (sa->_tria->OMT_domain != NULL) {
      // The function vlaue will be interpolated from a background mesh (.mtr).
      // ....
    } else {
      if (func->text().isEmpty()) {
        sa->_tria->Func = NULL;
        return;
      }
      User_Func.SetExpr(std::string(func->text().toStdString()));
      User_Func.DefineVar("x", &user_x);
      User_Func.DefineVar("y", &user_y);
      sa->_tria->Func = &callback_user_func;

      if (!func_x->text().isEmpty()) {
        User_Func_X.SetExpr(std::string(func_x->text().toStdString()));
        User_Func_X.DefineVar("x", &user_x);
        User_Func_X.DefineVar("y", &user_y);
        sa->_tria->GradX = &callback_user_func_X;
      } else {
        sa->_tria->GradX = NULL;
      }

      if (!func_y->text().isEmpty()) {
        User_Func_Y.SetExpr(std::string(func_y->text().toStdString()));
        User_Func_Y.DefineVar("x", &user_x);
        User_Func_Y.DefineVar("y", &user_y);
        sa->_tria->GradY = &callback_user_func_Y;
      } else {
        sa->_tria->GradY = NULL;
      }
    }

    sa->_tria->op_use_coarsening = use_coarsen->isChecked();
    sa->_tria->op_use_splitting = use_refine->isChecked();
    sa->_tria->op_use_smoothing = use_smooth->isChecked();

    sa->_tria->op_hde_s1 = hde_s1->text().toDouble();
    sa->_tria->op_hde_s2 = hde_s2->text().toDouble();
    sa->_tria->op_target_length = target_length->text().toDouble();
    sa->_tria->op_smooth_deltat = 0.1;

    sa->_tria->op_db_verbose = 10; // Debug

    sa->_tria->op_metric = METRIC_HDE;

    sa->detri2_HDE_act();

    // Reset parameters.
    sa->_tria->op_metric = METRIC_Euclidean;
    //sa->_tria->Func = NULL;
    //sa->_tria->GradX = NULL;
    //sa->_tria->GradY = NULL;

    sa->_tria->op_use_coarsening = 0;
    sa->_tria->op_use_splitting = 0;
    sa->_tria->op_use_smoothing = 0;

    sa->_tria->op_hde_s1 = 1.0;
    sa->_tria->op_hde_s2 = 1.0;
    sa->_tria->op_target_length = 0.0;
    sa->_tria->op_smooth_deltat = 0.0;

    sa->_tria->op_db_verbose = 0;
    */
}

