#include "imageview.h"
#include <QHBoxLayout>
#include <QMouseEvent>
#include <QTimer>
#include <QToolBar>
#include <QGroupBox>
#include <QDebug>

ImageView::ImageView(QWidget *parent) : QWidget(parent), imageLabel(new QLabel)
{
    num = 0;
    images = NULL;
    img_idx = 0;

    imageLabel->setBackgroundRole(QPalette::Base);
    imageLabel->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Ignored);
    imageLabel->setScaledContents(false);

    _slider = new QSlider(Qt::Horizontal);

    // Create actions
    QPixmap pix_play(":/images/icon-play.png");
    QPixmap pix_pause(":/images/icon-pause.png");
    QPixmap pix_first(":/images/icon-gotofirst.png");
    QPixmap pix_last(":/images/icon-gotolast.png");
    QPixmap pix_next(":/images/icon-playnext.png");
    QPixmap pix_prev(":/images/icon-playprev.png");

    playAct = new QAction(QIcon(pix_play), tr("Play"), this);
    playAct->setShortcut(QKeySequence(tr("P")));
    playAct->setCheckable(true);
    connect(playAct, &QAction::triggered, this, &ImageView::play_act);

    pauseAct = new QAction(QIcon(pix_pause), tr("Pause"), this);
    pauseAct->setShortcut(QKeySequence(tr("T")));
    pauseAct->setCheckable(true);
    connect(pauseAct, &QAction::triggered, this, &ImageView::play_act);

    gotofirstAct = new QAction(QIcon(pix_first), tr("Goto Begin"), this);
    gotofirstAct->setShortcut(QKeySequence(tr("B")));
    gotofirstAct->setCheckable(true);
    connect(gotofirstAct, &QAction::triggered, this, &ImageView::play_act);

    gotolastAct = new QAction(QIcon(pix_last), tr("Goto End"), this);
    gotolastAct->setShortcut(QKeySequence(tr("E")));
    gotolastAct->setCheckable(true);
    connect(gotolastAct, &QAction::triggered, this, &ImageView::play_act);

    forwardAct = new QAction(QIcon(pix_next), tr("Play Next"), this);
    forwardAct->setShortcut(QKeySequence(tr("N")));
    forwardAct->setCheckable(true);
    connect(forwardAct, &QAction::triggered, this, &ImageView::play_act);

    backwardAct = new QAction(QIcon(pix_prev), tr("Play Prev"), this);
    backwardAct->setShortcut(QKeySequence(tr("V")));
    backwardAct->setCheckable(true);
    connect(backwardAct, &QAction::triggered, this, &ImageView::play_act);

    // Mouse actions
    QActionGroup *actionGroup = new QActionGroup(this);
    actionGroup->addAction(playAct);
    actionGroup->addAction(pauseAct);
    actionGroup->addAction(gotofirstAct);
    actionGroup->addAction(gotolastAct);
    actionGroup->addAction(forwardAct);
    actionGroup->addAction(backwardAct);

    QToolBar *_toolbar_top = new QToolBar();
    //addToolBar(Qt::TopToolBarArea, _toolbar_top);
    _toolbar_top->addAction(gotofirstAct);
    _toolbar_top->addAction(backwardAct);
    _toolbar_top->addAction(playAct);
    _toolbar_top->addAction(pauseAct);
    _toolbar_top->addAction(forwardAct);
    _toolbar_top->addAction(gotolastAct);

    _toolbar_top->addSeparator();
    QLabel *MaxCountLabel = new QLabel(tr("#Images "));
    imageMaxCount = new QLabel(tr("0"));
    QLabel *CurCountLabel = new QLabel(tr("Current Index "));
    imageCountLabel = new QLabel(tr("0"));
    _toolbar_top->addWidget(MaxCountLabel);
    _toolbar_top->addWidget(imageMaxCount);
    _toolbar_top->addWidget(CurCountLabel);
    _toolbar_top->addWidget(imageCountLabel);

    //setCentralWidget(imageLabel);
    QVBoxLayout *layout = new QVBoxLayout();
    layout->addWidget(_toolbar_top);
    layout->addWidget(imageLabel);
    layout->addWidget(_slider);
    setLayout(layout);

    timer = NULL;
    //timer = new QTimer(this);
    //connect(timer, SIGNAL(timeout()), this, SLOT(update_image()));
    connect(_slider, SIGNAL(valueChanged(int)), this, SLOT(setSliderValue(int)));
    connect(_slider, SIGNAL(sliderMoved(int)), this, SLOT(setSliderValue(int)));

    resize(200,200); // default
}

void ImageView::loadImages(detri2::arraypool *imgarray)
{
    if (num > 0) {
      delete [] images;
    }

    num = imgarray->objects;
    images = new QImage[num];

    for (int i = 0; i < num; i++) {
      QImage **pimg = (QImage **) imgarray->get(i);
      images[i] = *(*pimg);
    }

    img_idx = 0;
    setImage(&(images[0]));
    //img_idx++;

    _slider->setMinimum(0);
    _slider->setMaximum(num-1);
    setSliderValue(0);
    imageMaxCount->setText(QString::number(num));
    imageCountLabel->setText(QString::number(img_idx));
    qDebug()<<" slider maximum ="<<num-1;

    //timer->start(100); // set a 0.1 sec. interval
    resize(images[0].width(), images[0].height());
}

void ImageView::setSliderValue(int i)
{
    //qDebug()<<" set value i="<<i;
    if (i < num) {
      img_idx = i;
      setImage(&(images[img_idx]));
      imageCountLabel->setText(QString::number(img_idx));
    }
}

void ImageView::setImage(const QImage *newImage)
{
    imageLabel->setPixmap(QPixmap::fromImage(*newImage));
}

void ImageView::update_image()
{
    //qDebug()<<" update_image, idx="<<img_idx;
    if (img_idx < num) {
      setImage(&(images[img_idx]));
      _slider->setValue(img_idx);
      imageCountLabel->setText(QString::number(img_idx));
      img_idx++;
    } else {
      //hide(); // end of playing.
      if (timer) {
        timer->stop();
        delete timer;
        timer = NULL;
      }
    }
}

void ImageView::play_act()
{
    if (playAct->isChecked()) {
      qDebug()<< "play";
      timer = new QTimer(this);
      connect(timer, SIGNAL(timeout()), this, SLOT(update_image()));
      timer->start(100); // set a 0.1 sec. interval
    }
    if (pauseAct->isChecked()) {
      qDebug()<< "pause";
      if (timer) {
        timer->stop();
        delete timer;
        timer = NULL;
      }
    }
    if (gotofirstAct->isChecked()) {
      qDebug()<< "go to first";
      img_idx = 0;
      //update_image();
      setImage(&(images[img_idx]));
      _slider->setValue(img_idx);
      imageCountLabel->setText(QString::number(img_idx));
    }
    if (gotolastAct->isChecked()) {
      qDebug()<< "go to last";
      img_idx = num - 1;
      //update_image();
      setImage(&(images[img_idx]));
      _slider->setValue(img_idx);
      imageCountLabel->setText(QString::number(img_idx));
    }
    if (forwardAct->isChecked()) {
      //qDebug()<< "play next";
      if (img_idx < (num - 1)) {
        img_idx++;
        setImage(&(images[img_idx]));
        _slider->setValue(img_idx);
        imageCountLabel->setText(QString::number(img_idx));
      }
    }
    if (backwardAct->isChecked()) {
      //qDebug()<< "play prev";
      if (img_idx > 0) {
        img_idx--;
        setImage(&(images[img_idx]));
        _slider->setValue(img_idx);
        imageCountLabel->setText(QString::number(img_idx));
      }
    }
}

/*
void ImageView::mousePressEvent(QMouseEvent *event)
{
    if (event->button() == Qt::LeftButton) {
      if (timer->isActive()) {
        timer->stop();
      } else {
        timer->start();
      }
    }
}
*/
