
//#include "scribblearea.h"
#include "metricpanel.h"
#include <QInputDialog>
#include <QMessageBox>
#include <QDebug>
#include <math.h>

#include <QPainter>

#include "../detri2/detri2.h" // metric, get_orthocenter

RenderArea::RenderArea(QWidget *parent)
    : QWidget(parent)
{
    shape = Polygon;
    antialiased = false;
    transformed = false;
    pixmap.load(":/images/qt-logo.png");

    setBackgroundRole(QPalette::Base);
    setAutoFillBackground(true);

    ctx = show_metric; // default

    a11 = 1.0;
    a21 = 0.0;
    a22 = 1.0;

    _px = 0.0; _py = 0.0; _wp = 0.;
    _qx = 1.0; _qy = 0.0; _wq = 0.;
    _rx = 0.0; _ry = 1.0; _wr = 0.;
}

QSize RenderArea::minimumSizeHint() const
{
    return QSize(100, 100);
}

QSize RenderArea::sizeHint() const
{
    return QSize(400, 200);
}

void RenderArea::setShape(Shape shape)
{
    this->shape = shape;
    update();
}

void RenderArea::setPen(const QPen &pen)
{
    this->pen = pen;
    update();
}

void RenderArea::setBrush(const QBrush &brush)
{
    this->brush = brush;
    update();
}

void RenderArea::setAntialiased(bool antialiased)
{
    this->antialiased = antialiased;
    update();
}

void RenderArea::setTransformed(bool transformed)
{
    this->transformed = transformed;
    update();
}

void RenderArea::redraw()
{
    update();
}

//===================================================================================
/*
static void get_MatVVP2x2(double a11, double a21, double a22,
               double *lambda1, double *lambda2, double vec1[2], double vec2[2])
{
    double c11 = a11*a11, c22 = a22*a22, c21= a21*a21;
    double b=-a11-a22,c=-c21+a11*a22;
    double delta = b*b - 4 * c ;
    double n2=(c11+c22+c21);

    if ( n2 < 1.e-30) {
       *lambda1 = *lambda2 = 0.;
       vec1[0] = 1; vec1[1] = 0; // v.x=1,v.y=0;
      }
    else if (delta < 1.e-5*n2) // eps=1e-5
      {
        *lambda1 = *lambda2 = -b/2;
        vec1[0] = 1; vec1[1] = 0; // v.x=1,v.y=0;
      }
    else
      {  //    ---  construction  de 2 vecteur dans (Im ( A - D(i) Id) ortogonal
         // construction of 2 vectors in (Im ( A - D(i) Id) ortogonal
        delta = sqrt(delta);
        *lambda1 = (-b-delta)/2.0;
        *lambda2 = (-b+delta)/2.0;
        double v0 = a11- *lambda1, v1 = a21, v2 = a22 - *lambda1;
        double s0 = v0*v0 + v1*v1, s1 = v1*v1 +v2*v2;

        if (s1 < s0) {
          s0 = sqrt(s0);
          vec1[0]=v1/s0; vec1[1]=-v0/s0;  //v.x=v1/s0,v.y=-v0/s0;
        }
        else
        {
          s1=sqrt(s1);
          vec1[0]=v2/s1; vec1[1]=-v1/s1; //v.x=v2/s1,v.y=-v1/s1;
        }
    }
    // The eigenvector of lambda2.
    vec2[0] =  vec1[1];
    vec2[1] = -vec1[0];
}

const double PI = 3.14159265358979323846;

static double get_rotation_angle(double vec[2])
{
    double theta;

    if (fabs(vec[0]) > 1.e-8) {
      theta = atan(fabs(vec[1]/vec[0]));
      if (vec[0] > 0.) {
        if (vec[1] >= 0.) {
          theta += 0.;
        } else {
          theta += (PI / 2. * 3.);
        }
      } else {
        if (vec[1] >= 0) {
          theta += (PI / 2.);
        } else {
          theta += PI;
        }
      }
    } else {
      if (vec[1] >= 0.) {
        theta = (PI / 2.);
      } else {
        theta = (PI / 2. * 3.);
      }
    }

    return theta;
}
*/

//======================================================================================

/*
void ScribbleArea::calc_trans(double xmin, double xmax, double ymin, double ymax)
{
    // remember the viewport.
    _viewport_min[0] = xmin;
    _viewport_min[1] = ymin;
    _viewport_max[0] = xmax;
    _viewport_max[1] = ymax;


    double w = (double)width()-20;
    double h = (double)height()-20;

    double dx = xmax - xmin;
    double dy = ymax - ymin;
    double min_hw = h < w ? h : w;
    if (dy < dx) {
      _scale_xy = min_hw / dy;
    } else {
      _scale_xy = min_hw / dx;
    }
    _cx = w / 2. - (xmin + xmax) / 2.0 * _scale_xy + 10;
    _cy = h / 2. + (ymin + ymax) / 2.0 * _scale_xy + 10;

    qDebug()<<"width="<<w<<", height="<<h;
    qDebug()<<"xmin="<<xmin<<", xmax="<<xmax<<", ymin="<<ymin<<", ymax="<<ymax;
    qDebug()<<"Cx="<<_cx<<", Cy="<<_cy<<", Sxy="<<_scale_xy;
}
*/

void RenderArea::draw_metric()
{
    double lambda1, lambda2, v1[2], v2[2];
    detri2::get_MatVVP2x2(a11, a21, a22,
                  &lambda1, &lambda2, v1, v2);
    double theta = detri2::get_rotation_angle(v1) / detri2::PI * 180.0;

    QPainter painter(this);

    QPen textPen;
    textPen.setColor(Qt::black);
    textPen.setWidth(2);
    painter.setPen(textPen);

    QFont serifFont("Times", 20, QFont::Normal); // QFont::Bold
    //QFont sansFont("Helvetica [Cronyx]", 12);
    painter.setFont(serifFont);
    //painter.setFont(_font_vertex_label);

    QString text;
    text.sprintf("lambda1 = %f", lambda1);
    painter.drawText(QPoint(10,20), text);
    text.sprintf("lambda2 = %f", lambda2);
    painter.drawText(QPoint(10,45), text);
    text.sprintf("theta = %.2f", theta);
    painter.drawText(QPoint(10,70), text);

    if (lambda1 <= 0. || lambda2 <= 0.0) {
      return; // no draw.
    }

    painter.setPen(pen);
    painter.setBrush(brush);
    if (antialiased) {
      painter.setRenderHint(QPainter::Antialiasing, true);
    }

    double r1 = sqrt(1.0 / lambda1);
    double r2 = sqrt(1.0 / lambda2);

    double max_r = r1 > r2 ? r1 : r2;
    max_r *= 1.2; // set the margin space.

    // Calculate the transformation.
    double cx, cy, scale_xy;

    double xmin = -max_r;
    double xmax =  max_r;
    double ymin = -max_r;
    double ymax =  max_r;

    double cct[2];
    cct[0] = 0.5 * (xmin + xmax);
    cct[1] = 0.5 * (ymin + ymax);

    double w = (double)width()-20;
    double h = (double)height()-20;
    double dx = xmax - xmin;
    double dy = ymax - ymin;

    double min_hw = h < w ? h : w;
    if (dy < dx) {
      scale_xy = min_hw / dy;
    } else {
      scale_xy = min_hw / dx;
    }
    cx = w / 2. - (xmin + xmax) / 2.0 * scale_xy + 10;
    cy = h / 2. + (ymin + ymax) / 2.0 * scale_xy + 10;

    // Draw the ellipse.
    int dr1 = (int) ( scale_xy * r1);
    int dr2 = (int) ( scale_xy * r2);

    int px = (int) ( scale_xy * cct[0] + cx);
    int py = (int) (-scale_xy * cct[1] + cy);

    painter.save();
    painter.translate(px, py);
    painter.rotate(-theta); // cw rotation
    painter.translate(-px, -py);
    painter.drawEllipse(QPoint(px,py), dr1, dr2);
    painter.restore();

    painter.setRenderHint(QPainter::Antialiasing, false);
    painter.setPen(palette().dark().color());
    painter.setBrush(Qt::NoBrush);
    painter.drawRect(QRect(0, 0, width() - 1, height() - 1));
}

void RenderArea::draw_transform()
{
    double lambda1, lambda2, v1[2], v2[2];
    detri2::get_MatVVP2x2(a11, a21, a22,
                  &lambda1, &lambda2, v1, v2);
    double theta = detri2::get_rotation_angle(v1) / detri2::PI * 180.0;

    QPainter painter(this);

    QPen textPen;
    textPen.setColor(Qt::black);
    textPen.setWidth(2);
    painter.setPen(textPen);

    QFont serifFont("Times", 20, QFont::Normal); // QFont::Bold
    //QFont sansFont("Helvetica [Cronyx]", 12);
    painter.setFont(serifFont);
    //painter.setFont(_font_vertex_label);

    QString text;
    text.sprintf("lambda1 = %f", lambda1);
    painter.drawText(QPoint(10,20), text);
    text.sprintf("lambda2 = %f", lambda2);
    painter.drawText(QPoint(10,45), text);
    text.sprintf("theta = %.2f", theta);
    painter.drawText(QPoint(10,70), text);

    if (lambda1 <= 0. || lambda2 <= 0.0) {
      return; // no draw.
    }

    // Caluclate the transformation.
    double px = v1[0]*lambda1*_px + v2[0]*lambda2*_py;
    double py = v1[1]*lambda1*_px + v2[1]*lambda2*_py;

    double qx = v1[0]*lambda1*_qx + v2[0]*lambda2*_qy;
    double qy = v1[1]*lambda1*_qx + v2[1]*lambda2*_qy;

    double rx = v1[0]*lambda1*_rx + v2[0]*lambda2*_ry;
    double ry = v1[1]*lambda1*_rx + v2[1]*lambda2*_ry;

    //============================================================
    // The bounding box.
    double xmin = px, xmax = px;
    double ymin = py, ymax = py;
    xmin = xmin < qx ? xmin : qx;
    xmax = xmax > qx ? xmax : qx;
    xmin = xmin < rx ? xmin : rx;
    xmax = xmax > rx ? xmax : rx;
    ymin = ymin < qy ? ymin : qy;
    ymax = ymax > qy ? ymax : qy;
    ymin = ymin < ry ? ymin : ry;
    ymax = ymax > ry ? ymax : ry;

    double max_d = xmax - xmin;
    max_d = max_d > (ymax - ymin) ? max_d : (ymax - ymin);
    double dd = max_d * 0.2;
    xmin -= dd;
    xmax += dd;
    ymin -= dd;
    ymax += dd;

    // Calculate the transformation.
    double cx, cy, scale_xy;

    //double cct[2];
    //cct[0] = 0.5 * (xmin + xmax);
    //cct[1] = 0.5 * (ymin + ymax);

    double w = (double)width()-20;
    double h = (double)height()-20;
    double dx = xmax - xmin;
    double dy = ymax - ymin;

    double min_hw = h < w ? h : w;
    if (dy < dx) {
      scale_xy = min_hw / dy;
    } else {
      scale_xy = min_hw / dx;
    }
    cx = w / 2. - (xmin + xmax) / 2.0 * scale_xy + 10;
    cy = h / 2. + (ymin + ymax) / 2.0 * scale_xy + 10;

    //======================================================
    painter.setPen(pen);    
    if (antialiased) {
      painter.setRenderHint(QPainter::Antialiasing, true);
    }

    // Draw three vertices.
    brush.setStyle(Qt::SolidPattern);
    painter.setBrush(brush);

    int PX, PY, QX, QY, RX, RY;
    PX = (int) ( scale_xy * px + cx);
    PY = (int) (-scale_xy * py + cy);
    painter.drawEllipse(QPoint(PX,PY), 4, 4);
    QX = (int) ( scale_xy * qx + cx);
    QY = (int) (-scale_xy * qy + cy);
    painter.drawEllipse(QPoint(QX,QY), 4, 4);
    RX = (int) ( scale_xy * rx + cx);
    RY = (int) (-scale_xy * ry + cy);
    painter.drawEllipse(QPoint(RX,RY), 4, 4);

    brush.setStyle(Qt::NoBrush);
    painter.setBrush(brush);

    // Draw triangle.
    painter.drawLine(QPoint(PX, PY), QPoint(QX, QY));
    painter.drawLine(QPoint(PX, PY), QPoint(RX, RY));
    painter.drawLine(QPoint(RX, RY), QPoint(QX, QY));

    // Draw orthocircle
    double Cx, Cy, radius2;
    _wp = _wq = _wr = 0.0;
    detri2::get_orthocenter(px, py, _wp,
                            qx, qy, _wq,
                            rx, ry, _wr,
                            &Cx, &Cy, &radius2,
                            a11, a21, a22);
    double r1 = sqrt(1.0 / lambda1 * radius2);
    double r2 = sqrt(1.0 / lambda2 * radius2);

    int CX, CY;
    CX = (int) ( scale_xy * Cx + cx);
    CY = (int) (-scale_xy * Cy + cy);
    int dr1 = (int) ( scale_xy * r1);
    int dr2 = (int) ( scale_xy * r2);

    painter.save();
    painter.translate(CX, CY);
    painter.rotate(-theta); // cw rotation
    painter.translate(-CX, -CY);
    painter.drawEllipse(QPoint(CX,CY), dr1, dr2);
    painter.restore();
}

//======================================================================================
void RenderArea::paintEvent(QPaintEvent * /* event */)
{
    if (ctx == show_metric) {
      draw_metric();
    } else if (ctx == show_transform) {
      draw_transform();
    }
    return;

    /*
    static const QPoint points[4] = {
        QPoint(10, 80),
        QPoint(20, 10),
        QPoint(80, 30),
        QPoint(90, 70)
    };

    QRect rect(10, 20, 80, 60);

    QPainterPath path;
    path.moveTo(20, 80);
    path.lineTo(20, 30);
    path.cubicTo(80, 0, 50, 50, 80, 80);

    int startAngle = 20 * 16;
    int arcLength = 120 * 16;

    QPainter painter(this);
    painter.setPen(pen);
    painter.setBrush(brush);
    if (antialiased)
        painter.setRenderHint(QPainter::Antialiasing, true);
    */

    /*
    for (int x = 0; x < width(); x += 100) {
        for (int y = 0; y < height(); y += 100) {
            painter.save();
            painter.translate(x, y);

            if (transformed) {
                painter.translate(50, 50);
                painter.rotate(60.0);
                painter.scale(0.6, 0.9);
                painter.translate(-50, -50);
            }

            switch (shape) {
            case Line:
                painter.drawLine(rect.bottomLeft(), rect.topRight());
                break;
            case Points:
                painter.drawPoints(points, 4);
                break;
            case Polyline:
                painter.drawPolyline(points, 4);
                break;
            case Polygon:
                painter.drawPolygon(points, 4);
                break;
            case Rect:
                painter.drawRect(rect);
                break;
            case RoundedRect:
                painter.drawRoundedRect(rect, 25, 25, Qt::RelativeSize);
                break;
            case Ellipse:
                painter.drawEllipse(rect);
                break;
            case Arc:
                painter.drawArc(rect, startAngle, arcLength);
                break;
            case Chord:
                painter.drawChord(rect, startAngle, arcLength);
                break;
            case Pie:
                painter.drawPie(rect, startAngle, arcLength);
                break;
            case Path:
                painter.drawPath(path);
                break;
            case Text:
                painter.drawText(rect,
                                 Qt::AlignCenter,
                                 tr("Qt by\nThe Qt Company"));
                break;
            case Pixmap:
                painter.drawPixmap(10, 10, pixmap);
            }
            painter.restore();
        }
    }
    */    
}

//==================================================================================

static double get_double(double input=0)
{
    double value = input;
    bool ok;
    QString text = QInputDialog::getText(NULL, // Qwidget
                                               QString("get double"), // title
                                               QString("value = "), // hint,  // label
                                               QLineEdit::Normal, QString::number(value), &ok);

    if (ok && !text.isEmpty()) {
      value = text.toDouble();
    }

    return value;
}

void metricDialog::set_a11()
{
  double value = get_double(a11Label->text().toDouble());
  a11Label->setText(QString::number(value));

  /*
  //qDebug()<<"set_a11";
  double value = a11Label->text().toDouble();

  bool ok;
  QString text = QInputDialog::getText(this, // Qwidget
                                             tr("set a11"), // title
                                             tr("a11 = "), // hint,  // label
                                             QLineEdit::Normal, QString::number(value), &ok);

  if (ok && !text.isEmpty()) {
    value = text.toDouble();
    a11Label->setText(QString::number(value));
  }
  */
}

void metricDialog::set_a21()
{
  //qDebug()<<"set_a21";
  double value = a21Label->text().toDouble();
  bool ok;
  //double d = QInputDialog::getDouble(this, tr("set a21"),
  //                                   tr("a21 = "), value, -10000, 10000, 2, &ok);
  //if (ok)
  //    a21Label->setText(QString::number(d));

  QString text = QInputDialog::getText(this, // Qwidget
                                             tr("set a21"), // title
                                             tr("a21 = "), // hint,  // label
                                             QLineEdit::Normal, QString::number(value), &ok);

  if (ok && !text.isEmpty()) {
    value = text.toDouble();
    a21Label->setText(QString::number(value));
  }
}

void metricDialog::set_a22()
{
  //qDebug()<<"set_a22";
  double value = a22Label->text().toDouble();
  bool ok;
  QString text = QInputDialog::getText(this, // Qwidget
                                             tr("set a22"), // title
                                             tr("a22 = "), // hint,  // label
                                             QLineEdit::Normal, QString::number(value), &ok);

  if (ok && !text.isEmpty()) {
    value = text.toDouble();
    a22Label->setText(QString::number(value));
  }
}

void metricDialog::set_radius1()
{
    double value = r1Label->text().toDouble();
    bool ok;
    QString text = QInputDialog::getText(this, // Qwidget
                                               tr("set r1"), // title
                                               tr("r1 = "), // hint,  // label
                                               QLineEdit::Normal, QString::number(value), &ok);

    if (ok && !text.isEmpty()) {
      value = text.toDouble();
      r1Label->setText(QString::number(value));
    }
}

void metricDialog::set_radius2()
{
    double value = r2Label->text().toDouble();
    bool ok;
    QString text = QInputDialog::getText(this, // Qwidget
                                               tr("set r2"), // title
                                               tr("r2 = "), // hint,  // label
                                               QLineEdit::Normal, QString::number(value), &ok);

    if (ok && !text.isEmpty()) {
      value = text.toDouble();
      r2Label->setText(QString::number(value));
    }
}

void metricDialog::set_theta1()
{
    double value = thetaLabel->text().toDouble();
    bool ok;
    QString text = QInputDialog::getText(this, // Qwidget
                                               tr("set theta1"), // title
                                               tr("theta1 = "), // hint,  // label
                                               QLineEdit::Normal, QString::number(value), &ok);

    if (ok && !text.isEmpty()) {
      value = text.toDouble();
      thetaLabel->setText(QString::number(value));
    }
}

void metricDialog::on_draw()
{
  renderArea->a11 = a11Label->text().toDouble();
  renderArea->a21 = a21Label->text().toDouble();
  renderArea->a22 = a22Label->text().toDouble();

  // Update r1, r2, theta1
  double lambda1, lambda2, v1[2], v2[2];
  detri2::get_MatVVP2x2(renderArea->a11, renderArea->a21, renderArea->a22,
                &lambda1, &lambda2, v1, v2);
  double theta = detri2::get_rotation_angle(v1) / detri2::PI * 180.0;

  if (lambda1 > 0) {
    r1Label->setText(QString::number(sqrt(1./lambda1)));
  } else if (lambda1 < 0) {
    r1Label->setText(tr("!! negative eigenvalue"));
  } else {
    r1Label->setText(tr("!! Zero eigenvalue"));
  }
  if (lambda2 > 0) {
    r2Label->setText(QString::number(sqrt(1./lambda2)));
  } else if (lambda2 < 0) {
    r2Label->setText(tr("!! negative eigenvalue"));
  } else {
    r2Label->setText(tr("!! Zero eigenvalue"));
  }
  thetaLabel->setText(QString::number(theta));

  renderArea->redraw();
}

void metricDialog::on_draw_ellipse()
{
    //renderArea->a11 = a11Label->text().toDouble();
    //renderArea->a21 = a21Label->text().toDouble();
    //renderArea->a22 = a22Label->text().toDouble();

    // Update r1, r2, theta1
    double r1 = r1Label->text().toDouble();
    double r2 = r2Label->text().toDouble();
    double theta = thetaLabel->text().toDouble() / 180. * detri2::PI;
    double v1[2], v2[2];
    // Rotate [1,0]^T theta degree CCW direction
    v1[0] =  cos(theta);
    v1[1] =  sin(theta);
    v2[0] = -sin(theta);
    v2[1] =  cos(theta);

    double lambda1 = 1. / (r1*r1);
    double lambda2 = 1. / (r2*r2);

    // get a11, a21, a22
    renderArea->a11 = lambda1*v1[0]*v1[0]+lambda2*v2[0]*v2[0];
    renderArea->a21 = lambda1*v1[0]*v1[1]+lambda2*v2[0]*v2[1];
    renderArea->a22 = lambda1*v1[1]*v1[1]+lambda2*v2[1]*v2[1];

    a11Label->setText(QString::number(renderArea->a11));
    a21Label->setText(QString::number(renderArea->a21));
    a22Label->setText(QString::number(renderArea->a22));

    renderArea->redraw();
}

void metricDialog::on_reset()
{
  renderArea->a11 = 1.0;
  renderArea->a21 = 0.0;
  renderArea->a22 = 1.0;

  a11Label->setText(QString::number(renderArea->a11));
  a21Label->setText(QString::number(renderArea->a21));
  a22Label->setText(QString::number(renderArea->a22));

  r1Label->setText(QString::number(1.0));
  r2Label->setText(QString::number(1.0));
  thetaLabel->setText(QString::number(0.0));
  renderArea->redraw();
}

void metricDialog::on_apply()
{
    QString text;
    text.sprintf("a11=%f, a21=%f, a22=%f",
                 a11Label->text().toDouble(),
                 a21Label->text().toDouble(),
                 a22Label->text().toDouble());

    QMessageBox::StandardButton reply;
    reply = QMessageBox::question(this, tr("Use current metric?"),
                                  text,
                                  QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel);
    if (reply == QMessageBox::Yes) {
        //questionLabel->setText(tr("Yes"));
        *_a11 = a11Label->text().toDouble();
        *_a21 = a21Label->text().toDouble();
        *_a22 = a22Label->text().toDouble();
    } else if (reply == QMessageBox::No) {
        //questionLabel->setText(tr("No"));
    } else {
        //questionLabel->setText(tr("Cancel"));
    }
    close();
}

//==================================================================================

const int IdRole = Qt::UserRole;

metricDialog::metricDialog(double *A11, double *A21, double *A22)
{
    renderArea = new RenderArea;
    renderArea->ctx = RenderArea::show_metric;

    _a11 = A11, _a21 = A21, _a22 = A22;
    renderArea->a11 = *_a11;
    renderArea->a21 = *_a21;
    renderArea->a22 = *_a22;

    double lambda1, lambda2, v1[2], v2[2], theta;
    detri2::get_MatVVP2x2(*A11, *A21, *A22, &lambda1, &lambda2, v1, v2);
    theta = detri2::get_rotation_angle(v1);

    int frameStyle = QFrame::Sunken | QFrame::Panel;
    //a11Label->setStyleSheet("QLineEdit { background: rgb(0, 255, 255); selection-background-color: rgb(233, 99, 0); }");

    a11Label = new QLabel;
    a11Label->setFrameStyle(frameStyle);    
    a11Label->setText(QString::number(*_a11));
    QPushButton *btn_a11 = new QPushButton(tr("a11"));
    a21Label = new QLabel;
    a21Label->setFrameStyle(frameStyle);
    a21Label->setText(QString::number(*_a21));
    QPushButton *btn_a21 = new QPushButton(tr("a21"));
    a22Label = new QLabel;
    a22Label->setFrameStyle(frameStyle);
    a22Label->setText(QString::number(*_a22));
    QPushButton *btn_a22 = new QPushButton(tr("a22"));

    r1Label = new QLabel;
    r1Label->setFrameStyle(frameStyle);
    r1Label->setText(QString::number(sqrt(1./lambda1)));
    QPushButton *btn_r1 = new QPushButton(tr("r1"));
    r2Label = new QLabel;
    r2Label->setFrameStyle(frameStyle);
    r2Label->setText(QString::number(sqrt(1./lambda2)));
    QPushButton *btn_r2 = new QPushButton(tr("r2"));
    thetaLabel = new QLabel;
    thetaLabel->setFrameStyle(frameStyle);
    thetaLabel->setText(QString::number(theta / detri2::PI * 180.));
    QPushButton *btn_theta = new QPushButton(tr("theta"));

    QPushButton *btn_show  = new QPushButton(tr("draw metric"));
    QPushButton *btn_show_ellipse  = new QPushButton(tr("draw ellipse"));
    QPushButton *btn_apply = new QPushButton(tr("apply"));
    QPushButton *btn_reset = new QPushButton(tr("reset"));

    shapeComboBox = new QComboBox;
    shapeComboBox->addItem(tr("Polygon"), RenderArea::Polygon);
    shapeComboBox->addItem(tr("Rectangle"), RenderArea::Rect);
    shapeComboBox->addItem(tr("Rounded Rectangle"), RenderArea::RoundedRect);
    shapeComboBox->addItem(tr("Ellipse"), RenderArea::Ellipse);
    shapeComboBox->addItem(tr("Pie"), RenderArea::Pie);
    shapeComboBox->addItem(tr("Chord"), RenderArea::Chord);
    shapeComboBox->addItem(tr("Path"), RenderArea::Path);
    shapeComboBox->addItem(tr("Line"), RenderArea::Line);
    shapeComboBox->addItem(tr("Polyline"), RenderArea::Polyline);
    shapeComboBox->addItem(tr("Arc"), RenderArea::Arc);
    shapeComboBox->addItem(tr("Points"), RenderArea::Points);
    shapeComboBox->addItem(tr("Text"), RenderArea::Text);
    shapeComboBox->addItem(tr("Pixmap"), RenderArea::Pixmap);

    shapeLabel = new QLabel(tr("&Shape:"));
    shapeLabel->setBuddy(shapeComboBox);
//! [1]

//! [2]
    penWidthSpinBox = new QSpinBox;
    penWidthSpinBox->setRange(0, 20);
    penWidthSpinBox->setSpecialValueText(tr("0 (cosmetic pen)"));

    penWidthLabel = new QLabel(tr("Pen &Width:"));
    penWidthLabel->setBuddy(penWidthSpinBox);
//! [2]

//! [3]
    penStyleComboBox = new QComboBox;
    penStyleComboBox->addItem(tr("Solid"), static_cast<int>(Qt::SolidLine));
    penStyleComboBox->addItem(tr("Dash"), static_cast<int>(Qt::DashLine));
    penStyleComboBox->addItem(tr("Dot"), static_cast<int>(Qt::DotLine));
    penStyleComboBox->addItem(tr("Dash Dot"), static_cast<int>(Qt::DashDotLine));
    penStyleComboBox->addItem(tr("Dash Dot Dot"), static_cast<int>(Qt::DashDotDotLine));
    penStyleComboBox->addItem(tr("None"), static_cast<int>(Qt::NoPen));

    penStyleLabel = new QLabel(tr("&Pen Style:"));
    penStyleLabel->setBuddy(penStyleComboBox);

    penCapComboBox = new QComboBox;
    penCapComboBox->addItem(tr("Flat"), Qt::FlatCap);
    penCapComboBox->addItem(tr("Square"), Qt::SquareCap);
    penCapComboBox->addItem(tr("Round"), Qt::RoundCap);

    penCapLabel = new QLabel(tr("Pen &Cap:"));
    penCapLabel->setBuddy(penCapComboBox);

    penJoinComboBox = new QComboBox;
    penJoinComboBox->addItem(tr("Miter"), Qt::MiterJoin);
    penJoinComboBox->addItem(tr("Bevel"), Qt::BevelJoin);
    penJoinComboBox->addItem(tr("Round"), Qt::RoundJoin);

    penJoinLabel = new QLabel(tr("Pen &Join:"));
    penJoinLabel->setBuddy(penJoinComboBox);
//! [3]

//! [4]
    brushStyleComboBox = new QComboBox;
    brushStyleComboBox->addItem(tr("Linear Gradient"),
            static_cast<int>(Qt::LinearGradientPattern));
    brushStyleComboBox->addItem(tr("Radial Gradient"),
            static_cast<int>(Qt::RadialGradientPattern));
    brushStyleComboBox->addItem(tr("Conical Gradient"),
            static_cast<int>(Qt::ConicalGradientPattern));
    brushStyleComboBox->addItem(tr("Texture"), static_cast<int>(Qt::TexturePattern));
    brushStyleComboBox->addItem(tr("Solid"), static_cast<int>(Qt::SolidPattern));
    brushStyleComboBox->addItem(tr("Horizontal"), static_cast<int>(Qt::HorPattern));
    brushStyleComboBox->addItem(tr("Vertical"), static_cast<int>(Qt::VerPattern));
    brushStyleComboBox->addItem(tr("Cross"), static_cast<int>(Qt::CrossPattern));
    brushStyleComboBox->addItem(tr("Backward Diagonal"), static_cast<int>(Qt::BDiagPattern));
    brushStyleComboBox->addItem(tr("Forward Diagonal"), static_cast<int>(Qt::FDiagPattern));
    brushStyleComboBox->addItem(tr("Diagonal Cross"), static_cast<int>(Qt::DiagCrossPattern));
    brushStyleComboBox->addItem(tr("Dense 1"), static_cast<int>(Qt::Dense1Pattern));
    brushStyleComboBox->addItem(tr("Dense 2"), static_cast<int>(Qt::Dense2Pattern));
    brushStyleComboBox->addItem(tr("Dense 3"), static_cast<int>(Qt::Dense3Pattern));
    brushStyleComboBox->addItem(tr("Dense 4"), static_cast<int>(Qt::Dense4Pattern));
    brushStyleComboBox->addItem(tr("Dense 5"), static_cast<int>(Qt::Dense5Pattern));
    brushStyleComboBox->addItem(tr("Dense 6"), static_cast<int>(Qt::Dense6Pattern));
    brushStyleComboBox->addItem(tr("Dense 7"), static_cast<int>(Qt::Dense7Pattern));
    brushStyleComboBox->addItem(tr("None"), static_cast<int>(Qt::NoBrush));

    brushStyleLabel = new QLabel(tr("&Brush:"));
    brushStyleLabel->setBuddy(brushStyleComboBox);
//! [4]

//! [5]
    otherOptionsLabel = new QLabel(tr("Options:"));
//! [5] //! [6]
    antialiasingCheckBox = new QCheckBox(tr("&Antialiasing"));
//! [6] //! [7]
    transformationsCheckBox = new QCheckBox(tr("&Transformations"));
//! [7]

//! [8]
    connect(shapeComboBox, SIGNAL(activated(int)),
            this, SLOT(shapeChanged()));
    connect(penWidthSpinBox, SIGNAL(valueChanged(int)),
            this, SLOT(penChanged()));
    connect(penStyleComboBox, SIGNAL(activated(int)),
            this, SLOT(penChanged()));
    connect(penCapComboBox, SIGNAL(activated(int)),
            this, SLOT(penChanged()));
    connect(penJoinComboBox, SIGNAL(activated(int)),
            this, SLOT(penChanged()));
    connect(brushStyleComboBox, SIGNAL(activated(int)),
            this, SLOT(brushChanged()));
    connect(antialiasingCheckBox, SIGNAL(toggled(bool)),
            renderArea, SLOT(setAntialiased(bool)));
    connect(transformationsCheckBox, SIGNAL(toggled(bool)),
            renderArea, SLOT(setTransformed(bool)));
//! [8]

    connect(btn_a11, &QAbstractButton::clicked, this, &metricDialog::set_a11);
    connect(btn_a21, &QAbstractButton::clicked, this, &metricDialog::set_a21);
    connect(btn_a22, &QAbstractButton::clicked, this, &metricDialog::set_a22);

    connect(btn_r1, &QAbstractButton::clicked, this, &metricDialog::set_radius1);
    connect(btn_r2, &QAbstractButton::clicked, this, &metricDialog::set_radius2);
    connect(btn_theta, &QAbstractButton::clicked, this, &metricDialog::set_theta1);

    connect(btn_show, &QAbstractButton::clicked, this, &metricDialog::on_draw);
    connect(btn_show_ellipse, &QAbstractButton::clicked, this, &metricDialog::on_draw_ellipse);

    connect(btn_reset, &QAbstractButton::clicked, this, &metricDialog::on_reset);    
    connect(btn_apply, &QAbstractButton::clicked, this, &metricDialog::on_apply);

//! [9]
    QGridLayout *mainLayout = new QGridLayout;
//! [9] //! [10]
    mainLayout->setColumnStretch(0, 1);
    mainLayout->setColumnStretch(3, 1);
    mainLayout->addWidget(renderArea, 0, 0, 1, 4);
    mainLayout->addWidget(shapeLabel, 2, 0, Qt::AlignRight);
    mainLayout->addWidget(shapeComboBox, 2, 1);
    mainLayout->addWidget(penWidthLabel, 3, 0, Qt::AlignRight);
    mainLayout->addWidget(penWidthSpinBox, 3, 1);
    mainLayout->addWidget(penStyleLabel, 4, 0, Qt::AlignRight);
    mainLayout->addWidget(penStyleComboBox, 4, 1);
    mainLayout->addWidget(penCapLabel, 3, 2, Qt::AlignRight);
    mainLayout->addWidget(penCapComboBox, 3, 3);
    mainLayout->addWidget(penJoinLabel, 2, 2, Qt::AlignRight);
    mainLayout->addWidget(penJoinComboBox, 2, 3);
    mainLayout->addWidget(brushStyleLabel, 4, 2, Qt::AlignRight);
    mainLayout->addWidget(brushStyleComboBox, 4, 3);
    mainLayout->addWidget(otherOptionsLabel, 5, 0, Qt::AlignRight);
    mainLayout->addWidget(antialiasingCheckBox, 5, 1, 1, 1, Qt::AlignRight);
    mainLayout->addWidget(transformationsCheckBox, 5, 2, 1, 2, Qt::AlignRight);

    // set metric
    mainLayout->addWidget(btn_a11, 6,0,1,1);
    mainLayout->addWidget(a11Label, 6,1,1,1);
    mainLayout->addWidget(btn_a21, 7,0,1,1);
    mainLayout->addWidget(a21Label, 7,1,1,1);
    mainLayout->addWidget(btn_a22, 7,2,1,1);
    mainLayout->addWidget(a22Label, 7,3,1,1);
    mainLayout->addWidget(btn_show,  6,3,1,1);
    mainLayout->addWidget(btn_r1, 8,0,1,1);
    mainLayout->addWidget(r1Label, 8,1,1,1);
    mainLayout->addWidget(btn_r2, 8,2,1,1);
    mainLayout->addWidget(r2Label, 8,3,1,1);
    mainLayout->addWidget(btn_theta, 9,0,1,1);
    mainLayout->addWidget(thetaLabel, 9,1,1,1);
    mainLayout->addWidget(btn_show_ellipse,  9,3,1,1);
    mainLayout->addWidget(btn_apply, 10,3,1,1);
    mainLayout->addWidget(btn_reset, 10,2,1,1);

    setLayout(mainLayout);

    shapeChanged();
    penChanged();
    brushChanged();
    antialiasingCheckBox->setChecked(true);

    renderArea->setBrush(Qt::NoBrush); // default option

    penWidthSpinBox->setValue(3);
    int width = penWidthSpinBox->value();
    Qt::PenStyle style = Qt::PenStyle(penStyleComboBox->itemData(
            penStyleComboBox->currentIndex(), IdRole).toInt());
    Qt::PenCapStyle cap = Qt::PenCapStyle(penCapComboBox->itemData(
            penCapComboBox->currentIndex(), IdRole).toInt());
    Qt::PenJoinStyle join = Qt::PenJoinStyle(penJoinComboBox->itemData(
            penJoinComboBox->currentIndex(), IdRole).toInt());
    renderArea->setPen(QPen(Qt::blue, width, style, cap, join));

    setWindowTitle(tr("Basic Drawing"));
}
//! [10]

//! [11]
void metricDialog::shapeChanged()
{
    RenderArea::Shape shape = RenderArea::Shape(shapeComboBox->itemData(
            shapeComboBox->currentIndex(), IdRole).toInt());
    renderArea->setShape(shape);
}
//! [11]

//! [12]
void metricDialog::penChanged()
{
    int width = penWidthSpinBox->value();
    Qt::PenStyle style = Qt::PenStyle(penStyleComboBox->itemData(
            penStyleComboBox->currentIndex(), IdRole).toInt());
    Qt::PenCapStyle cap = Qt::PenCapStyle(penCapComboBox->itemData(
            penCapComboBox->currentIndex(), IdRole).toInt());
    Qt::PenJoinStyle join = Qt::PenJoinStyle(penJoinComboBox->itemData(
            penJoinComboBox->currentIndex(), IdRole).toInt());

    renderArea->setPen(QPen(Qt::blue, width, style, cap, join));
}
//! [12]

//! [13]
void metricDialog::brushChanged()
{
    Qt::BrushStyle style = Qt::BrushStyle(brushStyleComboBox->itemData(
//! [13]
            brushStyleComboBox->currentIndex(), IdRole).toInt());

//! [14]
    if (style == Qt::LinearGradientPattern) {
        QLinearGradient linearGradient(0, 0, 100, 100);
        linearGradient.setColorAt(0.0, Qt::white);
        linearGradient.setColorAt(0.2, Qt::green);
        linearGradient.setColorAt(1.0, Qt::black);
        renderArea->setBrush(linearGradient);
//! [14] //! [15]
    } else if (style == Qt::RadialGradientPattern) {
        QRadialGradient radialGradient(50, 50, 50, 70, 70);
        radialGradient.setColorAt(0.0, Qt::white);
        radialGradient.setColorAt(0.2, Qt::green);
        radialGradient.setColorAt(1.0, Qt::black);
        renderArea->setBrush(radialGradient);
    } else if (style == Qt::ConicalGradientPattern) {
        QConicalGradient conicalGradient(50, 50, 150);
        conicalGradient.setColorAt(0.0, Qt::white);
        conicalGradient.setColorAt(0.2, Qt::green);
        conicalGradient.setColorAt(1.0, Qt::black);
        renderArea->setBrush(conicalGradient);
//! [15] //! [16]
    } else if (style == Qt::TexturePattern) {
        renderArea->setBrush(QBrush(QPixmap(":/images/brick.png")));
//! [16] //! [17]
    } else {
        renderArea->setBrush(QBrush(Qt::green, style));
    }
}

//===================================================================================

transformDialog::transformDialog(double A11, double A21, double A22)
{
    //setGeometry(geometry().x(), geometry().y(), 600, 400);

    renderArea  = new RenderArea;
    renderArea2 = new RenderArea;

    renderArea->ctx = RenderArea::show_transform;
    renderArea2->ctx = RenderArea::show_transform;

    renderArea->a11 = A11;
    renderArea->a21 = A21;
    renderArea->a22 = A22;

    renderArea2->a11 = A11;
    renderArea2->a21 = A21;
    renderArea2->a22 = A22;

    double lambda1, lambda2, v1[2], v2[2], theta;
    detri2::get_MatVVP2x2(A11, A21, A22, &lambda1, &lambda2, v1, v2);
    theta = detri2::get_rotation_angle(v1);

    int frameStyle = QFrame::Sunken | QFrame::Panel;

    a11Label = new QLabel;
    a11Label->setFrameStyle(frameStyle);
    a11Label->setText(QString::number(A11));
    a21Label = new QLabel;
    a21Label->setFrameStyle(frameStyle);
    a21Label->setText(QString::number(A21));
    a22Label = new QLabel;
    a22Label->setFrameStyle(frameStyle);
    a22Label->setText(QString::number(A22));

    r1Label = new QLabel;
    r1Label->setFrameStyle(frameStyle);
    r1Label->setText(QString::number(sqrt(1./lambda1)));
    QPushButton *btn_r1 = new QPushButton(tr("r1"));

    r2Label = new QLabel;
    r2Label->setFrameStyle(frameStyle);
    r2Label->setText(QString::number(sqrt(1./lambda2)));
    QPushButton *btn_r2 = new QPushButton(tr("r2"));

    thetaLabel = new QLabel;
    thetaLabel->setFrameStyle(frameStyle);
    thetaLabel->setText(QString::number(theta / detri2::PI * 180.));
    QPushButton *btn_theta = new QPushButton(tr("theta"));

    PexLabel = new QLabel;
    PexLabel->setFrameStyle(frameStyle);
    PexLabel->setText(tr("0"));
    QPushButton *btn_Pex = new QPushButton(tr("Pe_x"));

    PeyLabel = new QLabel;
    PeyLabel->setFrameStyle(frameStyle);
    PeyLabel->setText(tr("0"));
    QPushButton *btn_Pey = new QPushButton(tr("Pe_y"));

    QexLabel = new QLabel;
    QexLabel->setFrameStyle(frameStyle);
    QexLabel->setText(tr("1"));
    QPushButton *btn_Qex = new QPushButton(tr("Qe_x"));

    QeyLabel = new QLabel;
    QeyLabel->setFrameStyle(frameStyle);
    QeyLabel->setText(tr("0"));
    QPushButton *btn_Qey = new QPushButton(tr("Qe_y"));

    RexLabel = new QLabel;
    RexLabel->setFrameStyle(frameStyle);
    RexLabel->setText(tr("0"));
    QPushButton *btn_Rex = new QPushButton(tr("Re_x"));

    ReyLabel = new QLabel;
    ReyLabel->setFrameStyle(frameStyle);
    ReyLabel->setText(tr("0"));
    QPushButton *btn_Rey = new QPushButton(tr("Re_y"));


    PtxLabel = new QLabel;
    PtxLabel->setFrameStyle(frameStyle);
    PtxLabel->setText(tr("0"));
    QPushButton *btn_Ptx = new QPushButton(tr("Pt_x"));

    QtxLabel = new QLabel;
    QtxLabel->setFrameStyle(frameStyle);
    QtxLabel->setText(tr("1"));
    QPushButton *btn_Qtx = new QPushButton(tr("Qt_x"));

    RtxLabel = new QLabel;
    RtxLabel->setFrameStyle(frameStyle);
    RtxLabel->setText(tr("0"));
    QPushButton *btn_Rtx = new QPushButton(tr("Re_x"));

    PtyLabel = new QLabel;
    PtyLabel->setFrameStyle(frameStyle);
    PtyLabel->setText(tr("0"));
    QPushButton *btn_Pty = new QPushButton(tr("Pt_y"));

    QtyLabel = new QLabel;
    QtyLabel->setFrameStyle(frameStyle);
    QtyLabel->setText(tr("1"));
    QPushButton *btn_Qty = new QPushButton(tr("Qt_y"));

    RtyLabel = new QLabel;
    RtyLabel->setFrameStyle(frameStyle);
    RtyLabel->setText(tr("0"));
    QPushButton *btn_Rty = new QPushButton(tr("Re_y"));

    QPushButton *btn_transform = new QPushButton(tr("Tranform"));
    QPushButton *btn_inv_transform = new QPushButton(tr("Inv tranform"));

    connect(btn_r1, &QAbstractButton::clicked, this, &transformDialog::set_radius1);
    connect(btn_r2, &QAbstractButton::clicked, this, &transformDialog::set_radius2);
    connect(btn_theta, &QAbstractButton::clicked, this, &transformDialog::set_theta1);

    connect(btn_Pex, &QAbstractButton::clicked, this, &transformDialog::set_Pex);
    connect(btn_Qex, &QAbstractButton::clicked, this, &transformDialog::set_Qex);
    connect(btn_Rex, &QAbstractButton::clicked, this, &transformDialog::set_Rex);

    connect(btn_Ptx, &QAbstractButton::clicked, this, &transformDialog::set_Ptx);
    connect(btn_Qtx, &QAbstractButton::clicked, this, &transformDialog::set_Qtx);
    connect(btn_Rtx, &QAbstractButton::clicked, this, &transformDialog::set_Rtx);

    connect(btn_transform, &QAbstractButton::clicked, this, &transformDialog::on_transform);
    connect(btn_inv_transform, &QAbstractButton::clicked, this, &transformDialog::on_inv_transform);

    QGridLayout *mainLayout = new QGridLayout;
    //mainLayout->setColumnStretch(0, 1);
    //mainLayout->setColumnStretch(3, 1);
    mainLayout->addWidget(renderArea,  0, 0, 1, 2);
    mainLayout->addWidget(renderArea2, 0, 2, 1, 2);

    mainLayout->addWidget(btn_r1,  2,0,1,1);
    mainLayout->addWidget(r1Label, 2,1,1,1);
    mainLayout->addWidget(btn_r2,  3,0,1,1);
    mainLayout->addWidget(r2Label, 3,1,1,1);
    mainLayout->addWidget(btn_theta,  4,0,1,1);
    mainLayout->addWidget(thetaLabel, 4,1,1,1);

    mainLayout->addWidget(a11Label, 2,3,1,1);
    mainLayout->addWidget(a21Label, 3,3,1,1);
    mainLayout->addWidget(a22Label, 4,3,1,1);

    mainLayout->addWidget(btn_Pex,  5,0,1,1);
    mainLayout->addWidget(PexLabel, 5,1,1,1);
    mainLayout->addWidget(btn_Pey,  5,2,1,1);
    mainLayout->addWidget(PeyLabel, 5,3,1,1);

    mainLayout->addWidget(btn_Qex,  6,0,1,1);
    mainLayout->addWidget(QexLabel, 6,1,1,1);
    mainLayout->addWidget(btn_Qey,  6,2,1,1);
    mainLayout->addWidget(QeyLabel, 6,3,1,1);

    mainLayout->addWidget(btn_Rex,  7,0,1,1);
    mainLayout->addWidget(RexLabel, 7,1,1,1);
    mainLayout->addWidget(btn_Rey,  7,2,1,1);
    mainLayout->addWidget(ReyLabel, 7,3,1,1);

    mainLayout->addWidget(btn_Ptx,  5,2,1,1);
    mainLayout->addWidget(PtxLabel, 5,3,1,1);
    mainLayout->addWidget(btn_Qtx,  6,2,1,1);
    mainLayout->addWidget(QtxLabel, 6,3,1,1);
    mainLayout->addWidget(btn_Rtx,  7,2,1,1);
    mainLayout->addWidget(RtxLabel, 7,3,1,1);

    mainLayout->addWidget(btn_transform, 8,0,1,1);
    mainLayout->addWidget(btn_inv_transform, 8,1,1,1);

    setLayout(mainLayout);

    renderArea->setBrush(Qt::NoBrush); // default option
    renderArea2->setBrush(Qt::NoBrush); // default option

    //penWidthSpinBox->setValue(3);
    int width = 3;// penWidthSpinBox->value();
    Qt::PenStyle style = Qt::SolidLine; // Qt::PenStyle(penStyleComboBox->itemData(
    //        penStyleComboBox->currentIndex(), IdRole).toInt());
    Qt::PenCapStyle cap = Qt::FlatCap; // Qt::PenCapStyle(penCapComboBox->itemData(
    //        penCapComboBox->currentIndex(), IdRole).toInt());
    Qt::PenJoinStyle join = Qt::MiterJoin; // Qt::PenJoinStyle(penJoinComboBox->itemData(
    //        penJoinComboBox->currentIndex(), IdRole).toInt());
    renderArea->setPen(QPen(Qt::blue, width, style, cap, join));
    renderArea2->setPen(QPen(Qt::blue, width, style, cap, join));
}

void transformDialog::set_radius1()
{
    double value = get_double(r1Label->text().toDouble());
    r1Label->setText(QString::number(value));
}
void transformDialog::set_radius2()
{
    double value = get_double(r2Label->text().toDouble());
    r2Label->setText(QString::number(value));
}
void transformDialog::set_theta1()
{
    double value = get_double(thetaLabel->text().toDouble());
    thetaLabel->setText(QString::number(value));
}
void transformDialog::set_Pex()
{
    double value = get_double(PexLabel->text().toDouble());
    PexLabel->setText(QString::number(value));
}
void transformDialog::set_Qex()
{
    double value = get_double(QexLabel->text().toDouble());
    QexLabel->setText(QString::number(value));
}
void transformDialog::set_Rex()
{
    double value = get_double(RexLabel->text().toDouble());
    RexLabel->setText(QString::number(value));
}

void transformDialog::set_Pey()
{
    double value = get_double(PeyLabel->text().toDouble());
    PeyLabel->setText(QString::number(value));
}
void transformDialog::set_Qey()
{
    double value = get_double(QeyLabel->text().toDouble());
    QeyLabel->setText(QString::number(value));
}
void transformDialog::set_Rey()
{
    double value = get_double(ReyLabel->text().toDouble());
    ReyLabel->setText(QString::number(value));
}

void transformDialog::set_Ptx()
{
    double value = get_double(PtxLabel->text().toDouble());
    PtxLabel->setText(QString::number(value));
}
void transformDialog::set_Qtx()
{
    double value = get_double(QtxLabel->text().toDouble());
    QtxLabel->setText(QString::number(value));
}
void transformDialog::set_Rtx()
{
    double value = get_double(RtxLabel->text().toDouble());
    RtxLabel->setText(QString::number(value));
}
void transformDialog::set_Pty()
{
    double value = get_double(PtyLabel->text().toDouble());
    PtyLabel->setText(QString::number(value));
}
void transformDialog::set_Qty()
{
    double value = get_double(QtyLabel->text().toDouble());
    QtyLabel->setText(QString::number(value));
}
void transformDialog::set_Rty()
{
    double value = get_double(RtyLabel->text().toDouble());
    RtyLabel->setText(QString::number(value));
}

//void on_draw_ellipse(); // use the eigen values form
void transformDialog::on_transform()
{

}
void transformDialog::on_inv_transform()
{

}

void transformDialog::on_reset()
{

}
void transformDialog::on_close()
{

}
