#ifndef METRICPANEL_H
#define METRICPANEL_H

#include <QObject>
#include <QWidget>
#include <QDialog>

#include <QGridLayout>
#include <QVBoxLayout>
#include <QHBoxLayout>
#include <QGroupBox>
#include <QRadioButton>
#include <QCheckBox>
#include <QLabel>
#include <QLineEdit>
#include <QPushButton>
#include <QListWidget>
#include <QComboBox>
#include <QSpinBox>

class RenderArea : public QWidget
{
    Q_OBJECT

public:
    enum Shape { Line, Points, Polyline, Polygon, Rect, RoundedRect, Ellipse, Arc,
                 Chord, Pie, Path, Text, Pixmap };

    enum ShowContent {show_metric, show_transform};

    RenderArea(QWidget *parent = 0);

    QSize minimumSizeHint() const override;
    QSize sizeHint() const override;

public slots:
    void setShape(Shape shape);
    void setPen(const QPen &pen);
    void setBrush(const QBrush &brush);
    void setAntialiased(bool antialiased);
    void setTransformed(bool transformed);
    void redraw();

protected:
    void paintEvent(QPaintEvent *event) override;

private:
    Shape shape;
    QPen pen;
    QBrush brush;
    bool antialiased;
    bool transformed;
    QPixmap pixmap;

public:
    ShowContent ctx;

    double a11, a21, a22;

    // three points.
    double _px, _py, _pz, _wp;
    double _qx, _qy, _qz, _wq;
    double _rx, _ry, _rz, _wr;

    void draw_metric();
    void draw_transform();
};

class metricDialog : public QDialog
{
    Q_OBJECT

public:
    metricDialog(double *A11, double *A21, double *A22);

public slots:
    void set_a11();
    void set_a21();
    void set_a22();
    void set_radius1();
    void set_radius2();
    void set_theta1();
    void on_draw(); // use the quadratic form
    void on_draw_ellipse(); // use the eigen values form
    void on_reset();
    void on_apply();

private slots:
    void shapeChanged();
    void penChanged();
    void brushChanged();

private:
    RenderArea *renderArea;

    double *_a11, *_a21, *_a22;

    QLabel *a11Label;
    QLabel *a21Label;
    QLabel *a22Label;
    QLabel *r1Label;
    QLabel *r2Label;
    QLabel *thetaLabel;

    QLabel *shapeLabel;
    QLabel *penWidthLabel;
    QLabel *penStyleLabel;
    QLabel *penCapLabel;
    QLabel *penJoinLabel;
    QLabel *brushStyleLabel;
    QLabel *otherOptionsLabel;
    QComboBox *shapeComboBox;
    QSpinBox *penWidthSpinBox;
    QComboBox *penStyleComboBox;
    QComboBox *penCapComboBox;
    QComboBox *penJoinComboBox;
    QComboBox *brushStyleComboBox;
    QCheckBox *antialiasingCheckBox;
    QCheckBox *transformationsCheckBox;
};

class transformDialog : public QDialog
{
    Q_OBJECT

public:
    transformDialog(double A11, double A21, double A22);

public slots:
    void set_radius1();
    void set_radius2();
    void set_theta1();

    void set_Pex();
    void set_Qex();
    void set_Rex();

    void set_Pey();
    void set_Qey();
    void set_Rey();

    void set_Ptx();
    void set_Qtx();
    void set_Rtx();

    void set_Pty();
    void set_Qty();
    void set_Rty();

    //void on_draw_ellipse(); // use the eigen values form
    void on_transform();
    void on_inv_transform();

    void on_reset();
    void on_close();

private:
    RenderArea *renderArea;  // Euclidean
    RenderArea *renderArea2; // Transformed

    QLabel *a11Label; // quadratic form
    QLabel *a21Label;
    QLabel *a22Label;

    QLabel *r1Label; // the transformation matrix
    QLabel *r2Label;
    QLabel *thetaLabel;

    QLabel *PexLabel;  // A Euclidean triangle
    QLabel *QexLabel;
    QLabel *RexLabel;

    QLabel *PeyLabel;  // A Euclidean triangle
    QLabel *QeyLabel;
    QLabel *ReyLabel;

    QLabel *PtxLabel; // The transformed triangle
    QLabel *QtxLabel;
    QLabel *RtxLabel;

    QLabel *PtyLabel; // The transformed triangle
    QLabel *QtyLabel;
    QLabel *RtyLabel;
};

#endif // METRICPANEL_H

