#include <QtWidgets>

#include <QFileDialog>
#include <QColorDialog>
#include <QFontDialog>

#include "view3d.h"
#include "scribblearea.h"
//#include "preferdialog.h"

//==============================================================================

class DialogOptionsWidget : public QGroupBox
{
public:
    explicit DialogOptionsWidget(QWidget *parent = 0);

    void addCheckBox(const QString &text, int value);
    void addSpacer();
    int value() const;

private:
    typedef QPair<QCheckBox *, int> CheckBoxEntry;
    QVBoxLayout *layout;
    QList<CheckBoxEntry> checkBoxEntries;
};

DialogOptionsWidget::DialogOptionsWidget(QWidget *parent) :
    QGroupBox(parent) , layout(new QVBoxLayout)
{
    setTitle(PreferDialog::tr("Options"));
    setLayout(layout);
}

void DialogOptionsWidget::addCheckBox(const QString &text, int value)
{
    QCheckBox *checkBox = new QCheckBox(text);
    layout->addWidget(checkBox);
    checkBoxEntries.append(CheckBoxEntry(checkBox, value));
}

void DialogOptionsWidget::addSpacer()
{
    layout->addItem(new QSpacerItem(0, 0, QSizePolicy::Ignored, QSizePolicy::MinimumExpanding));
}

int DialogOptionsWidget::value() const
{
    int result = 0;
    foreach (const CheckBoxEntry &checkboxEntry, checkBoxEntries)
        if (checkboxEntry.first->isChecked())
            result |= checkboxEntry.second;
    return result;
}

//==============================================================================

PreferDialog::PreferDialog(void *ptr, QWidget *parent)
    : QWidget(parent)
{
    _scribblearea = ptr;
    ScribbleArea *sptr = (ScribbleArea *) _scribblearea;

    /////////////////////////////////////////////////////////////

    QVBoxLayout *verticalLayout;
    verticalLayout = new QVBoxLayout(this);

    QToolBox *toolbox = new QToolBox;
    verticalLayout->addWidget(toolbox);

    _applybtn  = new QPushButton(tr("Apply"));
    _savebtn   = new QPushButton(tr("Save..."));
    _importbtn = new QPushButton(tr("Import..."));

    QHBoxLayout *BtnLayout = new QHBoxLayout;
    BtnLayout->addWidget(_savebtn);   // , 0, Qt::AlignLeft
    BtnLayout->addWidget(_importbtn); // , 0, Qt::AlignLeft
    BtnLayout->addWidget(_applybtn);  // , 0, Qt::AlignLeft

    QGroupBox *BtnBox = new QGroupBox;
    BtnBox->setLayout(BtnLayout);

    connect(_applybtn, SIGNAL(clicked()), this, SLOT(on_apply()));

    verticalLayout->addWidget(BtnBox);

    /////////////////////////////////////////////////////////////

    int frameStyle = QFrame::Sunken | QFrame::Panel;

    // Geometry tolerances
    rel_min_gap_tolLabel = new QLabel;
    rel_min_gap_tolLabel->setFrameStyle(frameStyle);
    rel_min_gap_tolLabel->setText(QString::number(sptr->_geom->io_tol_rel_gap));
    QPushButton *rel_min_gap_tolButton = new QPushButton(tr("Relative minimum gap >"));

    //abs_min_gap_tolLabel = new QLabel;
    //abs_min_gap_tolLabel->setFrameStyle(frameStyle);
    //abs_min_gap_tolLabel->setText(QString::number(0.0));
    //QPushButton *abs_min_gap_tolButton = new QPushButton(tr("Absolute minimum gap > "));

    min_angle_tolLabel = new QLabel;
    min_angle_tolLabel->setFrameStyle(frameStyle);
    min_angle_tolLabel->setText(QString::number(sptr->_geom->io_tol_minangle / detri2::PI * 180.0));
    QPushButton *min_angle_tolButton = new QPushButton(tr("Minimum collinear angle (degree) <"));

    //enable_debugLabel = new QCheckBox(tr("Enable debug"));
    //enable_debugLabel->setChecked(false);

    // Draw options
    vertex_sizeLabel = new QLabel;
    vertex_sizeLabel->setFrameStyle(frameStyle);
    vertex_sizeLabel->setText(QString::number(sptr->_draw_vertex_size));
    QPushButton *vertex_sizeButton = new QPushButton(tr("Vertex size"));

    segment_widthLabel = new QLabel;
    segment_widthLabel->setFrameStyle(frameStyle);
    segment_widthLabel->setText(QString::number(sptr->_draw_segment_width));
    QPushButton *segment_widthButton = new QPushButton(tr("Segment width"));

    edge_widthLabel = new QLabel;
    edge_widthLabel->setFrameStyle(frameStyle);
    edge_widthLabel->setText(QString::number(sptr->_draw_edge_width));
    QPushButton *edge_widthButton = new QPushButton(tr("Mesh edge width"));

    dual_edge_widthLabel = new QLabel;
    dual_edge_widthLabel->setFrameStyle(frameStyle);
    dual_edge_widthLabel->setText(QString::number(sptr->_draw_dual_edge_width));
    QPushButton *dual_edge_widthButton = new QPushButton(tr("Voronoi edge width"));

    ccircle_widthLabel = new QLabel;
    ccircle_widthLabel->setFrameStyle(frameStyle);
    ccircle_widthLabel->setText(QString::number(sptr->_draw_ccircle_width));
    QPushButton *ccircle_widthButton = new QPushButton(tr("Circumcircle width"));

    //view3d======================================================================
    v3d_vertex_sizeLabel = new QLabel;
    v3d_vertex_sizeLabel->setFrameStyle(frameStyle);
    v3d_vertex_sizeLabel->setText(QString::number(sptr->_v3d_vertex_size));
    QPushButton *v3d_vertex_sizeButton = new QPushButton(tr("view3d Vertex size"));

    v3d_edge_widthLabel = new QLabel;
    v3d_edge_widthLabel->setFrameStyle(frameStyle);
    v3d_edge_widthLabel->setText(QString::number(sptr->_v3d_edge_width));
    QPushButton *v3d_edge_widthButton = new QPushButton(tr("view3d edge width"));

    v3d_segment_widthLabel = new QLabel;
    v3d_segment_widthLabel->setFrameStyle(frameStyle);
    v3d_segment_widthLabel->setText(QString::number(sptr->_v3d_segment_width));
    QPushButton *v3d_segment_widthButton = new QPushButton(tr("view3d Segment width"));
    //view3d======================================================================

    // Color panel
    background_colorLabel = new QLabel;
    background_colorLabel->setFrameStyle(frameStyle);
    background_colorLabel->setText(sptr->_color_canvas_background.name());
    background_colorLabel->setPalette(QPalette(sptr->_color_canvas_background));
    background_colorLabel->setAutoFillBackground(true);
    QPushButton *background_colorButton = new QPushButton(tr("Background color"));

    vertex_colorLabel = new QLabel;
    vertex_colorLabel->setFrameStyle(frameStyle);
    vertex_colorLabel->setText(sptr->_color_vertex.name());
    vertex_colorLabel->setPalette(QPalette(sptr->_color_vertex));
    vertex_colorLabel->setAutoFillBackground(true);
    QPushButton *vertex_colorButton = new QPushButton(tr("Vertex color"));

    segment_colorLabel = new QLabel;
    segment_colorLabel->setFrameStyle(frameStyle);
    segment_colorLabel->setText(sptr->_color_segment.name());
    segment_colorLabel->setPalette(QPalette(sptr->_color_segment));
    segment_colorLabel->setAutoFillBackground(true);
    QPushButton *segment_colorButton = new QPushButton(tr("Segment color"));

    selected_colorLabel = new QLabel;
    selected_colorLabel->setFrameStyle(frameStyle);
    selected_colorLabel->setText(sptr->_color_selected.name());
    selected_colorLabel->setPalette(QPalette(sptr->_color_selected));
    selected_colorLabel->setAutoFillBackground(true);
    QPushButton *selected_colorButton = new QPushButton(tr("Selected color"));

    edge_colorLabel = new QLabel;
    edge_colorLabel->setFrameStyle(frameStyle);
    edge_colorLabel->setText(sptr->_color_edge.name());
    edge_colorLabel->setPalette(QPalette(sptr->_color_edge));
    edge_colorLabel->setAutoFillBackground(true);
    QPushButton *edge_colorButton = new QPushButton(tr("Mesh edge color"));

    dual_edge_colorLabel = new QLabel;
    dual_edge_colorLabel->setFrameStyle(frameStyle);
    dual_edge_colorLabel->setText(sptr->_color_dual_edge.name());
    dual_edge_colorLabel->setPalette(QPalette(sptr->_color_dual_edge));
    dual_edge_colorLabel->setAutoFillBackground(true);
    QPushButton *dual_edge_colorButton = new QPushButton(tr("Voronoi edge color"));

    bgmesh_edge_colorLabel = new QLabel;
    bgmesh_edge_colorLabel->setFrameStyle(frameStyle);
    bgmesh_edge_colorLabel->setText(sptr->_color_bgmesh_edge.name());
    bgmesh_edge_colorLabel->setPalette(QPalette(sptr->_color_bgmesh_edge));
    bgmesh_edge_colorLabel->setAutoFillBackground(true);
    QPushButton *bgmesh_edge_colorButton = new QPushButton(tr("Bg mesh edge color"));

    mesh_size_colorLabel = new QLabel;
    mesh_size_colorLabel->setFrameStyle(frameStyle);
    mesh_size_colorLabel->setText(sptr->_color_dual_edge.name());
    mesh_size_colorLabel->setPalette(QPalette(sptr->_color_dual_edge));
    mesh_size_colorLabel->setAutoFillBackground(true);
    QPushButton *mesh_size_colorButton = new QPushButton(tr("Mesh size color"));

    ccircle_colorLabel = new QLabel;
    ccircle_colorLabel->setFrameStyle(frameStyle);
    ccircle_colorLabel->setText(sptr->_color_ccircle.name());
    ccircle_colorLabel->setPalette(QPalette(sptr->_color_ccircle));
    ccircle_colorLabel->setAutoFillBackground(true);
    QPushButton *ccircle_colorButton = new QPushButton(tr("Circumcircle color"));

    //view3d======================================================================
    v3d_background_colorLabel = new QLabel;
    v3d_background_colorLabel->setFrameStyle(frameStyle);
    v3d_background_colorLabel->setText(sptr->_v3d_bgcolor.name());
    v3d_background_colorLabel->setPalette(QPalette(sptr->_v3d_bgcolor));
    v3d_background_colorLabel->setAutoFillBackground(true);
    QPushButton *v3d_background_colorButton = new QPushButton(tr("view3d Background color"));

    v3d_vertex_colorLabel = new QLabel;
    v3d_vertex_colorLabel->setFrameStyle(frameStyle);
    v3d_vertex_colorLabel->setText(sptr->_v3d_vertexcolor.name());
    v3d_vertex_colorLabel->setPalette(QPalette(sptr->_v3d_vertexcolor));
    v3d_vertex_colorLabel->setAutoFillBackground(true);
    QPushButton *v3d_vertex_colorButton = new QPushButton(tr("view3d Vertex color"));

    v3d_edge_colorLabel = new QLabel;
    v3d_edge_colorLabel->setFrameStyle(frameStyle);
    v3d_edge_colorLabel->setText(sptr->_v3d_edgecolor.name());
    v3d_edge_colorLabel->setPalette(QPalette(sptr->_v3d_edgecolor));
    v3d_edge_colorLabel->setAutoFillBackground(true);
    QPushButton *v3d_edge_colorButton = new QPushButton(tr("view3d edge color"));

    v3d_facet_colorLabel = new QLabel;
    v3d_facet_colorLabel->setFrameStyle(frameStyle);
    v3d_facet_colorLabel->setText(sptr->_v3d_facetcolor.name());
    v3d_facet_colorLabel->setPalette(QPalette(sptr->_v3d_facetcolor));
    v3d_facet_colorLabel->setAutoFillBackground(true);
    QPushButton *v3d_facet_colorButton = new QPushButton(tr("view3d facet color"));

    v3d_polecolorLabel = new QLabel;
    v3d_polecolorLabel->setFrameStyle(frameStyle);
    v3d_polecolorLabel->setText(sptr->_v3d_polecolor.name());
    v3d_polecolorLabel->setPalette(QPalette(sptr->_v3d_polecolor));
    v3d_polecolorLabel->setAutoFillBackground(true);
    QPushButton *v3d_polecolorButton = new QPushButton(tr("view3d pole color"));

    v3d_polarcolorLabel = new QLabel;
    v3d_polarcolorLabel->setFrameStyle(frameStyle);
    v3d_polarcolorLabel->setText(sptr->_v3d_polarcolor.name());
    v3d_polarcolorLabel->setPalette(QPalette(sptr->_v3d_polarcolor));
    v3d_polarcolorLabel->setAutoFillBackground(true);
    QPushButton *v3d_polarcolorButton = new QPushButton(tr("view3d polar color"));
    //view3d======================================================================

    bbox_fontLabel = new QLabel;
    bbox_fontLabel->setFrameStyle(frameStyle);
    bbox_fontLabel->setText(sptr->_font_bbox_label.key());
    bbox_fontLabel->setFont(sptr->_font_bbox_label);
    QPushButton *bbox_fontButton = new QPushButton(tr("Bounding box label font"));

    vertex_fontLabel = new QLabel;
    vertex_fontLabel->setFrameStyle(frameStyle);
    vertex_fontLabel->setText(sptr->_font_vertex_label.key());
    vertex_fontLabel->setFont(sptr->_font_vertex_label);
    QPushButton *vertex_fontButton = new QPushButton(tr("Vertex label font"));

    segment_fontLabel = new QLabel;
    segment_fontLabel->setFrameStyle(frameStyle);
    segment_fontLabel->setText(sptr->_font_segment_label.key());
    segment_fontLabel->setFont(sptr->_font_segment_label);
    QPushButton *segment_fontButton = new QPushButton(tr("Segment label font"));

    subdomain_fontLabel = new QLabel;
    subdomain_fontLabel->setFrameStyle(frameStyle);
    subdomain_fontLabel->setText(sptr->_font_subdomain_label.key());
    subdomain_fontLabel->setFont(sptr->_font_subdomain_label);
    QPushButton *subdomain_fontButton = new QPushButton(tr("Subdomain label font"));

    //fontLabel = new QLabel;
    //fontLabel->setFrameStyle(frameStyle);
    //QPushButton *fontButton = new QPushButton(tr("QFontDialog::get&Font()"));

    backimageLabel = new QLabel;
    backimageLabel->setFrameStyle(frameStyle);
    QPushButton *backimageButton = new QPushButton(tr("Use background image")); // open file

    directoryLabel = new QLabel;
    directoryLabel->setFrameStyle(frameStyle);
    QPushButton *directoryButton = new QPushButton(tr("QFileDialog::getE&xistingDirectory()"));

    openFileNameLabel = new QLabel;
    openFileNameLabel->setFrameStyle(frameStyle);
    QPushButton *openFileNameButton = new QPushButton(tr("QFileDialog::get&OpenFileName()"));

    saveFileNameLabel = new QLabel;
    saveFileNameLabel->setFrameStyle(frameStyle);
    QPushButton *saveFileNameButton = new QPushButton(tr("QFileDialog::get&SaveFileName()"));


    // signals

    QSignalMapper* signalMapperTol = new QSignalMapper (this) ;
    connect (rel_min_gap_tolButton, SIGNAL(clicked()), signalMapperTol, SLOT(map()));
    //connect (abs_min_gap_tolButton, SIGNAL(clicked()), signalMapperTol, SLOT(map()));
    connect (min_angle_tolButton, SIGNAL(clicked()), signalMapperTol, SLOT(map()));
    signalMapperTol->setMapping(rel_min_gap_tolButton, 1);
    //signalMapperTol->setMapping(abs_min_gap_tolButton, 2);
    signalMapperTol->setMapping(min_angle_tolButton, 2);
    connect (signalMapperTol, SIGNAL(mapped(int)), this, SLOT(setDouble(int)));

    QSignalMapper* signalMapperInt = new QSignalMapper (this) ;
    connect (vertex_sizeButton, SIGNAL(clicked()), signalMapperInt, SLOT(map()));
    connect (segment_widthButton, SIGNAL(clicked()), signalMapperInt, SLOT(map()));
    connect (edge_widthButton, SIGNAL(clicked()), signalMapperInt, SLOT(map()));
    connect (dual_edge_widthButton, SIGNAL(clicked()), signalMapperInt, SLOT(map()));
    connect (ccircle_widthButton, SIGNAL(clicked()), signalMapperInt, SLOT(map()));
    // view3d
    connect (v3d_vertex_sizeButton, SIGNAL(clicked()), signalMapperInt, SLOT(map()));
    connect (v3d_edge_widthButton, SIGNAL(clicked()), signalMapperInt, SLOT(map()));
    connect (v3d_segment_widthButton, SIGNAL(clicked()), signalMapperInt, SLOT(map()));

    signalMapperInt->setMapping(vertex_sizeButton, 1);
    signalMapperInt->setMapping(segment_widthButton, 2);
    signalMapperInt->setMapping(edge_widthButton, 3);
    signalMapperInt->setMapping(dual_edge_widthButton, 4);
    signalMapperInt->setMapping(ccircle_widthButton, 5);
    // view3d
    signalMapperInt->setMapping(v3d_vertex_sizeButton, 6);
    signalMapperInt->setMapping(v3d_edge_widthButton, 7);
    signalMapperInt->setMapping(v3d_segment_widthButton, 8);

    connect (signalMapperInt, SIGNAL(mapped(int)), this, SLOT(setInteger(int)));

    QSignalMapper* signalMapperColor = new QSignalMapper (this) ;
    connect (background_colorButton, SIGNAL(clicked()), signalMapperColor, SLOT(map()));
    connect (vertex_colorButton, SIGNAL(clicked()), signalMapperColor, SLOT(map()));
    connect (segment_colorButton, SIGNAL(clicked()), signalMapperColor, SLOT(map()));
    connect (selected_colorButton, SIGNAL(clicked()), signalMapperColor, SLOT(map()));
    connect (edge_colorButton, SIGNAL(clicked()), signalMapperColor, SLOT(map()));
    connect (dual_edge_colorButton, SIGNAL(clicked()), signalMapperColor, SLOT(map()));
    connect (bgmesh_edge_colorButton, SIGNAL(clicked()), signalMapperColor, SLOT(map()));
    connect (mesh_size_colorButton, SIGNAL(clicked()), signalMapperColor, SLOT(map()));
    // view3d
    connect (v3d_background_colorButton, SIGNAL(clicked()), signalMapperColor, SLOT(map()));
    connect (v3d_vertex_colorButton, SIGNAL(clicked()), signalMapperColor, SLOT(map()));
    connect (v3d_edge_colorButton, SIGNAL(clicked()), signalMapperColor, SLOT(map()));
    connect (v3d_facet_colorButton, SIGNAL(clicked()), signalMapperColor, SLOT(map()));
    connect (v3d_polecolorButton, SIGNAL(clicked()), signalMapperColor, SLOT(map()));
    connect (v3d_polarcolorButton, SIGNAL(clicked()), signalMapperColor, SLOT(map()));
    signalMapperColor->setMapping(background_colorButton, 0);
    signalMapperColor->setMapping(vertex_colorButton, 1);
    signalMapperColor->setMapping(segment_colorButton, 2);
    signalMapperColor->setMapping(selected_colorButton, 3);
    signalMapperColor->setMapping(edge_colorButton, 4);
    signalMapperColor->setMapping(dual_edge_colorButton, 5);
    signalMapperColor->setMapping(bgmesh_edge_colorButton, 6);
    signalMapperColor->setMapping(mesh_size_colorButton, 7);
    signalMapperColor->setMapping(ccircle_colorButton, 8);
    // view3d
    signalMapperColor->setMapping(v3d_background_colorButton, 9); // v3d
    signalMapperColor->setMapping(v3d_vertex_colorButton, 10);
    signalMapperColor->setMapping(v3d_edge_colorButton, 11);
    signalMapperColor->setMapping(v3d_facet_colorButton, 12);
    signalMapperColor->setMapping(v3d_polecolorButton, 13);
    signalMapperColor->setMapping(v3d_polarcolorButton, 14);
    connect (signalMapperColor, SIGNAL(mapped(int)), this, SLOT(setColor(int)));

    QSignalMapper* signalMapperFont = new QSignalMapper (this) ;
    connect (bbox_fontButton, SIGNAL(clicked()), signalMapperFont, SLOT(map()));
    connect (vertex_fontButton, SIGNAL(clicked()), signalMapperFont, SLOT(map()));
    connect (segment_fontButton, SIGNAL(clicked()), signalMapperFont, SLOT(map()));
    connect (subdomain_fontButton, SIGNAL(clicked()), signalMapperFont, SLOT(map()));
    signalMapperFont->setMapping(bbox_fontButton, 0);
    signalMapperFont->setMapping(vertex_fontButton, 1);
    signalMapperFont->setMapping(segment_fontButton, 2);
    signalMapperFont->setMapping(subdomain_fontButton, 3);
    connect (signalMapperFont, SIGNAL(mapped(int)), this, SLOT(setFont(int)));

    connect(backimageButton, &QAbstractButton::clicked, this, &PreferDialog::setOpenFileName);

    //connect(colorButton, &QAbstractButton::clicked, this, &PreferDialog::setColor);
    //connect(fontButton, &QAbstractButton::clicked, this, &PreferDialog::setFont);
    connect(directoryButton, &QAbstractButton::clicked, this, &PreferDialog::setExistingDirectory);
    connect(openFileNameButton, &QAbstractButton::clicked, this, &PreferDialog::setOpenFileName);
    connect(saveFileNameButton, &QAbstractButton::clicked, this, &PreferDialog::setSaveFileName);

    // Pages and layouts

    QWidget *page = new QWidget;
    QGridLayout *layout = new QGridLayout(page);
    layout->setColumnStretch(1, 1);
    layout->setColumnMinimumWidth(1, 250);
    layout->addWidget(rel_min_gap_tolButton, 0, 0);
    layout->addWidget(rel_min_gap_tolLabel, 0, 1);
    //layout->addWidget(abs_min_gap_tolButton, 1, 0);
    //layout->addWidget(abs_min_gap_tolLabel, 1, 1);
    layout->addWidget(min_angle_tolButton, 1, 0);
    layout->addWidget(min_angle_tolLabel, 1, 1);
    layout->addItem(new QSpacerItem(0, 0, QSizePolicy::Ignored, QSizePolicy::MinimumExpanding), 5, 0);
    toolbox->addItem(page, tr("Geometry tolerances"));

    page = new QWidget;
    layout = new QGridLayout(page);
    layout->setColumnStretch(1, 1);
    //layout->setColumnMinimumWidth(1, 250);
    layout->addWidget(vertex_sizeButton, 0, 0);
    layout->addWidget(vertex_sizeLabel, 0, 1);
    layout->addWidget(segment_widthButton, 1, 0);
    layout->addWidget(segment_widthLabel, 1, 1);
    layout->addWidget(edge_widthButton, 2, 0);
    layout->addWidget(edge_widthLabel, 2, 1);
    layout->addWidget(dual_edge_widthButton, 3, 0);
    layout->addWidget(dual_edge_widthLabel, 3, 1);
    layout->addWidget(ccircle_widthButton, 4, 0);
    layout->addWidget(ccircle_widthLabel, 4, 1);
    // view3d
    layout->addWidget(v3d_vertex_sizeButton, 5, 0);
    layout->addWidget(v3d_vertex_sizeLabel, 5, 1);
    layout->addWidget(v3d_edge_widthButton, 6, 0);
    layout->addWidget(v3d_edge_widthLabel, 6, 1);
    layout->addWidget(v3d_segment_widthButton, 7, 0);
    layout->addWidget(v3d_segment_widthLabel, 7, 1);

    layout->addItem(new QSpacerItem(0, 0, QSizePolicy::Ignored, QSizePolicy::MinimumExpanding), 5, 0);
    toolbox->addItem(page, tr("Draw options"));

    const QString doNotUseNativeDialog = tr("Do not use native dialog");

    page = new QWidget;
    layout = new QGridLayout(page);
    layout->setColumnStretch(1, 1);
    layout->addWidget(background_colorButton, 0, 0);
    layout->addWidget(background_colorLabel, 0, 1);
    layout->addWidget(vertex_colorButton, 1, 0);
    layout->addWidget(vertex_colorLabel, 1, 1);
    layout->addWidget(segment_colorButton, 2, 0);
    layout->addWidget(segment_colorLabel, 2, 1);
    layout->addWidget(selected_colorButton, 3, 0);
    layout->addWidget(selected_colorLabel, 3, 1);
    layout->addWidget(edge_colorButton, 4, 0);
    layout->addWidget(edge_colorLabel, 4, 1);
    layout->addWidget(dual_edge_colorButton, 5, 0);
    layout->addWidget(dual_edge_colorLabel, 5, 1);
    layout->addWidget(bgmesh_edge_colorButton, 6, 0);
    layout->addWidget(bgmesh_edge_colorLabel, 6, 1);
    layout->addWidget(mesh_size_colorButton, 7, 0);
    layout->addWidget(mesh_size_colorLabel, 7, 1);
    layout->addWidget(ccircle_colorButton, 8, 0);
    layout->addWidget(ccircle_colorLabel, 8, 1);
    // view3d
    layout->addWidget(v3d_background_colorButton, 9, 0);
    layout->addWidget(v3d_background_colorLabel, 9, 1);
    layout->addWidget(v3d_vertex_colorButton, 10, 0);
    layout->addWidget(v3d_vertex_colorLabel, 10, 1);
    layout->addWidget(v3d_edge_colorButton, 11, 0);
    layout->addWidget(v3d_edge_colorLabel, 11, 1);
    layout->addWidget(v3d_facet_colorButton, 12, 0);
    layout->addWidget(v3d_facet_colorLabel, 12, 1);
    layout->addWidget(v3d_polecolorButton, 13, 0);
    layout->addWidget(v3d_polecolorLabel, 13, 1);
    layout->addWidget(v3d_polarcolorButton, 14, 0);
    layout->addWidget(v3d_polarcolorLabel, 14, 1);

    colorDialogOptionsWidget = new DialogOptionsWidget;
    colorDialogOptionsWidget->addCheckBox(doNotUseNativeDialog, QColorDialog::DontUseNativeDialog);
    colorDialogOptionsWidget->addCheckBox(tr("Show alpha channel") , QColorDialog::ShowAlphaChannel);
    colorDialogOptionsWidget->addCheckBox(tr("No buttons") , QColorDialog::NoButtons);
    layout->addItem(new QSpacerItem(0, 0, QSizePolicy::Ignored, QSizePolicy::MinimumExpanding), 1, 0);
    layout->addWidget(colorDialogOptionsWidget, 15, 0, 1 ,2);
    toolbox->addItem(page, tr("Color Dialog"));

    page = new QWidget;
    layout = new QGridLayout(page);
    layout->setColumnStretch(1, 1);
    //layout->addWidget(fontButton, 0, 0);
    //layout->addWidget(fontLabel, 0, 1);
    layout->addWidget(bbox_fontButton, 0, 0);
    layout->addWidget(bbox_fontLabel, 0, 1);
    layout->addWidget(vertex_fontButton, 1, 0);
    layout->addWidget(vertex_fontLabel, 1, 1);
    layout->addWidget(segment_fontButton, 2, 0);
    layout->addWidget(segment_fontLabel, 2, 1);
    layout->addWidget(subdomain_fontButton, 3, 0);
    layout->addWidget(subdomain_fontLabel, 3, 1);
    fontDialogOptionsWidget = new DialogOptionsWidget;
    fontDialogOptionsWidget->addCheckBox(doNotUseNativeDialog, QFontDialog::DontUseNativeDialog);
    fontDialogOptionsWidget->addCheckBox(tr("Show scalable fonts"), QFontDialog::ScalableFonts);
    fontDialogOptionsWidget->addCheckBox(tr("Show non scalable fonts"), QFontDialog::NonScalableFonts);
    fontDialogOptionsWidget->addCheckBox(tr("Show monospaced fonts"), QFontDialog::MonospacedFonts);
    fontDialogOptionsWidget->addCheckBox(tr("Show proportional fonts"), QFontDialog::ProportionalFonts);
    fontDialogOptionsWidget->addCheckBox(tr("No buttons") , QFontDialog::NoButtons);
    layout->addItem(new QSpacerItem(0, 0, QSizePolicy::Ignored, QSizePolicy::MinimumExpanding), 1, 0);
    layout->addWidget(fontDialogOptionsWidget, 4, 0, 1 ,2);
    toolbox->addItem(page, tr("Font Dialog"));

    page = new QWidget;
    layout = new QGridLayout(page);
    layout->setColumnStretch(1, 1);
    layout->addWidget(directoryButton, 0, 0);
    layout->addWidget(directoryLabel, 0, 1);
    layout->addWidget(backimageButton, 1, 0);
    layout->addWidget(backimageLabel, 1, 1);
    layout->addWidget(saveFileNameButton, 2, 0);
    layout->addWidget(saveFileNameLabel, 2, 1);
    //layout->addWidget(enable_debugLabel, 3, 0);
    fileDialogOptionsWidget = new DialogOptionsWidget;
    fileDialogOptionsWidget->addCheckBox(doNotUseNativeDialog, QFileDialog::DontUseNativeDialog);
    fileDialogOptionsWidget->addCheckBox(tr("Show directories only"), QFileDialog::ShowDirsOnly);
    fileDialogOptionsWidget->addCheckBox(tr("Do not resolve symlinks"), QFileDialog::DontResolveSymlinks);
    fileDialogOptionsWidget->addCheckBox(tr("Do not confirm overwrite"), QFileDialog::DontConfirmOverwrite);
    fileDialogOptionsWidget->addCheckBox(tr("Do not use sheet"), QFileDialog::DontUseSheet);
    fileDialogOptionsWidget->addCheckBox(tr("Readonly"), QFileDialog::ReadOnly);
    fileDialogOptionsWidget->addCheckBox(tr("Hide name filter details"), QFileDialog::HideNameFilterDetails);
    fileDialogOptionsWidget->addCheckBox(tr("Do not use custom directory icons (Windows)"), QFileDialog::DontUseCustomDirectoryIcons);
    layout->addItem(new QSpacerItem(0, 0, QSizePolicy::Ignored, QSizePolicy::MinimumExpanding), 5, 0);
    layout->addWidget(fileDialogOptionsWidget, 6, 0, 1 ,2);
    toolbox->addItem(page, tr("File Dialogs"));

    setWindowTitle(tr("Preferences"));

    // see https://stackoverflow.com/questions/20491864/how-close-and-delete-a-modeless-qt-dialog
    //setAttribute(Qt::WA_DeleteOnClose);
}

void PreferDialog::setDouble(int ii)
{
    bool ok;
    if (ii == 1) {
      // minimum value 0.0,
      // maximum value 1.0
      // Use maximum 10 decimals
      double d = QInputDialog::getDouble(this, // parent Widget
                                         tr("Set relative min gap"), // title
                                         tr("relative minimu gap >"), // label
                                         rel_min_gap_tolLabel->text().toDouble(), // value
                                         0.0, // min
                                         1.0, // max
                                         10,  // maimum number of decimal
                                         &ok);
      if (ok)
          rel_min_gap_tolLabel->setText(QString::number(d));
    } else if (ii == 2) {
      //double d = QInputDialog::getDouble(this, tr("Set absolute min gap"),
      //                                   tr("absolute minimu gap >"),
      //                                   abs_min_gap_tolLabel->text().toDouble(),
      //                                   0.0, 10000.0, 10, &ok);
      //if (ok)
      //    abs_min_gap_tolLabel->setText(QString::number(d));
    } else if (ii == 2) {
      double d = QInputDialog::getDouble(this, tr("Set minimum angle"),
          tr("Minimu angle (degree) >"), min_angle_tolLabel->text().toDouble(), 0.0, 180.0, 10, &ok);
      if (ok)
          min_angle_tolLabel->setText(QString::number(d));
    }
}

void PreferDialog::setInteger(int ii)
{
    printf("setIneger %d\n", ii);
    bool ok;
    if (ii == 1) {
      int dd = vertex_sizeLabel->text().toInt();
      int i = QInputDialog::getInt(this, tr("Change geometry point size"),
                                 tr("point size"), dd, 1, 100, 1, &ok);
      if (ok) {
        vertex_sizeLabel->setText(QString::number(i)); // tr("%1%").arg(i)
      }
    } else if (ii == 2) {
      int dd = segment_widthLabel->text().toInt();
      int i = QInputDialog::getInt(this, tr("Change boundary segment width"),
                                   tr("line width"), dd, 1, 100, 1, &ok);
      if (ok) {
        segment_widthLabel->setText(QString::number(i)); // tr("%1%").arg(i)
      }
    } else if (ii == 3) {
      int dd = edge_widthLabel->text().toInt();;
      int i = QInputDialog::getInt(this, tr("Change mesh edge width"),
                                   tr("line width"), dd, 1, 100, 1, &ok);
      if (ok) {
        edge_widthLabel->setText(QString::number(i)); // tr("%1%").arg(i)
      }
    } else if (ii == 4) {
      int dd = dual_edge_widthLabel->text().toInt();
      int i = QInputDialog::getInt(this, tr("Change Voronoi edge width"),
                                   tr("line width"), dd, 1, 100, 1, &ok);
      if (ok) {
        dual_edge_widthLabel->setText(QString::number(i)); // tr("%1%").arg(i)
      }
    } else if (ii == 5) {
      int dd = ccircle_widthLabel->text().toInt();
      int i = QInputDialog::getInt(this, tr("Change circumcircle width"),
                                   tr("line width"), dd, 1, 100, 1, &ok);
      if (ok) {
        ccircle_widthLabel->setText(QString::number(i)); // tr("%1%").arg(i)
      }
    } else if (ii == 6) {
      int dd = v3d_vertex_sizeLabel->text().toInt();
      int i = QInputDialog::getInt(this, tr("Change view3d vertex size"),
                                   tr("line width"), dd, 1, 100, 1, &ok);
      if (ok) {
        v3d_vertex_sizeLabel->setText(QString::number(i)); // tr("%1%").arg(i)
      }
    } else if (ii == 7) {
      int dd = v3d_edge_widthLabel->text().toInt();
      int i = QInputDialog::getInt(this, tr("Change view3d edge size"),
                                   tr("line width"), dd, 1, 100, 1, &ok);
      if (ok) {
        v3d_edge_widthLabel->setText(QString::number(i)); // tr("%1%").arg(i)
      }
    } else if (ii == 8) {
      int dd = v3d_segment_widthLabel->text().toInt();
      int i = QInputDialog::getInt(this, tr("Change view3d segment size"),
                                   tr("line width"), dd, 1, 100, 1, &ok);
      if (ok) {
        v3d_segment_widthLabel->setText(QString::number(i)); // tr("%1%").arg(i)
      }
    }
}

void PreferDialog::setColor(int i)
{
    //ScribbleArea * sptr = (ScribbleArea *) _scribblearea;
    const QColorDialog::ColorDialogOptions options = QFlag(colorDialogOptionsWidget->value());
    if (i == 0) {
      const QColor color = QColorDialog::getColor(background_colorLabel->palette().color(QPalette::Window), this, "Select Color", options);
      if (color.isValid()) {
          background_colorLabel->setText(color.name());
          background_colorLabel->setPalette(QPalette(color));
          background_colorLabel->setAutoFillBackground(true);
      }
    } else if (i == 1) {
      const QColor color = QColorDialog::getColor(vertex_colorLabel->palette().color(QPalette::Window), this, "Select Color", options);
      if (color.isValid()) {
          vertex_colorLabel->setText(color.name());
          vertex_colorLabel->setPalette(QPalette(color));
          vertex_colorLabel->setAutoFillBackground(true);
      }
    } else if (i == 2) {
      const QColor color = QColorDialog::getColor(segment_colorLabel->palette().color(QPalette::Window), this, "Select Color", options);
      if (color.isValid()) {
          segment_colorLabel->setText(color.name());
          segment_colorLabel->setPalette(QPalette(color));
          segment_colorLabel->setAutoFillBackground(true);
      }
    } else if (i == 3) {
      const QColor color = QColorDialog::getColor(selected_colorLabel->palette().color(QPalette::Window), this, "Select Color", options);
      if (color.isValid()) {
          selected_colorLabel->setText(color.name());
          selected_colorLabel->setPalette(QPalette(color));
          selected_colorLabel->setAutoFillBackground(true);
      }
    } else if (i == 4) {
      const QColor color = QColorDialog::getColor(edge_colorLabel->palette().color(QPalette::Window), this, "Select Color", options);
      if (color.isValid()) {
          edge_colorLabel->setText(color.name());
          edge_colorLabel->setPalette(QPalette(color));
          edge_colorLabel->setAutoFillBackground(true);
      }
    } else if (i == 5) {
      const QColor color = QColorDialog::getColor(dual_edge_colorLabel->palette().color(QPalette::Window), this, "Select Color", options);
      if (color.isValid()) {
          dual_edge_colorLabel->setText(color.name());
          dual_edge_colorLabel->setPalette(QPalette(color));
          dual_edge_colorLabel->setAutoFillBackground(true);
      }
    } else if (i == 6) {
      const QColor color = QColorDialog::getColor(bgmesh_edge_colorLabel->palette().color(QPalette::Window), this, "Select Color", options);
      if (color.isValid()) {
          bgmesh_edge_colorLabel->setText(color.name());
          bgmesh_edge_colorLabel->setPalette(QPalette(color));
          bgmesh_edge_colorLabel->setAutoFillBackground(true);
      }
    } else if (i == 7) {
      const QColor color = QColorDialog::getColor(mesh_size_colorLabel->palette().color(QPalette::Window), this, "Select Color", options);
      if (color.isValid()) {
          mesh_size_colorLabel->setText(color.name());
          mesh_size_colorLabel->setPalette(QPalette(color));
          mesh_size_colorLabel->setAutoFillBackground(true);
      }
    } else if (i == 8) {
      const QColor color = QColorDialog::getColor(ccircle_colorLabel->palette().color(QPalette::Window), this, "Select Color", options);
      if (color.isValid()) {
          ccircle_colorLabel->setText(color.name());
          ccircle_colorLabel->setPalette(QPalette(color));
          ccircle_colorLabel->setAutoFillBackground(true);
      }
    // view3d
    } else if (i == 9) {
      const QColor color = QColorDialog::getColor(v3d_background_colorLabel->palette().color(QPalette::Window), this, "Select Color", options);
      if (color.isValid()) {
        v3d_background_colorLabel->setText(color.name());
        v3d_background_colorLabel->setPalette(QPalette(color));
        v3d_background_colorLabel->setAutoFillBackground(true);
      }
    } else if (i == 10) {
      const QColor color = QColorDialog::getColor(v3d_vertex_colorLabel->palette().color(QPalette::Window), this, "Select Color", options);
      if (color.isValid()) {
        v3d_vertex_colorLabel->setText(color.name());
        v3d_vertex_colorLabel->setPalette(QPalette(color));
        v3d_vertex_colorLabel->setAutoFillBackground(true);
      }
    } else if (i == 11) {
      const QColor color = QColorDialog::getColor(v3d_edge_colorLabel->palette().color(QPalette::Window), this, "Select Color", options);
      if (color.isValid()) {
        v3d_edge_colorLabel->setText(color.name());
        v3d_edge_colorLabel->setPalette(QPalette(color));
        v3d_edge_colorLabel->setAutoFillBackground(true);
      }
    } else if (i == 12) {
      const QColor color = QColorDialog::getColor(v3d_facet_colorLabel->palette().color(QPalette::Window), this, "Select Color", options);
      if (color.isValid()) {
        v3d_facet_colorLabel->setText(color.name());
        v3d_facet_colorLabel->setPalette(QPalette(color));
        v3d_facet_colorLabel->setAutoFillBackground(true);
      }
    } else if (i == 13) {
      const QColor color = QColorDialog::getColor(v3d_polecolorLabel->palette().color(QPalette::Window), this, "Select Color", options);
      if (color.isValid()) {
        v3d_polecolorLabel->setText(color.name());
        v3d_polecolorLabel->setPalette(QPalette(color));
        v3d_polecolorLabel->setAutoFillBackground(true);
      }
    } else if (i == 14) {
      const QColor color = QColorDialog::getColor(v3d_polarcolorLabel->palette().color(QPalette::Window), this, "Select Color", options);
      if (color.isValid()) {
        v3d_polarcolorLabel->setText(color.name());
        v3d_polarcolorLabel->setPalette(QPalette(color));
        v3d_polarcolorLabel->setAutoFillBackground(true);
      }
    }
}

void PreferDialog::setFont(int ii)
{   /*
    const QFontDialog::FontDialogOptions options = QFlag(fontDialogOptionsWidget->value());
    bool ok;
    QFont font = QFontDialog::getFont(&ok, QFont(fontLabel->text()), this, "Select Font", options);
    if (ok) {
        fontLabel->setText(font.key());
        fontLabel->setFont(font);
    }
    */

    //ScribbleArea * sptr = (ScribbleArea *) _scribblearea;
    const QFontDialog::FontDialogOptions options = QFlag(fontDialogOptionsWidget->value());
    bool ok;

    if (ii == 0) {
      QFont font = QFontDialog::getFont(&ok, QFont(bbox_fontLabel->text()), this, "Select Font", options);
      if (ok) {
          bbox_fontLabel->setText(font.key());
          bbox_fontLabel->setFont(font);
      }
    } else if (ii == 1) {
      QFont font = QFontDialog::getFont(&ok, QFont(vertex_fontLabel->text()), this, "Select Font", options);
      if (ok) {
          vertex_fontLabel->setText(font.key());
          vertex_fontLabel->setFont(font);
      }
    } else if (ii == 2) {
      QFont font = QFontDialog::getFont(&ok, QFont(segment_fontLabel->text()), this, "Select Font", options);
      if (ok) {
        segment_fontLabel->setText(font.key());
        segment_fontLabel->setFont(font);
      }
    } else if (ii == 3) {
      QFont font = QFontDialog::getFont(&ok, QFont(subdomain_fontLabel->text()), this, "Select Font", options);
      if (ok) {
        subdomain_fontLabel->setText(font.key());
        subdomain_fontLabel->setFont(font);
      }
    }

}

void PreferDialog::setExistingDirectory()
{
    QFileDialog::Options options = QFlag(fileDialogOptionsWidget->value());
    options |= QFileDialog::DontResolveSymlinks | QFileDialog::ShowDirsOnly;
    QString directory = QFileDialog::getExistingDirectory(this,
                                tr("QFileDialog::getExistingDirectory()"),
                                directoryLabel->text(),
                                options);
    if (!directory.isEmpty())
        directoryLabel->setText(directory);
}

void PreferDialog::setOpenFileName()
{
    const QFileDialog::Options options = QFlag(fileDialogOptionsWidget->value());
    QString selectedFilter;
    QString fileName = QFileDialog::getOpenFileName(this,
                                tr("Load background image from file"),
                                openFileNameLabel->text(),
                                tr("*.png;;*.jpg"),
                                &selectedFilter,
                                options);
    if (!fileName.isEmpty()) {
        backimageLabel->setText(fileName);
    }
}

void PreferDialog::setSaveFileName()
{
    const QFileDialog::Options options = QFlag(fileDialogOptionsWidget->value());
    QString selectedFilter;
    QString fileName = QFileDialog::getSaveFileName(this,
                                tr("QFileDialog::getSaveFileName()"),
                                saveFileNameLabel->text(),
                                tr("All Files (*);;Text Files (*.txt)"),
                                &selectedFilter,
                                options);
    if (!fileName.isEmpty())
        saveFileNameLabel->setText(fileName);
}

void PreferDialog::on_apply()
{
    qDebug()<<"apply!";

    ScribbleArea *sa = (ScribbleArea *)_scribblearea;

    sa->_geom->io_tol_rel_gap = rel_min_gap_tolLabel->text().toDouble();
    //sa->_geom->io_tol_mingap = abs_min_gap_tolLabel->text().toDouble();
    sa->_geom->io_tol_minangle = min_angle_tolLabel->text().toDouble() * detri2::PI / 180.;

    sa->_draw_vertex_size = vertex_sizeLabel->text().toInt();
    sa->_draw_segment_width = segment_widthLabel->text().toInt();
    sa->_draw_edge_width = edge_widthLabel->text().toInt();
    sa->_draw_dual_edge_width = dual_edge_widthLabel->text().toInt();
    // view3d
    sa->_v3d_vertex_size = v3d_vertex_sizeLabel->text().toInt();
    sa->_v3d_edge_width = v3d_edge_widthLabel->text().toInt();
    sa->_v3d_segment_width = v3d_segment_widthLabel->text().toInt();

    sa->_color_canvas_background = background_colorLabel->palette().color(QPalette::Window);
    sa->_color_vertex = vertex_colorLabel->palette().color(QPalette::Window);
    sa->_color_segment = segment_colorLabel->palette().color(QPalette::Window);
    sa->_color_selected = selected_colorLabel->palette().color(QPalette::Window);
    sa->_color_edge = edge_colorLabel->palette().color(QPalette::Window);
    sa->_color_dual_edge = dual_edge_colorLabel->palette().color(QPalette::Window);
    sa->_color_bgmesh_edge = bgmesh_edge_colorLabel->palette().color(QPalette::Window);
    sa->_color_mesh_size = mesh_size_colorLabel->palette().color(QPalette::Window);
    sa->_color_ccircle = ccircle_colorLabel->palette().color(QPalette::Window);
    // view3d
    sa->_v3d_bgcolor = v3d_background_colorLabel->palette().color(QPalette::Window);
    sa->_v3d_vertexcolor = v3d_vertex_colorLabel->palette().color(QPalette::Window);
    sa->_v3d_edgecolor = v3d_edge_colorLabel->palette().color(QPalette::Window);
    sa->_v3d_facetcolor = v3d_facet_colorLabel->palette().color(QPalette::Window);
    sa->_v3d_polecolor = v3d_polarcolorLabel->palette().color(QPalette::Window);
    sa->_v3d_polarcolor = v3d_polarcolorLabel->palette().color(QPalette::Window);

    sa->_font_bbox_label = bbox_fontLabel->font();
    sa->_font_vertex_label = vertex_fontLabel->font();
    sa->_font_segment_label = segment_fontLabel->font();
    sa->_font_subdomain_label = subdomain_fontLabel->font();

    if (!backimageLabel->text().isEmpty()) {
      sa->_background_img = new QImage(backimageLabel->text());
      if (sa->_background_img == NULL) {
        qDebug()<<"Wrong background image file.";
      } //else {
        //sa->_background_img->save("/Users/si/Desktop/dump.png", 0, 50);
      //}
    }

    sa->_object_valid = 0;
    sa->update();

    if (sa->_view3dWindow != NULL) {
      sa->_view3dWindow->_glpanel->view_update_act();
    }

    hide();
}

/*
Dialog::Dialog(QWidget *parent)
    : QWidget(parent)
{
    QVBoxLayout *verticalLayout;
    if (QGuiApplication::styleHints()->showIsFullScreen() || QGuiApplication::styleHints()->showIsMaximized()) {
        QHBoxLayout *horizontalLayout = new QHBoxLayout(this);
        QGroupBox *groupBox = new QGroupBox(QGuiApplication::applicationDisplayName(), this);
        horizontalLayout->addWidget(groupBox);
        verticalLayout = new QVBoxLayout(groupBox);
    } else {
        verticalLayout = new QVBoxLayout(this);
    }

    QToolBox *toolbox = new QToolBox;
    verticalLayout->addWidget(toolbox);

    errorMessageDialog = new QErrorMessage(this);

    int frameStyle = QFrame::Sunken | QFrame::Panel;

    integerLabel = new QLabel;
    integerLabel->setFrameStyle(frameStyle);
    QPushButton *integerButton =
            new QPushButton(tr("QInputDialog::get&Int()"));

    doubleLabel = new QLabel;
    doubleLabel->setFrameStyle(frameStyle);
    QPushButton *doubleButton =
            new QPushButton(tr("QInputDialog::get&Double()"));

    itemLabel = new QLabel;
    itemLabel->setFrameStyle(frameStyle);
    QPushButton *itemButton = new QPushButton(tr("QInputDialog::getIte&m()"));

    textLabel = new QLabel;
    textLabel->setFrameStyle(frameStyle);
    QPushButton *textButton = new QPushButton(tr("QInputDialog::get&Text()"));

    multiLineTextLabel = new QLabel;
    multiLineTextLabel->setFrameStyle(frameStyle);
    QPushButton *multiLineTextButton = new QPushButton(tr("QInputDialog::get&MultiLineText()"));

    colorLabel = new QLabel;
    colorLabel->setFrameStyle(frameStyle);
    QPushButton *colorButton = new QPushButton(tr("QColorDialog::get&Color()"));

    fontLabel = new QLabel;
    fontLabel->setFrameStyle(frameStyle);
    QPushButton *fontButton = new QPushButton(tr("QFontDialog::get&Font()"));

    directoryLabel = new QLabel;
    directoryLabel->setFrameStyle(frameStyle);
    QPushButton *directoryButton =
            new QPushButton(tr("QFileDialog::getE&xistingDirectory()"));

    openFileNameLabel = new QLabel;
    openFileNameLabel->setFrameStyle(frameStyle);
    QPushButton *openFileNameButton =
            new QPushButton(tr("QFileDialog::get&OpenFileName()"));

    openFileNamesLabel = new QLabel;
    openFileNamesLabel->setFrameStyle(frameStyle);
    QPushButton *openFileNamesButton =
            new QPushButton(tr("QFileDialog::&getOpenFileNames()"));

    saveFileNameLabel = new QLabel;
    saveFileNameLabel->setFrameStyle(frameStyle);
    QPushButton *saveFileNameButton =
            new QPushButton(tr("QFileDialog::get&SaveFileName()"));

    criticalLabel = new QLabel;
    criticalLabel->setFrameStyle(frameStyle);
    QPushButton *criticalButton =
            new QPushButton(tr("QMessageBox::critica&l()"));

    informationLabel = new QLabel;
    informationLabel->setFrameStyle(frameStyle);
    QPushButton *informationButton =
            new QPushButton(tr("QMessageBox::i&nformation()"));

    questionLabel = new QLabel;
    questionLabel->setFrameStyle(frameStyle);
    QPushButton *questionButton =
            new QPushButton(tr("QMessageBox::&question()"));

    warningLabel = new QLabel;
    warningLabel->setFrameStyle(frameStyle);
    QPushButton *warningButton = new QPushButton(tr("QMessageBox::&warning()"));

    errorLabel = new QLabel;
    errorLabel->setFrameStyle(frameStyle);
    QPushButton *errorButton =
            new QPushButton(tr("QErrorMessage::showM&essage()"));

    connect(integerButton, &QAbstractButton::clicked, this, &Dialog::setInteger);
    connect(doubleButton, &QAbstractButton::clicked, this, &Dialog::setDouble);
    connect(itemButton, &QAbstractButton::clicked, this, &Dialog::setItem);
    connect(textButton, &QAbstractButton::clicked, this, &Dialog::setText);
    connect(multiLineTextButton, &QAbstractButton::clicked, this, &Dialog::setMultiLineText);
    connect(colorButton, &QAbstractButton::clicked, this, &Dialog::setColor);
    connect(fontButton, &QAbstractButton::clicked, this, &Dialog::setFont);
    connect(directoryButton, &QAbstractButton::clicked,
            this, &Dialog::setExistingDirectory);
    connect(openFileNameButton, &QAbstractButton::clicked,
            this, &Dialog::setOpenFileName);
    connect(openFileNamesButton, &QAbstractButton::clicked,
            this, &Dialog::setOpenFileNames);
    connect(saveFileNameButton, &QAbstractButton::clicked,
            this, &Dialog::setSaveFileName);
    connect(criticalButton, &QAbstractButton::clicked, this, &Dialog::criticalMessage);
    connect(informationButton, &QAbstractButton::clicked,
            this, &Dialog::informationMessage);
    connect(questionButton, &QAbstractButton::clicked, this, &Dialog::questionMessage);
    connect(warningButton, &QAbstractButton::clicked, this, &Dialog::warningMessage);
    connect(errorButton, &QAbstractButton::clicked, this, &Dialog::errorMessage);

    QWidget *page = new QWidget;
    QGridLayout *layout = new QGridLayout(page);
    layout->setColumnStretch(1, 1);
    layout->setColumnMinimumWidth(1, 250);
    layout->addWidget(integerButton, 0, 0);
    layout->addWidget(integerLabel, 0, 1);
    layout->addWidget(doubleButton, 1, 0);
    layout->addWidget(doubleLabel, 1, 1);
    layout->addWidget(itemButton, 2, 0);
    layout->addWidget(itemLabel, 2, 1);
    layout->addWidget(textButton, 3, 0);
    layout->addWidget(textLabel, 3, 1);
    layout->addWidget(multiLineTextButton, 4, 0);
    layout->addWidget(multiLineTextLabel, 4, 1);
    layout->addItem(new QSpacerItem(0, 0, QSizePolicy::Ignored, QSizePolicy::MinimumExpanding), 5, 0);
    toolbox->addItem(page, tr("Input Dialogs"));

    const QString doNotUseNativeDialog = tr("Do not use native dialog");

    page = new QWidget;
    layout = new QGridLayout(page);
    layout->setColumnStretch(1, 1);
    layout->addWidget(colorButton, 0, 0);
    layout->addWidget(colorLabel, 0, 1);
    colorDialogOptionsWidget = new DialogOptionsWidget;
    colorDialogOptionsWidget->addCheckBox(doNotUseNativeDialog, QColorDialog::DontUseNativeDialog);
    colorDialogOptionsWidget->addCheckBox(tr("Show alpha channel") , QColorDialog::ShowAlphaChannel);
    colorDialogOptionsWidget->addCheckBox(tr("No buttons") , QColorDialog::NoButtons);
    layout->addItem(new QSpacerItem(0, 0, QSizePolicy::Ignored, QSizePolicy::MinimumExpanding), 1, 0);
    layout->addWidget(colorDialogOptionsWidget, 2, 0, 1 ,2);

    toolbox->addItem(page, tr("Color Dialog"));

    page = new QWidget;
    layout = new QGridLayout(page);
    layout->setColumnStretch(1, 1);
    layout->addWidget(fontButton, 0, 0);
    layout->addWidget(fontLabel, 0, 1);
    fontDialogOptionsWidget = new DialogOptionsWidget;
    fontDialogOptionsWidget->addCheckBox(doNotUseNativeDialog, QFontDialog::DontUseNativeDialog);
    fontDialogOptionsWidget->addCheckBox(tr("Show scalable fonts"), QFontDialog::ScalableFonts);
    fontDialogOptionsWidget->addCheckBox(tr("Show non scalable fonts"), QFontDialog::NonScalableFonts);
    fontDialogOptionsWidget->addCheckBox(tr("Show monospaced fonts"), QFontDialog::MonospacedFonts);
    fontDialogOptionsWidget->addCheckBox(tr("Show proportional fonts"), QFontDialog::ProportionalFonts);
    fontDialogOptionsWidget->addCheckBox(tr("No buttons") , QFontDialog::NoButtons);
    layout->addItem(new QSpacerItem(0, 0, QSizePolicy::Ignored, QSizePolicy::MinimumExpanding), 1, 0);
    layout->addWidget(fontDialogOptionsWidget, 2, 0, 1 ,2);
    toolbox->addItem(page, tr("Font Dialog"));

    page = new QWidget;
    layout = new QGridLayout(page);
    layout->setColumnStretch(1, 1);
    layout->addWidget(directoryButton, 0, 0);
    layout->addWidget(directoryLabel, 0, 1);
    layout->addWidget(openFileNameButton, 1, 0);
    layout->addWidget(openFileNameLabel, 1, 1);
    layout->addWidget(openFileNamesButton, 2, 0);
    layout->addWidget(openFileNamesLabel, 2, 1);
    layout->addWidget(saveFileNameButton, 3, 0);
    layout->addWidget(saveFileNameLabel, 3, 1);
    fileDialogOptionsWidget = new DialogOptionsWidget;
    fileDialogOptionsWidget->addCheckBox(doNotUseNativeDialog, QFileDialog::DontUseNativeDialog);
    fileDialogOptionsWidget->addCheckBox(tr("Show directories only"), QFileDialog::ShowDirsOnly);
    fileDialogOptionsWidget->addCheckBox(tr("Do not resolve symlinks"), QFileDialog::DontResolveSymlinks);
    fileDialogOptionsWidget->addCheckBox(tr("Do not confirm overwrite"), QFileDialog::DontConfirmOverwrite);
    fileDialogOptionsWidget->addCheckBox(tr("Do not use sheet"), QFileDialog::DontUseSheet);
    fileDialogOptionsWidget->addCheckBox(tr("Readonly"), QFileDialog::ReadOnly);
    fileDialogOptionsWidget->addCheckBox(tr("Hide name filter details"), QFileDialog::HideNameFilterDetails);
    fileDialogOptionsWidget->addCheckBox(tr("Do not use custom directory icons (Windows)"), QFileDialog::DontUseCustomDirectoryIcons);
    layout->addItem(new QSpacerItem(0, 0, QSizePolicy::Ignored, QSizePolicy::MinimumExpanding), 4, 0);
    layout->addWidget(fileDialogOptionsWidget, 5, 0, 1 ,2);
    toolbox->addItem(page, tr("File Dialogs"));

    page = new QWidget;
    layout = new QGridLayout(page);
    layout->setColumnStretch(1, 1);
    layout->addWidget(criticalButton, 0, 0);
    layout->addWidget(criticalLabel, 0, 1);
    layout->addWidget(informationButton, 1, 0);
    layout->addWidget(informationLabel, 1, 1);
    layout->addWidget(questionButton, 2, 0);
    layout->addWidget(questionLabel, 2, 1);
    layout->addWidget(warningButton, 3, 0);
    layout->addWidget(warningLabel, 3, 1);
    layout->addWidget(errorButton, 4, 0);
    layout->addWidget(errorLabel, 4, 1);
    layout->addItem(new QSpacerItem(0, 0, QSizePolicy::Ignored, QSizePolicy::MinimumExpanding), 5, 0);
    toolbox->addItem(page, tr("Message Boxes"));

    setWindowTitle(QGuiApplication::applicationDisplayName());
}

void Dialog::setInteger()
{
//! [0]
    bool ok;
    int i = QInputDialog::getInt(this, tr("QInputDialog::getInteger()"),
                                 tr("Percentage:"), 25, 0, 100, 1, &ok);
    if (ok)
        integerLabel->setText(tr("%1%").arg(i));
//! [0]
}

void Dialog::setDouble()
{
//! [1]
    bool ok;
    double d = QInputDialog::getDouble(this, tr("QInputDialog::getDouble()"),
                                       tr("Amount:"), 37.56, -10000, 10000, 2, &ok);
    if (ok)
        doubleLabel->setText(QString("$%1").arg(d));
//! [1]
}

void Dialog::setItem()
{
//! [2]
    QStringList items;
    items << tr("Spring") << tr("Summer") << tr("Fall") << tr("Winter");

    bool ok;
    QString item = QInputDialog::getItem(this, tr("QInputDialog::getItem()"),
                                         tr("Season:"), items, 0, false, &ok);
    if (ok && !item.isEmpty())
        itemLabel->setText(item);
//! [2]
}

void Dialog::setText()
{
//! [3]
    bool ok;
    QString text = QInputDialog::getText(this, tr("QInputDialog::getText()"),
                                         tr("User name:"), QLineEdit::Normal,
                                         QDir::home().dirName(), &ok);
    if (ok && !text.isEmpty())
        textLabel->setText(text);
//! [3]
}

void Dialog::setMultiLineText()
{
//! [4]
    bool ok;
    QString text = QInputDialog::getMultiLineText(this, tr("QInputDialog::getMultiLineText()"),
                                                  tr("Address:"), "John Doe\nFreedom Street", &ok);
    if (ok && !text.isEmpty())
        multiLineTextLabel->setText(text);
//! [4]
}

void Dialog::setColor()
{
    const QColorDialog::ColorDialogOptions options = QFlag(colorDialogOptionsWidget->value());
    const QColor color = QColorDialog::getColor(Qt::green, this, "Select Color", options);

    if (color.isValid()) {
        colorLabel->setText(color.name());
        colorLabel->setPalette(QPalette(color));
        colorLabel->setAutoFillBackground(true);
    }
}

void Dialog::setFont()
{
    const QFontDialog::FontDialogOptions options = QFlag(fontDialogOptionsWidget->value());
    bool ok;
    QFont font = QFontDialog::getFont(&ok, QFont(fontLabel->text()), this, "Select Font", options);
    if (ok) {
        fontLabel->setText(font.key());
        fontLabel->setFont(font);
    }
}

void Dialog::setExistingDirectory()
{
    QFileDialog::Options options = QFlag(fileDialogOptionsWidget->value());
    options |= QFileDialog::DontResolveSymlinks | QFileDialog::ShowDirsOnly;
    QString directory = QFileDialog::getExistingDirectory(this,
                                tr("QFileDialog::getExistingDirectory()"),
                                directoryLabel->text(),
                                options);
    if (!directory.isEmpty())
        directoryLabel->setText(directory);
}

void Dialog::setOpenFileName()
{
    const QFileDialog::Options options = QFlag(fileDialogOptionsWidget->value());
    QString selectedFilter;
    QString fileName = QFileDialog::getOpenFileName(this,
                                tr("QFileDialog::getOpenFileName()"),
                                openFileNameLabel->text(),
                                tr("All Files (*);;Text Files (*.txt)"),
                                &selectedFilter,
                                options);
    if (!fileName.isEmpty())
        openFileNameLabel->setText(fileName);
}

void Dialog::setOpenFileNames()
{
    const QFileDialog::Options options = QFlag(fileDialogOptionsWidget->value());
    QString selectedFilter;
    QStringList files = QFileDialog::getOpenFileNames(
                                this, tr("QFileDialog::getOpenFileNames()"),
                                openFilesPath,
                                tr("All Files (*);;Text Files (*.txt)"),
                                &selectedFilter,
                                options);
    if (files.count()) {
        openFilesPath = files[0];
        openFileNamesLabel->setText(QString("[%1]").arg(files.join(", ")));
    }
}

void Dialog::setSaveFileName()
{
    const QFileDialog::Options options = QFlag(fileDialogOptionsWidget->value());
    QString selectedFilter;
    QString fileName = QFileDialog::getSaveFileName(this,
                                tr("QFileDialog::getSaveFileName()"),
                                saveFileNameLabel->text(),
                                tr("All Files (*);;Text Files (*.txt)"),
                                &selectedFilter,
                                options);
    if (!fileName.isEmpty())
        saveFileNameLabel->setText(fileName);
}

void Dialog::criticalMessage()
{
    QMessageBox::StandardButton reply;
    reply = QMessageBox::critical(this, tr("QMessageBox::critical()"),
                                    MESSAGE,
                                    QMessageBox::Abort | QMessageBox::Retry | QMessageBox::Ignore);
    if (reply == QMessageBox::Abort)
        criticalLabel->setText(tr("Abort"));
    else if (reply == QMessageBox::Retry)
        criticalLabel->setText(tr("Retry"));
    else
        criticalLabel->setText(tr("Ignore"));
}

void Dialog::informationMessage()
{
    QMessageBox::StandardButton reply;
    reply = QMessageBox::information(this, tr("QMessageBox::information()"), MESSAGE);
    if (reply == QMessageBox::Ok)
        informationLabel->setText(tr("OK"));
    else
        informationLabel->setText(tr("Escape"));
}

void Dialog::questionMessage()
{
    QMessageBox::StandardButton reply;
    reply = QMessageBox::question(this, tr("QMessageBox::question()"),
                                    MESSAGE,
                                    QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel);
    if (reply == QMessageBox::Yes)
        questionLabel->setText(tr("Yes"));
    else if (reply == QMessageBox::No)
        questionLabel->setText(tr("No"));
    else
        questionLabel->setText(tr("Cancel"));
}

void Dialog::warningMessage()
{
    QMessageBox msgBox(QMessageBox::Warning, tr("QMessageBox::warning()"),
                       MESSAGE, 0, this);
    msgBox.setDetailedText(MESSAGE_DETAILS);
    msgBox.addButton(tr("Save &Again"), QMessageBox::AcceptRole);
    msgBox.addButton(tr("&Continue"), QMessageBox::RejectRole);
    if (msgBox.exec() == QMessageBox::AcceptRole)
        warningLabel->setText(tr("Save Again"));
    else
        warningLabel->setText(tr("Continue"));

}

void Dialog::errorMessage()
{
    errorMessageDialog->showMessage(
            tr("This dialog shows and remembers error messages. "
               "If the checkbox is checked (as it is by default), "
               "the shown message will be shown again, "
               "but if the user unchecks the box the message "
               "will not appear again if QErrorMessage::showMessage() "
               "is called with the same message."));
    errorLabel->setText(tr("If the box is unchecked, the message "
                           "won't appear again."));
}
*/
