#include <QtWidgets>

#include "scribblearea.h"
//#include "querydialog.h"

QueryDialog::QueryDialog(void *win, QWidget *parent)
    : QDialog(parent)
{
    _scribblearea = win;

    int frameStyle = QFrame::Sunken | QFrame::Panel;

    op_show_vertex   = new QRadioButton(tr("Vertex"));
    op_show_edge     = new QRadioButton(tr("Edge"));
    op_show_triangle = new QRadioButton(tr("Triangle"));
    op_show_location = new QRadioButton(tr("Show location (i,j)"));
    op_show_vertex->setChecked(true);

    _i = new QLineEdit;
    _i->setText(tr("0"));
    QLabel *iButton = new QLabel(tr("i"));
    _j = new QLineEdit;
    _j->setText(tr("0"));
    QLabel *jButton = new QLabel(tr("j"));
    _k = new QLineEdit;
    _k->setText(tr("0"));
    QLabel *kButton = new QLabel(tr("k"));

    _queryMesh = new QCheckBox(tr("Query in mesh"));
    _queryMesh->setChecked(true); // default

    applyButton = new QPushButton(tr("apply"));
    connect(applyButton, &QAbstractButton::clicked, this, &QueryDialog::eval);

    QGridLayout *layout = new QGridLayout(this);

    layout->addWidget(op_show_vertex, 0,0,1,2);
    layout->addWidget(op_show_edge, 1,0,1,2);
    layout->addWidget(op_show_triangle, 2,0,1,2);
    layout->addWidget(op_show_location, 3,0,1,2);
    layout->addWidget(_queryMesh, 4,0,1,2);

    layout->addWidget(iButton, 5, 0,1,1);
    layout->addWidget(_i, 5, 1,1,1);
    layout->addWidget(jButton, 6, 0,1,1);
    layout->addWidget(_j, 6, 1,1,1);
    layout->addWidget(kButton, 7, 0,1,1);
    layout->addWidget(_k, 7, 1,1,1);

    layout->addWidget(applyButton, 8, 0);
}

/*
void MuParserDialog::setText()
{
    //qDebug()<<"clicked here";
    bool ok;
    QString text = QInputDialog::getText(this, tr("Input formula"),
                                         tr("formula:"), QLineEdit::Normal,
                                         QString(""), &ok);
    if (ok && !text.isEmpty())
        textLabel->setText(text);
}
*/

void QueryDialog::eval()
{
    ScribbleArea *sa = (ScribbleArea *) _scribblearea;
    if (op_show_vertex->isChecked()) {
      int idx = _i->text().toInt();
      detri2::Vertex *v = NULL;
      if (_queryMesh->isChecked()) {
        if (sa->_tria != NULL) {
          qDebug()<<"Query vertex idx="<<idx<<" in mesh";
          v = sa->_tria->db_vrt(idx);
        } else {
          qDebug()<<"Mesh not exists!";
        }
      } else {
        qDebug()<<"Query vertex idx="<<idx<<" in geom";
        v = sa->_geom->db_vrt(idx);
      }
      if (v != NULL) {
        // Draw (highlit) this vertex, Draw a circle centered at it.
        int U = (int) ( sa->_scale_xy * v->crd[0] + sa->_cx);
        int V = (int) (-sa->_scale_xy * v->crd[1] + sa->_cy);
        QPainter painter(&sa->_canvas[0]);
        QPen pen(qRgb(255,0,0)); // red
        pen.setWidth(2);
        painter.setPen(pen);
        painter.drawEllipse(QPoint(U,V), 6, 6);
        sa->update();
      } else {
        qDebug()<<"Did not find vertex idx="<<idx;
      }
    }
    else if (op_show_edge->isChecked()) {
      int e1 = _i->text().toInt();
      int e2 = _j->text().toInt();
      detri2::TriEdge E;
      if (_queryMesh->isChecked()) {
        if (sa->_tria != NULL) {
          qDebug()<<"Query edge ["<<e1<<","<<e2<<"] in mesh";
          E = sa->_tria->db_seg(e1, e2);
        } else {
          qDebug()<<"Mesh not exists!";
        }
      } else {
        qDebug()<<"Query edge ["<<e1<<","<<e2<<"] in geom";
        E = sa->_geom->db_seg(e1, e2);
      }
      if (E.tri != NULL) {
        // Draw (highlit) this edge
        QPainter painter(&sa->_canvas[0]);
        QPen pen;
        pen.setColor(qRgb(0, 255, 255));
        pen.setWidth(2); // default
        painter.setPen(pen);
        detri2::Vertex *pa = E.org();
        detri2::Vertex *pb = E.dest();
        detri2_draw_line(pa, pb, &painter, sa->_cx, sa->_cy, sa->_scale_xy);
        sa->update();
      } else {
        qDebug()<<"Did not find edge ["<<e1<<","<<e2<<"]";
      }
    }
    else if (op_show_triangle->isChecked()) {
      int e1 = _i->text().toInt();
      int e2 = _j->text().toInt();
      int e3 = _k->text().toInt();
      detri2::TriEdge E;
      if (_queryMesh->isChecked()) {
        if (sa->_tria != NULL) {
          qDebug()<<"Query triangle ["<<e1<<","<<e2<<","<<e3<<"] in mesh";
          E = sa->_tria->db_tri(e1, e2, e3);
        } else {
          qDebug()<<"Mesh not exists!";
        }
      } else {
        qDebug()<<"Query triangle ["<<e1<<","<<e2<<","<<e3<<"] in geom";
        E = sa->_geom->db_tri(e1, e2, e3);
      }
      if (E.tri != NULL) {
        // Draw (highlit) this triangle
        QPainter painter(&sa->_canvas[0]);
        QPen pen;
        pen.setColor(qRgb(0, 255, 255));
        pen.setWidth(2); // default
        painter.setPen(pen);
        detri2::Vertex *pa = E.org();
        detri2::Vertex *pb = E.dest();
        detri2::Vertex *pc = E.apex();
        detri2_draw_line(pa, pb, &painter, sa->_cx, sa->_cy, sa->_scale_xy);
        detri2_draw_line(pb, pc, &painter, sa->_cx, sa->_cy, sa->_scale_xy);
        detri2_draw_line(pc, pa, &painter, sa->_cx, sa->_cy, sa->_scale_xy);
        sa->update();
      } else {
        qDebug()<<"Did not find triangle ["<<e1<<","<<e2<<","<<e3<<"]";
      }
    }
    else if (op_show_location->isChecked()) {
      double x = _i->text().toDouble();
      double y = _j->text().toDouble();
      // Draw (highlit) this vertex, Draw a circle centered at it.
      int U = (int) ( sa->_scale_xy * x + sa->_cx);
      int V = (int) (-sa->_scale_xy * y + sa->_cy);
      QPainter painter(&sa->_canvas[0]);
      QPen pen(qRgb(0,255,0)); // green
      pen.setWidth(2);
      painter.setPen(pen);
      painter.drawEllipse(QPoint(U,V), 6, 6);
      sa->update();
    }
}

