#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "detri2qt.h"
#include "HDE/hde.h"

#include <QPainter>
#include <QDebug>

using namespace detri2;

//=============================================================================

void detri2_init_with_bbox(Triangulation **pTr, double xmin, double xmax, double ymin, double ymax)
{
  qDebug()<<"Init geometry with bbox ["<<xmin<<", "<<ymin<<"]x["<<xmax<<", "<<ymax<<"]";

  Triangulation *Tr = *pTr;
  if (Tr != NULL) delete Tr;
  Tr = new detri2::Triangulation();

  Tr->io_firstindex = 1; // indices starting from 1
  strcpy(Tr->io_outfilename, "dump_bbox"); // Only for debug

  Tr->ct_in_vrts = 4;
  Tr->in_vrts = new Vertex[4];
  Tr->ct_unused_vrts = 4;

  Vertex *v; double x, y;
  v = &(Tr->in_vrts[0]);
  v->init();
  x = v->crd[0] = xmin;
  y = v->crd[1] = ymin;
  v->crd[2] = x*x+y*y;
  //v->crd[2] = Tr->op_lambda1 * x*x + Tr->op_lambda2 * y*y;
  v->idx = -11;

  v = &(Tr->in_vrts[1]);
  v->init();
  x = v->crd[0] = xmax;
  y = v->crd[1] = ymin;
  v->crd[2] = x*x+y*y;
  //v->crd[2] = Tr->op_lambda1 * x*x + Tr->op_lambda2 * y*y;
  v->idx = -12;

  v = &(Tr->in_vrts[2]);
  v->init();
  x = v->crd[0] = xmax;
  y = v->crd[1] = ymax;
  v->crd[2] = x*x+y*y;
  //v->crd[2] = Tr->op_lambda1 * x*x + Tr->op_lambda2 * y*y;
  v->idx = -13;

  v = &(Tr->in_vrts[3]);
  v->init();
  x = v->crd[0] = xmin;
  y = v->crd[1] = ymax;
  v->crd[2] = x*x+y*y;
  //v->crd[2] = Tr->op_lambda1 * x*x + Tr->op_lambda2 * y*y;
  v->idx = -14;

  Tr->io_xmin = xmin;
  Tr->io_ymin = ymin;
  Tr->io_xmax = xmax;
  Tr->io_ymax = ymax;
  double dx = Tr->io_xmax - Tr->io_xmin;
  double dy = Tr->io_ymax - Tr->io_ymin;
  Tr->io_diagonal = sqrt(dx*dx + dy*dy);

  exactinit(0, 0, 0, 0, Tr->io_xmax - Tr->io_xmin, Tr->io_ymax - Tr->io_ymax, 0.0);

  Tr->incremental_delaunay();

  assert(Tr->tr_tris != NULL);
  Tr->tr_steiners = new arraypool(sizeof(Vertex), 10);
  Tr->tr_segs     = new arraypool(sizeof(Triang), 5);

  if (Tr->check_mesh(0,0) > 0) {
    qDebug()<<"BUG: detri2_init_with_bbox()";
    assert(0);
  }
  //Tr->save_triangulation();
  *pTr = Tr;
}

//=============================================================================

int detri2_reconstruct_geometry(Triangulation** pGeom, Triangulation* Tr)
{
  // Update _geom
  double xmin = Tr->io_xmin;
  double xmax = Tr->io_xmax;
  double ymin = Tr->io_ymin;
  double ymax = Tr->io_ymax;
  // Englarge the bounding box by 10%.
  double dmax = xmax - xmin;
  if (dmax < ymax - ymin) dmax = ymax - ymin;
  double dd = dmax * 0.1;
  xmin -= dd;
  ymin -= dd;
  xmax += dd;
  ymax += dd;

  detri2_init_with_bbox(pGeom, xmin, xmax, ymin, ymax);

  TriEdge E; int loc = 0;
  for (int i = 0; i < Tr->ct_in_vrts; i++) {    
    E.tri = NULL; // do detri2_locate_point();
    //if (i == Tr->ct_in_vrts-1) {
    //  printf("debug\n");
    //}
    Vertex *pp = &(Tr->in_vrts[i]);    
    if (Tr->tr_tris != NULL) {
      // Only insert geometry vertices.
      if (pp->on_bd.tri == NULL) continue;
    }

    double x = pp->crd[0];
    double y = pp->crd[1];
    double w = pp->wei; // weight
    //double dummy = 0;
    //loc = detri2_locate_point(*pGeom, x, y, E, dummy, 0); // Use exact locating.
    //if (loc != LOC_IN_OUTSIDE) {
    if (detri2_insert_point(*pGeom, x, y, E, loc)) { // weight=0
      //detri2_insert_point(*pGeom, x, y, E, loc);
      Vertex *newpt = E.org();
      newpt->crd[2] = x*x+y*y; // no weight (ensure every point appears in _geom).
      //newpt->crd[2] = Tr->op_lambda1 * x*x + Tr->op_lambda2 * y*y;
      newpt->wei = w; // weight
      newpt->Pair = pp; // point to this mesh vertex
      pp->Pair = newpt; // point to this geom vertex (for importing segments)
      newpt->idx = pp->idx;
      newpt->val = pp->val; // Vertex metric
      newpt->typ = STEINERVERTEX;
    } else {
      qDebug()<<"  Warning:  point is not inserted i="<<i << "E.org() = "<<E.org()->idx;
      //assert(0); // for debug
      //pp->Pair = NULL;
      pp->Pair = E.org();
    }
    /*
    // Debug segment vertex
    int j;
    for (j = 0; j < Tr->tr_segs->objects; j++) {
      Triang *seg = (Triang *)  Tr->tr_segs->get(j);
      Vertex *np1 = seg->vrt[0]->Pair;
      Vertex *np2 = seg->vrt[1]->Pair;
      if (np1 != NULL) {
        if (seg->vrt[0]->idx != np1->idx) {
          printf("  !!!! Check this ****************************** i=%d.\n", i);
          printf(" i=%d, v1=%d, v2=%d\n", i, seg->vrt[0]->idx, seg->vrt[1]->idx);
          break;
        }
      }
      if (np2 != NULL) {
        if (seg->vrt[1]->idx != np2->idx) {
          printf("  !!!! Check this ******************************* i=%d.\n", i);
          printf(" i=%d, v1=%d, v2=%d\n", i, seg->vrt[0]->idx, seg->vrt[1]->idx);
          break;
        }
      }
    }
    if (j < Tr->tr_segs->objects) break;
    */
  }

  /* // Debug
  printf("\n============================\n");
  for (int i = 0; i < Tr->tr_segs->objects; i++) {
    Triang *seg = (Triang *)  Tr->tr_segs->get(i);
    printf(" i=%d, v1=%d (%d), v2=%d (%d)\n", i,
           seg->vrt[0]->idx, (seg->vrt[0]->Pair)->idx,
           seg->vrt[1]->idx, (seg->vrt[1]->Pair)->idx);

  }
  */

  if (Tr->tr_segs) {
    //qDebug()<<"adding segments: "<< Tr->tr_segs->objects;
    //Tr->op_db_verbose = 4; // Debug
    assert((*pGeom)->tr_segs != NULL); // Created in detri2_init_bbox().
    //arraypool *fqueue = new arraypool(sizeof(TriEdge), 4);
    // Insert all segments into Geometry.
    for (int i = 0; i < Tr->tr_segs->objects; i++) {
      Triang *seg = (Triang *)  Tr->tr_segs->get(i);
      if (seg->is_deleted()) continue;
      if (seg->tag != 0) {
        //qDebug()<<" seg->tag="<<seg->tag<< " i="<<i;
        Vertex *e1 = seg->vrt[0]->Pair;
        Vertex *e2 = seg->vrt[1]->Pair;
        int tag = seg->tag;
        REAL val = seg->val;
        /*
        if ((e1 != NULL) && (e2 != NULL) && (e1 != e2)) {
          //detri2_insert_segment(*pGeom, e1, e2, tag, E);
          if ((*pGeom)->recover_edge(e1, e2, E, fqueue) == INTERSECT_SHARE_EDGE) {
            Triang *newseg = NULL;
            (*pGeom)->insert_segment(E, &newseg);
            newseg->tag = tag;
          } else {
            printf("!! Failed to recover segment [%d,%d] tga(%d).\n",
                   e1->idx, e2->idx, tag);
          }
          if (fqueue->objects > 0) {
            (*pGeom)->lawson_flip(NULL, 0, fqueue);
          }
        } else {
          qDebug()<<"  Warning:  skip segment i="<<i;
        }
        */
        if (!detri2_insert_segment(*pGeom, e1, e2, tag, val)) {
          printf("!! Failed to recover segment [%d,%d] tag(%d).\n", e1->idx, e2->idx, tag);
        }
      }
    }
    //delete fqueue;
    qDebug()<<"added "<<(*pGeom)->tr_segs->objects<< " segments";
  }

  if (Tr->ct_in_sdms > 0) {
    //qDebug()<<"adding "<<Tr->ct_in_sdms<< " subdomains";
    (*pGeom)->in_sdms = new Vertex[Tr->ct_in_sdms];
    int idx = 0;
    for (int i = 0; i < Tr->ct_in_sdms; i++) {
      // Make sure this vertex lies inside the geometry.
      E.tri = NULL;
      double x = Tr->in_sdms[i].crd[0];
      double y = Tr->in_sdms[i].crd[1];
      double dummy;
      loc = detri2_locate_point(*pGeom, x, y, E, dummy, 1);
      if (loc != LOC_IN_OUTSIDE) {
        (*pGeom)->in_sdms[idx].init();
        (*pGeom)->in_sdms[idx].crd[0] = x;
        (*pGeom)->in_sdms[idx].crd[1] = y;
        (*pGeom)->in_sdms[idx].tag = Tr->in_sdms[i].tag;
        (*pGeom)->in_sdms[idx].val = Tr->in_sdms[i].val;
        (*pGeom)->in_sdms[idx].adj = E; // Remember its location.
        idx++;
      }
    }
    (*pGeom)->ct_in_sdms = idx;
    qDebug()<<"added "<<(*pGeom)->ct_in_sdms<< " subdomains";
  }

  // debug
  //if ((*pGeom)->check_mesh(0,0) > 0) {
  //  qDebug() << "!!! BUG: Reconstruct mesh failed.";
  //}

  (*pGeom)->mesh_statistics();

  return 1;
}

//==============================================================================
// [2019-10-21]
// This function will be dropped in the futrue. Do not use it in openFile().

int detri2_reconstruct(Triangulation** pGeom, Triangulation** pTr, const char* filename)
{
  int argc = 2;
  char str1[100], str2[256];
  strcpy(str1, "detri2");
  sprintf(str2, "%s", filename);
  char *argv[2];
  argv[0] = str1;
  argv[1] = str2;

  // read the input files
  Triangulation *Tr=new Triangulation();

  Tr->parse_commands(argc, argv);
  if (!Tr->read_mesh()) {
    //qDebug()<<" Failed to read file "<< filename;
    delete Tr;
    return 0; // do nothing.
  }

  if (Tr->tr_tris != NULL) {
    Tr->reconstruct_mesh(0); // delaunay flag=0
    if (Tr->check_mesh(0,0) > 0) {
      printf("debug: detri2_reconstruct() Tr->reconstruct_mesh(0) failed.\n");
      return 0; //assert(0);
    }
  }

  if (*pTr != NULL) delete *pTr;
  *pTr = Tr;

  // Update _geom
  double xmin = Tr->io_xmin;
  double xmax = Tr->io_xmax;
  double ymin = Tr->io_ymin;
  double ymax = Tr->io_ymax;
  // Englarge the bounding box by 10%.
  double dmax = xmax - xmin;
  if (dmax < ymax - ymin) dmax = ymax - ymin;
  double dd = dmax * 0.1;
  xmin -= dd;
  ymin -= dd;
  xmax += dd;
  ymax += dd;

  detri2_init_with_bbox(pGeom, xmin, xmax, ymin, ymax);
  //strcpy((*pGeom)->io_outfilename, "/Users/si/tmp/dump"); // for debug only

  // Insert all Steiner points.
  qDebug()<<"Inserting "<<Tr->ct_in_vrts<<" Steiner points.";

  // Debug
  //for (int i = 0; i < Tr->tr_segs->objects; i++) {
  //  Triang *seg = (Triang *)  Tr->tr_segs->get(i);
  //  printf(" i=%d, v1=%d, v2=%d\n", i, seg->vrt[0]->idx, seg->vrt[1]->idx);
  //}

  TriEdge E; int loc = 0;
  for (int i = 0; i < Tr->ct_in_vrts; i++) {
    E.tri = NULL; // do detri2_locate_point();
    //if (i == Tr->ct_in_vrts-1) {
    //  printf("debug\n");
    //}
    Vertex *pp = &(Tr->in_vrts[i]);
    double x = pp->crd[0];
    double y = pp->crd[1];
    double w = pp->wei; // weight
    double dummy = 0;
    //loc = detri2_locate_point(*pGeom, x, y, E, dummy, 0); // Use exact locating.
    //if (loc != LOC_IN_OUTSIDE) {
    if (detri2_insert_point(*pGeom, x, y, E, loc)) { // weight=0
      //detri2_insert_point(*pGeom, x, y, E, loc);
      Vertex *newpt = E.org();
      newpt->crd[2] = x*x+y*y; // no weight (ensure every point appears in _geom).
      //newpt->crd[2] = Tr->op_lambda1 * x*x + Tr->op_lambda2 * y*y;
      newpt->wei = w; // weight
      newpt->Pair = pp; // point to this mesh vertex
      pp->Pair = newpt; // point to this geom vertex (for importing segments)
      newpt->idx = pp->idx;
      newpt->val = pp->val; // Vertex metric
      newpt->typ = STEINERVERTEX;
    } else {
      qDebug()<<"  Warning:  point is not inserted i="<<i << "E.org() = "<<E.org()->idx;
      //assert(0); // for debug
      //pp->Pair = NULL;
      pp->Pair = E.org();
    }
    /*
    // Debug segment vertex
    int j;
    for (j = 0; j < Tr->tr_segs->objects; j++) {
      Triang *seg = (Triang *)  Tr->tr_segs->get(j);
      Vertex *np1 = seg->vrt[0]->Pair;
      Vertex *np2 = seg->vrt[1]->Pair;
      if (np1 != NULL) {
        if (seg->vrt[0]->idx != np1->idx) {
          printf("  !!!! Check this ****************************** i=%d.\n", i);
          printf(" i=%d, v1=%d, v2=%d\n", i, seg->vrt[0]->idx, seg->vrt[1]->idx);
          break;
        }
      }
      if (np2 != NULL) {
        if (seg->vrt[1]->idx != np2->idx) {
          printf("  !!!! Check this ******************************* i=%d.\n", i);
          printf(" i=%d, v1=%d, v2=%d\n", i, seg->vrt[0]->idx, seg->vrt[1]->idx);
          break;
        }
      }
    }
    if (j < Tr->tr_segs->objects) break;
    */
  }

  /* // Debug
  printf("\n============================\n");
  for (int i = 0; i < Tr->tr_segs->objects; i++) {
    Triang *seg = (Triang *)  Tr->tr_segs->get(i);
    printf(" i=%d, v1=%d (%d), v2=%d (%d)\n", i,
           seg->vrt[0]->idx, (seg->vrt[0]->Pair)->idx,
           seg->vrt[1]->idx, (seg->vrt[1]->Pair)->idx);

  }
  */

  if (Tr->tr_segs) {
    //qDebug()<<"adding segments: "<< Tr->tr_segs->objects;
    //Tr->op_db_verbose = 4; // Debug
    assert((*pGeom)->tr_segs != NULL); // Created in detri2_init_bbox().
    //arraypool *fqueue = new arraypool(sizeof(TriEdge), 4);
    // Insert all segments into Geometry.
    for (int i = 0; i < Tr->tr_segs->objects; i++) {
      Triang *seg = (Triang *)  Tr->tr_segs->get(i);
      if (seg->is_deleted()) continue;
      if (seg->tag != 0) {
        //qDebug()<<" seg->tag="<<seg->tag<< " i="<<i;
        Vertex *e1 = seg->vrt[0]->Pair;
        Vertex *e2 = seg->vrt[1]->Pair;
        int tag = seg->tag;
        REAL val = seg->val;
        /*
        if ((e1 != NULL) && (e2 != NULL) && (e1 != e2)) {
          //detri2_insert_segment(*pGeom, e1, e2, tag, E);
          if ((*pGeom)->recover_edge(e1, e2, E, fqueue) == INTERSECT_SHARE_EDGE) {
            Triang *newseg = NULL;
            (*pGeom)->insert_segment(E, &newseg);
            newseg->tag = tag;
          } else {
            printf("!! Failed to recover segment [%d,%d] tga(%d).\n",
                   e1->idx, e2->idx, tag);
          }
          if (fqueue->objects > 0) {
            (*pGeom)->lawson_flip(NULL, 0, fqueue);
          }
        } else {
          qDebug()<<"  Warning:  skip segment i="<<i;
        }
        */
        if (!detri2_insert_segment(*pGeom, e1, e2, tag, val)) {
          printf("!! Failed to recover segment [%d,%d] tag(%d).\n", e1->idx, e2->idx, tag);
        }
      }
    }
    //delete fqueue;
    qDebug()<<"added "<<(*pGeom)->tr_segs->objects<< " segments";
  }

  if (Tr->ct_in_sdms > 0) {
    //qDebug()<<"adding "<<Tr->ct_in_sdms<< " subdomains";
    (*pGeom)->in_sdms = new Vertex[Tr->ct_in_sdms];
    int idx = 0;
    for (int i = 0; i < Tr->ct_in_sdms; i++) {
      // Make sure this vertex lies inside the geometry.
      E.tri = NULL;
      double x = Tr->in_sdms[i].crd[0];
      double y = Tr->in_sdms[i].crd[1];
      double dummy;
      loc = detri2_locate_point(*pGeom, x, y, E, dummy, 1);
      if (loc != LOC_IN_OUTSIDE) {
        (*pGeom)->in_sdms[idx].init();
        (*pGeom)->in_sdms[idx].crd[0] = x;
        (*pGeom)->in_sdms[idx].crd[1] = y;
        (*pGeom)->in_sdms[idx].tag = Tr->in_sdms[i].tag;
        (*pGeom)->in_sdms[idx].val = Tr->in_sdms[i].val;
        (*pGeom)->in_sdms[idx].adj = E; // Remember its location.
        idx++;
      }
    }
    (*pGeom)->ct_in_sdms = idx;
    qDebug()<<"added "<<(*pGeom)->ct_in_sdms<< " subdomains";
  }

  // debug
  //if ((*pGeom)->check_mesh(0,0) > 0) {
  //  qDebug() << "!!! BUG: Reconstruct mesh failed.";
  //}

  (*pGeom)->mesh_statistics();

  return 1;
}

//=============================================================================
// Get the orthognoal projection of p onto the line e1->e2.

void detri2_projpt2edge(Vertex *p, Vertex* e1, Vertex* e2, Vertex* prj)
{
  double v1[3], v2[3];
  double len, l_p;

  v1[0] = e2->crd[0] - e1->crd[0];
  v1[1] = e2->crd[1] - e1->crd[1];
  v1[2] = 0.0; // e2[2] - e1[2];
  v2[0] = p->crd[0] - e1->crd[0];
  v2[1] = p->crd[1] - e1->crd[1];
  v2[2] = 0.0; // p[2] - e1[2];

  len = sqrt(v1[0]*v1[0]+v1[1]*v1[1]+v1[2]*v1[2]); //sqrt(dot(v1, v1));
// \iffalse
  assert(len != 0.0);
// \fi
  v1[0] /= len;
  v1[1] /= len;
  v1[2] /= len;
  l_p = (v1[0]*v2[0]+v1[1]*v2[1]+v1[2]*v2[2]); //dot(v1, v2);

  prj->crd[0] = e1->crd[0] + l_p * v1[0];
  prj->crd[1] = e1->crd[1] + l_p * v1[1];
  prj->crd[2] = 0.0; // e1[2] + l_p * v1[2];
}

//=============================================================================

void detri2_get_vertex_typ(Vertex *p, char strtyp[64])
{
    if (p->typ == UNUSEDVERTEX) {
      strcpy(strtyp, "UNUSED VERTEX");
    } else if (p->typ == INFVERTEX) {
      strcpy(strtyp, "INF VERTEX");
    } else if (p->typ == FREEVERTEX) {
      strcpy(strtyp, "FREE VERTEX");
    } else if (p->typ == SEGMENTVERTEX) {
      strcpy(strtyp, "SEGMENT VERTEX");
      if (p->is_fixed()) {
        strcat(strtyp, " (fixed)");
      }
    } else if (p->typ == STEINERVERTEX) {
      strcpy(strtyp, "STEINER VERTEX");
      if (p->is_fixed()) {
        strcat(strtyp, " (fixed)");
      }
    } else if (p->typ == DEADVERTEX) {
      strcpy(strtyp, "DEAD VERTEX");
    } else {
      strcpy(strtyp, "UNKNOWN ERTEX");
      assert(0); // should be a bug
    }
}

//=============================================================================
// Locte the point (x, y) in the triangulation.
//   Return LOC_IN_OUTSIDE (0). E returns the hull triangle contains (x,y).
//   Return LOC_IN_TRI (in triangle), LOC_ON_EDGE (on edge), LOC_ON_VERT (on vertex),
//     and E returns the triangle, moreover,
//     -- E.org() is the nearest vertex to (x,y),
//     -- dist returns the distance from its neartest vertex E.org().
int detri2_locate_point(Triangulation* Tr, double x, double y, TriEdge &E,
                        double &dist, int rndflag)
{
  dist = 0.;

  Vertex pt; pt.init();
  pt.crd[0] = x;
  pt.crd[1] = y;

  //qDebug()<<" locate point from E.tri="<<E.tri;
  //qDebug()<<" tr_recnttri="<<Tr->tr_recnttri;

  //int loc = Tr->locate_point(&pt, E, 0, 0); // exact location
  int loc = Tr->locate_point(&pt, E, 0); // exact location

  //printf("loc = %d\n", loc);

  if (loc == LOC_IN_OUTSIDE) {
    assert(E.tri->is_hulltri());
    assert(E.apex() == Tr->tr_infvrt);
    // Return the nearest vertex [E.org(), E.dest()].
    double da = Tr->get_distance(&pt, E.org());
    double db = Tr->get_distance(&pt, E.dest());
    if (da < db) {
      // E.dest() is closer.
      // Search another hull edge at E.dest().
      TriEdge N = E.enext();
      N = N.esym_enext();
      assert(N.org() == E.dest());
      assert(N.tri->is_hulltri());
      assert(!N.esym().tri->is_hulltri());
      E = N;
      dist = db;
    } else {
      dist = da;
    }
    return loc; // return LOC_IN_OUTSIDE; return 0;
  }

  // The point lies inside or on edge. Check the two cases:
  //   - if it is too close to an existing vertex (Tr->io_tol_mingap).
  //   - if it is too close to an existing segment (Tr->io_tol_minangle).
  if (rndflag && (loc == LOC_IN_TRI)) {
    // Find the nearest edge by searching the max angle at vertex.
    for (int i = 0; i < 3; i++) {
      double ang = Tr->get_angle(E.org(), &pt, E.dest());
      if (fabs(ang) < Tr->io_tol_minangle) {
        //assert(ang > 0.);
        loc = LOC_ON_EDGE;
        break;
      }
      E = E.enext();
    }
  }

  if (loc == LOC_IN_TRI) {
    // get the nearest distance.
    double da = Tr->get_distance(&pt, E.org());
    double db = Tr->get_distance(&pt, E.dest());
    double dc = Tr->get_distance(&pt, E.apex());
    if (da > db) {
      if (db > dc) {
        dist = dc;
        E = E.eprev(); // E.apex();
      } else {
        dist = db;
        E = E.enext(); // E.dest();
      }
    } else if (da > dc) {
      if (db > dc) {
        dist = dc;
        E = E.eprev(); // E.apex();
      } else {
        dist = db;
        E = E.enext(); // E.dest();
      }
    } else {
      dist = da; // E.org();
    }
    // Check if it is too close to this vertex.
    if (rndflag && ((dist / Tr->io_diagonal) < Tr->io_tol_rel_gap)) {
      // assert(dist > 0.);
      //printf("rrrrrrrrrronding to on vertex\n");
      loc = LOC_ON_VERT;
    }
  } else if (loc == LOC_ON_EDGE) {
    // On edge.
    // get the nearest distance.
    double da = Tr->get_distance(&pt, E.org());
    double db = Tr->get_distance(&pt, E.dest());
    if (da > db) {
      dist = db;
      E = E.esym(); //E = E.enext(); BUGGG
    } else {
      dist = da;
    }
    // Check if it is too close to this vertex.
    if (rndflag && ((dist / Tr->io_diagonal) < Tr->io_tol_rel_gap)) { // 1.e-3
      // assert(dist > 0.);
      loc = LOC_ON_VERT;
    }
  } else if (loc == LOC_ON_VERT) {
    // On vertex.
    dist = 0;
  } else {
    assert(0); // unknown case
  }

  return loc;
}

//=============================================================================
// Insert the point (x,y) into the triangulation.
//   If E.tri != NULL, then E and loc are returned from the locate_point();
//   Otherwise, detri2_locate_point() will be called.
int detri2_insert_point(Triangulation* Tr, double x, double y, TriEdge &E,
                        int& loc, int rndflag)
{
  //printf("debug detri2: insert point\n");
  if ((E.tri==NULL) || E.tri->is_hulltri() || (loc == LOC_IN_OUTSIDE)) {
    //printf("  locating point ...\n");
    double dist = 0.;
    loc = detri2_locate_point(Tr, x, y, E, dist, rndflag);
    // These cases are not possible to insert a vertex.
    //printf("  result = %d\n", loc);
    if (loc == LOC_IN_OUTSIDE) {
      if (!Tr->tr_nonconvex) {
        // Enlarge the convex hull
        loc = LOC_IN_TRI;
      }
    }
    if ((loc == LOC_IN_OUTSIDE) ||
        (loc == LOC_ON_VERT) ||
        (loc == LOC_ENC_SEG)) {
      return 0; // Do not insert.
    }

  }

  if (Tr->tr_steiners == NULL) {
    Tr->tr_steiners = new arraypool(sizeof(Vertex), 8);
  }

  Vertex *newpt = (Vertex *) Tr->tr_steiners->alloc();
  newpt->init();

  newpt->crd[0] = x;
  newpt->crd[1] = y;
  newpt->crd[2] = x*x + y*y; // - w;
  //newpt->crd[2] = Tr->op_lambda1 * x*x + Tr->op_lambda2 * y*y;
  newpt->typ = STEINERVERTEX;
  newpt->idx = (/*Tr->ct_in_vrts + */Tr->tr_steiners->objects);

  //qDebug()<<"detri2 insert point: at ("<<x<<", "<<y<<")";
  arraypool *fqueue = new arraypool(sizeof(TriEdge), 4);
  TriEdge tt[4];
  tt[0] = E;

  /*
  if ((loc == LOC_ON_EDGE) && E.is_segment()) {
    Tr->split_segment(newpt, tt, fqueue);
    // Debug only
    printf("   hhhhhhhhhhhhh\n");
    if (Tr->check_mesh(0,0) > 0) {
      //qDebug()<<" Failed after splitting segment ["<<e1->idx<<","<<e2->idx<<"]";
      assert(0);
    }
  } else
  */
  {
    if (loc == LOC_IN_TRI) {
      int fflag = FLIP_13;
      Tr->flip(tt, &newpt, fflag, fqueue);
    } else if (loc == LOC_ON_EDGE) {
      printf("  Inserting point on edge [%d,%d]\n", tt[0].org()->idx, tt[0].dest()->idx);
      //assert(!tt[0].is_segment());
      int fflag = FLIP_24;
      Tr->flip(tt, &newpt, fflag, fqueue);
    } else if (loc == LOC_ON_VERT) {
      // Do not insert this point.
      assert(0);
    } else {
      assert(0); // No possible.
    }
  }

  Tr->lawson_flip(newpt, 1, fqueue);
  delete fqueue;

  E = newpt->adj; // Return the new point by E.org();
  return 1;
}

//=============================================================================
// Locate the neartest edge to the point (x,y).
//   It should always find an edge (Even (x,y) lies outside).
//   If loc != LOC_IN_OUTSIDE, the returned triangle E must contain (x,y), moreover,
//   If loc == LOC_IN_TRI, ang returns the value at vertex E.org(), E.dest(), and pt(x,y).
//     It is used to check if we should split the edge (segment) by (x,y).

int detri2_locate_edge(Triangulation* Tr, double x, double y, detri2::TriEdge &E)
{
    double dist = 0.;

    int loc = detri2_locate_point(Tr, x, y, E, dist, 1); // rndflag = 1

    //if ((loc == LOC_IN_OUTSIDE) ||
    //    (loc == LOC_ON_EDGE) ||
    //    (loc == LOC_ON_VERT)) {
    //  return loc;
    //}

    if (loc == LOC_IN_TRI) { // in triangle.
      // Find the nearest edge by calculating barycenter coordinates.
      Vertex pt; pt.init();
      pt.crd[0] = x;
      pt.crd[1] = y;
      double da = Tr->get_tri_area(&pt, E.org(), E.dest());
      double db = Tr->get_tri_area(&pt, E.dest(), E.apex());
      double dc = Tr->get_tri_area(&pt, E.apex(), E.org());
      if (da > db) {
        if (db > dc) {
          dist = dc;
          E = E.eprev(); // E.apex(), E.org()
        } else {
          dist = db;
          E = E.enext(); // E.dest(), E.apex()
        }
      } else if (da > dc) {
        if (db > dc) {
          dist = dc;
          E = E.eprev(); // E.apex(), E.org();
        } else {
          dist = db;
          E = E.enext(); // E.dest(), E.apex();
        }
      } else {
        dist = da; // E.org(), E.dest();
      }
    } else if (loc == LOC_ON_EDGE) { // on edge
      dist = 0;
    } else if (loc == LOC_ON_VERT) { // on vertex
      dist = 0; // Return any arbitrary edge at this vertex.
    }

    return loc;
}

//=============================================================================
/*
int detri2_get_segment(TriEdge &E, Triang** pseg)
{
    *pseg = NULL;

    if (!E.is_segment()) {
      return 0;
    }

    Vertex *e1 = E.org();
    if ((e1->typ != RIDGEVERTEX) && !e1->is_fixed()) {
      *pseg = e1->on_bd.tri;
      return 1;
    }
    Vertex *e2 = E.dest();
    if ((e2->typ != RIDGEVERTEX) && !e2->is_fixed()) {
      *pseg = e2->on_bd.tri;
      return 1;
    }

    // Both endpoints must be RIDGEVERTEX
    // Search the segment from the vertex-to-segment map.
    TriEdge N = e1->on_bd;
    do {
      assert(N.tri->vrt[N.ver] == e1);
      if (N.tri->vrt[1 - N.ver] == e2) {
        *pseg = N.tri; // break; // Found
        return 1;
      }
      N = N.tri->nei[N.ver]; // Get the next segment.
    } while (N.tri != NULL);

    return 0; // Not found.
}
*/
//=============================================================================
/*
// Check if the edge [e1,e2] is intersected by any segment or vertex.
// The assumption is that both e1 and e2 must be vertices of this triangulation.
// The return value is one of the following:
//   - INTERSECT_NONE, it is not intersected by any segment.
//   - INTERSECT_SHARE_EDGE, it is coincident with an edge, returned by [S.org(), S.dest()].
//   - INTERSECT_VERTEX, is intersected with a vertex, returned by S.org().
int detri2_detect_intersection(Triangulation* Tr, Vertex *e1, Vertex *e2, TriEdge &S)
{
  // Perform a line walk from e1 to e2.
  TriEdge E = e1->adj;
  int loc = Tr->find_direction(E, e2); // rndflag = 1 check Tr->op_tol_minangle

  assert((loc == 1) || (loc == 2));
  // loc == 1, means that [e1,e2] crosses the edge [E.org(), E.dest()].
  // loc == 2, means that e1, E.dest(), e2 are collinear.
  //           It may be E.dest == e2, if so, the [e1, e2] = [E.org(), E.dest()].
  if (loc == 1) {
    assert(E.apex() == e1);
    // Detect whether if there is a nearly collinear endpoint.
    double ang = Tr->get_angle(e1, e2, E.org());
    if (ang < Tr->io_tol_minangle) {
      // Round to collinear with E.org();
      qDebug()<<" rounding find_direction: collinear to E.org(), ang="<<ang/PI*180<<" degree";
      loc = 2;
    } else {
      ang = Tr->get_angle(e1, e2, E.dest());
      if (ang < Tr->io_tol_minangle) {
        // Round to collinear with E.dest();
        Vertex *dbv = E.dest(); // debug only
        qDebug()<<" rounding find_direction: collinear to E.dest() ang="<<ang/PI*180<<" degree";
        E = (E.enext_esym()).enext();
        assert(E.apex() == e1);
        assert(E.org() == dbv);
        loc = 2;
      }
    }
  }

  if (loc == 2) {
    assert(E.org() == e1);
    if (E.dest() == e2) {
      // The edge exists.
      S = E;
      if (E.is_segment()) {
        return INTERSECT_CONFLICT; // A segment already exists
      } else {
        return INTERSECT_SHARE_EDGE;
      }
    } else {
      // Found a vertex E.dest() intersect this edge.
      S = E.enext();
      return INTERSECT_VERTEX;
    }
  } else if (loc == 1) {
    // crossing an edge.
    if (E.is_segment()) {
      S = E;
      return INTERSECT_SEGMENT;
    }
  } else {
    assert(0); // not possible.
  }

  // Continue the search for the next intersected edge of [e1, e2].
  while (true) {
    E = E.esym();
    // E is [a,b,c], where e2 lies in the halfspace of [a,b] which contains c.
    Vertex *pc = E.apex();
    if (pc == e2) {
      // End of the walk. No intersection is found.
      break; //return INTERSECT_NONE;
    }

    double s3 = Orient2d(e1, e2, pc);
    if (s3 != 0) {
      double ang = Tr->get_angle(pc, e1, e2);
      if (ang < Tr->io_tol_minangle) {
        qDebug()<<"  ang="<<ang/detri2::PI*180<<" degree";
        s3 = 0; // Round to collinear case.
      }
    }

    if (s3 > 0) {
      E = E.enext(); // intersect with [b,c]
    } else if (s3 < 0) {
      E = E.eprev(); // intersect with [c,a]
    } else { // if (s3 == 0) {
      // collinear with pc.
      E = E.enext(); // [c,a]
      return INTERSECT_VERTEX;
    }

    if (E.is_segment()) {
      S = E;
      return INTERSECT_SEGMENT;
    }
  } // while (true)

  return INTERSECT_NONE;
}
*/
//=============================================================================
// Insert a segment e1->e2 into the triangulation (geometry).

int detri2_insert_segment(Triangulation* Tr, Vertex *e1, Vertex *e2, int tag, REAL val)
{
  if ((e1 != NULL) && (e2 != NULL) && (e1 != e2)) {
    if (Tr->tr_segs == NULL) {
      Tr->tr_segs = new arraypool(sizeof(Triang), 8);
    }
    arraypool *fqueue = new arraypool(sizeof(TriEdge), 4);
    int succeed = 1;
    TriEdge E;
    if (Tr->recover_edge(e1, e2, E, fqueue) == INTERSECT_SHARE_EDGE) {
      if (!E.is_segment()) {
        Tr->insert_segment(E, tag, val, NULL);
      } else {        
        succeed = 1; // A segment already exists.
      }
    } else {
      succeed = 0;
    }
    if (fqueue && (fqueue->objects > 0)) {
      Tr->lawson_flip(NULL, 0, fqueue);
    }
    delete fqueue;
    return succeed;
  } else {
    return 0;
  }
}

/*
// Insert a segment e1->e2 into the triangulation (geometry).
// Note: This function assume that this segment does not intersect any existing segment.
//   and both endpoints of this segment exist.
//   If it is inserted, [E.org,E.dest] = [e1,e2].
int detri2_insert_segment(Triangulation* Tr, Vertex *e1, Vertex *e2, int tag, TriEdge &E)
{
  Triang *seg = (Triang *) Tr->tr_segs->alloc();
  seg->init();
  seg->vrt[0] = e1;
  seg->vrt[1] = e2;
  seg->vrt[2] = NULL;
  seg->tag = tag; // must be non-zero.

  qDebug()<< "  detri2 insert segment ["<<e1->idx<<", "<<e2->idx<<")";

  arraypool *fqueue = new arraypool(sizeof(TriEdge), 4); // for CDT
  //Vertex *e1 = seg->vrt[0];
  //Vertex *e2 = seg->vrt[1];
  //seg->vrt[2] = NULL; // it is a segment

  E.tri = NULL; E.ver = 0; int insflag = 0;
  if (!Tr->recover_edge(e1, e2, E, insflag, fqueue)) {
    qDebug()<< "  failed insflag = "<<insflag;
    assert(0); // do detect_intersect
    delete fqueue;
    return 0;
  }

  qDebug()<< "  fqueue size = "<<fqueue->objects;

  if (!E.is_segment()) {
    E.set_segment();
    E.esym().set_segment();
    Tr->ct_segments++;
  } else {
    assert(0); // already exists?
    delete fqueue;
    return 0;
  }

  if (e1->typ == STEINERVERTEX) {
    e1->set_fix(); // it cannot be simply deleted.
  } else {
    // e1 is an input vertex
    e1->typ = RIDGEVERTEX;
  }

  if (e2->typ == STEINERVERTEX) {
    e2->set_fix(); // it cannot be simply deleted.
  } else {
    // e2 is an input vertex
    e2->typ = RIDGEVERTEX;
  }

  // Link the segment into the ring of its endpoints.
  TriEdge nextseg = seg->vrt[0]->on_bd;
  e1->on_bd = TriEdge(seg, 0);
  e1->on_bd.tri->nei[0] = nextseg;

  nextseg = seg->vrt[1]->on_bd;
  e2->on_bd = TriEdge(seg, 1);
  e2->on_bd.tri->nei[1] = nextseg;

  if (fqueue->objects > 0) {
    Tr->lawson_flip(NULL, 0, fqueue);
  }

  delete fqueue;
  return 1;
}
*/

/*
// Insert segments from e1 to e2, may be intersect with existing segments.
int detri2_insert_segment_intersect(Triangulation* Tr, Vertex *e1, Vertex *e2, int tag, TriEdge &S)
{
  Vertex *startpt = e1;

  while (startpt != NULL) {
    qDebug()<<"Insert segment ["<<startpt->idx<<","<<e2->idx<<"]";

    int chk = detri2_detect_intersection(Tr, startpt, e2, S);
    qDebug() << "  detect_intersection result = "<<chk;

    if ((chk == INTERSECT_NONE) || (chk == INTERSECT_SHARE_EDGE)) {
      // The edge of this segment exists.
      detri2_insert_segment(Tr, startpt, e2, tag, S);
      return 1;
    }

    if (chk == INTERSECT_CONFLICT) {
      return 0;
    }

    if (chk == INTERSECT_VERTEX) {
      qDebug()<<"  - Cutting by vertex "<<S.org()->idx;
      Vertex *cutpt = S.org();
      detri2_insert_segment(Tr, startpt, cutpt, tag, S);
      startpt = cutpt;
      continue;
    }

    if (chk == INTERSECT_SEGMENT) {
      // S must be an existing segment of Tr, [e1,e2] intersects with S.
      //   (1) Calculate the intersection point.
      //   (2) Remove the segment S from the triangulation (Tr).
      //   (3) Insert the intersection point into Tr.
      //   (4) Insert four line segments at this intersection point to Tr.
      assert(S.is_segment());
      // (1) Calculate the intersecting point.
      REAL xx, yy;
      Vertex *v1 = S.org();
      Vertex *v2 = S.dest();
      double X0 = v1->crd[0];
      double Y0 = v1->crd[1];
      double X1 = v2->crd[0];
      double Y1 = v2->crd[1];
      double X2 = e1->crd[0];
      double Y2 = e1->crd[1];
      double X3 = e2->crd[0];
      double Y3 = e2->crd[1];
      double t1, t2;
      if (Tr->line_line_intersection(X0, Y0, X1, Y1, X2, Y2, X3, Y3, &t1, &t2)) {
        xx = X0 + t1 * (X1 - X0);
        yy = Y0 + t1 * (Y1 - Y0);
      } else {
        printf("!! Warning: Failed at calculating line-line intersection.\n");
        return 0;
      }
      qDebug()<<"  - Intersection point ("<<xx<<","<<yy<<")";
      // (2) Remove the segment S from Tr.
      Triang *seg = NULL;
      detri2_get_segment(S, &seg);
      assert(seg != NULL);
      int stag1 = seg->tag;
      // see ScribbleArea::edit_delete_act()
      //Vertex *v1 = seg->tri->vrt[0];
      //Vertex *v2 = seg->tri->vrt[1];
      detri2_remove_segment_at_vertex(v1, seg);
      detri2_remove_segment_at_vertex(v2, seg);
      // remove segment from Tr.
      S.clear_segment();
      TriEdge N = S.esym();
      N.clear_segment();
      seg->set_deleted();
      Tr->tr_segs->dealloc(seg);
      Tr->ct_segments--;
      qDebug()<<"  - Removed segment ("<<v1->idx<<","<<v2->idx<<") from Geom";
      // (3) Insert a new vertex on edge S.
      int loc = LOC_ON_EDGE;
      if (!detri2_insert_point(Tr, xx, yy, S, loc)) {
        printf("  Failed to insert the new point.\n");
        assert(0); // Debug
      }
      // The new point is returned by S.org().
      Vertex *newpt = S.org();
      qDebug()<<"  - New point: idx="<<newpt->idx<<" ("<<newpt->crd[0]<<","<<newpt->crd[1]<<")";
      // (4) Insert three new segments [v1, newpt], [v2, newpt], [e1, newpt]
      detri2_insert_segment(Tr, v1, newpt, stag1, S);
      detri2_insert_segment(Tr, v2, newpt, stag1, S);
      detri2_insert_segment(Tr, startpt, newpt, tag, S);
      startpt = newpt;
    } // if (chk == INTERSECT_SEGMENT)
  } // while (startpt != NULL)

  return 0; // Should not be here
}
*/

//=============================================================================
/*
// Removing a segment at vertex, update the vertex-to-segment ring.
//   'v' must be an endpoint of a segment => v->typ is either RIDGEVERTEX or v->is_fixed().
int detri2_remove_segment_at_vertex(Vertex *v, Triang *seg)
{
    TriEdge nextseg;
    if (seg->vrt[0] == v) {
      nextseg = seg->nei[0];
    } else if (seg->vrt[1] == v) {
      nextseg = seg->nei[1];
    } else {
      assert(0); // seg does not contain v.
    }
    if (nextseg.tri != NULL) {
      assert(nextseg.tri->vrt[nextseg.ver] == v);
    }

    TriEdge prevseg = v->on_bd;
    assert(prevseg.tri->vrt[prevseg.ver] == v);
    if (prevseg.tri == seg) {
      // remove the seg by restting the vertex-to-seg map.
      v->on_bd = nextseg;
      if (nextseg.tri == NULL) {
        // There is no segment at this vertex.
        // v is not a restricted vertex anymore.
        if (v->typ == RIDGEVERTEX) {
          v->typ = FACETVERTEX; // a free input vertex.
        } else if (v->typ == STEINERVERTEX) {
          v->clear_fix();
        } else {
          assert(0); // should not be possible.
        }
      }
      return 1;
    }

    // Search the seg in the vertex-to-segment map.
    TriEdge searchseg = prevseg.tri->nei[prevseg.ver];
    while (searchseg.tri != NULL) {
      assert(searchseg.tri->vrt[searchseg.ver] == v);
      if (searchseg.tri == seg) {
        // Found. Remove seg from the list.
        prevseg.tri->nei[prevseg.ver] = nextseg;
        return 1;
      }
      // Go the next of the next segment.
      TriEdge tmp = searchseg.tri->nei[searchseg.ver];
      searchseg = tmp;
    }

    return 0; // not found.
}
*/
//=============================================================================
// Move the vertex (*ppt) of the triangulation to the new position (x,y).
int detri2_move_vertex(Triangulation* Tr, Vertex **ppt, double x, double y)
{
  Vertex *mv = *(ppt);

  qDebug() << "detri2 move vertex idx="<< mv->idx
    <<", from ("<<mv->crd[0]<<","<<mv->crd[1]<<")"<< " to ("<<x <<","<< y<<")";

  // Check if this location has already a vertex.
  TriEdge Eloc; Eloc.tri = NULL;
  double dist = 0.;
  int loc = detri2_locate_point(Tr, x, y, Eloc, dist, 1); // rndflag = 1
  // Do not move it to outside the bbox.
  if (loc == LOC_IN_OUTSIDE) return 0;
  if (Eloc.org() != mv) {
    // The nearest vertex is not the moving vertex.
    if ((loc == LOC_ON_VERT) ||
        ((dist / Tr->io_diagonal) < (Tr->io_tol_rel_gap * 10.0))) {
      qDebug()<<"detri2 quit the move due to a close vertex.";
      return 0;
    }
  }

  //if (((*ppt)->typ == STEINERVERTEX) && (!(*ppt)->is_fixed()) &&
  //     ((*ppt)->on_bd.tri != NULL)) {
  //  // A Stiner point on a segment. Can be moved along this segment.
  //  // Not implemented yet.
  //  return 0;
  //}

  // First check if all triangles at the new position (x,y) are valid.
  bool bflag = true;
  Vertex testpt;
  testpt.crd[0] = x;
  testpt.crd[1] = y;
  TriEdge E = mv->adj;
  do {
    if (!E.tri->is_hulltri()) {
      if (Orient2d(E.dest(), E.apex(), &testpt) <= 0.) {
        bflag = false; break;
      }
    }
    E = E.eprev_esym(); // ccw
  } while (E.tri != mv->adj.tri);

  if (bflag) {
    qDebug()<<"  Directly move the vertex to (x,y)";
    mv->crd[0] = x;
    mv->crd[1] = y;
    mv->crd[2] = x*x + y*y;// - mv->wei;
    // Do flips to recover Delaunayness.
    arraypool *fqueue = new arraypool(sizeof(TriEdge), 4);
    E = mv->adj;
    do {
      if (!E.tri->is_hulltri()) {
        E.set_edge_infect();
      * (TriEdge *) fqueue->alloc() = E;
      }
      E = E.eprev_esym(); // ccw
    } while (E.tri != mv->adj.tri);
    Tr->lawson_flip(NULL, 0, fqueue);
    delete fqueue;
    // The type of this vertex might change.
    if (Tr->is_ridge_vertex(mv)) {
      mv->set_fix();
    } else {
      mv->clear_fix();
    }
    return 1;
  }

  // We only do this if this vertex is totally free, i.e., no segment contain it.
  if ((*ppt)->is_fixed()) {
    return 0;
  }

  qDebug()<<"  detri2 moves the vertex by insertion and deletion";

  arraypool *fqueue = new arraypool(sizeof(TriEdge), 4);

  // Insert a new vertex at (x,y).
  Vertex *newpt = (Vertex *) Tr->tr_steiners->alloc();
  newpt->init();
  newpt->crd[0] = x;
  newpt->crd[1] = y;
  newpt->crd[2] = x*x + y*y; // - mv->wei;
  //newpt->crd[2] = Tr->op_lambda1 * x*x + Tr->op_lambda2 * y*y;
  newpt->idx = mv->idx;
  newpt->Pair = mv->Pair;
  newpt->wei = mv->wei; // the weight
  newpt->val = mv->val; // mesh size
  newpt->typ = STEINERVERTEX;

  TriEdge tt[4];
  tt[0] = Eloc;
  if (loc == LOC_IN_TRI) {
    int fflag = FLIP_13;
    Tr->flip(tt, &newpt, fflag, fqueue);
  } else if (loc == LOC_ON_EDGE) {
    int fflag = FLIP_24;
    Tr->flip(tt, &newpt, fflag, fqueue);
  } else {
    assert(0); // Not possible.
  }
  Tr->lawson_flip(newpt, 0, fqueue); // hullflag = 0

  if (Tr->is_ridge_vertex(newpt)) {
    newpt->set_fix();
  }

  // Remove this vertex from the triangulation.
  // assert(!mv->is_fixed()); // already checked in above
  if (!Tr->remove_point(mv, fqueue)) {
    assert(0); // This should always succeed.
    delete fqueue;
    return 0;
  }
  assert(mv->is_deleted());
  Tr->lawson_flip(newpt, 0, fqueue); // hullflag = 0

  delete fqueue;

  *ppt = newpt;
  return 1;
}

//==============================================================================

int detri2_get_trilist(Triang *tri, arraypool *tri_list)
{
  assert(tri_list->objects == 0);
  assert(!tri->is_hulltri() && !tri->is_infected());

  tri->set_infect();
  Triang** ptri = (Triang **) tri_list->alloc();
  *ptri = tri;

  TriEdge E, N;
  for (int i = 0; i < tri_list->objects; i++) {
    E.tri = * (Triang **) tri_list->get(i);
    for (E.ver = 0; E.ver  < 3; E.ver++) {
      if (!E.is_segment()) {
        N = E.esym();
        if (!N.tri->is_hulltri() && !N.tri->is_infected()) {
          N.tri->set_infect();
          ptri = (Triang **) tri_list->alloc();
          *ptri = N.tri;
        }
      }
    }
  }
  // Uninfect all collected triangles.
  for (int i = 0; i < tri_list->objects; i++) {
    E.tri = * (Triang **) tri_list->get(i);
    E.tri->clear_infect();
  }

  qDebug()<< "  Collected "<<tri_list->objects<<" in subdomain";
  return 1;
}

int detri2_is_exterior_trilist(arraypool *tri_list)
{
  for (int i = 0; i < tri_list->objects; i++) {
    Triang *tri = * (Triang **) tri_list->get(i);
    for (int j = 0; j < 3; j++) {
      if (tri->vrt[j]->typ != STEINERVERTEX) {
        return 1; // It is exterior
      }
    }
  }
  return 0;
}

int detri2_is_tri_in_trilist(arraypool *tri_list, Triang* testtri)
{
  for (int i = 0; i < tri_list->objects; i++) {
    Triang *tri = * (Triang **) tri_list->get(i);
    if (tri == testtri) return 1;
  }
  return 0;
}

int detri2_get_subdomain(Triangulation* Tr, arraypool *tri_list, Vertex **sdm_vertex)
{
  if (Tr->ct_in_sdms == 0) {
    *sdm_vertex = NULL;
    return 0;
  }

  TriEdge E;
  int loc = 0, i;

  for (i = 0; i < tri_list->objects; i++) {
    E.tri = * (Triang **) tri_list->get(i);
    E.tri->set_infect();
  }

  for (i = 0; i < Tr->ct_in_sdms; i++) {
    E = Tr->in_sdms[i].adj;
    //Tr->locate_point(&(Tr->in_sdms[i]), E, 0, 0);
    Tr->locate_point(&(Tr->in_sdms[i]), E, 0);
    if (E.tri->is_infected()) {
      // Found the domain.
      Tr->in_sdms[i].adj = E;
      *sdm_vertex = &(Tr->in_sdms[i]);
      loc = 1;
      break;
    }
  }

  for (i = 0; i < tri_list->objects; i++) {
    E.tri = * (Triang **) tri_list->get(i);
    E.tri->clear_infect();
  }

  return loc;
}

int detri2_add_subdomain(Triangulation** pGeom, arraypool *tri_list, int tag, double maxarea)
{
  int old_ct_sdm = (*pGeom)->ct_in_sdms;
  Vertex *newsdms = new Vertex[old_ct_sdm + 1];
  if (old_ct_sdm > 0) {
    for (int i = 0; i < old_ct_sdm; i++) {
      Vertex *ov = &((*pGeom)->in_sdms[i]);
      Vertex *v = &(newsdms[i]);
      v->init();
      v->crd[0] = ov->crd[0];
      v->crd[1] = ov->crd[1];
      v->tag    = ov->tag;
      v->val    = ov->val;
    }
    delete [] (*pGeom)->in_sdms;
  }
  (*pGeom)->ct_in_sdms = old_ct_sdm + 1;
  (*pGeom)->in_sdms = newsdms;
  Vertex *v = &((*pGeom)->in_sdms[old_ct_sdm]); // The last one.
  // Randomly choose a triangle from trilist.
  int rndidx = (rand() % (tri_list->objects));
  Triang *tri = * (Triang **) tri_list->get(rndidx);
  v->crd[0] = (tri->vrt[0]->crd[0]+tri->vrt[1]->crd[0]+tri->vrt[2]->crd[0]) / 3.0;
  v->crd[1] = (tri->vrt[0]->crd[1]+tri->vrt[1]->crd[1]+tri->vrt[2]->crd[1]) / 3.0;
  v->tag = tag;
  v->val = maxarea;

  return 1;
}

//==============================================================================
// Enlarge the bounding box to include the point (x,y). It does the following:
//   - delete the current triangulation in Tr,
//   - enlarge the current bbox to include the point (x,y),
//   - create the inital triangulation of the bbox,
//   - insert all Steiner points of Tr.
int detri2_enlarge_bbox(Triangulation** pTr, double x, double y)
{
    Triangulation *Tr = *pTr;
    double xmin = x < Tr->io_xmin ? x : Tr->io_xmin;
    double xmax = x > Tr->io_xmax ? x : Tr->io_xmax;
    double ymin = y < Tr->io_ymin ? y : Tr->io_ymin;
    double ymax = y > Tr->io_ymax ? y : Tr->io_ymax;

    return detri2_enlarge_bbox(pTr, xmin, xmax, ymin, ymax);
}

int detri2_enlarge_bbox(Triangulation** pTr,
                        double xmin, double xmax, double ymin, double ymax)
{
  Triangulation *Tr = *pTr;
  /*
  double xmin = x < Tr->io_xmin ? x : Tr->io_xmin;
  double xmax = x > Tr->io_xmax ? x : Tr->io_xmax;
  double ymin = y < Tr->io_ymin ? y : Tr->io_ymin;
  double ymax = y > Tr->io_ymax ? y : Tr->io_ymax;
  */

  // Englarge the bounding box by 10%.
  double dmax = xmax - xmin;
  if (dmax < ymax - ymin) dmax = ymax - ymin;
  double dd = dmax * 0.1;
  xmin -= dd;
  ymin -= dd;
  xmax += dd;
  ymax += dd;

  // Backup all Steiner points.
  int ns = Tr->tr_steiners->objects;
  Vertex *pp = new Vertex[ns];
  int idx = 0;
  for (int i = 0; i < Tr->tr_steiners->used_items; i++) {
    Vertex *v = (Vertex *) Tr->tr_steiners->get(i);
    if (v->is_deleted()) continue;
    pp[idx].init();
    pp[idx].crd[0] = v->crd[0];
    pp[idx].crd[1] = v->crd[1];
    pp[idx].crd[2] = v->crd[2]; // contains weight
    pp[idx].Pair = v->Pair;
    v->Pair = &(pp[idx]); // for transfering segments.
    pp[idx].idx = v->idx;
    pp[idx].typ = v->typ;
    pp[idx].wei = v->wei; // weight
    idx++;
  }

  // Bakup all segments.
  int gs = Tr->tr_segs->objects;
  Triang *baksegs = new Triang[gs+1]; // gs might be 0.
  idx = 0;
  for (int i = 0; i < Tr->tr_segs->used_items; i++) {
    Triang *seg = (Triang *) Tr->tr_segs->get(i);
    if (seg->is_deleted()) continue;
    baksegs[idx].init();
    baksegs[idx].vrt[0] = seg->vrt[0]->Pair;
    baksegs[idx].vrt[1] = seg->vrt[1]->Pair;
    baksegs[idx].tag    = seg->tag;
    idx++;
  }

  // Create an empty Geometry with an initial triangulation.
  detri2_init_with_bbox(pTr, xmin, xmax, ymin, ymax);

  // Insert all Steiner points.
  qDebug()<<"Inserting "<<ns<<" Steiner points.";
  TriEdge E; int loc = 0;
  for (int i = 0; i < ns; i++) {
    //qDebug()<<"i="<<i;
    E.tri = NULL;
    if (detri2_insert_point(*pTr, pp[i].crd[0], pp[i].crd[1], E, loc)) { // weight=0
      Vertex *newpt = E.org();
      newpt->crd[2] = pp[i].crd[2];
      newpt->Pair = pp[i].Pair;
      pp[i].Pair = newpt; // for transfering segments.
      newpt->idx = pp[i].idx;
      newpt->typ = pp[i].typ;
      newpt->wei = pp[i].wei; // weight
    } else {
      qDebug()<<"Warning: a vertex idx="<<pp[i].idx<<" is lost.";
      assert(0); // a bug.
    }
  }

  // Insert all segments.
  //arraypool *fqueue = new arraypool(sizeof(TriEdge), 4);

  for (int i = 0; i < gs; i++) {
    Triang *bakseg = &(baksegs[i]);
    Vertex *e1 = bakseg->vrt[0]->Pair;
    Vertex *e2 = bakseg->vrt[1]->Pair;
    int tag = bakseg->tag;
    REAL val = bakseg->val;
    /*
    //detri2_insert_segment(*pTr, e1, e2, tag, E);
    if ((e1 != NULL) && (e2 != NULL) && (e1 != e2)) {
      if ((*pTr)->recover_edge(e1, e2, E, fqueue) == INTERSECT_SHARE_EDGE) {
        Triang *newseg = NULL;
        (*pTr)->insert_segment(E, &newseg);
        newseg->tag = tag;
      } else {
        printf("!! Failed to recover segment [%d,%d] tga(%d).\n",
               e1->idx, e2->idx, tag);
      }
      if (fqueue->objects > 0) {
        (*pTr)->lawson_flip(NULL, 0, fqueue);
      }
    }
    */
    if (!detri2_insert_segment(*pTr, e1, e2, tag, val)) {
      printf("!! Failed to recover segment [%d,%d] tga(%d).\n", e1->idx, e2->idx, tag);
    }
  }

  //delete fqueue;

  // Debug only
  if ((*pTr)->check_mesh(0,0) > 0) {
    qDebug()<<"!!! Failed after enlarge bbox";
    assert(0);
  }

  delete [] pp;
  delete [] baksegs;
  return 1;
}

//=============================================================================
#include "3rdparty/rbox.h"
//#include <sstream>   // std::stringstream

int detri2_rbox(Triangulation **pTr, int dim, int numpoints)
{
  qDebug()<<"detri2 rbox: dim="<<dim<<", number of points="<<numpoints;

  // Create the command line.
  int argc = 3;
  char *argv[3];
  char str1[100], str2[100], str3[100];
  strcpy (str1, "rbox");
  sprintf(str2, "D%d", dim);
  sprintf(str3, "%d", numpoints);
  argv[0] = str1;
  argv[1] = str2;
  argv[2] = str3;

  qDebug()<<" rbox commands: "<<str2<<" "<<str3;

  return detri2_rbox(pTr, argc, argv);
}

int detri2_rbox(Triangulation **pTr, int argc, char *argv[])
{
  int dim = 0;
  int ptnum = 0;
  double *ptlist = NULL;

  if (!rbox_main(argc, argv, &dim, &ptnum, &ptlist)) {
    // Something is wrong!
    if (ptlist != NULL) {
      delete [] ptlist;
    }
    return 0;
  }

  if (dim != 2) {
    // Not handle 3d case.
    delete [] ptlist;
    return 0;
  }

  /*
  fprintf(stderr, "generated points...\n");

  for (int i = 0; i < ptnum; i++) {
    fprintf(stderr, "%g %g", ptlist[i*dim], ptlist[i*dim+1]);
    if (dim == 3) {
      fprintf(stderr, "%g", ptlist[i*dim+2]);
    }
    fprintf(stderr, "\n");
  }
  */

  detri2_add_geometry(pTr, ptnum, ptlist, 0, NULL, NULL);

  delete [] ptlist;
  return 1;
}

//==============================================================================
// Add a PSLG into the triangulaiton of an existing PSLG

int detri2_add_geometry(Triangulation** pGeom, int np, double *ptlist,
                        int ne, int *elist, int *etaglist)
{
  qDebug()<<" add geometry "<< np << " vertices"<< ne << "segments";

  // First get the bbox of the set of added points.
  double xmin = +1.e30;
  double xmax = -1.e30;
  double ymin = +1.e30;
  double ymax = -1.e30;

  for (int i = 0; i < np; i++) {
    double x = ptlist[i*2];
    double y = ptlist[i*2+1];
    xmin = (x < xmin) ? x : xmin;
    xmax = (x > xmax) ? x : xmax;
    ymin = (y < ymin) ? y : ymin;
    ymax = (y > ymax) ? y : ymax;
  }

  if (*pGeom == NULL) {
    qDebug() << "  Initialize a new geometry";
    // Create an empty Geometry.
    *pGeom = new Triangulation();
    /*
    double xmin = +1.e30;
    double xmax = -1.e30;
    double ymin = +1.e30;
    double ymax = -1.e30;
    if (np > 0) {
      for (int i = 0; i < np; i++) {
        double x = ptlist[i*2];
        double y = ptlist[i*2+1];
        xmin = (x < xmin) ? x : xmin;
        xmax = (x > xmax) ? x : xmax;
        ymin = (y < ymin) ? y : ymin;
        ymax = (y > ymax) ? y : ymax;
      }
    } else {
      xmin = -1.0;
      xmax =  1.0;
      ymin = -1.0;
      ymax =  1.0;
    }
    */

    // Englarge the bounding box by 10%.
    double dmax = xmax - xmin;
    if (dmax < ymax - ymin) dmax = ymax - ymin;
    double dd = dmax * 0.1;
    xmin -= dd;
    ymin -= dd;
    xmax += dd;
    ymax += dd;

    detri2_init_with_bbox(pGeom, xmin, xmax, ymin, ymax);
  } else {
    // To make sure the current bbox is larger enough, otherwise, we enlarge it.
    if (((*pGeom)->io_xmin > xmin) ||
        ((*pGeom)->io_xmax < xmax) ||
        ((*pGeom)->io_ymin > ymin) ||
        ((*pGeom)->io_ymax < ymax)) {
      // We need enlarge the bbox.
      xmin = xmin < (*pGeom)->io_xmin ? xmin : (*pGeom)->io_xmin;
      xmax = xmax > (*pGeom)->io_xmax ? xmax : (*pGeom)->io_xmax;
      ymin = ymin < (*pGeom)->io_ymin ? ymin : (*pGeom)->io_ymin;
      ymax = ymax > (*pGeom)->io_ymax ? ymax : (*pGeom)->io_ymax;

      detri2_enlarge_bbox(pGeom, xmin, xmax, ymin, ymax);
    }
  }

  // Debug, check if the triangulation is valid before adding new Geom.
  if ((*pGeom)->check_mesh(0,0) > 0) {
    printf("!!! Triangulation is invalid before adding Geom\n");
    assert(0);
  } else {
    printf("Triangulation is valid before adding new Geom.\n");
  }

  TriEdge E; double dist = 0.;
  Vertex **vlist = new Vertex*[np+1]; // Avoid the case np = 0

  for (int i = 0; i < np; i++) {
    //qDebug()<<" vertex i="<<i;
    E.tri = NULL; // do detri2_locate_point();
    double x = ptlist[i*2];
    double y = ptlist[i*2+1];
    //double w = 0.0; // weight
    int loc = detri2_locate_point(*pGeom, x, y, E, dist, 1);
    assert(loc != LOC_IN_OUTSIDE);
    /*
    if (loc == LOC_IN_OUTSIDE) {
      detri2_enlarge_bbox(pGeom, x, y);
      E.tri = NULL; // Re-locate point.
      loc = detri2_locate_point(*pGeom, x, y, E, dist, 1);
    }
    */
    if (loc != LOC_ON_VERT) {
      detri2_insert_point(*pGeom, x, y, E, loc);
    }
    Vertex *newpt = E.org();
    vlist[i] = newpt;
  }

  // Insert segments.
  if (ne > 0) {
    if ((*pGeom)->tr_segs == NULL) {
      (*pGeom)->tr_segs = new arraypool(sizeof(Triang), 8);
    }
    for (int i = 0; i < ne; i++) {
      qDebug()<<"seg i="<<i << " v1="<<elist[i*2]<<", v2="<<elist[i*2+1];
      Vertex *e1 = vlist[elist[i*2]];
      Vertex *e2 = vlist[elist[i*2+1]];
      int tag = etaglist[i];
      if ((e1 != NULL) && (e2 != NULL)) {
        //detri2_insert_segment_intersect(*pGeom, e1, e2, tag, E);
        if (!(*pGeom)->insert_segment_intersect(e1, e2, tag, 0)) {
          printf("!! Failed to recover segment [%d,%d] tag(%d).\n",
                 e1->idx, e2->idx, tag);
        }
      }
    }
  }

  delete [] vlist;

  // Debug, check if the triangulation is valid before adding new Geom.
  if ((*pGeom)->check_mesh(0,0) > 0) {
    printf("!!! Triangulation is invalid after adding Geom\n");
    assert(0);
  } else {
    printf("Triangulation is valid after adding new Geom.\n");
  }

  // Debug special triangle
  //TriEdge tE = (*pGeom)->db_tri(-13,-14,7);
  //if (tE.tri == NULL) {
  //  printf("Not found\n");
  //}

  return 1;
}

// (xx,yy) is the center point of this rectangle.
int detri2_add_rectangle(Triangulation** pGeom,
                         double xx, double yy, double width, double height, int etag,
                         bool transflag,
                         double rcx, double rcy, double angle, double tx, double ty)
{
  if ((width <= 0) || (height <= 0)) {
    return 0;
  }
  double dx = width / 2.;
  double dy = height / 2.;

  double ptlist[8];
  ptlist[0] = xx - dx;
  ptlist[1] = yy - dy;
  ptlist[2] = xx + dx;
  ptlist[3] = yy - dy;
  ptlist[4] = xx + dx;
  ptlist[5] = yy + dy;
  ptlist[6] = xx - dx;
  ptlist[7] = yy + dy;

  if (transflag) {
    double ang = angle / 180.0 * detri2::PI;
    for (int i = 0; i < 4; i++) {
      double *pp = &(ptlist[i*2]);
      double x = pp[0];
      double y = pp[1];
      // Translate it to the origin.
      double x1 = x - rcx;
      double y1 = y - rcy;
      // Rotate it
      double x2 = cos(ang) * x1 - sin(ang) * y1;
      double y2 = sin(ang) * x1 + cos(ang) * y1;
      // Translate it back
      double x3 = x2 + rcx;
      double y3 = y2 + rcy;
      // Translate it to the final position.
      pp[0] = x3 + tx;
      pp[1] = y3 + ty;
    }
  }

  int elist[8];
  elist[0] = 0;
  elist[1] = 1;
  elist[2] = 1;
  elist[3] = 2;
  elist[4] = 2;
  elist[5] = 3;
  elist[6] = 3;
  elist[7] = 0;

  int etaglist[4];
  for (int i = 0; i < 4; i++) etaglist[i] = etag;

  return detri2_add_geometry(pGeom, 4, ptlist, 4, elist, etaglist);
}

int detri2_add_circle(detri2::Triangulation** pGeom,
                      double ctx, double cty, double radius,
                      double ang1, double ang2, int segnum, int etag,
                      bool transflag,
                      double rcx, double rcy, double angle, double tx, double ty)
{
    if (segnum < 0) return 0;

    double ang_start = ang1 / 180. * detri2::PI;
    double dfi = fabs(ang2 - ang1) / segnum;
    dfi = dfi / 180.0 * detri2::PI;

    int np = segnum + 1;
    double *ptlist = new double[np * 2];
    for (int i = 0; i < np; i++) {
      double *pp = &(ptlist[i*2]);
      pp[0] = radius * cos(ang_start + dfi * (double) i) + ctx;
      pp[1] = radius * sin(ang_start + dfi * (double) i) + cty;
    }
    int *elist = new int[segnum * 2];
    for (int i = 0; i < segnum; i++) {
      elist[i*2] = i;
      elist[i*2+1] = i+1;
    }

    if (transflag) {
      double ang = angle / 180.0 * detri2::PI;
      for (int i = 0; i < np; i++) {
        double *pp = &(ptlist[i*2]);
        double x = pp[0];
        double y = pp[1];
        // Translate it to the origin.
        double x1 = x - rcx;
        double y1 = y - rcy;
        // Rotate it
        double x2 = cos(ang) * x1 - sin(ang) * y1;
        double y2 = sin(ang) * x1 + cos(ang) * y1;
        // Translate it back
        double x3 = x2 + rcx;
        double y3 = y2 + rcy;
        // Translate it to the final position.
        pp[0] = x3 + tx;
        pp[1] = y3 + ty;
      }
    }

    int *etaglist = new int[segnum];
    for (int i = 0; i < segnum; i++) etaglist[i] = etag;

    detri2_add_geometry(pGeom, np, ptlist, segnum, elist, etaglist);

    delete [] ptlist;
    delete [] elist;
    delete [] etaglist;

    return 1;
}

// Scale the vertices into the range [bbxmin, bbxmax]x[bbymin, bbymax]
int detri2_scale_points(Triangulation* Geom, double bxmin, double bxmax,
                        double bymin, double bymax)
{
    double scale_xy, cx, cy;

    double w = bxmax - bxmin;
    double h = bymax - bymin;
    double min_hw = h < w ? h : w;

    double dx = Geom->io_xmax - Geom->io_xmin;
    double dy = Geom->io_ymax - Geom->io_ymin;
    if (dy < dx) {
      scale_xy = min_hw / dy;
    } else {
      scale_xy = min_hw / dx;
    }

    // The following one is use the lower-left corner a
    // Use the both mid points as the translation center.
    cx = bxmin + w/2. - (Geom->io_xmin + dx/2.) * scale_xy;
    cy = bymin + h/2. - (Geom->io_ymin + dy/2.) * scale_xy;

    int i;
    for (i = 0; i < Geom->ct_in_vrts; i++) {
      if (Geom->in_vrts[i].typ == UNUSEDVERTEX) continue;
      double px =  scale_xy * Geom->in_vrts[i].crd[0] + cx;
      double py =  scale_xy * Geom->in_vrts[i].crd[1] + cy;
      Geom->in_vrts[i].crd[0] = px;
      Geom->in_vrts[i].crd[1] = py;
      Geom->in_vrts[i].crd[2] = px*px + py*py;
      //Geom->in_vrts[i].crd[2] = Geom->op_lambda1 * px*px + Geom->op_lambda2 * py*py;
    }
    if (Geom->tr_steiners != NULL) {
      for (i = 0; i < Geom->tr_steiners->used_items; i++) {
        Vertex *vrt = (Vertex *) Geom->tr_steiners->get(i);
        if (vrt->is_deleted()) continue;
        double px =  scale_xy * vrt->crd[0] + cx;
        double py =  scale_xy * vrt->crd[1] + cy;
        vrt->crd[0] = px;
        vrt->crd[1] = py;
        vrt->crd[2] = px*px + py*py;
        //vrt->crd[2] = Geom->op_lambda1 * px*px + Geom->op_lambda2 * py*py;
      }
    }

    Geom->io_xmin = bxmin;
    Geom->io_xmax = bxmax;
    Geom->io_ymin = bymin;
    Geom->io_ymax = bymax;
    return 1;
}

//==============================================================================

void detri2_sweepline_sort(Vertex **ptlist, int ptnum, double vx, double vy)
{
  int n = 0, i;
  while (n < ptnum) {
    double min = ptlist[n]->crd[0] * vx + ptlist[n]->crd[1] * vy;
    int min_idx = n;
    for (i = n + 1; i < ptnum; i++) {
      double value = ptlist[i]->crd[0] * vx + ptlist[i]->crd[1] * vy;
      if (value < min) {
        min = value;
        min_idx = i;
      }
    }
    // Swap i and n;
    Vertex *swappt = ptlist[min_idx];
    ptlist[min_idx] = ptlist[n];
    ptlist[n] = swappt;
    n++;
  }
}

//=============================================================================

// Create the (nearest or furthest) weighted Delaunay triangulation.
int detri2_Delaunay(Triangulation* _geom, Triangulation** pTr,
                    int furthest, int weighted, int cdt,
                    double a11, double a21, double a22)
{
  if (_geom->tr_steiners->objects < 3) {
    //qDebug() <<"A triangulation needs at least 3 points!";
    return 0;
  }

  qDebug() << "Create triangulation from "<<_geom->tr_steiners->objects<<" vertices, "
           << _geom->tr_segs->objects << " segments.";

  //Triangulation *_tria = *pTr;
  if (*pTr != NULL) delete *pTr;
  *pTr = new Triangulation();

  // Euclidean metric only
  (*pTr)->_a11 = a11;
  (*pTr)->_a21 = a21;
  (*pTr)->_a22 = a22;

  if (furthest > 0) (*pTr)->op_dt_nearest=-1; // furthest Delaunay

  int firstindex = 1; // Set it to be 0 or 1.
  (*pTr)->io_firstindex = firstindex;
  Vertex *vrt, *gvrt;
  int i, j;

  // Read the input points and weights.
  double x, y;
  int nv = _geom->tr_steiners->objects;
  (*pTr)->ct_in_vrts = nv;
  (*pTr)->in_vrts = new Vertex[nv];
  for (i = 0, j = 0; i < nv; i++) {
    do {
      gvrt = (Vertex *) _geom->tr_steiners->get(j);
      j++;
    } while (gvrt->is_deleted());
    vrt = &((*pTr)->in_vrts[i]);
    vrt->init();
    x = vrt->crd[0] = gvrt->crd[0];
    y = vrt->crd[1] = gvrt->crd[1];
    vrt->crd[2] = x*x + y*y;
    //vrt->crd[2] = (*pTr)->op_lambda1 * x*x + (*pTr)->op_lambda2 * y*y;
    if (weighted) {
   //   vrt->crd[2] = x*x+y*y-gvrt->wei; // its height
        vrt->crd[2] -= gvrt->wei;
    } else {
      vrt->crd[2] = x*x+y*y; // ignore weight
    }
    vrt->wei = gvrt->wei;
    vrt->idx = i + firstindex;
    vrt->val = gvrt->val;
    gvrt->Pair = vrt; // Remember this vertex.
  }
  (*pTr)->ct_unused_vrts=nv;
  (*pTr)->io_xmin = _geom->io_xmin;
  (*pTr)->io_xmax = _geom->io_xmax;
  (*pTr)->io_ymin = _geom->io_ymin;
  (*pTr)->io_ymax = _geom->io_ymax;
  (*pTr)->io_diagonal = _geom->io_diagonal;

  (*pTr)->incremental_delaunay();

  if ((cdt > 0) && _geom->tr_segs) {
    // Read segments
    Triang *seg, *gseg;
    (*pTr)->tr_segs = new arraypool(sizeof(Triang), 8);
    for (i = 0; i < _geom->tr_segs->objects; i++) {
      seg =  (Triang *) (*pTr)->tr_segs->alloc();
      gseg = (Triang *) _geom->tr_segs->get(i);
      seg->init();
      seg->vrt[0] = gseg->vrt[0]->Pair;
      seg->vrt[1] = gseg->vrt[1]->Pair;
      seg->tag = gseg->tag;
      seg->val = gseg->val;
    }

    if ((*pTr)->tr_segs->objects > 0) {
      (*pTr)->recover_segments();
      //(*pTr)->set_subdomains();
    }
  } // cdt

  (*pTr)->mesh_statistics();

  if ((*pTr)->tr_tris->objects > 0) {
    return 1;
  } else {
    qDebug()<<"!! No triangulation is created.";
    delete (*pTr);
    *pTr = NULL;
    return 0;
  }
}

//==============================================================================
// Generate mesh (*pTr) from _geom.
//   *pTr is an empty Triangulaiton which contains options.
//   (*pTr->clean() and *pTr->initalise()).

int detri2_generate(Triangulation* _geom, Triangulation** pTr, 
                    int weighted, int cdt, int nonconvex,
                    double a11, double a21, double a22)
{
  if (_geom->tr_steiners->objects < 3) {
    return 0;
  }

  // *pTr is an empty mesh which contains mesh options.
  assert((*pTr)->ct_in_vrts == 0);
  assert((*pTr)->tr_nonconvex == false);
  int firstindex = 1; // Set it to be 0 or 1.
  (*pTr)->io_firstindex = firstindex;

  (*pTr)->_a11 = a11;
  (*pTr)->_a21 = a21;
  (*pTr)->_a22 = a22;

  Vertex *vrt, *gvrt;
  int i, j;

  // Read the input points and weights.
  double x, y;
  int nv = _geom->tr_steiners->objects;
  (*pTr)->ct_in_vrts = nv;
  (*pTr)->in_vrts = new Vertex[nv];
  for (i = 0, j = 0; i < nv; i++) {
    do {
      gvrt = (Vertex *) _geom->tr_steiners->get(j); j++;
    } while (gvrt->is_deleted()); // skip deleted Steiner points.
    vrt = &((*pTr)->in_vrts[i]);
    vrt->init();
    x = vrt->crd[0] = gvrt->crd[0];
    y = vrt->crd[1] = gvrt->crd[1];
    vrt->crd[2] = x*x+y*y;
    //vrt->crd[2] = (*pTr)->op_lambda1 * x*x + (*pTr)->op_lambda2 * y*y;
    if (weighted) {
      //vrt->crd[2] = x*x+y*y-gvrt->wei; // its height
      vrt->crd[2] -= gvrt->wei;
    } else {
      //vrt->crd[2] = x*x+y*y;
    }
    vrt->wei = gvrt->wei;
    vrt->val = gvrt->val; // vertex mesh size
    vrt->idx = i + firstindex;
    gvrt->Pair = vrt; // Remember this vertex.
  }
  (*pTr)->ct_unused_vrts=nv;
  (*pTr)->io_xmin = _geom->io_xmin;
  (*pTr)->io_xmax = _geom->io_xmax;
  (*pTr)->io_ymin = _geom->io_ymin;
  (*pTr)->io_ymax = _geom->io_ymax;
  (*pTr)->io_diagonal = _geom->io_diagonal;

  (*pTr)->incremental_delaunay();

  //(*pTr)->op_db_verbose = 4; // for debug

  if ((cdt > 0) || ((*pTr)->op_quality > 0)) {
    // Read segments
    Triang *seg, *gseg;
    (*pTr)->tr_segs = new arraypool(sizeof(Triang), 8);
    for (i = 0; i < _geom->tr_segs->used_items; i++) {
      gseg = (Triang *) _geom->tr_segs->get(i);
      if (gseg->is_deleted()) continue;
      seg =  (Triang *) (*pTr)->tr_segs->alloc();
      seg->init();
      seg->vrt[0] = (gseg->vrt[0]->Pair);
      seg->vrt[1] = (gseg->vrt[1]->Pair);
      seg->tag = gseg->tag;
      seg->val = gseg->val;
      //printf("  get segment %d,%d, tag=%d, val=%g\n", seg->vrt[0]->idx, seg->vrt[1]->idx, seg->tag, seg->val);
    }

    if ((*pTr)->tr_segs->objects > 0) {
      (*pTr)->recover_segments();

      if (_geom->ct_in_sdms > 0) {
        (*pTr)->ct_in_sdms = _geom->ct_in_sdms;
        (*pTr)->in_sdms = new Vertex[(*pTr)->ct_in_sdms];
        for (i = 0; i < (*pTr)->ct_in_sdms; i++) {
          Vertex *ov = &(_geom->in_sdms[i]);
          Vertex *v = &((*pTr)->in_sdms[i]);
          v->init();
          v->crd[0] = ov->crd[0];
          v->crd[1] = ov->crd[1];
          v->tag    = ov->tag;
          v->val    = ov->val;
          //printf("  set domain %d, val=%g\n", v->tag, v->val);
        }
      }
      (*pTr)->set_subdomains();

      if ((*pTr)->op_quality || (*pTr)->op_metric) {
        //(*pTr)->op_db_verbose = 4; // for debug
        if ((*pTr)->ct_exteriors > 0) {
          (*pTr)->remove_exteriors(); // Remove exterior triangles.
        }

        if ((*pTr)->op_metric) {
          (*pTr)->set_vertex_metrics();
          (*pTr)->coarsen_mesh();
        }
        (*pTr)->delaunay_refinement();
      } else {
        if (nonconvex) {
          if ((*pTr)->ct_exteriors > 0) {
            (*pTr)->remove_exteriors(); // Remove exterior triangles.
          }
        }
      }
    }

    //if ((*pTr)->op_smooth_criterion) {
    // to do ...
    //}
  }

  (*pTr)->mesh_statistics();
  if ((*pTr)->op_quality) {
    (*pTr)->quality_statistics();
  }

  return 1;
}


//==============================================================================
/*
//#include "HDE/test_functions.h"

// Mesh adaptation through HDE
int detri2_HDE_mesh_adapt(detri2::Triangulation** MeshIO,
                          double s1,
                          double s2,
                          double min_len,
                          double min_angle,
                          double size_iter,
                          double qual_iter,
                          double use_intpoints,
                          double use_smooth,
                          double smooth_iter,
                          double smooth_alpha,
                          int report_error_only, // verbose,
                          int test_fun)
{
  // Step 1, transfer mesh from MeshIO to HDEMesher
  if (((*MeshIO)->tr_tris == NULL) ||
      ((*MeshIO)->tr_tris->objects == 0)) {
    qDebug()<<"No input mesh!";
    return 0; // no input mesh
  }

  qDebug()<<"HDE mesh adaptation!";
  HDE_Triangulation* HDEMesher=new HDE_Triangulation();

  HDEMesher->op_s1 = s1;
  HDEMesher->op_s2 = s2;
  HDEMesher->op_target_length = min_len;
  HDEMesher->op_minangle = min_angle;
  HDEMesher->op_max_size_iter = size_iter;
  HDEMesher->op_max_quality_iter = qual_iter;
  HDEMesher->op_use_intpoints = use_intpoints;
  HDEMesher->op_use_smoothing = use_smooth;
  HDEMesher->op_smooth_iter = smooth_iter;
  HDEMesher->op_smooth_alpha = smooth_alpha;
  //HDEMesher->op_db_verbose = verbose;
  HDEMesher->op_report_quality_only = report_error_only;

  //HDEMesher->Func = &fun_lennard;
  //HDEMesher->GradX = &funX_lennard;
  //HDEMesher->GradY = &funY_lennard;

  if (test_fun == 1) {
    HDEMesher->Func = &fun_gaussian;
    HDEMesher->GradX = &funX_gaussian;
    HDEMesher->GradY = &funY_gaussian;
  } else if (test_fun == 2) {
    HDEMesher->Func = &fun_lennard;
    HDEMesher->GradX = &funX_lennard;
    HDEMesher->GradY = &funY_lennard;
  } else if (test_fun == 3) {
    HDEMesher->Func = &fun_alauzet;
    HDEMesher->GradX = &funX_alauzet;
    HDEMesher->GradY = &funY_alauzet;
  } else if (test_fun == 4) {
    HDEMesher->Func = &fun_Huang;
    HDEMesher->GradX = &funX_Huang;
    HDEMesher->GradY = &funY_Huang;
  } else if (test_fun == 5) {
    //HDEMesher->Func = &fun_du;
    //HDEMesher->GradX = &funX_du;
    //HDEMesher->GradY = &funY_du;
    HDEMesher->Func = &fun_thanks;
    HDEMesher->GradX = &funX_thanks;
    HDEMesher->GradY = &funY_thanks;
  } else if (test_fun == 6) {
    HDEMesher->Func = &fun_Gaussian2;
    HDEMesher->GradX = &funX_Gaussian2;
    HDEMesher->GradY = &funY_Gaussian2;
  } else if (test_fun == 7) {
    // use unitcircle.1.ele
    HDEMesher->Func = &fun_bamg;
    HDEMesher->GradX = &funX_bamg;
    HDEMesher->GradY = &funY_bamg;
  } else if (test_fun == 8) {
    HDEMesher->Func = &fun_Rosenbrock;
    HDEMesher->GradX = &funX_Rosenbrock;
    HDEMesher->GradY = &funY_Rosenbrock;
  } else if (test_fun == 100) {
    HDEMesher->Func = &fun_Delaunay;
    HDEMesher->GradX = &funX_Delaunay;
    HDEMesher->GradY = &funY_Delaunay;
  } else if (test_fun == 101) {
    HDEMesher->Func = &fun_Saddle;
    HDEMesher->GradX = &funX_Saddle;
    HDEMesher->GradY = &funY_Saddle;
  }
  else {
    printf("Please choose a test function by set test_fun=#:\n");
    printf("  1, Gaussian.\n");
    printf("  2, Lennard.\n");
    printf("  3, Alauzet.\n");
    printf("  4, Huang.\n");
    printf("  5, Du.\n");
    printf("  6, Gaussian2.\n");
    printf("  7, BAMG.\n");
    printf("  8, Rosenbrock.\n");
    delete HDEMesher;
    return 0;
  }

  // Transfer the mesh vertices.
  int firstindex = 1;
  HDEMesher->io_firstindex = 1;
  Vertex *vrt, *mvrt;
  int i;

  // Read the input points.
  double x, y;
  int nv = (*MeshIO)->ct_in_vrts - (*MeshIO)->ct_unused_vrts;
  if ((*MeshIO)->tr_steiners) {
    nv += (*MeshIO)->tr_steiners->objects;
  }
  HDEMesher->ct_in_vrts = nv;
  HDEMesher->in_vrts = new Vertex[nv];
  int idx = 0;
  for (i = 0; i < (*MeshIO)->ct_in_vrts; i++) {
    mvrt = &((*MeshIO)->in_vrts[i]);
    if (mvrt->typ == UNUSEDVERTEX) continue;
    vrt = &(HDEMesher->in_vrts[idx]);
    vrt->init();
    x = vrt->crd[0] = mvrt->crd[0];
    y = vrt->crd[1] = mvrt->crd[1];
    vrt->crd[2] = x*x+y*y;
    vrt->idx = idx + firstindex;
    mvrt->Pair = vrt; // Remember this vertex for segments/triangles.
    idx++;
  }
  qDebug()<<" input idx="<<idx;
  if ((*MeshIO)->tr_steiners) {
    for (i = 0; i < (*MeshIO)->tr_steiners->used_items; i++) {
      mvrt = (Vertex *) (*MeshIO)->tr_steiners->get(i);
      if (mvrt->is_deleted()) continue;
      vrt = &(HDEMesher->in_vrts[idx]);
      vrt->init();
      x = vrt->crd[0] = mvrt->crd[0];
      y = vrt->crd[1] = mvrt->crd[1];
      vrt->crd[2] = x*x+y*y;
      vrt->idx = idx + firstindex;
      mvrt->Pair = vrt; // Remember this vertex for segments/triangles.
      idx++;
    }
  }
  qDebug()<<" total idx="<<idx<<", nv="<<nv;
  assert(idx == nv);
  HDEMesher->ct_unused_vrts = nv;
  HDEMesher->io_xmin = (*MeshIO)->io_xmin;
  HDEMesher->io_xmax = (*MeshIO)->io_xmax;
  HDEMesher->io_ymin = (*MeshIO)->io_ymin;
  HDEMesher->io_ymax = (*MeshIO)->io_ymax;
  HDEMesher->io_diagonal = (*MeshIO)->io_diagonal;

  // Transfer the mesh triangles.
  HDEMesher->ct_in_tris = (*MeshIO)->tr_tris->objects - (*MeshIO)->ct_hullsize - (*MeshIO)->ct_exteriors;
  HDEMesher->tr_tris = new arraypool(sizeof(Triang), 10);
  for (i = 0; i < (*MeshIO)->tr_tris->used_items; i++) {
    Triang *mtri = (Triang *) (*MeshIO)->tr_tris->get(i);
    if (mtri->is_deleted() || mtri->is_hulltri() || mtri->is_exterior()) continue;
    Triang *tri = (Triang *) HDEMesher->tr_tris->alloc();
    tri->init();
    for (int j = 0; j < 3; j++) {
      tri->vrt[j] = mtri->vrt[j]->Pair;
    }
    tri->tag = mtri->tag;
  }
  assert(HDEMesher->tr_tris->objects == HDEMesher->ct_in_tris);

  qDebug() << "HDE mesh reconstruction";
  HDEMesher->reconstruct_mesh(0); // Do not flip.
  qDebug() << "HDE mesh reconstruction end";

  strcpy(HDEMesher->io_outfilename, "dump_hde"); // for debugging
  //HDEMesher->save_triangulation(); // for debug

  qDebug() << "HDE mesh adaptation started....";
  HDEMesher->HDE_mesh_adapt();

  qDebug() << "HDE mesh adaptation finished";

  //strcpy(HDEMesher->io_outfilename, "dump_hde.1"); // for debugging
  //HDEMesher->save_triangulation(); // for debug

  //delete HDEMesher;
  //return 1;

  // Transfer the mesh back to MeshIO.
  //(*MeshIO)->clean();
  //(*MeshIO)->initialize();
  delete (*MeshIO);
  (*MeshIO) = new Triangulation();

  nv = HDEMesher->ct_in_vrts - HDEMesher->ct_unused_vrts;
  if (HDEMesher->tr_steiners) {
    nv += HDEMesher->tr_steiners->objects;
  }
  (*MeshIO)->ct_in_vrts = nv;
  (*MeshIO)->io_firstindex = 1;
  (*MeshIO)->in_vrts = new Vertex[nv];

  idx = 0;
  for (i = 0; i < HDEMesher->ct_in_vrts; i++) {
    mvrt = &(HDEMesher->in_vrts[i]);
    if (mvrt->typ == UNUSEDVERTEX) continue;
    vrt = &((*MeshIO)->in_vrts[idx]);
    vrt->init();
    x = vrt->crd[0] = mvrt->crd[0];
    y = vrt->crd[1] = mvrt->crd[1];
    vrt->crd[2] = x*x+y*y;
    vrt->wei = mvrt->wei;
    vrt->idx = idx + firstindex;
    mvrt->Pair = vrt; // Remember this vertex for segments/triangles.
    idx++;
  }
  if (HDEMesher->tr_steiners) {
    for (i = 0; i < HDEMesher->tr_steiners->used_items; i++) {
      mvrt = (Vertex *) HDEMesher->tr_steiners->get(i);
      if (mvrt->is_deleted()) continue;
      vrt = &((*MeshIO)->in_vrts[idx]);
      vrt->init();
      x = vrt->crd[0] = mvrt->crd[0];
      y = vrt->crd[1] = mvrt->crd[1];
      vrt->crd[2] = x*x+y*y;
      vrt->wei = mvrt->wei;
      vrt->idx = idx + firstindex;
      mvrt->Pair = vrt; // Remember this vertex for segments/triangles.
      idx++;
    }
  }
  qDebug()<<"  idx="<<idx<<", nv="<<nv;
  assert(idx == nv);
  (*MeshIO)->ct_unused_vrts = nv;
  (*MeshIO)->io_xmin = HDEMesher->io_xmin;
  (*MeshIO)->io_xmax = HDEMesher->io_xmax;
  (*MeshIO)->io_ymin = HDEMesher->io_ymin;
  (*MeshIO)->io_ymax = HDEMesher->io_ymax;
  (*MeshIO)->io_diagonal = HDEMesher->io_diagonal;

  // Transfer the mesh triangles.
  (*MeshIO)->ct_in_tris = HDEMesher->tr_tris->objects - HDEMesher->ct_hullsize - HDEMesher->ct_exteriors;
  (*MeshIO)->tr_tris = new arraypool(sizeof(Triang), 10);
  for (i = 0; i < HDEMesher->tr_tris->used_items; i++) {
    Triang *mtri = (Triang *) HDEMesher->tr_tris->get(i);
    if (mtri->is_deleted() || mtri->is_hulltri() || mtri->is_exterior()) continue;
    Triang *tri = (Triang *) (*MeshIO)->tr_tris->alloc();
    tri->init();
    for (int j = 0; j < 3; j++) {
      tri->vrt[j] = mtri->vrt[j]->Pair;
    }
    tri->tag = mtri->tag;
  }
  assert((*MeshIO)->tr_tris->objects == (*MeshIO)->ct_in_tris);

  (*MeshIO)->reconstruct_mesh(0); // Do not flip.

  delete HDEMesher;

  return 1;
}
*/
//==============================================================================

int detri2_function_interpolation(Triangulation* Geom, Triangulation* MeshIO)
{
    qDebug()<<"HDE mesh adaptation!";
    MeshIO->op_db_verbose = 10; // debug

    MeshIO->mesh_adapt();

    return 1;
}

//==============================================================================
/*
// mesh_smoothing.cpp
void mesh_smoothing(int d, int N, int Nv, int Nb, double c[], int tri[],
                    int bndry[], double tf, double abstol, double dt);
void mesh_smoothing2(int d, int N, int Nv, int Nb, double c[], int tri[],
                    int bndry[], double tf, double abstol, double dt);

// Mesh smoothing -- Huang's MMPDE approach
int detri2_mesh_smooth_mmpde(Triangulation** MeshIO, double tf, double abstol, double dt)
{
  // Step 1, transfer mesh from MeshIO to HDEMesher
  if (((*MeshIO)->tr_tris == NULL) ||
      ((*MeshIO)->tr_tris->objects == 0)) {
    qDebug()<<"No input mesh!";
    return 0; // no input mesh
  }

  int i, j, d, N, Nv, Nb;
  int *tri, *bndry;
  double *c, tcpu;

  int pnum = (*MeshIO)->ct_in_vrts - (*MeshIO)->ct_unused_vrts;
  pnum += ((*MeshIO)->tr_steiners != NULL ?
           (*MeshIO)->tr_steiners->objects : 0);

  Nv = pnum;
  d = 2;
  c = new double[Nv*d];
  bndry = new int[Nv]; // Not all will be used.
  Nb = 0; // count Nb

  qDebug()<<"Reading "<<Nv<<" mesh vertices";

  int shift = ((*MeshIO)->io_firstindex > 0 ? 1 : 0); // indices start from 0.
  int idx = (*MeshIO)->io_firstindex - shift; // Start from 0;
  for (i = 0; i < (*MeshIO)->ct_in_vrts; i++) {
    Vertex *v = &((*MeshIO)->in_vrts[i]);
    if (v->typ == UNUSEDVERTEX) continue;
    c[idx*2] = v->crd[0];
    c[idx*2 + 1] = v->crd[1];
    v->idx = idx;
    if ((v->typ == RIDGEVERTEX) ||
        (v->typ == SEGMENTVERTEX)) {
      bndry[Nb] = idx;
      Nb++;
    }
    idx++;
  }
  if ((*MeshIO)->tr_steiners != NULL) {
    for (i = 0; i < (*MeshIO)->tr_steiners->used_items; i++) {
      Vertex *v = (Vertex *) (*MeshIO)->tr_steiners->get(i);
      if (v->is_deleted()) continue;
      c[idx*2] = v->crd[0];
      c[idx*2 + 1] = v->crd[1];
      v->idx = idx;
      if (v->is_fixed() ||
          (v->on_bd.tri != NULL)) {
        bndry[Nb] = idx;
        Nb++;
      }
      idx++;
    }
  }
  
  qDebug()<<"Found "<<Nb<<" fixed vertices";
  
  int trinum = (*MeshIO)->tr_tris->objects - (*MeshIO)->ct_hullsize - (*MeshIO)->ct_exteriors;
  N = trinum;
  tri = new int[N*(d+1)];

  qDebug()<<"Reading "<<N<<" triangles";

  idx = 0;
  for (i = 0; i < (*MeshIO)->tr_tris->used_items; i++) {
    Triang *ele = (Triang *) (*MeshIO)->tr_tris->get(i);
    if (ele->is_deleted() || ele->is_hulltri() || ele->is_exterior()) continue;
    tri[3*idx]   = ele->vrt[0]->idx;
    tri[3*idx+1] = ele->vrt[1]->idx;
    tri[3*idx+2] = ele->vrt[2]->idx;
    idx++;
  }

  qDebug()<<"Call MMPDE mesh smoothing";
  
  //tf = 10;
  //abstol = 1e-6;
  //dt = 1e-6;
  mesh_smoothing(d, N, Nv, Nb, c, tri, bndry, tf, abstol, dt);

  qDebug()<<"Done! Update the mesh vertices";

  idx = (*MeshIO)->io_firstindex - shift; // Start from 0;
  for (i = 0; i < (*MeshIO)->ct_in_vrts; i++) {
    Vertex *v = &((*MeshIO)->in_vrts[i]);
    if (v->typ == UNUSEDVERTEX) continue;
    v->crd[0] = c[idx*2];
    v->crd[1] = c[idx*2 + 1];
    v->idx = idx + (*MeshIO)->io_firstindex;
    idx++;
  }
  if ((*MeshIO)->tr_steiners != NULL) {
    for (i = 0; i < (*MeshIO)->tr_steiners->used_items; i++) {
      Vertex *v = (Vertex *) (*MeshIO)->tr_steiners->get(i);
      if (v->is_deleted()) continue;
      v->crd[0] = c[idx*2];
      v->crd[1] = c[idx*2 + 1];
      v->idx = idx + (*MeshIO)->io_firstindex;
      idx++;
    }
  }

  (*MeshIO)->quality_statistics();

  return 1;
}
*/

//==============================================================================

void detri2_draw_point(Vertex* v, int r, void *Painter, double Cx, double Cy, double Sxy)
{
  QPainter* _pnt = (QPainter *) Painter;
  int x = (int) ( Sxy * v->crd[0] + Cx);
  int y = (int) (-Sxy * v->crd[1] + Cy);
  _pnt->drawEllipse(QPoint(x,y), r, r);
}

void detri2_draw_line(Vertex* e1, Vertex* e2, void *Painter, double Cx, double Cy, double Sxy)
{
  QPainter* _pnt = (QPainter *) Painter;
  // Only draw this line if none of its endpoints are infinite vertex.
  int x1 = (int) ( Sxy * e1->crd[0] + Cx);
  int y1 = (int) (-Sxy * e1->crd[1] + Cy);
  int x2 = (int) ( Sxy * e2->crd[0] + Cx);
  int y2 = (int) (-Sxy * e2->crd[1] + Cy);
  _pnt->drawLine(QPoint(x1, y1), QPoint(x2,y2));
}

//http://csharphelper.com/blog/2014/08/calculate-the-binomial-coefficient-n-choose-k-efficiently-in-c/
int fun_N_choose_K(int N, int K)
{
    int result = 1;
    for (int i = 1; i <= K; i++)
    {
      result *= N - (K - i);
      result /= i;
    }
    return result;
}

void detri2_draw_Bezier_curve(int n, Vertex *pt[], void* Painter, double Cx, double Cy, double Sxy)
{
    // n=1 has two points, it is linear
    // n=2 has three points, it is quadratic,
    if (n < 2) {
      return;
    }

    QPainter* _pnt = (QPainter *) Painter;

    int x1, y1, x2, y2;
    x1 = (int) ( Sxy * pt[0]->crd[0] + Cx);
    y1 = (int) (-Sxy * pt[0]->crd[1] + Cy);

    double xu = 0.0, yu = 0.0, u = 0.0 ;
    for(u = 0.01 ; u <= 0.99 ; u += 0.01) {
      xu = yu = 0.0;
      for (int i = 0; i <= n; i++) {
        double Bi = fun_N_choose_K(n, i) * pow(1-u, n-i) * pow(u, i);
        xu += pt[i]->crd[0] * Bi;
        yu += pt[i]->crd[1] * Bi;
      }
      x2 = (int) ( Sxy * xu + Cx);
      y2 = (int) (-Sxy * yu + Cy);
      _pnt->drawLine(QPoint(x1, y1), QPoint(x2,y2));
      x1 = x2;
      y1 = y2;
    }

    x2 = (int) ( Sxy * pt[n]->crd[0] + Cx);
    y2 = (int) (-Sxy * pt[n]->crd[1] + Cy);
    _pnt->drawLine(QPoint(x1, y1), QPoint(x2,y2));
}

//=============================================================================

void detri2_draw_bbox(double _xmin, double _xmax, double _ymin, double _ymax,
                      void* Painter, double Cx, double Cy, double Sxy)
{
  QPainter *painter = (QPainter *) Painter;
  //qDebug() << "detri2 draw bbox";
  //QPen pen;
  //pen.setColor(qRgb(204, 204, 204));
  //pen.setWidth(2);
  //painter.setPen(pen);

  int xmin = (int) ( Sxy * _xmin + Cx);
  int xmax = (int) ( Sxy * _xmax + Cx);
  int ymin = (int) (-Sxy * _ymin + Cy);
  int ymax = (int) (-Sxy * _ymax + Cy);

  //qDebug() << "  xmin="<< xmin<<", xmax="<<xmax;
  //qDebug() << "  ymin="<< ymin<<", ymax="<<ymax;

  painter->drawLine(QPoint(xmin, ymin), QPoint(xmax,ymin));
  painter->drawLine(QPoint(xmax, ymin), QPoint(xmax,ymax));
  painter->drawLine(QPoint(xmax, ymax), QPoint(xmin,ymax));
  painter->drawLine(QPoint(xmin, ymax), QPoint(xmin,ymin));

  //QFont serifFont("Times", 10, QFont::Bold);
  //QFont sansFont("Helvetica [Cronyx]", 12);
  //painter.setFont(sansFont);
  char str_min[100], str_max[100];
  sprintf(str_min, "(%.2f, %.2f)", _xmin, _ymin);
  sprintf(str_max, "(%.2f, %.2f)", _xmax, _ymax);
  painter->drawText(xmin, ymin, QString(str_min));
  painter->drawText(xmax-60, ymax+10, QString(str_max));
}

//=============================================================================
/*
void detri2_draw_geometry(Triangulation* Tr, void *Painter, double Cx, double Cy, double Sxy)
{
    QPainter *painter = (QPainter *) Painter;
    qDebug() << "detri2 draw geometry: "<< Tr->tr_steiners->objects << " points"
             << Tr->tr_segs->objects << " segments";

    // Draw Steiner points.
    //QPen pen;
    //pen.setColor(qRgb(102, 102, 102)); // vertex color
    //pen.setWidth(1); // default
    //painter.setPen(pen);
    //QBrush brush;
    //brush.setColor(qRgb(102, 102, 102));
    //brush.setStyle(Qt::SolidPattern);
    //painter.setBrush(brush);

    // Draw Steiner points.
    int ptsize=4;
    for (int i = 0; i < Tr->tr_steiners->used_items; i++) {
      Vertex *v = (Vertex *) Tr->tr_steiners->get(i);
      if (v->is_deleted()) continue;
      int x = (int) ( Sxy * v->crd[0] + Cx);
      int y = (int) (-Sxy * v->crd[1] + Cy);
      painter->drawEllipse(QPoint(x,y), ptsize, ptsize);
    }

    //painter.setBrush(Qt::NoBrush);
}
*/
//==============================================================================
/*
void detri2_draw_vertex_labels(Triangulation* Tr, void* Painter, double Cx, double Cy, double Sxy)
{
    QPainter *painter = (QPainter *) Painter;

    // Draw Steiner points.
    //QPen pen;
    //pen.setColor(qRgb(102, 102, 102)); // vertex color
    //pen.setWidth(1); // default
    //painter.setPen(pen);
    //QBrush brush;
    //brush.setColor(qRgb(102, 102, 102));
    //brush.setStyle(Qt::SolidPattern);
    //painter.setBrush(brush);

    //QFont serifFont("Times", 10, QFont::Bold);
    //QFont sansFont("Helvetica [Cronyx]", 12);
    //painter.setFont(sansFont);

    // Draw Steiner points.
    int ptsize=4;
    for (int i = 0; i < Tr->tr_steiners->used_items; i++) {
      Vertex *v = (Vertex *) Tr->tr_steiners->get(i);
      if (v->is_deleted()) continue;
      int x = (int) ( Sxy * v->crd[0] + Cx);
      int y = (int) (-Sxy * v->crd[1] + Cy);
      //painter->drawEllipse(QPoint(x,y), ptsize, ptsize);
      painter->drawText(x+ptsize, y-ptsize, QString::number(v->idx));
    }

    //painter.setBrush(Qt::NoBrush);
}
*/
//==============================================================================
/*
void detri2_draw_segments(Triangulation* Tr, void *Painter, double Cx, double Cy, double Sxy)
{
  QPainter *painter = (QPainter *) Painter;
      int x1, y1, x2, y2;
      for (int i = 0; i < Tr->tr_segs->used_items; i++) {
        Triang *seg = (Triang *) Tr->tr_segs->get(i);
        if (seg->is_deleted()) continue;
        x1 = (int) ( Sxy * seg->vrt[0]->crd[0] + Cx);
        y1 = (int) (-Sxy * seg->vrt[0]->crd[1] + Cy);
        x2 = (int) ( Sxy * seg->vrt[1]->crd[0] + Cx);
        y2 = (int) (-Sxy * seg->vrt[1]->crd[1] + Cy);
        painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
      }
}

void detri2_draw_segment_labels(Triangulation* Tr, void* Painter, double Cx, double Cy, double Sxy)
{
  QPainter *painter = (QPainter *) Painter;
  int x1, y1, x2, y2;
  for (int i = 0; i < Tr->tr_segs->used_items; i++) {
    Triang *seg = (Triang *) Tr->tr_segs->get(i);
    if (seg->is_deleted()) continue;
    x1 = (int) ( Sxy * seg->vrt[0]->crd[0] + Cx);
    y1 = (int) (-Sxy * seg->vrt[0]->crd[1] + Cy);
    x2 = (int) ( Sxy * seg->vrt[1]->crd[0] + Cx);
    y2 = (int) (-Sxy * seg->vrt[1]->crd[1] + Cy);
    //painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
    painter->drawText((x1+x2)/2, (y1+y2)/2, QString::number(seg->tag));
  }
}
*/
//=============================================================================
/*
void detri2_draw_subdomains(Triangulation* Tr, void *Painter, double Cx, double Cy, double Sxy)
{
    QPainter *painter = (QPainter *) Painter;
    //QPen pen;
    //pen.setColor(qRgb(255,140,0)); // dark orange
    //pen.setWidth(2); // default
    //painter.setPen(pen);
    int r = 4;
    for (int i = 0; i < Tr->ct_in_sdms; i++) {
      Vertex *v = &(Tr->in_sdms[i]);
      int x = (int) ( Sxy * v->crd[0] + Cx);
      int y = (int) (-Sxy * v->crd[1] + Cy);
      //painter->drawText(x, y, QString::number(v->tag));
      //QRect rect(x-10, y-10, x+10, y+10);
      painter->drawEllipse(QPoint(x,y),r,r);
      painter->drawText(x,y, QString::number(v->tag));
      //painter->drawText(rect, Qt::AlignCenter, QString::number(v->tag));
    }
}
*/
//=============================================================================

void detri2_draw_convex_hull(Triangulation* Tr, void* Painter, double Cx, double Cy, double Sxy)
{
    QPainter *painter = (QPainter *) Painter;
    //qDebug() << "detri2 draw convex hull: " << Tr->ct_hullsize << " edges";

    //QPen pen;
    //pen.setColor(qRgb(102, 102, 102)); // edge color
    //pen.setWidth(1); // default
    //painter.setPen(pen);

    int x1, y1, x2, y2;
    TriEdge E; //, N; //, S;
    int hcount = 0;
    for (int i = 0; i < Tr->tr_tris->used_items; i++) {
      E.tri = (Triang *) Tr->tr_tris->get(i);
      if (E.tri->is_deleted() || !E.tri->is_hulltri()) continue;
      // E.tri must be a hull triangle.
      for (E.ver = 0; E.ver < 3; E.ver++) {
        if (E.apex() == Tr->tr_infvrt) break;
      }
      // Draw this hull edge.
      x1 = (int) ( Sxy * E.org()->crd[0] + Cx);
      y1 = (int) (-Sxy * E.org()->crd[1] + Cy);
      x2 = (int) ( Sxy * E.dest()->crd[0] + Cx);
      y2 = (int) (-Sxy * E.dest()->crd[1] + Cy);
      painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
      hcount++;
    }

    assert(hcount == Tr->ct_hullsize);
}

void detri2_draw_Triangulation(Triangulation* Tr, void *Painter, double Cx, double Cy, double Sxy)
{
    if (Tr == NULL) return;
    if (Tr->tr_tris == NULL) return;

    QPainter *painter = (QPainter *) Painter;
    //qDebug() << "detri2 draw triangulation: " << Tr->tr_tris->objects << " triangles";

    //QPen pen;
    //pen.setColor(qRgb(102, 102, 102)); // edge color
    //pen.setWidth(1); // default
    //painter.setPen(pen);

    int x1, y1, x2, y2;
    TriEdge E, N; //, S;
    for (int i = 0; i < Tr->tr_tris->used_items; i++) {
      E.tri = (Triang *) Tr->tr_tris->get(i);
      if (E.tri->is_deleted() || E.tri->is_hulltri()) continue;
      for (E.ver = 0; E.ver < 3; E.ver++) {
        N = E.esym();
        if (N.tri->is_hulltri() ||
            (E.apex()->idx < N.apex()->idx)) {
          // Draw it if it is not in the exterior.
          //if (!(E.tri->is_exterior() && N.tri->is_exterior()) &&
          //    !(E.tri->is_exterior() && N.tri->is_hulltri()) &&
          //    !(E.tri->is_hulltri() && N.tri->is_exterior())) {
            x1 = (int) ( Sxy * E.org()->crd[0] + Cx);
            y1 = (int) (-Sxy * E.org()->crd[1] + Cy);
            x2 = (int) ( Sxy * E.dest()->crd[0] + Cx);
            y2 = (int) (-Sxy * E.dest()->crd[1] + Cy);
            //if (E.is_segment()) {
            //pen.setWidth(2);
            //painter.drawLine(QPoint(x1, y1), QPoint(x2,y2));
            //pen.setWidth(1);
            //} else {
            painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
            //}
          //}
        }
      }
    }
}

//=============================================================================
/*
void detri2_draw_weights(Triangulation* Tr, Triangulation* tmpTr, void *Painter, double Cx, double Cy, double Sxy)
{
    QPainter *painter = (QPainter *) Painter;
    //QPen pen;
    //pen.setColor(qRgb(255,140,0)); // dark orange
    //pen.setWidth(2); // default
    //painter.setPen(pen);

      for (int i = 0; i < Tr->tr_steiners->used_items; i++) {
        Vertex *v = (Vertex *) Tr->tr_steiners->get(i);
        if (v->is_deleted()) continue;
        int x = (int) ( Sxy * v->crd[0] + Cx);
        int y = (int) (-Sxy * v->crd[1] + Cy);
        if (v->wei > 0.0) {
          int r = (int) (sqrt(fabs(v->wei)) * Sxy);
          painter->drawEllipse(QPoint(x,y), r, r);
        } //else {
          QString text;
          text.sprintf("%.3f", v->wei);
          painter->drawText(x,y, text);
        //}
      }
}
*/
//=============================================================================

void detri2_draw_orthocircles(Triangulation* Tr, void *Painter, double Cx, double Cy, double Sxy)
{
  QPainter *painter = (QPainter *) Painter;

  // Always draw the nearest-point Delaunay circumcircles.
  //int bak_flag = Tr->op_dt_nearest;
  //Tr->op_dt_nearest = 1;

  /*
  // Weights (radius^2) might be negative.
  double minw=1.e+30;
  double maxw=-1.e+30;
  TriEdge E; //, S;
  for (int i = 0; i < Tr->tr_tris->used_items; i++) {
    E.tri = (Triang *) Tr->tr_tris->get(i);
    if (E.tri->is_deleted() || E.tri->is_hulltri()) continue;
    Tr->get_tri_orthocenter(E.tri);
    //qDebug()<<" i="<<i<<" cx="<<E.tri->cct[0]<<", cy="<<E.tri->cct[1]<<", r^2="<<E.tri->cct[2];
    if (E.tri->cct[2] < minw) minw =  E.tri->cct[2];
    if (E.tri->cct[2] > maxw) maxw =  E.tri->cct[2];
  }

  qDebug()<<" min weight ="<<minw <<", max weight ="<< maxw;
  if (minw < 0) {
    // There are negative weights (radius), which means some heights are above the paraboloid.
    qDebug()<<" Transforming negative weights into positive (for visualiation)";
    for (int i = 0; i < Tr->tr_tris->used_items; i++) {
      E.tri = (Triang *) Tr->tr_tris->get(i);
      if (E.tri->is_deleted() || E.tri->is_hulltri()) continue;
      double x = E.tri->cct[0];
      double y = E.tri->cct[1];
      double h = x*x+y*y - E.tri->cct[2];
      // Shift the height downwards by fabs(minw)
      double new_h = h - fabs(minw);
      // Calculate the new weight (muts be non-negaive).
      E.tri->cct[2] = x*x+y*y - new_h;
    }
  }
  */

  double lambda1, lambda2, v1[2], v2[2];
  get_MatVVP2x2(Tr->_a11,Tr->_a21,Tr->_a22,
                    &lambda1, &lambda2, v1, v2);

  //int x, y, r;
  double r1, r2, theta;
  int px, py, dr1, dr2;
  TriEdge E;  
  for (int i = 0; i < Tr->tr_tris->used_items; i++) {
    E.tri = (Triang *) Tr->tr_tris->get(i);
    if (E.tri->is_deleted() || E.tri->is_hulltri()) continue;
    Tr->get_tri_orthocenter(E.tri);
    //qDebug()<<" i="<<i<<" cx="<<E.tri->cct[0]<<", cy="<<E.tri->cct[1]<<", r^2="<<E.tri->cct[2];
    if (E.tri->cct[2] > 0) { // weight maybe negative
      px = (int) ( Sxy * E.tri->cct[0] + Cx);
      py = (int) (-Sxy * E.tri->cct[1] + Cy);
      //r = (int) ( Sxy * sqrt(E.tri->cct[2]));
      //painter->drawEllipse(QPoint(x,y), r, r);
      r1 = sqrt(E.tri->cct[2] / lambda1);
      r2 = sqrt(E.tri->cct[2] / lambda2);
      theta = get_rotation_angle(v1) / PI * 180.0;
      dr1 = (int) ( Sxy * r1);
      dr2 = (int) ( Sxy * r2);
      painter->save();
      painter->translate(px, py);
      painter->rotate(-theta); // CW direction (we need CCW)
      painter->translate(-px, -py);
      painter->drawEllipse(QPoint(px,py), dr1, dr2);
      painter->restore();
    } else {
      /*
      // A negative weight. Do not draw a circle.
      x = (int) ( Sxy * E.tri->cct[0] + Cx);
      y = (int) (-Sxy * E.tri->cct[1] + Cy);
      painter->drawEllipse(QPoint(x,y), 4, 4);
      QString text;
      text.sprintf("%.2f", E.tri->cct[2]);
      painter->drawText(x+4, y+4, text);
      */
    }
  }

  /*
   * //Tr->op_dt_nearest = bak_flag;
  if (Tr->op_dt_nearest < 0) {
    double vertical_shift = 0.0;
    if (minw < 0.) {
      vertical_shift = fabs(minw);
    }
    for (int i = 0; i < Tr->ct_in_vrts; i++) {
      Vertex *v = &(Tr->in_vrts[i]);
      if (v->typ == UNUSEDVERTEX) continue;
      double vx = v->crd[0];
      double vy = v->crd[1];
      double weight = vx*vx+vy*vy - v->crd[2];
      double vh_inv = -(vx*vx+vy*vy - weight); // The inverted distance.
      //double vw = vx*vx+vy*vy - vh_inv; // the weight of v in FDT
      double vh_fdt = vx*vx+vy*vy - vh_inv; // the height of v in FDT case.
      double new_vh = vh_fdt - vertical_shift; // shift the height in FDT case.
      double new_weight = vx*vx+vy*vy - new_vh;
      qDebug()<<"   vidx="<<v->idx<<" fdt weight="<<new_weight;
      if (new_weight > 0.) {
        int x = (int) ( Sxy * v->crd[0] + Cx);
        int y = (int) (-Sxy * v->crd[1] + Cy);
        int r = (int) (sqrt(new_weight) * Sxy);
        painter->drawEllipse(QPoint(x,y), r, r);
      }
    }
  }
  */
}

//=============================================================================

void detri2_draw_voronoi_1pt(Vertex* vt, Triangulation* OMT,
                             void* Painter, double Cx, double Cy, double Sxy)
{
    QPainter *painter = (QPainter *) Painter;
    qDebug()<<"detri2 draw voronoi for one point";

    // This is equivalent to draw the convex hull of the vertices of OMT.
    // Draw the convex hull.
    TriEdge E, N;
    int x1, y1, x2, y2;
    for (int i = 0; i < OMT->tr_tris->used_items; i++) {
      E.tri = (Triang *) OMT->tr_tris->get(i);
      if (E.tri->is_deleted()) continue;
      if (E.tri->is_hulltri()) continue;
      for (E.ver = 0; E.ver < 3; E.ver++) {
        N = E.esym();
        if (N.tri->is_hulltri()) {
          // Draw this hull edge.
          x1 = (int) ( Sxy * E.org()->crd[0] + Cx);
          y1 = (int) (-Sxy * E.org()->crd[1] + Cy);
          x2 = (int) ( Sxy * E.dest()->crd[0] + Cx);
          y2 = (int) (-Sxy * E.dest()->crd[1] + Cy);
          painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
        }
      }
    }
}

void detri2_draw_voronoi_2pts(Vertex* v1, Vertex* v2, Triangulation* OMT,
                              double a11, double a21, double a22,
                              void* Painter, double Cx, double Cy, double Sxy)
{
    QPainter *painter = (QPainter *) Painter;
    qDebug()<<"detri2 draw voronoi for two points";

    // The bisector of the line segment between [v1, v2].
    Vertex mid;
    mid.init();
    // temporary change the background metric
    OMT->_a11 = a11;
    OMT->_a21 = a21;
    OMT->_a22 = a22;

    //OMT->op_use_gmp = 0; // debug only
    get_bissector(v1->crd[0], v1->crd[1], v1->wei,
                       v2->crd[0], v2->crd[1], v2->wei,
                       &(mid.crd[0]), &(mid.crd[1]), &(mid.crd[2]),
                       a11, a21, a22);

    double outerN[2], len;
    /*
    outerN[0] =  (v1->crd[1] - v2->crd[1]);
    outerN[1] = -(v1->crd[0] - v2->crd[0]);
    len = sqrt(outerN[0]*outerN[0] + outerN[1]*outerN[1]);
    outerN[0] /= len;
    outerN[1] /= len;
    */

    // Calculate the bisector line (may not be orthogonal if metric is not Euclidean).
    // The line direction vector.
    double Ux = v1->crd[0], Uy = v1->crd[1];
    double Vx = v2->crd[0], Vy = v2->crd[1];
    outerN[0] =   2.*a22*(Uy-Vy) + 2.*a21*(Ux-Vx);
    outerN[1] = -(2.*a11*(Ux-Vx) + 2.*a21*(Uy-Vy));
    len = sqrt(outerN[0]*outerN[0] + outerN[1]*outerN[1]);
    outerN[0] /= len;
    outerN[1] /= len;
    // Choose the line direction towards exterior.
    Vertex tmp_pt; // Construct a point on the bisector line.
    tmp_pt.init();
    tmp_pt.crd[0] = mid.crd[0] + 2 * len * outerN[0];
    tmp_pt.crd[1] = mid.crd[1] + 2 * len * outerN[1];
    REAL check_sign = Orient2d(v1, v2, &tmp_pt);
    if (check_sign < 0) {
      // not ccw orientation. reverse the direction.
      outerN[0] = -outerN[0];
      outerN[1] = -outerN[1];
    } else if (check_sign == 0) {
      assert(0); // this is a bug.
    }

    Vertex far1, far2;
    far1.init();
    far2.init();
    double far_dist = OMT->io_diagonal * 2;
    far1.crd[0] = mid.crd[0] + outerN[0] * far_dist;
    far1.crd[1] = mid.crd[1] + outerN[1] * far_dist;
    far2.crd[0] = mid.crd[0] - outerN[0] * far_dist;
    far2.crd[1] = mid.crd[1] - outerN[1] * far_dist;

    /*
    int x1, y1, x2, y2;
    // Draw this edge.
    x1 = (int) ( Sxy * far1.crd[0] + Cx);
    y1 = (int) (-Sxy * far1.crd[1] + Cy);
    x2 = (int) ( Sxy * far2.crd[0] + Cx);
    y2 = (int) (-Sxy * far2.crd[1] + Cy);
    painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
    */

    // This is equivalent to draw the convex hull of the vertices of OMT.
    // Use hull edge to cut the bisector line.
    Vertex cut1, cut2;
    cut1.init();
    cut2.init();

    TriEdge S;
    int loc = OMT->locate_point(&mid, S, 0);
    loc = OMT->locate_point(&far1, S, 0);
    OMT->get_boundary_cut_dualedge(mid, far1, S);

    // Calculate the cut point.
    Vertex *e1 = S.org();
    Vertex *e2 = S.dest();
    double X0 = mid.crd[0];
    double Y0 = mid.crd[1];
    double X1 = far1.crd[0];
    double Y1 = far1.crd[1];
    double X2 = e1->crd[0];
    double Y2 = e1->crd[1];
    double X3 = e2->crd[0];
    double Y3 = e2->crd[1];
    double t1 = 0., t2 = 0.;
    line_line_intersection(X0, Y0, X1, Y1, X2, Y2, X3, Y3, &t1, &t2);
    cut1.crd[0] = X0 + t1 * (X1 - X0);
    cut1.crd[1] = Y0 + t1 * (Y1 - Y0);

    loc = OMT->locate_point(&mid, S, 0);
    loc = OMT->locate_point(&far2, S, 0);
    OMT->get_boundary_cut_dualedge(mid, far2, S);

    // Calculate the cut point.
    e1 = S.org();
    e2 = S.dest();
    X0 = mid.crd[0];
    Y0 = mid.crd[1];
    X1 = far2.crd[0];
    Y1 = far2.crd[1];
    X2 = e1->crd[0];
    Y2 = e1->crd[1];
    X3 = e2->crd[0];
    Y3 = e2->crd[1];
    line_line_intersection(X0, Y0, X1, Y1, X2, Y2, X3, Y3, &t1, &t2);
    cut2.crd[0] = X0 + t1 * (X1 - X0);
    cut2.crd[1] = Y0 + t1 * (Y1 - Y0);

    int x1, y1, x2, y2;
    // Draw this edge.
    x1 = (int) ( Sxy * cut1.crd[0] + Cx);
    y1 = (int) (-Sxy * cut1.crd[1] + Cy);
    x2 = (int) ( Sxy * cut2.crd[0] + Cx);
    y2 = (int) (-Sxy * cut2.crd[1] + Cy);
    painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));

    /*
    TriEdge E, N;
    int x1, y1, x2, y2;
    for (int i = 0; i < OMT->tr_tris->used_items; i++) {
      E.tri = (Triang *) OMT->tr_tris->get(i);
      if (E.tri->is_deleted()) continue;
      if (E.tri->is_hulltri()) continue;
      for (E.ver = 0; E.ver < 3; E.ver++) {
        N = E.esym();
        if (N.tri->is_hulltri()) {
          // Draw this hull edge.
          x1 = (int) ( Sxy * E.org()->crd[0] + Cx);
          y1 = (int) (-Sxy * E.org()->crd[1] + Cy);
          x2 = (int) ( Sxy * E.dest()->crd[0] + Cx);
          y2 = (int) (-Sxy * E.dest()->crd[1] + Cy);
          painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
          // Get the cut point.

        }
      }
    }
    */

    // Restore the metric
    OMT->_a11 = 1.0;
    OMT->_a21 = 0.0;
    OMT->_a22 = 1.0;
}

//=============================================================================

void detri2_draw_voronoi(Triangulation* Tr, Triangulation* OMT,
                         void *Painter, double Cx, double Cy, double Sxy)
{
  QPainter *painter = (QPainter *) Painter;
  qDebug()<<"detri2 draw voronoi";
  Tr->OMT_domain = OMT;

  qDebug()<<"  calculating voronoi vertices";
  int i, idx;
  idx = Tr->io_firstindex;
  // Calculate circumcenters for hull triangles.
  for (i = 0; i < Tr->tr_tris->used_items; i++) {
    Triang* tri = (Triang *) Tr->tr_tris->get(i);
    // Ignore exterior triangles.
    if (tri->is_deleted() || tri->is_hulltri()) continue;
    Tr->get_tri_orthocenter(tri);
    tri->idx = idx; // hulltri is also indexed.
    idx++;
  }
  //Tr->op_db_verbose=10; // debug
  // Calculate bisectors for hull triangles.
  for (i = 0; i < Tr->tr_tris->used_items; i++) {
    Triang* tri = (Triang *) Tr->tr_tris->get(i);
    // Ignore exterior triangles.
    if (tri->is_deleted()) continue;
    if (tri->is_hulltri()) { // A hull triangle.
      Tr->get_hulltri_orthocenter(tri);
    }
  }

  //qDebug()<<"\n========= Drawing voronoi edges ================\n";
  int x1, y1, x2, y2;
  TriEdge E, N;
  for (int i = 0; i < Tr->tr_tris->used_items; i++) {
    E.tri = (Triang *) Tr->tr_tris->get(i);
    if (E.tri->is_deleted() || E.tri->is_hulltri()) continue;
    for (E.ver = 0; E.ver < 3; E.ver++) {
      N = E.esym();
      if (N.tri->is_hulltri() || !N.tri->is_infected()) {
        // Draw the dual edge [E.tri->cct, N.tri->cct]
        if (E.tri->is_dual_in_exterior()) {
          if (N.tri->is_dual_in_exterior()) {
            if (E.tri->on.tri != N.tri->on.tri) {
              // They are on different boundary.
              //printf("  !!! Tow exterior dual vertices on different boundaries DEBUG!!!\n");
            }
          } else if (N.tri->is_dual_on_bdry()) {
            // Need to handle this case.
          } else {
            //qDebug()<<"  Calculating the dual of the edge ["<<E.org()->idx<<","<<E.dest()->idx<<"]";
            // N's dual vertex (N.tri->cct) lies inside the domain.
            // Calculating a cut vertex between this dual edge (from N.tri->cct to E.tri->cct)
            //    and a boundary edge (or a segment) S of the background mesh.
            //    We must ensure that the dual edge and S intersect in S's interior.
            TriEdge S = E.tri->on;
            Vertex In_pt, Out_pt;
            In_pt.init();
            Out_pt.init();
            In_pt.crd[0] = N.tri->cct[0];
            In_pt.crd[1] = N.tri->cct[1];
            Out_pt.crd[0] = E.tri->cct[0];
            Out_pt.crd[1] = E.tri->cct[1];

            if (!Tr->get_boundary_cut_dualedge(In_pt, Out_pt, S)) {
              continue; // Skip this edge.
            }
            /*
            //===================== Subroutine start =================================
            // see doc/voronoi-edge-boundary-cut-2018-11-1.pdf
            if (Tr->op_db_verbose > 3) {
              printf("  Dual edge: tri[%d,%d,%d] to tri[%d,%d,%d]\n",
                     N.org()->idx, N.dest()->idx, N.apex()->idx, E.org()->idx, E.dest()->idx, E.apex()->idx);
              printf("  In (%g,%g) - Out (%g,%g)\n", In_pt.crd[0], In_pt.crd[1], Out_pt.crd[0], Out_pt.crd[1]);
            }
            REAL ss = Orient2d(S.org(), S.dest(), &In_pt);
            assert(ss < 0); // CW orientation
            REAL s1 = Orient2d(&In_pt, &Out_pt, S.org());
            REAL s2 = Orient2d(&In_pt, &Out_pt, S.dest());
            if (Tr->op_db_verbose > 3) {
              printf("  [In, out, S.org(%d)]  s1 = %g\n", S.org()->idx, s1);
              printf("  [In, out, S.dest(%d)] s2 = %g\n", S.dest()->idx, s2);
            }
            if (s1 < 0) {
              do {
                // Must go left of S.org();
                TriEdge searchEdge = S.esym();
                //printf("DBG: adjust S to the left of S.org().\n");
                TriEdge workedge = searchEdge.enext();
                while (true) {
                  if (Tr->op_db_verbose > 3) {
                    printf("  workE: (%d,%d,%d)\n", workedge.org()->idx,
                           workedge.dest()->idx, workedge.apex()->idx);
                  }
                  assert(workedge.org() == S.org());
                  searchEdge = workedge.esym_enext(); // CW
                  if (searchEdge.is_segment()) break;
                  if ((searchEdge.esym()).tri->is_hulltri()) break;
                  workedge = searchEdge;
                }
                S = searchEdge.esym(); // Update S.
                s1 = Orient2d(&In_pt, &Out_pt, S.org());
              } while (s1 < 0);
            } else if (s2 > 0) {
              // Muts go right of S.dest();
              //printf("DBG: adjust S to the right of S.dest().\n");
              do {
                TriEdge searchEdge = S.esym();
                TriEdge workedge = searchEdge.eprev();
                while (true) {
                  if (Tr->op_db_verbose > 3) {
                    printf("  workE: (%d,%d,%d)\n", workedge.org()->idx,
                           workedge.dest()->idx, workedge.apex()->idx);
                  }
                  assert(workedge.dest() == S.dest());
                  searchEdge = workedge;
                  if (searchEdge.is_segment()) break;
                  if ((searchEdge.esym()).tri->is_hulltri()) break;
                  workedge = searchEdge.esym_eprev();
                }
                S = searchEdge.esym(); // Update S
                s2 = Orient2d(&In_pt, &Out_pt, S.dest());
              } while (s2 > 0);
            }
            if (Tr->op_db_verbose > 3) {
              printf("  Found background boundary S: [%d,%d,%d]\n\n",
                     S.org()->idx, S.dest()->idx, S.apex()->idx);
            }
            //===================== Subroutine end =================================
            */
            E.tri->on = S; // Update E.tri->on

            Vertex *e1 = E.tri->on.org();
            Vertex *e2 = E.tri->on.dest();
            assert((e1 != NULL) && (e2 != NULL));
            double X0 = E.tri->cct[0];
            double Y0 = E.tri->cct[1];
            double X1 = N.tri->cct[0];
            double Y1 = N.tri->cct[1];
            double X2 = e1->crd[0];
            double Y2 = e1->crd[1];
            double X3 = e2->crd[0];
            double Y3 = e2->crd[1];
            double t1 = 0, t2 = 0;
            if (line_line_intersection(X0, Y0, X1, Y1, X2, Y2, X3, Y3, &t1, &t2)) {
              double c0 = X0 + t1 * (X1 - X0);
              double c1 = Y0 + t1 * (Y1 - Y0);
              x1 = (int) ( Sxy * N.tri->cct[0] + Cx);
              y1 = (int) (-Sxy * N.tri->cct[1] + Cy);
              x2 = (int) ( Sxy * c0 + Cx);
              y2 = (int) (-Sxy * c1 + Cy);
              painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
            }
          }
        } else if (E.tri->is_dual_on_bdry()) {
          if (N.tri->is_dual_in_exterior()) {
            // Need to handle this case.
          } else if (N.tri->is_dual_on_bdry()) {
            if (E.tri->on.tri != N.tri->on.tri) {
              // They are on different boundary.
              //printf("  !!! Tow exterior dual vertices on different boundaries DEBUG!!!\n");
            }
          } else {
            // N's dual vertex lies inside the domain.
            // They are in the same subdomain.
            x1 = (int) ( Sxy * E.tri->cct[0] + Cx);
            y1 = (int) (-Sxy * E.tri->cct[1] + Cy);
            x2 = (int) ( Sxy * N.tri->cct[0] + Cx);
            y2 = (int) (-Sxy * N.tri->cct[1] + Cy);
            painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
          }
        } else {
          // E's dual vertex lies inside the domain (or subdomain).
          // There must exist a (or part of) dual edge.
          if (!N.tri->is_dual_in_exterior() || N.tri->is_dual_on_bdry()) {
            // An interior voronoi edge.
            x1 = (int) ( Sxy * E.tri->cct[0] + Cx);
            y1 = (int) (-Sxy * E.tri->cct[1] + Cy);
            x2 = (int) ( Sxy * N.tri->cct[0] + Cx);
            y2 = (int) (-Sxy * N.tri->cct[1] + Cy);
            painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
          } else {
            //qDebug()<<"  Calculating the dual of the edge ["<<E.org()->idx<<","<<E.dest()->idx<<"]";
            // The adjacent voronoi vertex is in exterior of the OMT.
            // Calculating a cut vertex between this dual edge (from E.tri->cct to N.tri->cct)
            //    and a boundary edge (or a segment) S of the background mesh.
            //    We must ensure that the dual edge and S intersect in S's interior.
            TriEdge S = N.tri->on;
            Vertex In_pt, Out_pt;
            In_pt.init();
            Out_pt.init();
            In_pt.crd[0] = E.tri->cct[0];
            In_pt.crd[1] = E.tri->cct[1];
            Out_pt.crd[0] = N.tri->cct[0];
            Out_pt.crd[1] = N.tri->cct[1];

            if (!Tr->get_boundary_cut_dualedge(In_pt, Out_pt, S)) {
              continue; // !! Skip this edge.
            }
            /*
            //===================== Subroutine start =================================
            // see doc/voronoi-edge-boundary-cut-2018-11-1.pdf
            if (Tr->op_db_verbose > 3) {
              printf("  Dual edge: tri[%d,%d,%d] to tri[%d,%d,%d]\n",
                     E.org()->idx, E.dest()->idx, E.apex()->idx, N.org()->idx, N.dest()->idx, N.apex()->idx);
              printf("  In (%g,%g) - Out (%g,%g)\n", In_pt.crd[0], In_pt.crd[1], Out_pt.crd[0], Out_pt.crd[1]);
            }
            REAL ss = Orient2d(S.org(), S.dest(), &In_pt);
            assert(ss < 0); // CW orientation
            REAL s1 = Orient2d(&In_pt, &Out_pt, S.org());
            REAL s2 = Orient2d(&In_pt, &Out_pt, S.dest());
            if (Tr->op_db_verbose > 3) {
              printf("  [In, out, S.org(%d)]  s1 = %g\n", S.org()->idx, s1);
              printf("  [In, out, S.dest(%d)] s2 = %g\n", S.dest()->idx, s2);
            }
            if (s1 < 0) {
              do {
                // Must go left of S.org();
                TriEdge searchEdge = S.esym();
                //printf("DBG: adjust S to the left of S.org().\n");
                TriEdge workedge = searchEdge.enext();
                while (true) {
                  if (Tr->op_db_verbose > 3) {
                    printf("  workE: (%d,%d,%d)\n", workedge.org()->idx,
                           workedge.dest()->idx, workedge.apex()->idx);
                  }
                  assert(workedge.org() == S.org());
                  searchEdge = workedge.esym_enext(); // CW
                  if (searchEdge.is_segment()) break;
                  if ((searchEdge.esym()).tri->is_hulltri()) break;
                  workedge = searchEdge;
                }
                S = searchEdge.esym(); // Update S.
                s1 = Orient2d(&In_pt, &Out_pt, S.org());
              } while (s1 < 0);
            } else if (s2 > 0) {
              // Muts go right of S.dest();
              //printf("DBG: adjust S to the right of S.dest().\n");
              do {
                TriEdge searchEdge = S.esym();
                TriEdge workedge = searchEdge.eprev();
                while (true) {
                  if (Tr->op_db_verbose > 3) {
                    printf("  workE: (%d,%d,%d)\n", workedge.org()->idx,
                           workedge.dest()->idx, workedge.apex()->idx);
                  }
                  assert(workedge.dest() == S.dest());
                  searchEdge = workedge;
                  if (searchEdge.is_segment()) break;
                  if ((searchEdge.esym()).tri->is_hulltri()) break;
                  workedge = searchEdge.esym_eprev();
                }
                S = searchEdge.esym(); // Update S
                s2 = Orient2d(&In_pt, &Out_pt, S.dest());
              } while (s2 > 0);
            }
            if (Tr->op_db_verbose > 3) {
              printf("  Found background boundary S: [%d,%d,%d]\n\n",
                     S.org()->idx, S.dest()->idx, S.apex()->idx);
            }
            //===================== Subroutine end =================================
            */
            N.tri->on = S; // Update N.tri->on

            Vertex *e1 = N.tri->on.org();
            Vertex *e2 = N.tri->on.dest();
            assert((e1 != NULL) && (e2 != NULL));
            double X0 = E.tri->cct[0];
            double Y0 = E.tri->cct[1];
            double X1 = N.tri->cct[0];
            double Y1 = N.tri->cct[1];
            double X2 = e1->crd[0];
            double Y2 = e1->crd[1];
            double X3 = e2->crd[0];
            double Y3 = e2->crd[1];
            double t1 = 0, t2 = 0;
            if (line_line_intersection(X0, Y0, X1, Y1, X2, Y2, X3, Y3, &t1, &t2)) {
              double c0 = X0 + t1 * (X1 - X0);
              double c1 = Y0 + t1 * (Y1 - Y0);
              x1 = (int) ( Sxy * E.tri->cct[0] + Cx);
              y1 = (int) (-Sxy * E.tri->cct[1] + Cy);
              x2 = (int) ( Sxy * c0 + Cx);
              y2 = (int) (-Sxy * c1 + Cy);
              painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
            }
          }
        } // if (!E.tri->is_dual_in_exterior())
      }
    }
    E.tri->set_infect();
  }

  // Uninfect all triangles.
  for (int i = 0; i < Tr->tr_tris->used_items; i++) {
    Triang* tri = (Triang *) Tr->tr_tris->get(i);
    if (tri->is_deleted() || tri->is_hulltri()) continue;
    tri->clear_infect();
  }

  Tr->OMT_domain=NULL;
}

void detri2_draw_voronoi_mass_centers(Triangulation* Tr, Triangulation* OMT,
                                      bool show_powe_cell, int center_size,
                                      void* Painter, double Cx, double Cy, double Sxy)
{
    QPainter *painter = (QPainter *) Painter;
    //qDebug()<<"detri2 draw voronoi";
    Tr->OMT_domain = OMT;

    qDebug()<<"  calculating voronoi vertices";
    int i, idx;
    idx = Tr->io_firstindex;
    // Calculate circumcenters for hull triangles.
    for (i = 0; i < Tr->tr_tris->used_items; i++) {
      Triang* tri = (Triang *) Tr->tr_tris->get(i);
      // Ignore exterior triangles.
      if (tri->is_deleted() || tri->is_hulltri()) continue;
      Tr->get_tri_orthocenter(tri);
      tri->idx = idx; // hulltri is also indexed.
      idx++;
    }

    printf("cccccccccccc detri2_draw_voronoi_mass_centers()\n");

    // Calculate bisectors for hull triangles.
    for (i = 0; i < Tr->tr_tris->used_items; i++) {
      Triang* tri = (Triang *) Tr->tr_tris->get(i);
      // Ignore exterior triangles.
      if (tri->is_deleted()) continue;
      if (tri->is_hulltri()) { // A hull triangle.
        Tr->get_hulltri_orthocenter(tri);
      }
    }

    printf("aaaaaaaaaaa detri2_draw_voronoi_mass_centers()\n");

    // Draw power cells and mass centers.
    for (int i = 0; i < Tr->ct_in_vrts; i++) {
      Vertex *mesh_vertex = &(Tr->in_vrts[i]);
      if (mesh_vertex->typ == UNUSEDVERTEX) continue;
      //printf("  idx = %d\n", i);
      Vertex *ptlist = NULL;
      int ptnum = 0;
      if (Tr->get_powercell(mesh_vertex, &ptlist, &ptnum)) {
        if (show_powe_cell) {
          // Draw the ptnum points and line segments.
          for (int i = 0; i < ptnum; i++) {
            Vertex *v1 = &(ptlist[i]);
            Vertex *v2 = &(ptlist[(i+1)%ptnum]);
            int x1 = (int) ( Sxy * v1->crd[0] + Cx);
            int y1 = (int) (-Sxy * v1->crd[1] + Cy);
            int x2 = (int) ( Sxy * v2->crd[0] + Cx);
            int y2 = (int) (-Sxy * v2->crd[1] + Cy);
            painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
          }
        }

        // Show mass center
        // The formula from Wikipedia.
        double Area = 0.; // signed area.
        for (int i = 0; i < ptnum; i++) {
          Vertex *v1 = &(ptlist[i]);
          Vertex *v2 = &(ptlist[(i+1)%ptnum]);
          double x0 = v1->crd[0];
          double y0 = v1->crd[1];
          double x1 = v2->crd[0];
          double y1 = v2->crd[1];
          Area += 0.5 * (x0 * y1 - x1 * y0);
        }
        double mcx = 0., mcy = 0.;
        for (int i = 0; i < ptnum; i++) {
          Vertex *v1 = &(ptlist[i]);
          Vertex *v2 = &(ptlist[(i+1)%ptnum]);
          double x0 = v1->crd[0];
          double y0 = v1->crd[1];
          double x1 = v2->crd[0];
          double y1 = v2->crd[1];
          mcx += (x0 + x1)*(x0 * y1 - x1 * y0);
          mcy += (y0 + y1)*(x0 * y1 - x1 * y0);
        }
        mcx /= (6 * Area);
        mcy /= (6 * Area);

        int px = (int) ( Sxy * mcx + Cx);
        int py = (int) (-Sxy * mcy + Cy);
        painter->drawEllipse(QPoint(px,py), center_size, center_size);

        delete [] ptlist;
      } // if (Tr->get_powercell
    }

    printf("bbbbbbbbbbbbb detri2_draw_voronoi_mass_centers()\n");

    if (Tr->tr_steiners == NULL) {
      return;
    }

    for (int i = 0; i < Tr->tr_steiners->used_items; i++) {
      Vertex *mesh_vertex = (Vertex *) Tr->tr_steiners->get(i);
      if (mesh_vertex->is_deleted()) continue;
      //printf("  idx = %d\n", i);
      Vertex *ptlist = NULL;
      int ptnum = 0;

      printf("%d ddddddddd detri2_draw_voronoi_mass_centers()\n", mesh_vertex->idx);

      if (Tr->get_powercell(mesh_vertex, &ptlist, &ptnum)) {
        if (show_powe_cell) {
          // Draw the ptnum points and line segments.
          for (int i = 0; i < ptnum; i++) {
            Vertex *v1 = &(ptlist[i]);
            Vertex *v2 = &(ptlist[(i+1)%ptnum]);
            int x1 = (int) ( Sxy * v1->crd[0] + Cx);
            int y1 = (int) (-Sxy * v1->crd[1] + Cy);
            int x2 = (int) ( Sxy * v2->crd[0] + Cx);
            int y2 = (int) (-Sxy * v2->crd[1] + Cy);
            painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
          }
        }

        // Show mass center
        // The formula from Wikipedia.
        double Area = 0.; // signed area.
        for (int i = 0; i < ptnum; i++) {
          Vertex *v1 = &(ptlist[i]);
          Vertex *v2 = &(ptlist[(i+1)%ptnum]);
          double x0 = v1->crd[0];
          double y0 = v1->crd[1];
          double x1 = v2->crd[0];
          double y1 = v2->crd[1];
          Area += 0.5 * (x0 * y1 - x1 * y0);
        }
        double mcx = 0., mcy = 0.;
        for (int i = 0; i < ptnum; i++) {
          Vertex *v1 = &(ptlist[i]);
          Vertex *v2 = &(ptlist[(i+1)%ptnum]);
          double x0 = v1->crd[0];
          double y0 = v1->crd[1];
          double x1 = v2->crd[0];
          double y1 = v2->crd[1];
          mcx += (x0 + x1)*(x0 * y1 - x1 * y0);
          mcy += (y0 + y1)*(x0 * y1 - x1 * y0);
        }
        mcx /= (6 * Area);
        mcy /= (6 * Area);

        int px = (int) ( Sxy * mcx + Cx);
        int py = (int) (-Sxy * mcy + Cy);
        painter->drawEllipse(QPoint(px,py), center_size, center_size);

        delete [] ptlist;
      } // if (Tr->get_powercell
    }

}

void detri2_draw_voronoi_cell(Triangulation* Tr, Triangulation* OMT, Vertex *mesh_vertex,
                              bool show_mass_center, int center_size,
                              void* Painter, double Cx, double Cy, double Sxy)
{
    QPainter *painter = (QPainter *) Painter;
    Tr->OMT_domain = OMT;

    // Calculate the circumcenters of triangles at this vertex.
    TriEdge E = mesh_vertex->adj;
    do {
      if (!E.tri->is_hulltri()) {
        Tr->get_tri_orthocenter(E.tri);
      }
      E = E.eprev_esym(); // ccw
    } while (E.tri != mesh_vertex->adj.tri);

    E = mesh_vertex->adj;
    do {
      if (E.tri->is_hulltri()) {
        Tr->get_hulltri_orthocenter(E.tri);
      }
      E = E.eprev_esym(); // ccw
    } while (E.tri != mesh_vertex->adj.tri);

    ////////////////////////////////////////////////////////////
    Vertex *ptlist = NULL;
    int ptnum = 0;

    if (Tr->get_powercell(mesh_vertex, &ptlist, &ptnum)) {
      // Draw the ptnum points and line segments.
      for (int i = 0; i < ptnum; i++) {
        Vertex *v1 = &(ptlist[i]);
        Vertex *v2 = &(ptlist[(i+1)%ptnum]);
        int x1 = (int) ( Sxy * v1->crd[0] + Cx);
        int y1 = (int) (-Sxy * v1->crd[1] + Cy);
        int x2 = (int) ( Sxy * v2->crd[0] + Cx);
        int y2 = (int) (-Sxy * v2->crd[1] + Cy);
        painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
      }

      if (show_mass_center) {
        // The formula from Wikipedia.
        double Area = 0.; // signed area.
        for (int i = 0; i < ptnum; i++) {
          Vertex *v1 = &(ptlist[i]);
          Vertex *v2 = &(ptlist[(i+1)%ptnum]);
          double x0 = v1->crd[0];
          double y0 = v1->crd[1];
          double x1 = v2->crd[0];
          double y1 = v2->crd[1];
          Area += 0.5 * (x0 * y1 - x1 * y0);
        }
        double mcx = 0., mcy = 0.;
        for (int i = 0; i < ptnum; i++) {
          Vertex *v1 = &(ptlist[i]);
          Vertex *v2 = &(ptlist[(i+1)%ptnum]);
          double x0 = v1->crd[0];
          double y0 = v1->crd[1];
          double x1 = v2->crd[0];
          double y1 = v2->crd[1];
          mcx += (x0 + x1)*(x0 * y1 - x1 * y0);
          mcy += (y0 + y1)*(x0 * y1 - x1 * y0);
        }
        mcx /= (6 * Area);
        mcy /= (6 * Area);

        int px = (int) ( Sxy * mcx + Cx);
        int py = (int) (-Sxy * mcy + Cy);
        painter->drawEllipse(QPoint(px,py), center_size, center_size);
        //detri2_draw_point(vt, 8, &painter, _cx, _cy, _scale_xy);
      }

      delete [] ptlist;
    }

    Tr->OMT_domain = NULL;
    return;
    ///////////////////////////////////////////////////////////////

    /*
    // Draw the Voronoi cell.
    int x1, y1, x2, y2;
    TriEdge N;
    E = mesh_vertex->adj;
    do {
      N = E.esym();

      if (E.tri->is_dual_in_exterior()) {
        if (N.tri->is_dual_in_exterior()) {
          if (E.tri->on.tri != N.tri->on.tri) {
            // They are on different boundary.
            //printf("  !!! Tow exterior dual vertices on different boundaries DEBUG!!!\n");
          }
        } else if (N.tri->is_dual_on_bdry()) {
          // This edge is from the OMT_domain boundary to exterior. Not visible.
          // No need to draw it.
        } else {
          // N's dual vertex lies inside the domain.
          Vertex *e1 = E.tri->on.org();
          Vertex *e2 = E.tri->on.dest();
          assert((e1 != NULL) && (e2 != NULL));
          double X0 = E.tri->cct[0];
          double Y0 = E.tri->cct[1];
          double X1 = N.tri->cct[0];
          double Y1 = N.tri->cct[1];
          double X2 = e1->crd[0];
          double Y2 = e1->crd[1];
          double X3 = e2->crd[0];
          double Y3 = e2->crd[1];
          double t1 = 0, t2 = 0;
          if (Tr->line_line_intersection(X0, Y0, X1, Y1, X2, Y2, X3, Y3, &t1, &t2)) {
            double c0 = X0 + t1 * (X1 - X0);
            double c1 = Y0 + t1 * (Y1 - Y0);
            x1 = (int) ( Sxy * N.tri->cct[0] + Cx);
            y1 = (int) (-Sxy * N.tri->cct[1] + Cy);
            x2 = (int) ( Sxy * c0 + Cx);
            y2 = (int) (-Sxy * c1 + Cy);
            painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
          }
        }
      } else if (E.tri->is_dual_on_bdry()) {
        if (N.tri->is_dual_in_exterior()) {
          // This edge is from the OMT_domain boundary to exterior. Not visible.
          // No need to draw this edge.
        } else if (N.tri->is_dual_on_bdry()) {
          if (E.tri->on.tri != N.tri->on.tri) {
            // They are on different boundary.
            //printf("  !!! Tow exterior dual vertices on different boundaries DEBUG!!!\n");
          }
        } else {
          // N's dual vertex lies inside the domain.
          // They are in the same subdomain.
          x1 = (int) ( Sxy * E.tri->cct[0] + Cx);
          y1 = (int) (-Sxy * E.tri->cct[1] + Cy);
          x2 = (int) ( Sxy * N.tri->cct[0] + Cx);
          y2 = (int) (-Sxy * N.tri->cct[1] + Cy);
          painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
        }
      } else {
        // E's dual vertex lies inside the domain (or subdomain).
        // There must exist a (or part of) dual edge.
        if (!N.tri->is_dual_in_exterior() || N.tri->is_dual_on_bdry()) {
          // An interior voronoi edge.
          x1 = (int) ( Sxy * E.tri->cct[0] + Cx);
          y1 = (int) (-Sxy * E.tri->cct[1] + Cy);
          x2 = (int) ( Sxy * N.tri->cct[0] + Cx);
          y2 = (int) (-Sxy * N.tri->cct[1] + Cy);
          painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
        } else {
          // The adjacent voronoi vertex is in exterior of the OMT.
          Vertex *e1 = N.tri->on.org();
          Vertex *e2 = N.tri->on.dest();
          assert((e1 != NULL) && (e2 != NULL));
          double X0 = E.tri->cct[0];
          double Y0 = E.tri->cct[1];
          double X1 = N.tri->cct[0];
          double Y1 = N.tri->cct[1];
          double X2 = e1->crd[0];
          double Y2 = e1->crd[1];
          double X3 = e2->crd[0];
          double Y3 = e2->crd[1];
          double t1 = 0, t2 = 0;
          if (Tr->line_line_intersection(X0, Y0, X1, Y1, X2, Y2, X3, Y3, &t1, &t2)) {
            double c0 = X0 + t1 * (X1 - X0);
            double c1 = Y0 + t1 * (Y1 - Y0);
            x1 = (int) ( Sxy * E.tri->cct[0] + Cx);
            y1 = (int) (-Sxy * E.tri->cct[1] + Cy);
            x2 = (int) ( Sxy * c0 + Cx);
            y2 = (int) (-Sxy * c1 + Cy);
            painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
          }
        }
      } // if (!E.tri->is_dual_in_exterior())

      // Go to the next edge.
      E = E.eprev_esym(); // ccw
    } while (E.tri != mesh_vertex->adj.tri);

    Tr->OMT_domain = NULL;
    */
}

// [2018-11-03] This function will be implemented in voronoi.cpp
void detri2_draw_voronoi_cell_debug(Triangulation* Tr, Triangulation* OMT, Vertex *mesh_vertex,
                              void* Painter, double Cx, double Cy, double Sxy)
{
    QPainter *painter = (QPainter *) Painter;
    //Tr->OMT_domain = OMT;

    //Vertex *ptlist = NULL;
    //int ptnum = 0;

    //if (!Tr->get_powercell(mesh_vertex, OMT, true, &ptlist, &ptnum)) {
    //  return; // Do not draw a boundary vertex.
    //}

    Tr->op_db_verbose = 10; // Debug

    //if (Tr->check_mesh(0, 0) > 0) {
    //  qDebug()<<"!!! Triangulation is not correct.";
    //  assert(0);
    //}
    // Debug special triangle
    //TriEdge tE = OMT->db_tri(-13,-14,7);
    //if (tE.tri == NULL) {
    //  printf("Not found\n");
    //}

    // A background mesh must be provided (maybe itself).
    if (OMT != NULL) {
      Tr->OMT_domain = OMT;
    }
    assert(Tr->OMT_domain != NULL);
    // Debug
    //Tr->OMT_domain->check_mesh(0,0);
    //strcpy(Tr->OMT_domain->io_outfilename, "/Users/si/Desktop/dumpOMT");
    //printf("%d input vertices.\n", Tr->OMT_domain->ct_in_vrts);
    //printf("%d steiner vertices.\n", Tr->OMT_domain->tr_steiners->objects);
    //printf("%d triangles.\n", Tr->OMT_domain->tr_tris->objects-Tr->OMT_domain->ct_hullsize);
    //Tr->OMT_domain->save_triangulation();

    // At the moment, we skip this operation for boundary vertices.
    int vcount = 0; // Count the number of Voronoi vertices.
    TriEdge E = mesh_vertex->adj;
    do {
      vcount++;
      if (E.is_segment()) { //if (E.tri->is_hulltri()) {
        if (Tr->op_db_verbose) {
          printf("  Vertex %d is a boundary vertex. Skipped.\n", mesh_vertex->idx);
        }
        return; //return 0;
      }
      E = E.eprev_esym(); // ccw
    } while (E.tri != mesh_vertex->adj.tri);

    if (Tr->op_db_verbose) {
      printf("  Found %d adjacent triangles at vertex %d.\n", vcount, mesh_vertex->idx);
    }

    if (true) { // if (calcflag) {
      // Calculate the circumcenters of triangles at this vertex.
      TriEdge E = mesh_vertex->adj;
      do {
        if (!E.tri->is_hulltri()) {
          Tr->get_tri_orthocenter(E.tri);
        }
        E = E.eprev_esym(); // ccw
      } while (E.tri != mesh_vertex->adj.tri);

      E = mesh_vertex->adj;
      do {
        if (E.tri->is_hulltri()) {
          Tr->get_hulltri_orthocenter(E.tri);
        }
        E = E.eprev_esym(); // ccw
      } while (E.tri != mesh_vertex->adj.tri);
    }

    TriEdge N, N_Last_Exit; // N_Last_Exit is a boundary or hull edge ([a,b,-1]) of the OMT_domain.
    N_Last_Exit.tri = NULL;

    // We need to start from an interior dual edge -- need two interior Voronoi vertices.
    // So the initial condition (N_Last_Exit.tri == NULL) is valid.
    //printf("\n DBG start: \n");
    E = mesh_vertex->adj;
    N = E.esym();
    bool is_interor_dual_edge = (!N.tri->is_dual_in_exterior() && !N.tri->is_dual_on_bdry());
    if (!is_interor_dual_edge) {
      do {
        //printf("  E: [%d,%d,%d] %d\n", E.org()->idx, E.dest()->idx, E.apex()->idx, E.tri->is_dual_in_exterior());
        //printf("  N: [%d,%d,%d] %d\n", N.org()->idx, N.dest()->idx, N.apex()->idx, N.tri->is_dual_in_exterior());
        E = E.eprev_esym(); // ccw
        N = E.esym();
        is_interor_dual_edge = (!N.tri->is_dual_in_exterior() && !N.tri->is_dual_on_bdry());
        if (is_interor_dual_edge) break;
      } while (E.tri != mesh_vertex->adj.tri);
      mesh_vertex->adj = E; // Must update this.
    }
    //printf(" DBG end: \n");
    if (!is_interor_dual_edge) {
      printf("All dual edges lie ouside of the OMT_domain.\n");
      return;
    }

    // Use an array to save the list of Voronoi or cut vertices (in ccw order).
    // [Remark] Due to the possible cut vertcies, the total number of vertcies
    //   of a power cell may be more than vcount. Make it large enough.
    Vertex *ptlist = new Vertex[vcount * 4];
    int ptcount = 0;

    do {
      if (Tr->op_db_verbose) {
        printf("  Get tri: [%d,%d,%d].\n", E.org()->idx, E.dest()->idx, E.apex()->idx);
      }
      N = E.esym();

      if (E.tri->is_dual_in_exterior()) {
        if (N.tri->is_dual_in_exterior() || N.tri->is_dual_on_bdry()) {
          // We're still walking in OUTSIDE.
          assert(N_Last_Exit.tri != NULL);
          /*
          if (E.tri->on.tri != N.tri->on.tri) {
            // They are on different boundary.
            printf("  !!! Searching background corners (1) DEBUG!!!\n");
            // Vertices on the boundary of OMT_domain are also corners of this cell.
            // Search them ==> see OMT_get_polygon_area()
            // Due to these boundary vertices, this cell may be non-convex.
            TriEdge startEdge, endEdge, searchEdge;
            startEdge = N.tri->on.esym(); // get the interior triangle
            endEdge = E.tri->on.esym();
            if (Tr->op_db_verbose > 3) {
              printf("  startE:   (%d,%d,%d)\n", startEdge.org()->idx,
                     startEdge.dest()->idx, startEdge.apex()->idx);
              printf("  endE:   (%d,%d,%d)\n", endEdge.org()->idx,
                     endEdge.dest()->idx, endEdge.apex()->idx);
            }
            // Start searching from startEdge towards endEdge.
            searchEdge = startEdge;
            while (searchEdge.tri != endEdge.tri) {
              if (Tr->op_db_verbose > 3) {
                printf("  searchE: (%d,%d,%d)\n", searchEdge.org()->idx,
                       searchEdge.dest()->idx, searchEdge.apex()->idx);
              }
              Vertex *pt = &(ptlist[ptcount]);
              pt->init();
              pt->crd[0] = searchEdge.dest()->crd[0];
              pt->crd[1] = searchEdge.dest()->crd[1];
              pt->crd[2] = searchEdge.dest()->crd[2]; // not used
              pt->tag = -2; // A background vertex.
              pt->idx = ptcount+1;
              ptcount++;
              // Go to the next boundary edge (around dest()).
              TriEdge workedge = searchEdge.enext();
              while (true) {
                searchEdge = workedge.esym_enext(); // CW
                if (searchEdge.is_segment()) break;
                if ((searchEdge.esym()).tri->is_hulltri()) break;
                if (Tr->op_db_verbose > 3) {
                  printf("  workE: (%d,%d,%d)\n", workedge.org()->idx,
                         workedge.dest()->idx, workedge.apex()->idx);
                }
                workedge = searchEdge;
              }
            }
            printf("Done!\n");
          }
          */
        } else {
          // N's dual vertex lies inside the domain.
          // We're walking from INSIDE to OUTSIDE.
          assert(N_Last_Exit.tri == NULL);
          // Calculating a cut vertex between this dual edge (from N.tri->cct to E.tri->cct)
          //    and a boundary edge (or a segment) S of the background mesh.
          //    We must ensure that the dual edge and S intersect in S's interior.
          TriEdge S = E.tri->on;
          Vertex In_pt, Out_pt;
          In_pt.init();
          Out_pt.init();
          In_pt.crd[0] = N.tri->cct[0];
          In_pt.crd[1] = N.tri->cct[1];
          Out_pt.crd[0] = E.tri->cct[0];
          Out_pt.crd[1] = E.tri->cct[1];
          Tr->get_boundary_cut_dualedge(In_pt, Out_pt, S);
          /*
          //===================== Subroutine start =================================
          // see doc/voronoi-edge-boundary-cut-2018-11-1.pdf
          if (Tr->op_db_verbose > 3) {
            printf("  Dual edge: tri[%d,%d,%d] to tri[%d,%d,%d]\n",
                   N.org()->idx, N.dest()->idx, N.apex()->idx, E.org()->idx, E.dest()->idx, E.apex()->idx);
            printf("  In (%g,%g) - Out (%g,%g)\n", In_pt.crd[0], In_pt.crd[1], Out_pt.crd[0], Out_pt.crd[1]);
          }
          REAL ss = Orient2d(S.org(), S.dest(), &In_pt);
          assert(ss < 0); // CW orientation
          REAL s1 = Orient2d(&In_pt, &Out_pt, S.org());
          REAL s2 = Orient2d(&In_pt, &Out_pt, S.dest());
          if (Tr->op_db_verbose > 3) {
            printf("  [In, out, S.org(%d)]  s1 = %g\n", S.org()->idx, s1);
            printf("  [In, out, S.dest(%d)] s2 = %g\n", S.dest()->idx, s2);
          }
          if (s1 < 0) {
            do {
              // Must go left of S.org();
              TriEdge searchEdge = S.esym();
              //printf("DBG: adjust S to the left of S.org().\n");
              TriEdge workedge = searchEdge.enext();
              while (true) {
                if (Tr->op_db_verbose > 3) {
                  printf("  workE: (%d,%d,%d)\n", workedge.org()->idx,
                         workedge.dest()->idx, workedge.apex()->idx);
                }
                assert(workedge.org() == S.org());
                searchEdge = workedge.esym_enext(); // CW
                if (searchEdge.is_segment()) break;
                if ((searchEdge.esym()).tri->is_hulltri()) break;
                workedge = searchEdge;
              }
              S = searchEdge.esym(); // Update S.
              s1 = Orient2d(&In_pt, &Out_pt, S.org());
            } while (s1 < 0);
          } else if (s2 > 0) {
            // Muts go right of S.dest();
            //printf("DBG: adjust S to the right of S.dest().\n");
            do {
              TriEdge searchEdge = S.esym();
              TriEdge workedge = searchEdge.eprev();
              while (true) {
                if (Tr->op_db_verbose > 3) {
                  printf("  workE: (%d,%d,%d)\n", workedge.org()->idx,
                         workedge.dest()->idx, workedge.apex()->idx);
                }
                assert(workedge.dest() == S.dest());
                searchEdge = workedge;
                if (searchEdge.is_segment()) break;
                if ((searchEdge.esym()).tri->is_hulltri()) break;
                workedge = searchEdge.esym_eprev();
              }
              S = searchEdge.esym(); // Update S
              s2 = Orient2d(&In_pt, &Out_pt, S.dest());
            } while (s2 > 0);
          }
          if (Tr->op_db_verbose > 3) {
            printf("  Found background boundary S: [%d,%d,%d]\n\n",
                   S.org()->idx, S.dest()->idx, S.apex()->idx);
          }
          //===================== Subroutine end =================================
          */
          E.tri->on = S; // Update E.tri->on
          //printf("  Last Exit [%d,%d,%d]\n", S.org()->idx, S.dest()->idx, S.apex()->idx);

          Vertex *e1 = E.tri->on.org();
          Vertex *e2 = E.tri->on.dest();
          assert((e1 != NULL) && (e2 != NULL));
          double X0 = E.tri->cct[0];
          double Y0 = E.tri->cct[1];
          double X1 = N.tri->cct[0];
          double Y1 = N.tri->cct[1];
          double X2 = e1->crd[0];
          double Y2 = e1->crd[1];
          double X3 = e2->crd[0];
          double Y3 = e2->crd[1];
          double t1 = 0, t2 = 0;
          if (line_line_intersection(X0, Y0, X1, Y1, X2, Y2, X3, Y3, &t1, &t2)) {
            double c0 = X0 + t1 * (X1 - X0);
            double c1 = Y0 + t1 * (Y1 - Y0);
            /*
            x1 = (int) ( Sxy * N.tri->cct[0] + Cx);
            y1 = (int) (-Sxy * N.tri->cct[1] + Cy);
            x2 = (int) ( Sxy * c0 + Cx);
            y2 = (int) (-Sxy * c1 + Cy);
            painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
            */
            Vertex *pt = &(ptlist[ptcount]);
            pt->init();
            pt->crd[0] = c0;
            pt->crd[1] = c1;
            pt->crd[2] = c0*c0 + c1*c1; // not used
            pt->tag = -1; // A cut vertex.
            pt->idx = ptcount+1;
            ptcount++;
          }

          // Remember this boundary edge.
          N_Last_Exit = E.tri->on;
          //printf("  2 Last Exit [%d,%d,%d]\n", N_Last_Exit.org()->idx,
          //       N_Last_Exit.dest()->idx, N_Last_Exit.apex()->idx);
        }
      } else if (E.tri->is_dual_on_bdry()) {
        if (N.tri->is_dual_in_exterior() || N.tri->is_dual_on_bdry()) {
          // We're walking from OUTSIDE to INSIDE
          assert(N_Last_Exit.tri != NULL);
          // Since E's dual vertex lies exactly on boundary, then E.tri->on must contain this dual vertex.
          //   This is guaranteed by the function get_hulltri_orthocenter().
          //   No need to call get_boundary_cut_dualedge().
          //printf("\n DBG start: \n");
          //printf("  N_Last_Exit: [%d,%d,%d]\n", N_Last_Exit.org()->idx, N_Last_Exit.dest()->idx, N_Last_Exit.apex()->idx);
          //printf("  E.tri->on:   [%d,%d,%d]\n", E.tri->on.org()->idx, E.tri->on.dest()->idx, E.tri->on.apex()->idx);
          //printf("\n DBG end: \n");
          if ((E.tri->on.tri != N_Last_Exit.tri) ||
              ((E.tri->on.tri == N_Last_Exit.tri) && (E.tri->on.ver != N_Last_Exit.ver))) {
            // They are on different boundary.
            printf("  !!! Searching background corners (2) DEBUG!!!\n");
            TriEdge startEdge, endEdge, searchEdge;
            startEdge = N_Last_Exit.esym();
            endEdge = E.tri->on.esym();
            //===================== Subroutine start =================================
            if (Tr->op_db_verbose) {
              printf("  Hull startE:   (%d,%d,%d)\n", startEdge.org()->idx,
                     startEdge.dest()->idx, startEdge.apex()->idx);
              printf("  endE:   (%d,%d,%d)\n", endEdge.org()->idx,
                     endEdge.dest()->idx, endEdge.apex()->idx);
            }
            // Start searching from startEdge towards endEdge.
            searchEdge = startEdge;
            while (searchEdge.tri != endEdge.tri) {
              if (Tr->op_db_verbose) {
                printf("  Hull searchE: (%d,%d,%d)\n", searchEdge.org()->idx,
                       searchEdge.dest()->idx, searchEdge.apex()->idx);
              }
              Vertex *pt = &(ptlist[ptcount]);
              pt->init();
              pt->crd[0] = searchEdge.dest()->crd[0];
              pt->crd[1] = searchEdge.dest()->crd[1];
              pt->crd[2] = searchEdge.dest()->crd[2]; // not used
              pt->tag = -2; // A background vertex.
              pt->idx = ptcount+1;
              ptcount++;
              // Go to the next boundary edge (around dest()).
              TriEdge workedge = searchEdge.enext();
              while (true) {
                searchEdge = workedge.esym_enext(); // CW
                if (searchEdge.is_segment()) break;
                if ((searchEdge.esym()).tri->is_hulltri()) break;
                if (Tr->op_db_verbose) {
                  printf("  Hull workE: (%d,%d,%d)\n", workedge.org()->idx,
                         workedge.dest()->idx, workedge.apex()->idx);
                }
                workedge = searchEdge;
              }
            }
            //printf("Done!\n");
            //===================== Subroutine end =================================
          }
          // Clear it. We're walking from OUTSIDE to INSIDE already.
          N_Last_Exit.tri = NULL;
        } else {
          // N's dual vertex lies inside the domain.
          // We're walking to the direction from INSIDE to OUTSIDE.
          // Remeber this boundary edge.
          assert(N_Last_Exit.tri == NULL);
          N_Last_Exit = E.tri->on;
          //printf("  2 Last Exit [%d,%d,%d]\n", N_Last_Exit.org()->idx,
          //       N_Last_Exit.dest()->idx, N_Last_Exit.apex()->idx);
          /*
          // They are in the same subdomain.
          x1 = (int) ( Sxy * E.tri->cct[0] + Cx);
          y1 = (int) (-Sxy * E.tri->cct[1] + Cy);
          x2 = (int) ( Sxy * N.tri->cct[0] + Cx);
          y2 = (int) (-Sxy * N.tri->cct[1] + Cy);
          painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
          */
        }

        Vertex *pt = &(ptlist[ptcount]);
        pt->init();
        pt->crd[0] = E.tri->cct[0];
        pt->crd[1] = E.tri->cct[1];
        pt->crd[2] = E.tri->cct[2];
        pt->tag = E.tri->idx; // A Voronoi vertex.
        pt->idx = ptcount+1;
        ptcount++;
      } else {
        // E's dual vertex lies inside the domain (or subdomain).
        // There must exist a (or part of) dual edge.
        if (!N.tri->is_dual_in_exterior() || N.tri->is_dual_on_bdry()) {
          // We're walking inside of the OMT_domain.
          assert(N_Last_Exit.tri == NULL);
          /*
          // An interior voronoi edge.
          x1 = (int) ( Sxy * E.tri->cct[0] + Cx);
          y1 = (int) (-Sxy * E.tri->cct[1] + Cy);
          x2 = (int) ( Sxy * N.tri->cct[0] + Cx);
          y2 = (int) (-Sxy * N.tri->cct[1] + Cy);
          painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
          */
        } else {
          // The adjacent voronoi vertex is in exterior of the OMT.
          // We're walking from OUTSIDE to INSIDE.
          assert(N_Last_Exit.tri != NULL);
          // Calculating a cut vertex between this dual edge (from N.tri->cct to E.tri->cct)
          //    and a boundary edge (or a segment) S of the background mesh.
          //    We must ensure that the dual edge and S intersect in S's interior.
          TriEdge S = N.tri->on;
          Vertex In_pt, Out_pt;
          In_pt.init();
          Out_pt.init();
          In_pt.crd[0] = E.tri->cct[0];
          In_pt.crd[1] = E.tri->cct[1];
          Out_pt.crd[0] = N.tri->cct[0];
          Out_pt.crd[1] = N.tri->cct[1];
          Tr->get_boundary_cut_dualedge(In_pt, Out_pt, S);
          /*
          //===================== Subroutine start =================================
          // see doc/voronoi-edge-boundary-cut-2018-11-1.pdf
          if (Tr->op_db_verbose > 3) {
            printf("  Dual edge: tri[%d,%d,%d] to tri[%d,%d,%d]\n",
                   E.org()->idx, E.dest()->idx, E.apex()->idx, N.org()->idx, N.dest()->idx, N.apex()->idx);
            printf("  In (%g,%g) - Out (%g,%g)\n", In_pt.crd[0], In_pt.crd[1], Out_pt.crd[0], Out_pt.crd[1]);
          }
          REAL ss = Orient2d(S.org(), S.dest(), &In_pt);
          assert(ss < 0); // CW orientation
          REAL s1 = Orient2d(&In_pt, &Out_pt, S.org());
          REAL s2 = Orient2d(&In_pt, &Out_pt, S.dest());
          if (Tr->op_db_verbose > 3) {
            printf("  [In, out, S.org(%d)]  s1 = %g\n", S.org()->idx, s1);
            printf("  [In, out, S.dest(%d)] s2 = %g\n", S.dest()->idx, s2);
          }
          if (s1 < 0) {
            do {
              // Must go left of S.org();
              TriEdge searchEdge = S.esym();
              //printf("DBG: adjust S to the left of S.org().\n");
              TriEdge workedge = searchEdge.enext();
              while (true) {
                if (Tr->op_db_verbose > 3) {
                  printf("  workE: (%d,%d,%d)\n", workedge.org()->idx,
                         workedge.dest()->idx, workedge.apex()->idx);
                }
                assert(workedge.org() == S.org());
                searchEdge = workedge.esym_enext(); // CW
                if (searchEdge.is_segment()) break;
                if ((searchEdge.esym()).tri->is_hulltri()) break;
                workedge = searchEdge;
              }
              S = searchEdge.esym(); // Update S.
              s1 = Orient2d(&In_pt, &Out_pt, S.org());
            } while (s1 < 0);
          } else if (s2 > 0) {
            // Muts go right of S.dest();
            //printf("DBG: adjust S to the right of S.dest().\n");
            do {
              TriEdge searchEdge = S.esym();
              TriEdge workedge = searchEdge.eprev();
              while (true) {
                if (Tr->op_db_verbose > 3) {
                  printf("  workE: (%d,%d,%d)\n", workedge.org()->idx,
                         workedge.dest()->idx, workedge.apex()->idx);
                }
                assert(workedge.dest() == S.dest());
                searchEdge = workedge;
                if (searchEdge.is_segment()) break;
                if ((searchEdge.esym()).tri->is_hulltri()) break;
                workedge = searchEdge.esym_eprev();
              }
              S = searchEdge.esym(); // Update S
              s2 = Orient2d(&In_pt, &Out_pt, S.dest());
            } while (s2 > 0);
          }
          if (Tr->op_db_verbose > 3) {
            printf("  Found background boundary S: [%d,%d,%d]\n\n",
                   S.org()->idx, S.dest()->idx, S.apex()->idx);
          }
          //===================== Subroutine end =================================
          */
          N.tri->on = S; // Update N.tri->on

          //printf("\n DBG start: \n");
          //printf("  N_Last_Exit: [%d,%d,%d]\n", N_Last_Exit.org()->idx, N_Last_Exit.dest()->idx, N_Last_Exit.apex()->idx);
          //printf("  N.tri->on:   [%d,%d,%d]\n", N.tri->on.org()->idx, N.tri->on.dest()->idx, N.tri->on.apex()->idx);
          //printf("\n DBG end: \n");
          if ((N.tri->on.tri != N_Last_Exit.tri) ||
              ((N.tri->on.tri == N_Last_Exit.tri) && (N.tri->on.ver != N_Last_Exit.ver))) {
            // They are on different boundary.
            printf("  !!! Searching background corners (2) DEBUG!!!\n");
            TriEdge startEdge, endEdge, searchEdge;
            startEdge = N_Last_Exit.esym();
            endEdge = N.tri->on.esym();
            //===================== Subroutine start =================================
            if (Tr->op_db_verbose) {
              printf("  startE:   (%d,%d,%d)\n", startEdge.org()->idx,
                     startEdge.dest()->idx, startEdge.apex()->idx);
              printf("  endE:   (%d,%d,%d)\n", endEdge.org()->idx,
                     endEdge.dest()->idx, endEdge.apex()->idx);
            }
            // Start searching from startEdge towards endEdge.
            searchEdge = startEdge;
            while (searchEdge.tri != endEdge.tri) {
              if (Tr->op_db_verbose) {
                printf("  Hull searchE: (%d,%d,%d)\n", searchEdge.org()->idx,
                       searchEdge.dest()->idx, searchEdge.apex()->idx);
              }
              Vertex *pt = &(ptlist[ptcount]);
              pt->init();
              pt->crd[0] = searchEdge.dest()->crd[0];
              pt->crd[1] = searchEdge.dest()->crd[1];
              pt->crd[2] = searchEdge.dest()->crd[2]; // not used
              pt->tag = -2; // A background vertex.
              pt->idx = ptcount+1;
              ptcount++;
              // Go to the next boundary edge (around dest()).
              TriEdge workedge = searchEdge.enext();
              while (true) {
                searchEdge = workedge.esym_enext(); // CW
                if (searchEdge.is_segment()) break;
                if ((searchEdge.esym()).tri->is_hulltri()) break;
                if (Tr->op_db_verbose) {
                  printf("  Hull workE: (%d,%d,%d)\n", workedge.org()->idx,
                         workedge.dest()->idx, workedge.apex()->idx);
                }
                workedge = searchEdge;
              }
            }
            //printf("Done!\n");
            //===================== Subroutine end =================================
          }

          Vertex *e1 = N.tri->on.org();
          Vertex *e2 = N.tri->on.dest();
          assert((e1 != NULL) && (e2 != NULL));
          double X0 = E.tri->cct[0];
          double Y0 = E.tri->cct[1];
          double X1 = N.tri->cct[0];
          double Y1 = N.tri->cct[1];
          double X2 = e1->crd[0];
          double Y2 = e1->crd[1];
          double X3 = e2->crd[0];
          double Y3 = e2->crd[1];
          double t1 = 0, t2 = 0;
          if (line_line_intersection(X0, Y0, X1, Y1, X2, Y2, X3, Y3, &t1, &t2)) {
            double c0 = X0 + t1 * (X1 - X0);
            double c1 = Y0 + t1 * (Y1 - Y0);
            /*
            x1 = (int) ( Sxy * E.tri->cct[0] + Cx);
            y1 = (int) (-Sxy * E.tri->cct[1] + Cy);
            x2 = (int) ( Sxy * c0 + Cx);
            y2 = (int) (-Sxy * c1 + Cy);
            painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
            */
            Vertex *pt = &(ptlist[ptcount]);
            pt->init();
            pt->crd[0] = c0;
            pt->crd[1] = c1;
            pt->crd[2] = c0*c0 + c1*c1; // not used
            pt->tag = -1; // A cut vertex.
            pt->idx = ptcount+1;
            ptcount++;
          }
        }

        Vertex *pt = &(ptlist[ptcount]);
        pt->init();
        pt->crd[0] = E.tri->cct[0];
        pt->crd[1] = E.tri->cct[1];
        pt->crd[2] = E.tri->cct[2];
        pt->tag = E.tri->idx; // A Voronoi vertex.
        pt->idx = ptcount+1;
        ptcount++;

        // Clear it, we're inside the OMT_domain again.
        N_Last_Exit.tri = NULL;
      } // if (!E.tri->is_dual_in_exterior())

      // Go to the next edge.
      E = E.eprev_esym(); // ccw
    } while (E.tri != mesh_vertex->adj.tri);

    //*pptlist = ptlist;
    int ptnum = ptcount;

    // Draw the ptnum points and line segments.
    for (int i = 0; i < ptnum; i++) {
      Vertex *v1 = &(ptlist[i]);
      Vertex *v2 = &(ptlist[(i+1)%ptnum]);
      int x1 = (int) ( Sxy * v1->crd[0] + Cx);
      int y1 = (int) (-Sxy * v1->crd[1] + Cy);
      int x2 = (int) ( Sxy * v2->crd[0] + Cx);
      int y2 = (int) (-Sxy * v2->crd[1] + Cy);
      painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
    }

    delete [] ptlist;
}

//=======================================================================

void detri2_draw_laplacian_center_2(Triangulation* Tr, Vertex *mesh_vertex, int center_size,
                                    void* Painter, double Cx, double Cy, double Sxy)
{
    QPainter *painter = (QPainter *) Painter;

    int x1, y1, x2, y2;
    double mcx = 0., mcy = 0.;
    int deg = 0; // debgree
    bool is_seg_vertex = false;
    TriEdge E = mesh_vertex->adj;
    x1 = (int) ( Sxy * mesh_vertex->crd[0] + Cx);
    y1 = (int) (-Sxy * mesh_vertex->crd[1] + Cy);
    do {
      if (E.is_segment()) {
        is_seg_vertex = true;
        break;
      }
      if (!E.tri->is_hulltri()) {
        deg++;
        mcx += (E.dest()->crd[0] - mesh_vertex->crd[0]);
        mcy += (E.dest()->crd[1] - mesh_vertex->crd[1]);
        x2 = (int) ( Sxy * E.dest()->crd[0] + Cx);
        y2 = (int) (-Sxy * E.dest()->crd[1] + Cy);
        painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
      }
      E = E.eprev_esym(); // CCW
    } while (E.tri != mesh_vertex->adj.tri);
    if (!is_seg_vertex) {
      //if (deg <= 2) continue;
      mcx /= deg;
      mcy /= deg;
      // Draw this point.
      int px = (int) ( Sxy * (mesh_vertex->crd[0] + mcx) + Cx);
      int py = (int) (-Sxy * (mesh_vertex->crd[1] + mcy) + Cy);
      painter->drawEllipse(QPoint(px,py), center_size, center_size);
    }
}

void detri2_draw_laplacian_center(Vertex *mesh_vertex, int center_size,
                                  void* Painter, double Cx, double Cy, double Sxy)
{
    QPainter *painter = (QPainter *) Painter;

    int x1, y1, x2, y2;
    double mcx = 0., mcy = 0.;
    int deg = 0; // debgree
    bool is_seg_vertex = false;
    TriEdge E = mesh_vertex->adj;
    x1 = (int) ( Sxy * mesh_vertex->crd[0] + Cx);
    y1 = (int) (-Sxy * mesh_vertex->crd[1] + Cy);
    do {
      if (E.is_segment()) {
        is_seg_vertex = true;
        break;
      }
      if (!E.tri->is_hulltri()) {
        deg++;
        mcx += E.dest()->crd[0];
        mcy += E.dest()->crd[1];
        x2 = (int) ( Sxy * E.dest()->crd[0] + Cx);
        y2 = (int) (-Sxy * E.dest()->crd[1] + Cy);
        painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
      }
      E = E.eprev_esym(); // CCW
    } while (E.tri != mesh_vertex->adj.tri);
    if (!is_seg_vertex) {
      //if (deg <= 2) continue;
      mcx /= deg;
      mcy /= deg;
      // Draw this point.
      int px = (int) ( Sxy * mcx + Cx);
      int py = (int) (-Sxy * mcy + Cy);
      painter->drawEllipse(QPoint(px,py), center_size, center_size);
    }
}

void detri2_draw_laplacian_centers(Triangulation* Tr, int center_size,
                                   void* Painter, double Cx, double Cy, double Sxy)
{
    QPainter *painter = (QPainter *) Painter;

    for (int i = 0; i < Tr->ct_in_vrts; i++) {
      Vertex *mesh_vertex = &(Tr->in_vrts[i]);
      if (mesh_vertex->typ == UNUSEDVERTEX) continue;
      //printf("  idx = %d fixed(%d)\n", i, mesh_vertex->is_fixed());
      if (mesh_vertex->is_fixed()) continue;
      detri2_draw_laplacian_center(mesh_vertex, center_size, Painter, Cx, Cy, Sxy);
    }

    //printf("bbbbbbbbbbbbb\n");
    if (Tr->tr_steiners == NULL) {
      return;
    }

    for (int i = 0; i < Tr->tr_steiners->used_items; i++) {
      Vertex *mesh_vertex = (Vertex *) Tr->tr_steiners->get(i);
      if (mesh_vertex->is_deleted()) continue;
      //printf("  idx = %d fixed(%d)\n", i, mesh_vertex->is_fixed());
      if (mesh_vertex->is_fixed()) continue;
      detri2_draw_laplacian_center(mesh_vertex, center_size, Painter, Cx, Cy, Sxy);
    }
}

//=======================================================================

void detri2_draw_distmesh_center(Triangulation *Tr, Vertex *mesh_vertex,
                                 int center_size, void* Painter, double Cx, double Cy, double Sxy)
{
    QPainter *painter = (QPainter *) Painter;

    int x1, y1, x2, y2;
    REAL L, F, barvec[2], Ftot[2];
    double mcx = 0., mcy = 0.;
    Ftot[0] = Ftot[1] = 0.;
    bool is_seg_vertex = false;
    TriEdge E = mesh_vertex->adj;
    x1 = (int) ( Sxy * mesh_vertex->crd[0] + Cx);
    y1 = (int) (-Sxy * mesh_vertex->crd[1] + Cy);
    do {
      if (E.is_segment()) {
        is_seg_vertex = true;
        break;
      }
      Vertex *pd = E.dest();
      if (pd != Tr->tr_infvrt) {
        //if (mesh_vertex->idx < pd->idx) {
          barvec[0] = pd->crd[0] - mesh_vertex->crd[0];
          barvec[1] = pd->crd[1] - mesh_vertex->crd[1];
        //} else {
        //  barvec[0] = mesh_vertex->crd[0] - pd->crd[0]; // test
        //  barvec[1] = mesh_vertex->crd[1] - pd->crd[1]; // test
        //}
        L = Tr->get_distance(mesh_vertex, pd);
        F = Tr->op_target_length - L;
        if (F < 0.) F = 0.;
        if (F > 0.) {
          // Ftot += Fvec
          Ftot[0] += (-barvec[0] * F / L);
          Ftot[1] += (-barvec[1] * F / L);
          x2 = (int) ( Sxy * E.dest()->crd[0] + Cx);
          y2 = (int) (-Sxy * E.dest()->crd[1] + Cy);
          painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
        }
      }
      E = E.eprev_esym(); // CCW
    } while (E.tri != mesh_vertex->adj.tri);

    if (!is_seg_vertex) {
      mcx = mesh_vertex->crd[0] + Tr->op_smooth_deltat * Ftot[0];
      mcy = mesh_vertex->crd[1] + Tr->op_smooth_deltat * Ftot[1];
      // Draw this point.
      int px = (int) ( Sxy * mcx + Cx);
      int py = (int) (-Sxy * mcy + Cy);
      painter->drawEllipse(QPoint(px,py), center_size, center_size);
    }
}

void detri2_draw_distmesh_points(Triangulation* Tr, int center_size,
                                 void* Painter, double Cx, double Cy, double Sxy)
{
    int ptnum = Tr->ct_in_vrts;
    if (Tr->tr_steiners != NULL) ptnum += Tr->tr_steiners->objects;
    Vertex *moveptlist = new Vertex[ptnum];

    qDebug()<<"  calculating distmesh points: len="<<Tr->op_target_length<<", dt="<<Tr->op_smooth_deltat;
    Tr->get_distmesh_points(moveptlist);

    QPainter *painter = (QPainter *) Painter;

    for (int i = 0; i < ptnum; i++) {
      Vertex *v = &(moveptlist[i]);
      if (v->Pair != NULL) {
        int px = (int) ( Sxy * v->crd[0] + Cx);
        int py = (int) (-Sxy * v->crd[1] + Cy);
        painter->drawEllipse(QPoint(px,py), center_size, center_size);
      }
    }

    delete [] moveptlist;
}

//=======================================================================

static int  transgc[8][3][8], tsb1mod3[8];

void detri2_draw_hilbert_init(int n)
{
  int gc[8], N, mask, travel_bit;
  int e, d, f, k, g;
  int v, c;
  int i;

  N = (n == 2) ? 4 : 8;
  mask = (n == 2) ? 3 : 7;

  // Generate the Gray code sequence.
  for (i = 0; i < N; i++) {
    gc[i] = i ^ (i >> 1);
  }

  for (e = 0; e < N; e++) {
    for (d = 0; d < n; d++) {
      // Calculate the end point (f).
      f = e ^ (1 << d);  // Toggle the d-th bit of 'e'.
      // travel_bit = 2**p, the bit we want to travel.
      travel_bit = e ^ f;
      for (i = 0; i < N; i++) {
        // // Rotate gc[i] left by (p + 1) % n bits.
        k = gc[i] * (travel_bit * 2);
        g = ((k | (k / N)) & mask);
        // Calculate the permuted Gray code by xor with the start point (e).
        transgc[e][d][i] = (g ^ e);
      }
// \iffalse
      assert(transgc[e][d][0] == e);
      assert(transgc[e][d][N - 1] == f);
// \fi
    } // d
  } // e

  // Output the table of transformed gray codes
  printf("transgc table (4 x 2 x 4):\n");
  printf("e(0) d(0) |  e(0) d(1) | e(1) d(0) |  e(1) d(1) | e(2) d(0) |  e(2) d(1) | e(3) d(0) |  e(3) d(1)\n");
  for (i = 0; i < 4; i++) {
    printf("%d %d | %d %d | %d %d | %d %d\n",
           transgc[0][0][i], transgc[0][1][i],
           transgc[1][0][i], transgc[1][1][i],
           transgc[2][0][i], transgc[2][1][i],
           transgc[3][0][i], transgc[3][1][i]);
  }

  // Count the consecutive '1' bits (trailing) on the right.
  tsb1mod3[0] = 0;
  for (i = 1; i < N; i++) {
    v = ~i; // Count the 0s.
    v = (v ^ (v - 1)) >> 1; // Set v's trailing 0s to 1s and zero rest
    for (c = 0; v; c++) {
      v >>= 1;
    }
    tsb1mod3[i] = c % n;
  }
}

void detri2_draw_Hilbert_curve(double bxmin, double bxmax, double bymin, double bymax,
                               int e, int d, int order, int depth,
                               double *EntryPt, double *ExitPt, int *hindex,
                               void* Painter, double Cx, double Cy, double Sxy)
{
  double x1, x2, y1, y2;
  int w, e_w, d_w, k, ei, di;
  int n = 2, mask = 3;

  double vrt[4][2];
  int xx[4], yy[4];

  if ((depth + 1) == order) {
    // Get the four centers of the subboxes.
    for (w = 0; w < 4; w++) {
      if (transgc[e][d][w] & 1) { // x-axis
        x1 = 0.5 * (bxmin + bxmax);
        x2 = bxmax;
      } else {
        x1 = bxmin;
        x2 = 0.5 * (bxmin + bxmax);
      }
      if (transgc[e][d][w] & 2) { // y-axis
        y1 = 0.5 * (bymin + bymax);
        y2 = bymax;
      } else {
        y1 = bymin;
        y2 = 0.5 * (bymin + bymax);
      }
      vrt[w][0] = 0.5 * (x1 + x2);
      vrt[w][1] = 0.5 * (y1 + y2);
    }
    // Draw the curve
    QPainter *painter = (QPainter *) Painter;
    for (w = 0; w < 4; w++) {
      xx[w] = (int) ( Sxy * vrt[w][0] + Cx);
      yy[w] = (int) (-Sxy * vrt[w][1] + Cy);
      if (hindex) {
        painter->drawText(QPoint(xx[w], yy[w]), QString::number(*hindex));
        (*hindex)++;
      }
    }
    for (w = 0; w < 3; w++) {
      painter->drawLine(QPoint(xx[w], yy[w]), QPoint(xx[w+1],yy[w+1]));
    }
    // Return the entry and exit points of this subbox.
    // for drawing the connecting line between two adjacent subboxes.
    if (EntryPt != NULL) {
      EntryPt[0] = vrt[0][0];
      EntryPt[1] = vrt[0][1];
      ExitPt[0]  = vrt[3][0];
      ExitPt[1]  = vrt[3][1];
    }
    return;
  } else if ((depth + 1) > order) {
    // Wrong depth and order parameters. Nothing to draw
    return;
  }

  double EntPt[4][2], ExtPt[4][2]; // Entry and Exit points.

  // Recursively sort the points in sub-boxes.
  //   See Theorem 2.10 (page 13) for e_w in [Hamilton 2006]
  //   See    Lemma 2.8 (page 12) for d_w in [Hamilton 2006]
  //   See my notes at page 18, formula for ei and di. [2017-06-27].
  for (w = 0; w < 4; w++) {
    // Calculcate the start point (ei) of the curve in this sub-box.
    //   update e = e ^ (e(w) left_rotate (d+1)).
    if (w == 0) {
      e_w = 0;
    } else {
      //   calculate e(w) = gc(2 * floor((w - 1) / 2)).
      k = 2 * ((w - 1) / 2);
      e_w = k ^ (k >> 1); // = gc(k).
    }
    k = e_w;
    e_w = ((k << (d+1)) & mask) | ((k >> (n-d-1)) & mask);
    ei = e ^ e_w;
    // Calulcate the direction (di) of the curve in this sub-box.
    //   update d = (d + d(w) + 1) % n
    if (w == 0) {
      d_w = 0;
    } else {
      // \iffalse
      /*if ((w % 2) == 0) {
        d_w = tsb1(w - 1) % n;
      } else {
        d_w = tsb1(w) % n;
      } */
      // \fi
      d_w = ((w % 2) == 0) ? tsb1mod3[w - 1] : tsb1mod3[w];
    }
    di = (d + d_w + 1) % n;

    // Calculate the bounding box of the sub-box.
    if (transgc[e][d][w] & 1) { // x-axis
      x1 = 0.5 * (bxmin + bxmax);
      x2 = bxmax;
    } else {
      x1 = bxmin;
      x2 = 0.5 * (bxmin + bxmax);
    }
    if (transgc[e][d][w] & 2) { // y-axis
      y1 = 0.5 * (bymin + bymax);
      y2 = bymax;
    } else {
      y1 = bymin;
      y2 = 0.5 * (bymin + bymax);
    }
    //if (transgc[e][d][w] & 4) { // z-axis
    //  z1 = 0.5 * (bzmin + bzmax);
    //  z2 = bzmax;
    //} else {
    //  z1 = bzmin;
    //  z2 = 0.5 * (bzmin + bzmax);
    //}

    detri2_draw_Hilbert_curve(x1,x2,y1,y2,
                              ei,di,order,depth+1,
                              EntPt[w], ExtPt[w], hindex,
                              Painter, Cx, Cy, Sxy);
  } // w

  // Draw the connection between two subbodxes.
  QPainter *painter = (QPainter *) Painter;
  xx[0] = (int) ( Sxy * ExtPt[0][0] + Cx);
  yy[0] = (int) (-Sxy * ExtPt[0][1] + Cy);
  xx[1] = (int) ( Sxy * EntPt[1][0] + Cx);
  yy[1] = (int) (-Sxy * EntPt[1][1] + Cy);
  painter->drawLine(QPoint(xx[0], yy[0]), QPoint(xx[1],yy[1]));
  xx[0] = (int) ( Sxy * ExtPt[1][0] + Cx);
  yy[0] = (int) (-Sxy * ExtPt[1][1] + Cy);
  xx[1] = (int) ( Sxy * EntPt[2][0] + Cx);
  yy[1] = (int) (-Sxy * EntPt[2][1] + Cy);
  painter->drawLine(QPoint(xx[0], yy[0]), QPoint(xx[1],yy[1]));
  xx[0] = (int) ( Sxy * ExtPt[2][0] + Cx);
  yy[0] = (int) (-Sxy * ExtPt[2][1] + Cy);
  xx[1] = (int) ( Sxy * EntPt[3][0] + Cx);
  yy[1] = (int) (-Sxy * EntPt[3][1] + Cy);
  painter->drawLine(QPoint(xx[0], yy[0]), QPoint(xx[1],yy[1]));

  // Return the entry and exit points
  if (EntryPt != NULL) {
    EntryPt[0] = EntPt[0][0];
    EntryPt[1] = EntPt[0][1];
    ExitPt[0]  = ExtPt[3][0];
    ExitPt[1]  = ExtPt[3][1];
  }
}
